// Watcom includes
#include <fstream.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <i86.h>
#include <dos.h>

// HGL includes
#include "multikey.h"
#include "sprite.h"
#include "mouse.h"
#include "font.h"
#include "vgl.h"

// Z2 includes
#include "compiler.h"
#include "resource.h"
#include "objects.h"
#include "console.h"
#include "kernel.h"
#include "global.h"
#include "editor.h"
#include "sound.h"
#include "user.h"

#include "midas.h"
#include "mconfig.h"

char *Cons=NULL, *ConsInput=NULL, *InputBuffer=NULL;
char *ConsBuffer=NULL, *Command=NULL;
int  Cursor, quit, fheight, InsertMode;
int  History, HistoryStart;
int  DispStart, DispFirst, DispHeight;
int  CursorColor, TextColor;
int  ConW=128, ConH=256;
int  worldloaded=0;

BITMAP Background, linebg;

void ConsoleBG(char *f)
{
  if (Background.Data) delete[] Background.Data;
  if (f)
  {
    if (stricmp(f, "internal")==0)
    {
      memcpy(&Background.x, Z2BG, sizeof(short));
      memcpy(&Background.y, Z2BG + 2, sizeof(short));
      Background.Data = Z2BG + 4;
    } else
    {
      if (g_LoadBitmap(f, &Background)<0)
        Background.Data = NULL;
    }
  } else Background.Data = NULL;

  // setup background
  if (Background.Data)
  {
    linebg.x = Background.x;
    linebg.y = FontMan.FontHeight(g_GetFont())+1;
    linebg.Data = Background.Data + (linebg.x * (Background.y-linebg.y));
  } else
  {
    linebg.x = 320;
    linebg.y = FontMan.FontHeight(g_GetFont())+1;
    linebg.Data = new char[linebg.x * linebg.y];
    if (!linebg.Data)
    {
      g_Done();
      printf("Z2 Console: Error allocating console updater.\n");
      exit(1);
    }
    memset(linebg.Data, 0, linebg.x*linebg.y);
  }
}

void AddConsLine(void)
{
  // bump console buffer up
  if (DispStart) DispStart--;
  memmove(Cons, Cons + ConW, (ConH-1)*ConW);
  memset(Cons + ((ConH-1)*ConW), 0, ConW);
}

void AddConsLine(char *fmt, ...)
{
  va_list arglist;

  // bump console buffer up
  if (DispStart) DispStart--;
  memmove(Cons, Cons + ConW, (ConH-1)*ConW);

  // handle extra parameters
  va_start(arglist, fmt);
  vsprintf(ConsBuffer, fmt, arglist);
  va_end(arglist);

  // copy line to console buffer
  strncpy(Cons + ((ConH-1)*ConW), ConsBuffer, ConW);
}

void AddConsLineM(char *fmt, ...)
{
  va_list arglist;

  // bump console buffer up
  if (DispStart) DispStart--;
  memmove(Cons, Cons + ConW, (ConH-1)*ConW);

  // handle extra parameters
  va_start(arglist, fmt);
  ConsBuffer[0] = 255;
  vsprintf(ConsBuffer+1, fmt, arglist);
  va_end(arglist);

  // copy line to console buffer
  strncpy(Cons + ((ConH-1)*ConW), ConsBuffer, ConW);
}

void ReAddConsLine(char *fmt, ...)
{
  va_list arglist;

  // handle extra parameters
  va_start(arglist, fmt);
  vsprintf(ConsBuffer, fmt, arglist);
  va_end(arglist);

  // copy line to console buffer
  strncpy(Cons + ((ConH-1)*ConW), ConsBuffer, ConW);
}

void DrawBackground(void)
{
  if (Background.Data)
    g_Bitmap(0, 0, &Background); else
    g_SuperBlack();
}

void DrawCons(void)
{
  DrawBackground();

  // Draw what's on screen
  g_Font(1);
  g_FontColor(TextColor);
  g_gotoxy(0, (g_ResY()-fheight)-(DispHeight*fheight));

  for (int i=DispFirst; i<DispFirst+DispHeight; i++)
  {
    if (i>=0)
    {
      if (Cons[i*ConW]==255)
        g_msprintf(Cons + (i*ConW) + 1); else
        g_sprintf(Cons + (i*ConW));
    }
    g_gotoxy(0, g_FontY()+linebg.y);
  }
}

void GetConsInput(void)
{
  int enter=0, i;
  char c;

  int SourceX = g_FontX();
  int SourceY = g_FontY();

  do {
    // Draw user's current input
    int offset = FontMan.CharWidth(g_GetFont(), ']')+1;
    for (i=0; i<Cursor; i++)
      offset += FontMan.CharWidth(g_GetFont(), ConsInput[i])+1;

    g_Bitmap(0, g_ResY()-linebg.y, &linebg);
    g_gotoxy(SourceX, SourceY);

    int x1 = offset;
    int x2 = offset+FontMan.CharWidth(g_GetFont(), ConsInput[Cursor]);
    int y1 = g_FontY();
    int y2 = g_FontY() + fheight - 1;

    if (InsertMode) y1 += (fheight/2); // cut cursor in half

    g_sprintf("]");
    g_sprintf(ConsInput);

    if (Cursor == strlen(ConsInput))
      g_ciRect(x1, y1, x1+4, y2, WinDark); else
      g_ciRect(x1, y1, x2, y2, WinDark);

    g_HSync();
    g_Update(ConsPage);
//    zm_Update();

    c = mk_PollChar();

    switch (c)
    {
      case mcNone : case mcF2 : case mcF4 : case mcF5 :
      case mcF6 : case mcF8 : case mcF9 : case mcF10 :
      case mcF11 : case mcF12 : break;

      case mcIns :
        InsertMode = !InsertMode;      // toggle insert mode
        break;
      case mcDel :
        if (Cursor<strlen(ConsInput))
          memmove(ConsInput + Cursor, ConsInput + Cursor + 1, strlen(ConsInput)-Cursor);
        break;
      case mcBackspace :
        if (Cursor>0)
        {
          if (Cursor<strlen(ConsInput))
          {
            Cursor--;
            memmove(ConsInput + Cursor, ConsInput + Cursor + 1, strlen(ConsInput)-Cursor);
          } else ConsInput[--Cursor] = 0;
        }
        break;
      case mcEnter :
        enter = 1;
        break;
      case mcEsc :
        memset(ConsInput, 0, ConW);
        Cursor = 0;
        break;
      case mcLeft :
        if (Cursor>0) Cursor--;
        break;
      case mcRight :
        if (Cursor<strlen(ConsInput)) Cursor++;
        break;
      case mcHome :
        Cursor = 0;
        break;
      case mcEnd :
        Cursor = strlen(ConsInput);
        break;
      case mcPgUp :
        DispFirst--;
        if (DispFirst<=DispStart-DispHeight)
          DispFirst = DispStart-DispHeight+1;

        UpdateConsole();
        break;
      case mcPgDn :
        DispFirst++;

        if (DispFirst>ConH-DispHeight)
          DispFirst = ConH-DispHeight;

        UpdateConsole();
        break;
      case mcUp :
        if (History>HistoryStart)
        {
          History--;
          memcpy(ConsInput, InputBuffer + (History * ConW), ConW);
          Cursor = strlen(ConsInput);
        }
        break;
      case mcDown :
        if (History<ConH-1)
        {
          History++;
          memcpy(ConsInput, InputBuffer + (History * ConW), ConW);
          Cursor = strlen(ConsInput);
        } else ClearConsInput();
        break;
      case mcF1 : // show context-sensitive help
        SendHelpMessage("ConsoleHelp");
        UpdateObj(Z2Help);
        break;
      case mcF3 : // show last command
        memcpy(ConsInput, InputBuffer + ((ConH-1) * ConW), ConW);
        Cursor = strlen(ConsInput);
        break;
      case mcF7 : // show command History
        AddConsLine("Command History:");
        AddConsLine("==================");
        for (i=HistoryStart; i<ConH; i++)
          AddConsLine("%d:%s", (i-HistoryStart)+1, InputBuffer + (i * ConW));
        UpdateConsole();
        break;
      default :
        if (InsertMode)
          memmove(ConsInput + Cursor + 1, ConsInput + Cursor, strlen(ConsInput)-Cursor+1);
        ConsInput[Cursor++] = c;

        break;
    }
  } while (!enter);
  mk_Clear();
}

int c_menu(void)
{
  if (stricmp(Command, "menu")) return 0;

  AddConsLine("Going to menu...");
  consolemode = 0;

  return 1;
}

int c_quit(void)
{
  if (stricmp(Command, "quit")) return 0;

  AddConsLine("Exiting Z2...");
  quit = 1;

  return 1;
}

int c_objdump(void)
{
  if (stricmp(Command, "objdump")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  if (!parm1 || !parm2)
  {
    AddConsLine("objdump <room> <object>");
    AddConsLine();
    AddConsLine("  <room>FRoom where object is located");
    AddConsLine("  <object>FObject to dump code from");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm1);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }
    pObject o = FindObj(parm2, r->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }

    AddConsLine("Object Code Dump");
    AddConsLine("==================");
    AddConsLine("Name:(%s", o->name);
    AddConsLine("Size:(%d bytes", o->codesize);
    AddConsLine();

    int max = 320/(FontMan.FontWidth(g_GetFont())+1);
    int size;
    char str[80];

    for (int i=0; i<o->codesize; i += max)
    {
      size = o->codesize-i;
      if (size>max) size = max;
      memcpy(str, o->code+i, size);
      str[max+1] = 0;
      for (int j=0; j<size; j++)
        if (str[j]<32) str[j] = '_';
      AddConsLineM(str);
    }
  }
  return 1;
}

int c_sysobjcopy(void)
{
  if (stricmp(Command, "sysobjcopy")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("sysobjcopy <object>");
    AddConsLine();
    AddConsLine("  <object>FSystem object to copy to current room");
  } else
  {
    pObject o = FindObj(parm, SystemRoom->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }
    CopyObj(o, RoomCur);
  }
  return 1;
}

int c_sysobjdump(void)
{
  if (stricmp(Command, "sysobjdump")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("sysobjdump <object>");
    AddConsLine();
    AddConsLine("  <object>FSystem object to dump code from");
  } else
  {
    pObject o = FindObj(parm, SystemRoom->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }

    AddConsLine("System Object Code Dump");
    AddConsLine("=========================");
    AddConsLine("Name:(%s", o->name);
    AddConsLine("Size:(%d bytes", o->codesize);
    AddConsLine();

    int max = 320/(FontMan.FontWidth(g_GetFont())+1);
    int size;
    char str[80];

    for (int i=0; i<o->codesize; i += max)
    {
      size = o->codesize-i;
      if (size>max) size = max;
      memcpy(str, o->code+i, size);
      str[max+1] = 0;
      for (int j=0; j<size; j++)
        if (str[j]<32) str[j] = '_';
      AddConsLineM(str);
    }
  }
  return 1;
}

int c_globaldump(void)
{
  if (stricmp(Command, "globaldump")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("globaldump <object>");
    AddConsLine();
    AddConsLine("  <object>FGlobal object to dump code from");
  } else
  {
    pObject o = FindObj(parm, GlobalRoom->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }

    AddConsLine("Global Object Code Dump");
    AddConsLine("=========================");
    AddConsLine("Name:(%s", o->name);
    AddConsLine("Size:(%d bytes", o->codesize);
    AddConsLine();

    int max = 320/(FontMan.FontWidth(g_GetFont())+1);
    int size;
    char str[80];

    for (int i=0; i<o->codesize; i += max)
    {
      size = o->codesize-i;
      if (size>max) size = max;
      memcpy(str, o->code+i, size);
      str[max+1] = 0;
      for (int j=0; j<size; j++)
        if (str[j]<32) str[j] = '_';
      AddConsLineM(str);
    }
  }
  return 1;
}

int c_mem(void)
{
  if (stricmp(Command, "mem")) return 0;

  // memory estimations, all overestimates to deter author from creating
  // super large, memory exploding worlds. :>

  int go, wr, so, vr, sv;         // World data
  int om=64000+(64*1024);         // estimate misc. to be 64k + background
  int vp=64000*4;                 // virtual pages (ouch!) minimize to 2?
  int ss=zm_modSize();            // current module size
  int sm=32*1024;                 // sound engine data

  AddConsLine("Free memory at start:%d", FreeMem);
  AddConsLine();
  AddConsLine("Graphics system:%d", vp);
  AddConsLine("Sound system:%d", ss+sm);
  AddConsLine("Sound effects:%d", sv=zm_sizeofWavList());
  AddConsLine("World Rooms:%d", wr=sizeofRoomList());
  AddConsLine("Global objects:%d", go=sizeofObjList(GlobalRoom->ObjHead));
  AddConsLine("System objects:%d", so=sizeofObjList(SystemRoom->ObjHead));
  AddConsLine("Variables:%d", vr=sizeofVarList());
  AddConsLine("Misc. (estimate):%d", om);
  AddConsLine();

  int totalmem = vp+ss+sm+go+wr+so+vr+om+sv;
  unsigned freememnow = FreeMem - totalmem;

  AddConsLine("Total memory used:%d", totalmem);
  AddConsLine("Free memory now:%d", freememnow);
  AddConsLine("Percent free:%d percent", (freememnow*100u)/(unsigned)FreeMem);
  AddConsLine();
  AddConsLine("World size does not include music and sound files.");
  AddConsLine("Compression typically yields 40-60 percent smaller files.");
  AddConsLine("Uncompressed world size:%d", go+wr+vr);

  return 1;
}

int c_ver(void)
{
  if (stricmp(Command, "ver")) return 0;

  // engine versions
  AddConsLine("Z2 engine:%s", Z2VERSION);
  AddConsLine("Graphics system:3.2");
  AddConsLine("Sound system:0.5.0");

  return 1;
}

int c_edit(void)
{
  if (stricmp(Command, "edit")) return 0;

#ifdef NOEDIT
  AddConsLine("Editor disabled in demo.");
#else
  char *parm = strtok(NULL, " ");

  AddConsLine("Entering Z2 Editor...");

  if (parm)
  {
    pRoom r = FindRoom(parm);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }

    AddConsLine("Editing room '%s'...", parm);
    zzt_Fade(zztPal, Black);

    Edit(r);
  } else
  {
    zzt_Fade(zztPal, Black);
    Edit();
  }
  UpdateConsole();
  zzt_Fade(Black, zztPal);
#endif

  return 1;
}

int c_play(void)
{
  if (stricmp(Command, "play")) return 0;

  char *parm = strtok(NULL, " ");

  AddConsLine("Entering Z2 Kernel...");

  if (parm)
  {
    pRoom r = FindRoom(parm);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }

    AddConsLine("Playing room '%s'...", parm);
    zzt_Fade(zztPal, Black);

    SaveWorld("$Z2$TEMP.$$$", "Temporary Z2 World");
    Play(r);
    LoadWorld("$Z2$TEMP.$$$");
    remove("$Z2$TEMP.$$$");
  } else
  {
    zzt_Fade(zztPal, Black);

    SaveWorld("$Z2$TEMP.$$$", "Temporary Z2 World");
    Play();
    LoadWorld("$Z2$TEMP.$$$");
    remove("$Z2$TEMP.$$$");
  }
  UpdateConsole();
  zzt_Fade(Black, zztPal);

  return 1;
}

int c_fade(void)
{
  if (stricmp(Command, "fade")) return 0;

  char *temp = strtok(NULL, " ");
  if (!temp)
  {
    AddConsLine("fade <0/1>");
    AddConsLine();
    if (Fade)
      AddConsLine("Screen fades are on."); else
      AddConsLine("Screen fades are off.");

    return 1;
  }

  Fade = strtol(temp, NULL, 10);

  if (Fade)
    AddConsLine("Screen fades are now on."); else
    AddConsLine("Screen fades are now off.");

  return 1;
}

int c_intro(void)
{
  if (stricmp(Command, "intro")) return 0;

//  AddConsLine("sysobjmsg Introduction WakeUp");
//  UpdateConsole();
  int result = SendHelpMessage("WakeUp");
  switch (result)
  {
    case SUCCESS :
      AddConsLine("Object found, and received message.");
      UpdateObj(Z2Help);
      break;
    case BAD_OBJ :
      AddConsLine("Object not found.");
      break;
    case BAD_MSG :
      AddConsLine("Object found, but won't take message.");
      break;
    default :
      AddConsLine("An unknown error occurred!");
      break;
  }

  return 1;
}

int c_cls(void)
{
  if (stricmp(Command, "cls")) return 0;

  memset(Cons, 0, ConW*ConH);
  memset(ConsBuffer, 0, ConW*4);
  memset(ConsInput, 0, ConW);
  Cursor = 0;

  UpdateConsole();

  return 1;
}

int c_sysobjmsg(void)
{
  if (stricmp(Command, "sysobjmsg")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  if (!parm1 || !parm2)
  {
    AddConsLine("sysobjmsg <object> <message>");
    AddConsLine();
    AddConsLine("  <object>FSystem object to send message to");
    AddConsLine("  <message>FMessage to send");
  } else
  {
    int result;
    UpdateConsole();
    result = SendSysMessage(parm1, parm2);
    switch (result)
    {
      case SUCCESS :
        AddConsLine("Object found, and received message.");
        UpdateObj(LastObj);
        break;
      case BAD_OBJ :
        AddConsLine("Object not found.");
        break;
      case BAD_MSG :
        AddConsLine("Object found, but won't take message.");
        break;
      default :    // should never happen :>
        AddConsLine("An unknown error occurred!");
        break;
    }
  }

  return 1;
}

int c_globalmsg(void)
{
  if (stricmp(Command, "globalmsg")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  if (!parm1 || !parm2)
  {
    AddConsLine("globalmsg <object> <message>");
    AddConsLine();
    AddConsLine("  <object>FGlobal object to send message to");
    AddConsLine("  <message>FMessage to send");
  } else
  {
    int result;
    UpdateConsole();
    result = SendMessage("global", parm1, parm2);
    switch (result)
    {
      case SUCCESS :
        AddConsLine("Object found, and received message.");
        UpdateObj(LastObj);
        break;
      case BAD_OBJ :
        AddConsLine("Object not found.");
        break;
      case BAD_MSG :
        AddConsLine("Object found, but won't take message.");
        break;
      default :    // should never happen :>
        AddConsLine("An unknown error occurred!");
        break;
    }
  }

  return 1;
}

int c_objmsg(void)
{
  if (stricmp(Command, "objmsg")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  char *parm3 = strtok(NULL, " ");
  if (!parm1 || !parm2 || !parm3)
  {
    AddConsLine("objmsg <room> <object> <message>");
    AddConsLine();
    AddConsLine("  <room>FRoom where object is located");
    AddConsLine("  <object>FObject to send message to");
    AddConsLine("  <message>FMessage to send");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    UpdateConsole();

    int result = SendMessage(parm1, parm2, parm3);
    switch (result)
    {
      case SUCCESS :
        AddConsLine("Object found, and received message.");
        UpdateObj(LastObj);
        break;
      case BAD_OBJ :
        AddConsLine("Object not found.");
        break;
      case BAD_ROOM :
        AddConsLine("Room not found.");
        break;
      case BAD_MSG :
        AddConsLine("Object found, but won't take message.");
        break;
      default :    // should never happen :>
        AddConsLine("An unknown error occurred!");
        break;
    }
  }

  return 1;
}

int c_comref(void)
{
  if (stricmp(Command, "comref")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("comref <topic>");
    AddConsLine();
    AddConsLine("  <topic>FGet command reference on these topics:");
    AddConsLine("FgeneralsGeneral commands");
    AddConsLine("FsoundsSound system commands");
    AddConsLine("FworldsWorld management commands");
  } else
  {
    AddConsLine("Z2 Console - Command Reference");

    if (strcmp(parm, "general")==0)
    {
      AddConsLine("General commands");
      AddConsLine("======================================");
      AddConsLine("cls<Clears console");
      AddConsLine("comref<Displays Z2 Console Command Reference");
      AddConsLine("fade<Turn screen fading on or off");
      AddConsLine("help<Displays Z2 Console Help");
      AddConsLine("intro<Displays 'Welcome to Z2' scroll");
      AddConsLine("load<Loads a world");
      AddConsLine("loadplay<Loads and plays world");
      AddConsLine("mem<Displays a memory report");
      AddConsLine("menu<Use the menu interface");
      AddConsLine("quit<Quits Z2");
      AddConsLine("ver<Displays version report");
      AddConsLine("zc<Compiles external script files");
    } else
    if (strcmp(parm, "sound")==0)
    {
      AddConsLine("Sound system commands");
      AddConsLine("======================================");
      AddConsLine("modinfo<Displays current module info");
      AddConsLine("modkill<Stops current module and deallocates it.");
      AddConsLine("modload<Kills current module and loads a new module");
      AddConsLine("modloop<Turn module looping on or off");
      AddConsLine("modplay<Begins or continues playing a module");
      AddConsLine("modstop<Pauses module");
      AddConsLine("soundinfo<Displays sound card & sound system info");
      AddConsLine("wavinfo<Displays wave sound info");
      AddConsLine("wavlist<Displays wave sound list");
      AddConsLine("wavload<Loads a wave file from disk");
      AddConsLine("wavplay<Plays a wave file");
    } else
    if (strcmp(parm, "world")==0)
    {
      AddConsLine("World management commands");
      AddConsLine("======================================");
      AddConsLine("edit<Launches Z2 Editor");
      AddConsLine("globaldump<Dumps global object's compiled code");
      AddConsLine("globalinfo<Displays specific global object info");
      AddConsLine("globallist<Displays global objects list");
      AddConsLine("globalmsg<Sends a global object a message");
      AddConsLine("iteminfo<Displays itemdef info");
      AddConsLine("itemlist<Displays the itemdef list");
      AddConsLine("objdump<Dumps room object's compiled code");
      AddConsLine("objinfo<Displays specific room object info");
      AddConsLine("objkill<Deallocates a specific room object");
      AddConsLine("objkillxy<Deallocates a specific room object at x,y");
      AddConsLine("objlist<Displays a room's objects list");
      AddConsLine("objload<Loads an object file from disk");
      AddConsLine("objmsg<Sends a room object a message");
      AddConsLine("play<Plays, er... test Z2 Kernel");
      AddConsLine("roomadd<Adds a room to the world");
      AddConsLine("roominfo<Displays specific room info");
      AddConsLine("roomkill<Deallocates a room (and all its objects)");
      AddConsLine("roomlist<Displays rooms list");
      AddConsLine("roomload<Loads a room from disk");
      AddConsLine("roomsave<Saves a room to disk");
      AddConsLine("sysobjcomp<Copies a system object to current room");
      AddConsLine("sysobjdump<Dumps system object's compiled code");
      AddConsLine("sysobjinfo<Displays specific system object info");
      AddConsLine("sysobjlist<Displays system objects list");
      AddConsLine("sysobjmsg<Sends a system object a message");
      AddConsLine("varkill<Deallocates a variable");
      AddConsLine("varlist<Displays variables list");
    } else AddConsLine("Unknown topic.");
  }
  return 1;
}

int c_help(void)
{
  if (stricmp(Command, "help")) return 0;

  AddConsLine("Z2 Console - Help");
  AddConsLine("======================");
  AddConsLine("Left,Right<Moves command line cursor");
  AddConsLine("Home,End<Goes to start/end of command line");
  AddConsLine("PgUp,PgDn<Scrolls console history");
  AddConsLine("Up,Down<Cycles through command history");
  AddConsLine("Backspace<Deletes char before cursor");
  AddConsLine("Enter<Accepts and processes command line");
  AddConsLine("Escape<Clears command line");
  AddConsLine("F3<Recalls last command");
  AddConsLine("F7<Displays command history");
  AddConsLine("comref<Displays command reference");
  AddConsLine("menu<Go to Z2 main menu");
  AddConsLine("quit<Quits Z2.");

  return 1;
}

int c_sysobjinfo(void)
{
  if (stricmp(Command, "sysobjinfo")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("sysobjinfo <object>");
    AddConsLine();
    AddConsLine("  <object>FSystem object to obtain info from");
  } else
  {
    pObject o = FindObj(parm, SystemRoom->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }

    AddConsLine("System Object Info");
    AddConsLine("===================");
    AddConsLine("Name:<%s", o->name);
    AddConsLine("Hash:<%d", o->NameHash);
    AddConsLine("Size:<%d bytes", sizeofObj(o));
    AddConsLine("Coords:<%d,%d", o->x, o->y);
    AddConsLine("Layer:<%d", o->z);
    AddConsLine("Properties:");
    AddConsLine(" Flight:<%d", o->objFlight);
    AddConsLine(" Item:<%d", o->objItem);
    AddConsLine(" Invisible:<%d", o->objInvisible);
    AddConsLine(" Push Player:<%d", o->objPushP);
    AddConsLine(" Push Horiz.:<%d", o->objPushNS);
    AddConsLine(" Push Vert.:<%d", o->objPushEW);
    AddConsLine(" Pusher:<%d", o->objPusher);
    AddConsLine(" Swimmer:<%d", o->objSwimmer);
    AddConsLine(" Breakable:<%d", o->objBreakable);
    AddConsLine(" Solid:<%d", o->objSolid);
    AddConsLine(" Player:<%d", o->objPlayer);
    AddConsLine(" Bound:<%d", o->objBound);
    AddConsLine(" Locked:<%d", o->objLocked);
    AddConsLine("Frame:<%d", o->frame);
    AddConsLine("CodePtr:<0x%X", o->code);
    AddConsLine("CodeSize:<%d bytes", o->codesize);
    AddConsLine("CodePos:<%d", o->codep);
    AddConsLine("Cycle:<%d/%d", o->cyclel, o->cycle);
    AddConsLine("Light?:<%d", o->light);
    AddConsLine("Walking:<%d", o->queue);
    if (o->next)
      AddConsLine("NextObj:<0x%X", o->next); else
      AddConsLine("NextObj:<NULL");
  }
  return 1;
}

int c_globalinfo(void)
{
  if (stricmp(Command, "globalinfo")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("globalinfo <object>");
    AddConsLine();
    AddConsLine("  <object>FGlobal object to obtain info from");
  } else
  {
    pObject o = FindObj(parm, GlobalRoom->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }

    AddConsLine("Global Object Info");
    AddConsLine("===================");
    AddConsLine("Name:<%s", o->name);
    AddConsLine("Hash:<%d", o->NameHash);
    AddConsLine("Size:<%d bytes", sizeofObj(o));
    AddConsLine("Coords:<%d,%d", o->x, o->y);
    AddConsLine("Layer:<%d", o->z);
    AddConsLine("Properties:");
    AddConsLine(" Flight:<%d", o->objFlight);
    AddConsLine(" Item:<%d", o->objItem);
    AddConsLine(" Invisible:<%d", o->objInvisible);
    AddConsLine(" Push Player:<%d", o->objPushP);
    AddConsLine(" Push Horiz.:<%d", o->objPushNS);
    AddConsLine(" Push Vert.:<%d", o->objPushEW);
    AddConsLine(" Pusher:<%d", o->objPusher);
    AddConsLine(" Swimmer:<%d", o->objSwimmer);
    AddConsLine(" Breakable:<%d", o->objBreakable);
    AddConsLine(" Solid:<%d", o->objSolid);
    AddConsLine(" Player:<%d", o->objPlayer);
    AddConsLine(" Bound:<%d", o->objBound);
    AddConsLine(" Locked:<%d", o->objLocked);
    AddConsLine("Frame:<%d", o->frame);
    AddConsLine("CodePtr:<0x%X", o->code);
    AddConsLine("CodeSize:<%d bytes", o->codesize);
    AddConsLine("CodePos:<%d", o->codep);
    AddConsLine("Cycle:<%d/%d", o->cyclel, o->cycle);
    AddConsLine("Light?:<%d", o->light);
    AddConsLine("Walking:<%d", o->queue);
    if (o->next)
      AddConsLine("NextObj:<0x%X", o->next); else
      AddConsLine("NextObj:<NULL");
  }
  return 1;
}

int c_objinfo(void)
{
  if (stricmp(Command, "objinfo")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  if (!parm1 || !parm2)
  {
    AddConsLine("objinfo <room> <object>");
    AddConsLine();
    AddConsLine("  <room>FRoom where object is located");
    AddConsLine("  <object>FObject to obtain info from");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm1);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }
    pObject o = FindObj(parm2, r->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }

    AddConsLine("Room Object Info");
    AddConsLine("=================");
    AddConsLine("Name:<%s", o->name);
    AddConsLine("Hash:<%d", o->NameHash);
    AddConsLine("Size:<%d bytes", sizeofObj(o));
    AddConsLine("Coords:<%d,%d", o->x, o->y);
    AddConsLine("Layer:<%d", o->z);
    AddConsLine("Properties:");
    AddConsLine(" Flight:<%d", o->objFlight);
    AddConsLine(" Item:<%d", o->objItem);
    AddConsLine(" Invisible:<%d", o->objInvisible);
    AddConsLine(" Push Player:<%d", o->objPushP);
    AddConsLine(" Push Horiz.:<%d", o->objPushNS);
    AddConsLine(" Push Vert.:<%d", o->objPushEW);
    AddConsLine(" Pusher:<%d", o->objPusher);
    AddConsLine(" Swimmer:<%d", o->objSwimmer);
    AddConsLine(" Breakable:<%d", o->objBreakable);
    AddConsLine(" Solid:<%d", o->objSolid);
    AddConsLine(" Player:<%d", o->objPlayer);
    AddConsLine(" Bound:<%d", o->objBound);
    AddConsLine(" Locked:<%d", o->objLocked);
    AddConsLine("Frame:<%d", o->frame);
    AddConsLine("CodePtr:<0x%X", o->code);
    AddConsLine("CodeSize:<%d bytes", o->codesize);
    AddConsLine("CodePos:<%d", o->codep);
    AddConsLine("Cycle:<%d/%d", o->cyclel, o->cycle);
    AddConsLine("Light?:<%d", o->light);
    AddConsLine("Walking:<%d", o->queue);
    if (o->next)
      AddConsLine("NextObj:<0x%X", o->next); else
      AddConsLine("NextObj:<NULL");
  }
  return 1;
}

int c_objkill(void)
{
  if (stricmp(Command, "objkill")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  if (!parm1 || !parm2)
  {
    AddConsLine("objkill <room> <object>");
    AddConsLine();
    AddConsLine("  <room>FRoom where object is located");
    AddConsLine("  <object>FObject to deallocate");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm1);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }
    pObject o = FindObj(parm2, r->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }

    if (KillObj(r, o)==SUCCESS)
      AddConsLine("Object deallocated."); else
      AddConsLine("Error deallocating object.");
  }
  return 1;
}

int c_objkillxy(void)
{
  if (stricmp(Command, "objkillxy")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  char *parm3 = strtok(NULL, " ");
  char *parm4 = strtok(NULL, " ");
  if (!parm1 || !parm2 || !parm3 || !parm4)
  {
    AddConsLine("objkillxy <room> <object> <x> <y>");
    AddConsLine();
    AddConsLine("  <room>FRoom where object is located");
    AddConsLine("  <object>FObject to deallocate");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm1);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }
    pObject o = FindObj(parm2, r->ObjHead);
    if (!o)
    {
      AddConsLine("Object not found.");
      return 1;
    }
    int x = strtol(parm3, NULL, 10);
    int y = strtol(parm4, NULL, 10);

    if (KillObjXY(x, y, r)==SUCCESS)
      AddConsLine("Object deallocated."); else
      AddConsLine("Error deallocating object.");
  }
  return 1;
}

int c_load(void)
{
  if (stricmp(Command, "load")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("load <file>");
    AddConsLine();
    AddConsLine("  <file>FWorld to load");
  } else
  {
    if (LoadWorld(parm) != SUCCESS)
    {
      AddConsLine("World file not found.");
    } else
    {
      SwitchToRoom(RoomHead);
      AddConsLine("World loaded.");
    }
  }
  return 1;
}

int c_loadplay(void)
{
  if (stricmp(Command, "loadplay")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("loadplay <file>");
    AddConsLine();
    AddConsLine("  <file>FWorld to load and play");
  } else
  {
    if (LoadWorld(parm) != SUCCESS)
    {
      AddConsLine("World file not found.");
    } else
    {
      SwitchToRoom(RoomHead);
      AddConsLine("World loaded.");
      zzt_Fade(zztPal, Black);

      SaveWorld("$Z2$TEMP.$$$", "Temporary Z2 World");
      Play();
      LoadWorld("$Z2$TEMP.$$$");
      remove("$Z2$TEMP.$$$");

      UpdateConsole();
      zzt_Fade(Black, zztPal);
    }
  }
  return 1;
}

int c_roominfo(void)
{
  if (stricmp(Command, "roominfo")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("roominfo <room>");
    AddConsLine();
    AddConsLine("  <room>FRoom to obtain info from");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }

    AddConsLine("Room Info");
    AddConsLine("==========");
    AddConsLine("Name:A%s", r->name);
    AddConsLine("Hash:A%d", r->NameHash);
    AddConsLine("Description:A%s", r->Desc);
    AddConsLine("Music file:A%s", r->Mod);
    AddConsLine("Size:A%d bytes", sizeofRoom(r));
    AddConsLine("Dimensions:A%d x %d", r->RoomW, r->RoomH);
    if (r->RoomFloor)
      AddConsLine("RoomFloor:A0x%X", r->RoomFloor); else
      AddConsLine("RoomFloor:ANULL");
    if (r->RoomObject)
      AddConsLine("RoomObject:A0x%X", r->RoomObject); else
      AddConsLine("RoomObject:ANULL");
    if (r->RoomCeiling)
      AddConsLine("RoomCeiling:A0x%X", r->RoomCeiling); else
      AddConsLine("RoomCeiling:ANULL");
    if (r->LinkN)
      AddConsLine("LinkN:A0x%X [%s]", r->LinkN, r->StrLinkN); else
      AddConsLine("LinkN:ANULL [None]");
    if (r->LinkS)
      AddConsLine("LinkS:A0x%X [%s]", r->LinkS, r->StrLinkS); else
      AddConsLine("LinkS:ANULL [None]");
    if (r->LinkE)
      AddConsLine("LinkE:A0x%X [%s]", r->LinkE, r->StrLinkE); else
      AddConsLine("LinkE:ANULL [None]");
    if (r->LinkW)
      AddConsLine("LinkW:A0x%X [%s]", r->LinkW, r->StrLinkW); else
      AddConsLine("LinkW:ANULL [None]");
    AddConsLine("Light?A%s", r->Light? "Yes" : "No");
    AddConsLine("Re-enter?A%s", r->ReEntry? "Yes" : "No");
    if (r->Time)
      AddConsLine("Time Limit:A%d sec.", r->Time); else
      AddConsLine("Time Limit:ANone");
    if (r->MaxShots)
      AddConsLine("Max Shots:A%d", r->MaxShots); else
      AddConsLine("Max Shots:AUnlimited");
    if (r->ObjHead)
      AddConsLine("ObjHead:A0x%X", r->ObjHead); else
      AddConsLine("ObjHead:ANULL");
    if (r->next)
      AddConsLine("NextRoom:A0x%X", r->next); else
      AddConsLine("NextRoom:ANULL");
  }
  return 1;
}

int c_roomadd(void)
{
  if (stricmp(Command, "roomadd")) return 0;

  char *parm1 = strtok(NULL, "\"");
  if (!parm1)
  {
    AddConsLine("roomadd <room> \"<desc>\"");
    AddConsLine();
    AddConsLine("  <room>FName of room to add");
    AddConsLine("  <desc>FRoom description");
    return 1;
  }
  char *parm2 = strtok(NULL, "\"");
  if (!parm2)
    strcpy(parm2, parm1); else
    parm1[strlen(parm1)-1] = 0;

  pRoom r = FindRoom(parm1);
  if (r)
  {
    AddConsLine("Duplicate name found, use another name.");
    return 1;
  }
  r = AddRoom(parm1, parm2);
  DefaultRoom(r);
  AddConsLine("Room added successfully.");

  return 1;
}

int c_roomkill(void)
{
  if (stricmp(Command, "roomkill")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("roomkill <room>");
    AddConsLine();
    AddConsLine("  <room>FRoom to deallocate");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm);
    if (!r)
    {
      AddConsLine("Room not found.");
      return 1;
    }

    if (KillRoom(r)==SUCCESS)
      AddConsLine("Room (and descendant objects) deallocated."); else
      AddConsLine("Error deallocating room.");
  }
  return 1;
}

int c_sysobjlist(void)
{
  if (stricmp(Command, "sysobjlist")) return 0;

  AddConsLine("System Object List");
  AddConsLine("Size: %d bytes", sizeofObjList(SystemRoom->ObjHead));
  AddConsLine();
  AddConsLine("[Name of Object]Z[Hash][MemPointer][NextPointer]");

  pObject Cur = SystemRoom->ObjHead;
  if (!Cur)
  {
    AddConsLine("No system objects.");
    return 1;
  }
  while (Cur)
  {
    if (Cur->next)
      AddConsLine("[%s]Z[%d][0x%X][0x%X]",
        Cur->name, Cur->NameHash, Cur, Cur->next); else
      AddConsLine("[%s]Z[%d][0x%X][NULL]",
        Cur->name, Cur->NameHash, Cur);
    Cur = Cur->next;
  }

  return 1;
}

int c_globallist(void)
{
  if (stricmp(Command, "globallist")) return 0;

  AddConsLine("Global Object List");
  AddConsLine("Size: %d bytes", sizeofObjList(GlobalRoom->ObjHead));
  AddConsLine();
  AddConsLine("[Name of Object]Z[Hash][MemPointer][NextPointer]");

  pObject Cur = GlobalRoom->ObjHead;
  if (!Cur)
  {
    AddConsLine("No system objects.");
    return 1;
  }
  while (Cur)
  {
    if (Cur->next)
      AddConsLine("[%s]Z[%d][0x%X][0x%X]",
        Cur->name, Cur->NameHash, Cur, Cur->next); else
      AddConsLine("[%s]Z[%d][0x%X][NULL]",
        Cur->name, Cur->NameHash, Cur);
    Cur = Cur->next;
  }

  return 1;
}

int c_objadd(void)
{
  if (stricmp(Command, "objadd")) return 0;

  char *parm1 = strtok(NULL, " ");
  char *parm2 = strtok(NULL, " ");
  if (!parm1 || !parm2)
  {
    AddConsLine("objadd <room> <script>");
    AddConsLine();
    AddConsLine("  <room>FRoom to add objects to");
    AddConsLine("  <script>FScript file to load (include extension)");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm1);
    if (!r)
    {
      AddConsLine("Room not found");
      return 1;
    }
    AddObj(parm2, r->ObjHead);
  }
  return 1;
}

int c_objlist(void)
{
  if (stricmp(Command, "objlist")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("objlist <room>");
    AddConsLine();
    AddConsLine("  <room>FRoom to list objects from");
    AddConsLine();
    AddConsLine("Use \"--\" to reference current room (%s)", RoomCur->name);
  } else
  {
    pRoom r = FindRoom(parm);
    if (!r)
    {
      AddConsLine("Room not found");
      return 1;
    }
    AddConsLine("Room Object List");
    AddConsLine("Size: %d bytes", sizeofObjList(r->ObjHead));
    AddConsLine();
    AddConsLine("[Name of Object]P[x,y]n[Hash][MemPointer][NextPointer]");

    pObject Cur = r->ObjHead;
    if (!Cur)
    {
      AddConsLine("No objects in this room.");
      return 1;
    }
    while (Cur)
    {
      if (Cur->next)
        AddConsLine("[%s]P[%d,%d]n[%d][0x%X][0x%X]",
          Cur->name, Cur->x, Cur->y, Cur->NameHash, Cur, Cur->next); else
        AddConsLine("[%s]P[%d,%d]n[%d][0x%X][NULL]",
          Cur->name, Cur->x, Cur->y, Cur->NameHash, Cur);
      Cur = Cur->next;
    }
  }
  return 1;
}

int c_roomlist(void)
{
  if (stricmp(Command, "roomlist")) return 0;

  AddConsLine("Room List");
  AddConsLine("Size: %d bytes", sizeofRoomList());
  AddConsLine("Current: %s", RoomCur->name);
  AddConsLine();
  AddConsLine("[Name of Room]Z[Hash][MemPointer][NextPointer]");

  pRoom Cur = RoomHead;
  if (!Cur)
  {
    AddConsLine("No rooms.");
    return 1;
  }
  while (Cur)
  {
    if (Cur->next)
      AddConsLine("[%s]Z[%d][0x%X][0x%X]",
        Cur->name, Cur->NameHash, Cur, Cur->next); else
      AddConsLine("[%s]Z[%d][0x%X][NULL]",
        Cur->name, Cur->NameHash, Cur);
    Cur = Cur->next;
  }

  return 1;
}

int c_varlist(void)
{
  if (stricmp(Command, "varlist")) return 0;

  AddConsLine("Variable List");
  AddConsLine("Size: %d bytes", sizeofVarList());
  AddConsLine();
  AddConsLine("[Name of Var]Z[Hash][Value][MemPointer][NextPointer]");

  pVar Cur = VarHead;
  if (!Cur)
  {
    AddConsLine("No variables.");
    return 1;
  }
  while (Cur)
  {
    if (Cur->next)
      AddConsLine("[%s]Z[%d][%d][0x%X][0x%X]",
        Cur->name, Cur->NameHash, Cur->value, Cur, Cur->next); else
      AddConsLine("[%s]Z[%d][%d][0x%X][NULL]",
        Cur->name, Cur->NameHash, Cur->value, Cur);
    Cur = Cur->next;
  }

  return 1;
}

int c_varkill(void)
{
  if (stricmp(Command, "varkill")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("varkill <var>");
    AddConsLine();
    AddConsLine("  <var>FVariable to deallocate");
  } else
  {
    pVar v = FindVar(parm);
    if (!v)
    {
      AddConsLine("Variable not found.");
      return 1;
    }

    if (KillVar(v)==SUCCESS)
      AddConsLine("Variable deallocated."); else
      AddConsLine("Error deallocating variable.");
  }
  return 1;
}

int c_soundinfo(void)
{
  if (stricmp(Command, "soundinfo")) return 0;

#ifndef NOMIDAS
  AddConsLine("Sound system info");
  AddConsLine("===================");
  AddConsLine("Soundcard name:d%s", zm_cardName());
  AddConsLine("Driver version:d%s", zm_cardDriver());
  AddConsLine("Sampling depth:d%d bit %s", zm_cardBits(),
              (zm_cardStereo()) ? "Stereo" : "Mono");
  AddConsLine("Digital filter:d%s", (zm_cardInterp()) ? "Yes" : "No");
  AddConsLine("Mixing rate:d%d Hz", zm_cardMixFreq());
  AddConsLine("Port:d0x%X", zm_cardPort());
  AddConsLine("IRQ:d%d", zm_cardIRQ());
  AddConsLine("DMA:d%d", zm_cardDMA());
  AddConsLine("Channels:d%d", zm_cardChannels());
#else
  AddConsLine("Sound system disabled.");
#endif

  return 1;
}

int c_modinfo(void)
{
  if (stricmp(Command, "modinfo")) return 0;

  if (zm_modLoaded())
  {
    AddConsLine("Music module info");
    AddConsLine("===================");
    AddConsLine("Song info:d%s", zm_modSongName());
    AddConsLine("Composed with:d%s", zm_modType());
    AddConsLine("Features:");
    AddConsLine(" No PT limits:d%s", zm_modPTLimits());
    AddConsLine(" Vol. slides:d%s", zm_modVolSlides());
    AddConsLine(" Ext. octaves:d%s", zm_modOctaves());
    AddConsLine(" Freq. table:d%s", zm_modFreqTable());
    AddConsLine("Volume:d%d percent", zm_modVolume());
    AddConsLine("Loop song:d%s", zm_modLooping()? "Yes" : "No");
  } else AddConsLine("No module in memory.");

  return 1;
}

int c_modkill(void)
{
  if (stricmp(Command, "modkill")) return 0;

  if (zm_modLoaded())
  {
    zm_modKill();
    AddConsLine("Module terminated.");
  } else AddConsLine("No module in memory.");

  return 1;
}

int c_modload(void)
{
  if (stricmp(Command, "modload")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("modload <module>");
    AddConsLine();
    AddConsLine("  <module>FModule to load (default .z2 extension)");
  } else
  {
    AutoAppend(parm, ".z2");
    switch (zm_modLoad(parm))
    {
      case 0 : // success
        AddConsLine("Module loaded. Type 'modplay' to play.");
        break;
      case -1 : // file not found
        AddConsLine("Module file not found.");
        break;
      case -2 : // invalid file
        AddConsLine("Invalid module file.");
        AddConsLine("Use MOD2GM to convert modules.");
        break;
    }
  }

  return 1;
}

int c_modloop(void)
{
  if (stricmp(Command, "modloop")) return 0;

  char *temp = strtok(NULL, " ");
  if (!temp)
  {
    AddConsLine("modloop <0/1>");
    AddConsLine();
    if (zm_modLooping())
      AddConsLine("Module looping is on."); else
      AddConsLine("Module looping is off.");

    return 1;
  }

  zm_modSetLooping(strtol(temp, NULL, 10));

  if (zm_modLooping())
    AddConsLine("Module looping is now on."); else
    AddConsLine("Module looping is now off.");

  return 1;
}

int c_modplay(void)
{
  if (stricmp(Command, "modplay")) return 0;

  if (zm_modLoaded())
    zm_modStart(); else
    AddConsLine("No module in memory.");

  return 1;
}

int c_modstop(void)
{
  if (stricmp(Command, "modstop")) return 0;

  if (zm_modLoaded())
    zm_modStop(); else
    AddConsLine("No module in memory.");

  return 1;
}

int c_wavinfo(void)
{
  if (stricmp(Command, "wavinfo")) return 0;

  char *temp = strtok(NULL, " ");
  if (!temp)
  {
    AddConsLine("wavinfo <handle>");
    AddConsLine();
    AddConsLine("  <handle>FSound file handle to obtain info from");
  } else
  {
    int parm = strtol(temp, NULL, 10);
    pSound s = zm_wavFind(parm);
    if (!s)
    {
      AddConsLine("Sound not found.");
      return 1;
    }

    AddConsLine("Sound File Info");
    AddConsLine("===================");
    AddConsLine("Name:<%s", s->name);
    AddConsLine("Handle:<%d", s->handle);
    AddConsLine("Size:<%d bytes", zm_sizeofWav(s));
    AddConsLine("Volume:<%d", s->volume);
    AddConsLine("Panning:<%d", s->panning);
    AddConsLine("Frequency:<%d", s->freq);
    AddConsLine("SamplePtr:<0x%X", s->snd);
    if (s->next)
      AddConsLine("NextSound:<0x%X", s->next); else
      AddConsLine("NextSound:<NULL");
  }

  return 1;
}

int c_wavplay(void)
{
  if (stricmp(Command, "wavplay")) return 0;

  char *temp = strtok(NULL, " ");
  if (!temp)
  {
    AddConsLine("wavplay <handle>");
    AddConsLine();
    AddConsLine("  <handle>FSound file handle to play");
  } else
  {
    int parm = strtol(temp, NULL, 10);
    pSound s = zm_wavFind(parm);
    if (!s)
    {
      AddConsLine("Sound not found.");
      return 1;
    }

    int c = zm_wavPlay(parm);
    AddConsLine("Played on channel %d.", c);
  }

  return 1;
}

int c_wavload(void)
{
  if (stricmp(Command, "wavload")) return 0;

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("wavload <filename>");
    AddConsLine();
    AddConsLine("  <filename>FFilename of .WAV to load");
    return 1;
  }

  AutoAppend(parm, ".sam");
  pSound s = zm_wavAdd(parm);
  if (s)
    AddConsLine("Wave file loaded successfully (Handle=%d).", s->handle); else
    AddConsLine("Error loading wave file.");

  return 1;
}

int c_wavlist(void)
{
  if (stricmp(Command, "wavlist")) return 0;

  AddConsLine("Sound List");
  AddConsLine("Size: %d bytes", zm_sizeofWavList());
  AddConsLine();
  AddConsLine("[Sound Handle]P[SamplePointer][MemPointer][NextPointer]");

  pSound Cur = SoundHead;
  if (!Cur)
  {
    AddConsLine("No sounds.");
    return 1;
  }
  while (Cur)
  {
    if (Cur->next)
      AddConsLine("[%d]P[0x%X][0x%X][0x%X]", Cur->handle, Cur->snd, Cur, Cur->next); else
      AddConsLine("[%d]P[0x%X][0x%X][NULL]", Cur->handle, Cur->snd, Cur);
    Cur = Cur->next;
  }

  return 1;
}

int c_itemlist(void)
{
  if (stricmp(Command, "itemlist")) return 0;

  AddConsLine("Item List");
  AddConsLine();
  AddConsLine("[Name]P[ID][Category]");

  for (int i=0; i<ItemNum; i++)
    AddConsLine("[%s]P[%d][%s]",
      ItemList[i].name, ItemList[i].id, CatList[ItemList[i].cat]);

  return 1;
}

int c_zc(void)
{
  if (stricmp(Command, "zc")) return 0;

  char spath[_MAX_PATH];
  char dpath[_MAX_PATH];
  char name[_MAX_FNAME];
  char ext[_MAX_EXT];

  AddConsLine("ZC 0.7 -- Z2 Script Compiler");
  AddConsLine();

  char *parm = strtok(NULL, " ");
  if (!parm)
  {
    AddConsLine("zc <filename>");
    AddConsLine();
    AddConsLine("  <filename>FFilename of script to compile");
    return 1;
  }

  strlwr(parm);
  _splitpath(parm, NULL, NULL, name, ext);
  if (!strlen(ext)) strcpy(ext, ".scr");
  _makepath(spath, NULL, NULL, name, ext);
  _makepath(dpath, NULL, NULL, name, ".oz");

  AddConsLine("Compiling '%s' to '%s'...", spath, dpath);

  if (!Compile(spath, dpath))
    AddConsLine("Error compiling. (Deleted object file)");
  AddConsLine("Compiling complete.");

  return 1;
}

void ConsParse()
{
  // bump display up
  DispFirst = ConH-DispHeight;

  // skip blank inputs (maybe a snide comment? :>)
  if (strcmp(ConsInput, "")==0) return;

  // update history
  if (HistoryStart) HistoryStart--;
  memmove(InputBuffer, InputBuffer + ConW, (ConH-1)*ConW);
  memcpy(InputBuffer + ((ConH-1)*ConW), ConsInput, ConW);

  int cmdp=0;

  // get command (first string terminated by space)
  Command = strtok(ConsInput, " ");

  // go through known commands chain
  if (!cmdp) cmdp = c_cls();
  if (!cmdp) cmdp = c_comref();
  if (!cmdp) cmdp = c_edit();
  if (!cmdp) cmdp = c_fade();
  if (!cmdp) cmdp = c_globaldump();
  if (!cmdp) cmdp = c_globalmsg();
  if (!cmdp) cmdp = c_globalinfo();
  if (!cmdp) cmdp = c_globallist();
  if (!cmdp) cmdp = c_help();
  if (!cmdp) cmdp = c_intro();
  if (!cmdp) cmdp = c_itemlist();
  if (!cmdp) cmdp = c_load();
  if (!cmdp) cmdp = c_loadplay();
  if (!cmdp) cmdp = c_mem();
  if (!cmdp) cmdp = c_menu();
  if (!cmdp) cmdp = c_modinfo();
  if (!cmdp) cmdp = c_modkill();
  if (!cmdp) cmdp = c_modload();
  if (!cmdp) cmdp = c_modloop();
  if (!cmdp) cmdp = c_modplay();
  if (!cmdp) cmdp = c_modstop();
  if (!cmdp) cmdp = c_objadd();
  if (!cmdp) cmdp = c_objdump();
  if (!cmdp) cmdp = c_objmsg();
  if (!cmdp) cmdp = c_objinfo();
  if (!cmdp) cmdp = c_objkill();
  if (!cmdp) cmdp = c_objkillxy();
  if (!cmdp) cmdp = c_objlist();
  if (!cmdp) cmdp = c_play();
  if (!cmdp) cmdp = c_quit();
  if (!cmdp) cmdp = c_roomadd();
  if (!cmdp) cmdp = c_roominfo();
  if (!cmdp) cmdp = c_roomkill();
  if (!cmdp) cmdp = c_roomlist();
  if (!cmdp) cmdp = c_soundinfo();
  if (!cmdp) cmdp = c_sysobjcopy();
  if (!cmdp) cmdp = c_sysobjdump();
  if (!cmdp) cmdp = c_sysobjmsg();
  if (!cmdp) cmdp = c_sysobjinfo();
  if (!cmdp) cmdp = c_sysobjlist();
  if (!cmdp) cmdp = c_varkill();
  if (!cmdp) cmdp = c_varlist();
  if (!cmdp) cmdp = c_ver();
  if (!cmdp) cmdp = c_wavinfo();
  if (!cmdp) cmdp = c_wavlist();
  if (!cmdp) cmdp = c_wavload();
  if (!cmdp) cmdp = c_wavplay();
  if (!cmdp) cmdp = c_zc();
  
  // no recognized commands
  if (!cmdp) AddConsLine("Unknown command \"%s\"", Command);
}

void InitConsole(void)
{
  // first-time initializations
  Cursor = quit = InsertMode = 0;
  History = HistoryStart = DispStart = ConH;

  CursorColor = g_rgb(0, 0, 200);      // a light blue
  TextColor   = 255;                   // a bright white

  fheight = FontMan.FontHeight(g_GetFont())+1;
  DispHeight = g_ResY()/fheight;
  DispFirst = ConH-DispHeight;

  // initialize memory, bomb if not enough
  if (!Cons) Cons = new char[ConW*ConH];
  if (!Cons)
  {
    g_Done();
    printf("Z2 Console: Error allocating console buffer.\n");
    exit(1);
  }
  if (!InputBuffer) InputBuffer = new char[ConW*ConH];
  if (!InputBuffer)
  {
    g_Done();
    printf("Z2 Console: Error allocating input history buffer.\n");
    exit(1);
  }
  if (!ConsInput) ConsInput = new char[ConW];
  if (!ConsInput)
  {
    g_Done();
    printf("Z2 Console: Error allocating input buffer.\n");
    exit(1);
  }
  if (!ConsBuffer) ConsBuffer = new char[ConW*2];
  if (!ConsBuffer)
  {
    g_Done();
    printf("Z2 Console: Error allocating console temp buffer.\n");
    exit(1);
  }

  memset(Cons, 0, ConW*ConH);
  memset(ConsBuffer, 0, ConW*2);
  memset(InputBuffer, 0, ConW*ConH);
  memset(ConsInput, 0, ConW);
}

void DoneConsole(void)
{
  if (Cons) delete[] Cons;
  if (ConsInput) delete[] ConsInput;
  if (InputBuffer) delete[] InputBuffer;
  if (ConsBuffer) delete[] ConsBuffer;
}

void UpdateConsole(void)
{
  g_SetCurPage(ConsPage);
  DrawCons();
  g_HSync();
  g_Update(ConsPage);
}

void SetupWindowConsole(void)
{
  g_SetCurPage(SpriPage);
  DrawBackground();
}

int ProcessOkCanConsole(char *str)
{
  Rect r;
  SetupWindowConsole();

  int len = g_strlen(str)+10;
  if (len<135) len = 135;
  r.Set(len, 40);
  Dialog Win(r);
  Win.SetTitle(str);
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+((Win.r.w-125)/2), Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  Win.Done();
  DrawBackground();
  mk_Clear();

  return (Result==1);
}

void ProcessMessageConsole(char *str)
{
  Rect r;
  SetupWindowConsole();

  int len = g_strlen(str)+10;
  if (len<70) len = 70;
  r.Set(len, 40);
  Dialog Win(r);
  Win.SetTitle(str);
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+((Win.r.w-60)/2), Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  Win.Add(&Ok);

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();
  } while (!Ok.WasPressed() || mk_Ctrl(mkEnter));

  Win.Done();
  DrawBackground();
  mk_Clear();
}

int ProcessMainMenu(void)
{
  Rect r;
  SetupWindowConsole();

  r.Set(160, 120);
  Dialog Win(r);
  Win.SetTitle("Welcome to Z2 v0.61!");
  Win.SetShade(Shadow);
  Win.SetHelp(1);

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Menu(r);
  Menu.SetTitle("");
  Menu.SetFont(1);
  Menu.Add("Play World");
  Menu.Add("Load World");
  Menu.Add("Restore World");
  Menu.Add("Editor");
  Menu.Add("Console");
  Menu.Add("Game Options");

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Quit Z2");

  Win.Add(&Menu);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("wakeup");
      UpdateObj(Z2Help);
    }
    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed()) Result = 2;
  } while (!Result);

  if (Result==1) Result = 10 + Menu.GetItem();

  Win.Done();
  DrawBackground();
  mk_Clear();

  return Result;
}

int ProcessPrefMenu(void)
{
  Rect r;
  SetupWindowConsole();

  r.Set(135, 90);
  Dialog Win(r);
  Win.SetTitle("Game Preferences");
  Win.SetShade(Shadow);
  Win.SetHelp(1);

  r.Set(Win.r.x+5, Win.r.y+24, 125, 12);
  Button Aest(r);
  Aest.SetTitle("Aesthetic Options");

  r.MoveRel(0, 15);
  Button Dir(r);
  Dir.SetTitle("Set Directories");

  r.MoveRel(0, 15);
  Button Sound(r);
  Sound.SetTitle("Sound Card Setup");

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  Win.Add(&Aest);
  Win.Add(&Dir);
  Win.Add(&Sound);
  Win.Add(&Ok);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Aest.WasPressed()) Result = 10;
    if (Dir.WasPressed()) Result = 11;
    if (Sound.WasPressed()) Result = 12;
  } while (!Result);

  Win.Done();
  DrawBackground();
  mk_Clear();

  return Result;
}

int ProcessModSelectConsole(char *title, char *msg, char *none, char *def)
{
  Rect r;
  SetupWindowConsole();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle(title);
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.MoveRel(65, 0);
  Button None(r);
  None.SetTitle("None");

  r.MoveRel(65, 0);
  Button Silence(r);
  Silence.SetTitle("Silence");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Mods(r);
  Mods.SetTitle(msg);
  Mods.SetNoMsg(none);

  // display wait
  char waitstr[] = "Scanning directory for music...";
  Rect r2;

  r2.Set(g_strlen(waitstr)+10, 20);
  Dialog Wait(r2);
  Wait.SetTitle(waitstr);
  Wait.SetHelp(0);

  g_SetCurPage(1);
  g_CopyPage(SpriPage, 1);
  Wait.Draw();
  g_HSync();
  g_Update();

  // scan files
  struct find_t fileinfo;
  ifstream rm;
  unsigned rc;
  pRoom Room;
  char *str;
  char buffer[256];
  char name[256];
  char desc[256];
  char id[20], ch;

  rc = _dos_findfirst("*.xm", _A_NORMAL, &fileinfo);
  while (rc==0)
  {
    if (fileinfo.size==0)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }
    rm.open(fileinfo.name, ios::in | ios::binary);
    rm.read(id, 16);
    if (strncmp(id, "Extended Module:", 16)==0)
    {
      strcpy(name, fileinfo.name);

      rm.seekg(17, ios::beg);
      str = desc;
      while ((ch = rm.get()) != 26)
      {
        if (ch=='\n' || ch==TABSTOP || ch==COLORCH) continue;
        *str++ = ch;
      }
      *str = 0;

      sprintf(buffer, "%sP%s (%s)", fileinfo.name, name, desc);
      Mods.Add(buffer, fileinfo.name);
    }
    rm.close();
    rc = _dos_findnext(&fileinfo);
  }

  Wait.Done();

  Win.Add(&Mods);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&None);
  Win.Add(&Silence);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (None.WasPressed()) Result = 3;
    if (Silence.WasPressed()) Result = 4;
  } while (!Result);

  int ret;

  switch (Result)
  {
    case 1 :
      strcpy(def, Mods.GetInfo(Mods.GetItem()));
      ret = 1;
      break;
    case 2 : ret = -1; break;
    case 3 : ret = -2; break;
    case 4 : ret = -3; break;
  }

  Win.Done();
  DrawBackground();
  mk_Clear();

  return ret;
}

void ProcessAesthetic(void)
{
  Rect r;
  SetupWindowConsole();

  r.Set(160, 100);
  Dialog Win(r);
  Win.SetTitle("Aesthetic Options");
  Win.SetShade(Shadow);
  Win.SetHelp(1);

  r.Move(Win.r.x+5, Win.r.y+25);
  r.SetLR(Win.r.x2-65, Win.r.y+35);
  Input ModName(r);
  ModName.SetTitle("Background music:");
  ModName.SetData(DefaultMOD);
  ModName.SetMax(FNAME_MAX);
  ModName.AllowSpaces(0);

  r.MoveRel(ModName.r.w+5, -2);
  r.Size(55, 12);
  Button Select(r);
  Select.SetTitle("Select");

  r.Move(Win.r.x+5, Win.r.y+47);
  r.SetLR(Win.r.x2-5, Win.r.y+80);
  Check Opt(r);
  Opt.SetFont(2);
  Opt.SetTitle("Options:");
  Opt.Add("Transmission fades between screens", Fade);
  Opt.Add("Do not use smooth-scrolling scrolls", OldScroll);
  Opt.Add("Dialog shadows as backgrounds", Shadow);
  Opt.Add("Go into console at game start", consolemode);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  Win.Add(&ModName);
  Win.Add(&Select);
  Win.Add(&Opt);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (Select.WasPressed())
    {
      char *mod = ModName.GetDataStr();
      int ret = ProcessModSelectConsole("Select Music",
        "Select to music to play on entry", "No music files found!", mod);

      switch (ret)
      {
        case 0 : // picked
          ModName.SetData(mod);
          break;
        case -2 : // none
          ModName.SetData("None");
          break;
        case -3 : // silence
          ModName.SetData("Silence");
          break;
      }

      Select.ResetPress();
    }
  } while (!Result);

  Win.Done();
  DrawBackground();
  mk_Clear();
}

void ProcessDir(void)
{
}

static int detectedSD;

int ProcessSound(void)
{
  Rect r;
  SetupWindowConsole();

  r.Set(240, 140);
  Dialog Win(r);
  Win.SetTitle("Sound Card Setup");
  Win.SetShade(Shadow);
  Win.SetHelp(1);

  char **cardNames;          // sound card names
  int  *cardSDs;             // Sound Device number corresponding
                             // to each sound card selection
  int  *cardTypes;           // SD sound card type corresponding
                             // to each sound card selection
  int  numCards;             // total number of card selections
  int  sdNum, cardNum, i, error;
  SoundDevice *SD;

  // find the total number of sound cards:
  numCards = 0;
  for (sdNum=0; sdNum<NUMSDEVICES; sdNum++)
    numCards += midasSoundDevices[sdNum]->numCardTypes;

  // allocate memory for card name pointers:
  if ((error = memAlloc(numCards * sizeof(char*), (void**) &cardNames)) != OK)
    midasError(error);

  // allocate memory for Sound Device numbers:
  if ((error = memAlloc(numCards * sizeof(int), (void**) &cardSDs)) != OK)
    midasError(error);

  // allocate memory for sound card type numbers:
  if ((error = memAlloc(numCards * sizeof(int), (void**) &cardTypes)) != OK)
    midasError(error);

  // set all entries in cardNames[], cardSDs[] and cardTypes[] to their
  // correct values:
  cardNum = 0;
  for (sdNum=0; sdNum<NUMSDEVICES; sdNum++)
  {
    // point SD to current Sound Device:
    SD = midasSoundDevices[sdNum];

    // do for all sound card possibilities for this SD:
    for (i=0; i<SD->numCardTypes; i++)
    {
      cardSDs[cardNum] = sdNum;
      cardNames[cardNum] = SD->cardNames[i];
      cardTypes[cardNum] = i+1;
      cardNum++;
    }
  }

  // attempt to autodetect sound card type:
  midasDetectSD();
  detectedSD = midasSDNumber;

  // if a sound device was detected, attempt to find a sound card from
  // the selections which correspond to the detected type:
  cardNum = 0;
  if (midasSDNumber != -1)
  {
    for (i=0; i<numCards; i++)
    {
      // if the Sound Device number and card type number match, this
      // is the correct sound card
      if ((cardSDs[i] == midasSDNumber) && (cardTypes[i] == midasSD->cardType))
        cardNum = i;
    }
  }

  // set up sound card selection:
  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Menu(r);
  Menu.SetTitle("");
  Menu.SetFont(1);
  for (i=0; i<numCards; i++)
    Menu.Add(cardNames[i]);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  Win.Add(&Menu);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    midasSDNumber = cardSDs[cardNum];
    midasSDCard = cardTypes[cardNum];
    midasSD = midasSoundDevices[midasSDNumber];
  }

  // deallocate sound card names:
  if ((error = memFree(cardNames)) != OK)
    midasError(error);

  // deallocate sound card Sound Device numbers:
  if ((error = memFree(cardSDs)) != OK)
    midasError(error);

  // deallocate sound card type numbers:
  if ((error = memFree(cardTypes)) != OK)
    midasError(error);

  Win.Done();
  DrawBackground();
  mk_Clear();

  return (Result==1);
}

int ProcessLoadWorldConsole(void)
{
  Rect r;
  SetupWindowConsole();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle("Load World");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Worlds(r);
  Worlds.SetTitle("Select a world to load:");
  Worlds.SetNoMsg("No worlds available to load.");

  // display wait
  char waitstr[] = "Scanning directory for worlds...";
  Rect r2;

  r2.Set(g_strlen(waitstr)+10, 20);
  Dialog Wait(r2);
  Wait.SetTitle(waitstr);
  Wait.SetHelp(0);

  g_SetCurPage(1);
  g_CopyPage(SpriPage, 1);
  Wait.Draw();
  g_HSync();
  g_Update();

  // scane files
  struct find_t fileinfo;
  ifstream rm;
  unsigned rc;
  pRoom Room;
  char *str;
  char buffer[256];
  char desc[256];
  char id[4], ch;

  rc = _dos_findfirst("*.z2", _A_NORMAL, &fileinfo);
  while (rc==0)
  {
    if (fileinfo.size==0)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }
    rm.open(fileinfo.name, ios::in | ios::binary);
    rm.read(id, 4);
    if (strncmp(id, WorldVer, 4)==0)
    {
      str = desc;
      while (ch = rm.get())
        *str++ = ch;
      *str = 0;

      sprintf(buffer, "%sP%s", fileinfo.name, desc);
      Worlds.Add(buffer, fileinfo.name);
    }
    rm.close();
    rc = _dos_findnext(&fileinfo);
  }

  Wait.Done();

  Win.Add(&Worlds);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    LoadWorld(Worlds.GetInfo(Worlds.GetItem()));
    SwitchToRoom(RoomHead);
  }

  Win.Done();
  DrawBackground();
  mk_Clear();

  return (Result==1);
}

void ClearConsInput(void)
{
  memset(ConsInput, 0, ConW);
  Cursor = 0;
  History = ConH;
  mk_Clear();
}

void Console(void)
{
  do {
    if (!consolemode)
    {
      int option, op2;

      option = ProcessMainMenu();
      switch (option)
      {
        case 2 : quit = 1; break;
        case 10 :
          if (worldloaded)
          {
            zzt_Fade(zztPal, Black);
            Play();
            g_SetCurPage(0);
            DrawBackground();
            zzt_Fade(Black, zztPal);
          } else
          {
            if (ProcessLoadWorldConsole())
            {
              worldloaded = 1;
              zzt_Fade(zztPal, Black);
              Play();
              g_SetCurPage(0);
              DrawBackground();
              zzt_Fade(Black, zztPal);
            }
            break;
          }
        case 11 :
          if (ProcessLoadWorldConsole())
          {
            worldloaded = 1;
            zzt_Fade(zztPal, Black);
            Play();
            g_SetCurPage(0);
            DrawBackground();
            zzt_Fade(Black, zztPal);
          }
          break; // load a world
        case 12 : break; // restore a saved world
        case 13 :
#ifdef NOEDIT
          ProcessMessageConsole("Editor disabled in demo.");
#else
          zzt_Fade(zztPal, Black);
          Edit();
          g_SetCurPage(0);
          DrawBackground();
          zzt_Fade(Black, zztPal);
#endif
          break;
        case 14 :
          consolemode = 1;
          UpdateConsole();
          break;
        case 15 :
          do {
            op2 = ProcessPrefMenu();
            switch (op2)
            {
              case 10 : ProcessAesthetic(); break;
              case 11 : break;
              case 12 : break; //ProcessSound(); break;
            }
          } while (op2>9);
          break; // game options
      }
    } else
    {
      GetConsInput();                    // Get console input
      AddConsLine("]%s", ConsInput);     // Display input in console screen
      ConsParse();                       // Parse console input
      ClearConsInput();                  // Clear for next input
      UpdateConsole();                   // Update screen
    }
  } while (!quit);
}

