// Watcom includes
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

// HGL includes
#include "multikey.h"
#include "sprite.h"
#include "mouse.h"
#include "font.h"
#include "tile.h"
#include "pal.h"
#include "vgl.h"

// Z2 includes
#include "compiler.h"
#include "objects.h"
#include "kernel.h"
#include "global.h"
#include "editor.h"
#include "sound.h"
#include "user.h"

#define ST_NORMAL   1
#define ST_TILESEL  2
#define ST_MOUSING  3
#define ST_BLOCKING 4

struct tClipboard {
  short w, h;
  short *data;
} Clip;

char EditStr[3][10] = {"Floor", "Object", "Ceiling"};
char *CatList[12];
int  CatNum, ItemNum;
tItem *ItemList;
char LastImportFile[80];
char LastWorldFile[80];
char LastWorldDesc[80];
char LastObjectFile[80];
char LastTileFile[80];
char LastObjFile[80];
int nomouse, ineditor=0;
int MAXLAYER = 2;

int quiteditor;
int DrawFloor   = 1;
int DrawObject  = 1;
int DrawCeiling = 0;
int EditLayer   = 0;

int MapSX, MapSY;
int MapX, MapY;
int MapW, MapH;
int CurX, CurY;
int BlockW, BlockH;
int PaintMode = 0;
int State = 1;
short TotTiles = 0;

void SwitchToRoom(pRoom r);
void DefaultRoom(pRoom r);
void InitEditor(void);
void DispTile(int x, int y, int n);
void Panel(int n);
void UpdateEditor(void);
void Paint(void);
void SelectTile(int x, int y, int w, int h);
void ProcessEditScript(pObject o, int n);
void ProcessMessage(char *str);
void PlotTile(int ct);

short CurTile, SeedVal;
short *Ptr, rw;

void SetPtr(void)
{
  rw = RoomCur->RoomW;

  switch (EditLayer)
  {
    case L_OBJECT :
      Ptr = RoomCur->RoomObject;
      break;
    case L_FLOOR :
      Ptr = RoomCur->RoomFloor;
      break;
    case L_CEILING :
      Ptr = RoomCur->RoomCeiling;
      break;
  }
}

int CheckBlock(void)
{
  return !(BlockW==1 && BlockH==1);
}

void PlotTile(int ct)
{
  short *SetPtr = Ptr + (CurY * rw + CurX);

  if (CheckBlock())                         
  {                                         // multi-cell
    for (int i=0; i<BlockH; i++)
      for (int j=0; j<BlockW; j++)
      {
        SetPtr = Ptr + ((CurY+i) * rw + (CurX+j));

        // if an objects resides in current position and layer, kill it
        if (GetObjXY(CurX+j, CurY+i, RoomCur, EditLayer))
          KillObjXY(CurX+j, CurY+i, RoomCur, EditLayer);

        if (ct<TILE_MAX)
          *SetPtr = ct; else
          AddObjXY(CurX+j, CurY+i, ItemList[ct-TILE_MAX].name, RoomCur, EditLayer);
      }
  } else
  {                                         // single-cell
    // if an objects resides in current position and layer, kill it
    if (GetObjXY(CurX, CurY, RoomCur, EditLayer))
      KillObjXY(CurX, CurY, RoomCur, EditLayer);

    if (ct<TILE_MAX)
      *SetPtr = ct; else
      AddObjXY(CurX, CurY, ItemList[ct-TILE_MAX].name, RoomCur, EditLayer);
  }
}

int CheckTile(void)
{
  short *SetPtr = Ptr + (CurY * rw + CurX);
  return (GetObjXY(CurX, CurY, RoomCur, EditLayer) || *SetPtr);
}

void KillXY(int x, int y)
{
  Ptr[y*rw+x] = 0;
  if (GetObjXY(x, y, RoomCur))
    KillObjXY(x, y, RoomCur);
}

int GrabXY(int x, int y)
{
  short *SetPtr = Ptr + (y*rw+x);

  if (*SetPtr) return *SetPtr;

  pObject o = GetObjXY(CurX, CurY, RoomCur);
  if (o)
  {
    pItem it = FindItem(o->NameHash);
    return TILE_MAX+it->id;
  }

  return 0;
}


void PlotXY(int x, int y)
{
  short *SetPtr = Ptr + (y * rw + x);

  if (CheckBlock())
  {                                         // multi-cell
    for (int i=0; i<BlockH; i++)
      for (int j=0; j<BlockW; j++)
      {
        SetPtr = Ptr + ((y+i) * rw + (x+j));

        KillXY(x+j, y+i);                   // kill any objects

        if (CurTile<TILE_MAX)
          *SetPtr = CurTile; else
          AddObjXY(x+j, y+i, ItemList[CurTile-TILE_MAX].name, RoomCur, EditLayer);
      }
  } else
  {                                         // single-cell
    KillXY(x, y);                           // kill any objects

    if (CurTile<TILE_MAX)
      *SetPtr = CurTile; else
      AddObjXY(x, y, ItemList[CurTile-TILE_MAX].name, RoomCur, EditLayer);
  }
}

int CheckXY(int x, int y)
{
  // first check the standard tiles
  if (Ptr[y*rw+x] == SeedVal) return 1;

  // if not, then you are screwed... Time for slowness...
  pObject o = GetObjXY(x, y, RoomCur, EditLayer);
  pItem it = FindItem(o->name);
  if (TILE_MAX+it->id == SeedVal) return 1;

  return 0;
}

int Fill(int x, int y, int d, int pxl, int pxr)
{
  int xl = x;
  int xr = x;
  int i, j;

  while (CheckXY(xl, y) && xl >= 0)
    xl--;
  xl++;
  while (CheckXY(xr, y) && xr < rw)
    xr++;

  for (j=xl; j<xr; j++)
    PlotXY(j, y);
  y += d;

  if (y < RoomCur->RoomH && y >= 0)
    for (x=xl; x<xr; x++)
      if (CheckXY(x, y))
      {
        x = Fill(x, y, d, xl, xr);
        if (x >= xr) break;
      }
  y -= d+d;
  d *= -1;

  if (y < RoomCur->RoomH && y >= 0)
  {
    for (x=xl; x<pxl; x++)
      if (CheckXY(x, y))
      {
        i = Fill(x, y, d, xl, xr);
        if (x >= pxl) break;
      }
    for (x=pxr; x<xr; x++)
      if (CheckXY(x, y))
      {
        i = Fill(x, y, d, xl, xr);
        if (x >= xr) break;
      }
  }

  return xr;
}

void FillBlock(int x, int y, int w, int h)
{
  SetPtr();

  for (int i=0; i<h; i++)
    for (int j=0; j<w; j++)
      Ptr[(i+y)*rw+(j+x)] = CurTile;
}

void CopyBlock(void)
{
  if (Clip.data) delete[] Clip.data;

  Clip.w = BlockW;
  Clip.h = BlockH;
  Clip.data = new short[Clip.w*Clip.h];

  SetPtr();

  for (int i=0; i<Clip.h; i++)
    for (int j=0; j<Clip.w; j++)
      Clip.data[i*Clip.w+j] = Ptr[(i+CurY)*rw+(j+CurX)];
}

void CutBlock(void)
{
  if (Clip.data) delete[] Clip.data;

  Clip.w = BlockW;
  Clip.h = BlockH;
  Clip.data = new short[Clip.w*Clip.h];

  SetPtr();

  for (int i=0; i<Clip.h; i++)
    for (int j=0; j<Clip.w; j++)
    {
      Clip.data[i*Clip.w+j] = Ptr[(i+CurY)*rw+(j+CurX)];
      KillXY(j+CurX, i+CurY);
    }
}

void TransPasteClip(void)
{
  if (!Clip.data) return;

  SetPtr();

  short tile;
  for (int i=0; i<Clip.h; i++)
    for (int j=0; j<Clip.w; j++)
    {
      tile = Clip.data[i*Clip.w+j];
      if (tile) Ptr[(i+CurY)*rw+(j+CurX)] = tile;
    }
}

void PasteClip(void)
{
  if (!Clip.data) return;

  SetPtr();

  for (int i=0; i<Clip.h; i++)
    for (int j=0; j<Clip.w; j++)
      Ptr[(i+CurY)*rw+(j+CurX)] = Clip.data[i*Clip.w+j];
}

int CheckBlock(int x, int y, int w, int h)
{
  SetPtr();

  for (int i=0; i<h; i++)
    for (int j=0; j<w; j++)
      if (Ptr[i*rw+j]) return 1;

  return 0;
}

void SwitchToRoom(pRoom r)
{
  if (r)
  {
    RoomCur = r;

    // keep cursor inbounds
    if (CurX>r->RoomW) CurX = r->RoomW;
    if (CurY>r->RoomH) CurY = r->RoomH;
  }
  RoomData.InitMap(RoomCur->RoomW, RoomCur->RoomH,
    RoomCur->RoomFloor, RoomCur->RoomObject, RoomCur->RoomCeiling);

  if (!r) return;

  if (stricmp(r->Mod, "None")==0) return;
  if (stricmp(r->Mod, "Silence")==0)
  {
    zm_modStop();
    return;
  }

  switch (zm_modLoad(r->Mod))
  {
    case 0 :
      zm_modStart();
      break;
    case -1 : case -2 :
      ProcessMessage("Error loading room music.");
      zm_modStop();
      break;
  }
}

void DefaultRoom(pRoom r)
{
  int maxx, maxy;
  int x, y;

  memset16(r->RoomObject, 0, r->RoomW*r->RoomH*sizeof(short));
  memset16(r->RoomFloor, 0, r->RoomW*r->RoomH*sizeof(short));

  maxx = r->RoomW-1;
  maxy = r->RoomH-1;
}

void InitEditor(void)
{
  RoomCur = AddRoom("IntroRoom", "Introduction Room");
  DefaultRoom(RoomCur);
  SwitchToRoom(RoomCur);

  MapSX = MapSY = 0;
  CurX = CurY = 0;
  MapX = MapY = 0;
  MapW = 241 / Tile_W;
  MapH = 193 / Tile_H;
  BlockW = BlockH = 1;
}

void ReInitEditor(void)
{
  MapSX = MapSY = 0;
  CurX = CurY = 0;
  MapX = MapY = 0;
  MapW = 241 / Tile_W;
  MapH = g_ResY() / Tile_H;
  BlockW = BlockH = 1;
}

void DispTile(int x, int y, int n)
{
  int AdjTileW = Tile_W/((Tile_W>24)+1);
  int AdjTileH = Tile_H/((Tile_H>24)+1);

  g_fRect(x, y, x+AdjTileW+3, y+AdjTileH+3, colPurple);
  g_cHLine(x, x+AdjTileW+2, y, colLPurp);
  g_VLine(x, y, y+AdjTileH+2, colLPurp);
  g_cHLine(x+1, x+AdjTileW+3, y+AdjTileH+3, colDPurp);
  g_VLine(x+AdjTileW+3, y+1, y+AdjTileH+3, colDPurp);

  if (n<0) return;

  int t;

  if (n<TotTiles)
    t = n; else
    t = ItemList[n-TILE_MAX].frame;

  g_ctBitmapS(x+2, y+2, AdjTileW, AdjTileH, &Tiles[t]);
}

void Panel(int n)
{
  int Left = 240;
  int LastFont, LastColor;
  int FloorTile, ObjTile, CeilingTile;
  pObject CurObj;

  // Save font id and color
  LastFont = g_GetFont();
  LastColor = g_GetFontColor();
  g_Font(1);
  g_FontColor(15);

  g_iRect(Left, 0, 319, 190, Dark);
  g_HLine(Left+1, 318, 12, 15);
  g_HLine(Left+2, 319, 13, 0);

  switch (n)
  {
    case ST_MOUSING : // General info (plus mouse help) panel
    case ST_NORMAL : // General info panel
      g_sbtwprintf(Left, 319, 2, "Z2 Editor");

      g_sprintf(Left+3, 15, "(%d,%d)", CurX, CurY);

      if (BlockW>1 || BlockH>1)
      {
        g_Font(2);
        g_sprintf(Left+3, 26, "W:%d  H:%d", BlockW, BlockH);
      }
      break;
    case ST_BLOCKING : // General info panel (with blocking info)
      g_sbtwprintf(Left, 319, 2, "Z2 Editor");

      g_sprintf(Left+3, 15, "(%d,%d)-(%d,%d)",
        CurX, CurY, CurX+BlockW-1, CurY+BlockH-1);

      g_Font(2);
      g_sprintf(Left+3, 26, "W:%d  H:%d", BlockW, BlockH);
      break;
  }

  switch (n)
  {
    case ST_MOUSING :  // General info (plus mouse help) panel
    case ST_NORMAL :   // General info panel
    case ST_BLOCKING : // General info panel (with blocking info)
      g_Font(1);
      g_sprintf(Left+3, 68, "Cur. Floor:");
      g_sprintf(Left+3, 90, "Cur. Ceiling:");
      g_sprintf(Left+3, 112, "Cur. Object:");

      g_Font(2);
      g_sprintf(Left+3, 37, "DrawFloor: %s", (DrawFloor) ? "Yes" : "No");
      g_sprintf(Left+3, 44, "DrawObject: %s", (DrawObject) ? "Yes" : "No");
      g_sprintf(Left+3, 51, "DrawCeiling: %s", (DrawCeiling) ? "Yes" : "No");
      g_sprintf(Left+3, 58, "EditLayer: %s", EditStr[EditLayer]);

      FloorTile = RoomCur->RoomFloor[CurY*RoomCur->RoomW+CurX];
      if (FloorTile) g_sprintf(Left+3, 79, "Tile #: %d", FloorTile);

      if (RoomCur->RoomCeiling)
      {
        CeilingTile = RoomCur->RoomCeiling[CurY*RoomCur->RoomW+CurX];
        if (CeilingTile) g_sprintf(Left+3, 101, "Tile #: %d", CeilingTile);
      } else g_sprintf(Left+3, 101, "Disabled");

      ObjTile = RoomCur->RoomObject[CurY*RoomCur->RoomW+CurX];
      if (ObjTile) g_sprintf(Left+3, 123, "Tile #: %d", ObjTile);

      CurObj = GetObjXY(CurX, CurY, RoomCur);
      if (CurObj)
      {
        g_sprintf(Left+3, 123, "%s", CurObj->name);
        g_sprintf(Left+3, 130, "Tile #: %d", CurObj->frame);
        g_sprintf(Left+3, 137, "Code: %d bytes", CurObj->codesize);
        g_sprintf(Left+3, 144, "Layer: %s", EditStr[CurObj->z]);
      }

      g_sprintf(Left+3, 154, "Alt+H: Editor Help");
      g_sprintf(Left+3, 161, "Alt+K: HotKey Help");

      g_sprintf(Left+3, 170, "Draw Tile: ");
      if (PaintMode)
        g_sprintf(Left+3, 177, "Painting"); else
        g_sprintf(Left+3, 177, "Plotting");
      DispTile(288, 170, CurTile);

      Paint();
      break;
    case ST_TILESEL : // Tile Select panel
      g_sbtwprintf(Left, 319, 2, "Select Tile");

      g_Font(2);
      g_sprintf(Left+3, 15, "Current: %d of %d", CurTile, TotTiles-1);
      g_sprintf(Left+3, 180, DefName);

      int AdjTileW = Tile_W/((Tile_W>24)+1);
      int AdjTileH = Tile_H/((Tile_H>24)+1);
      SelectTile(Left+3, 23, (320-Left)/(AdjTileW+6), 157/(AdjTileH+6));
      break;
  }

  g_iRect(0, 192, 319, 199, Dark);
  g_Font(2);

  switch (n)
  {
    case ST_NORMAL :
      g_sprintf(1, 192, "Editing: %s (%s)", RoomCur->name, RoomCur->Desc);
      break;
    case ST_MOUSING :
      g_sprintf(1, 192, "Left-Click: Select  Right-Drag: Scroll  Enter: Edit scroll/object");
      break;
    case ST_TILESEL :
      g_sprintf(1, 192, "Select tile with cursor keys, release Ctrl key to accept");
      break;
    case ST_BLOCKING :
      g_sprintf(1, 192, "Drag: Select Block  Let Go: Accept Block  Right-Click: Cancel Block");
      break;
  }

  // Restore previous font id and color
  g_Font(LastFont);
  g_FontColor(LastColor);
}

void UpdateEditor(void)
{
  int AbsMapX, AbsMapY;
  int AbsCurX, AbsCurY;

  g_SetCurPage(1);
  g_SuperBlack();

  AbsMapX = MapX*Tile_W;
  AbsMapY = MapY*Tile_H;
  AbsCurX = (CurX*Tile_W)-AbsMapX;
  AbsCurY = (CurY*Tile_H)-AbsMapY;

  RoomData.SetPos(AbsMapX, AbsMapY);
  if (DrawFloor) RoomData.DrawBase();
  if (DrawObject)
  {
    RoomData.DrawFringe();

    pObject Cur = RoomCur->ObjHead;
    while (Cur)
    {
      int x = RoomData.Tile2ScrX(Cur->x);
      int y = RoomData.Tile2ScrY(Cur->y);
      if (x>-Tile_W && x<320 && y>-Tile_H && y<200)
      {
        if (x>0 && x<320-Tile_W && y>0 && y<200-Tile_H)
          g_tBitmap(x, y, &Tiles[Cur->frame]); else
          g_ctBitmap(x, y, &Tiles[Cur->frame]);
      } // don't draw way out of bounds
      Cur = Cur->next;
    }
/*
    // draw separate objects
    pObject Cur = RoomCur->ObjHead;
    while (Cur)
    {
      g_ctBitmap(RoomData.Tile2ScrX(Cur->x),
                 RoomData.Tile2ScrY(Cur->y),
                 &Tiles[Cur->frame]);
      Cur = Cur->next;
    }
*/
  }
  if (DrawCeiling) RoomData.DrawRoof();

  g_cRect(AbsCurX, AbsCurY, AbsCurX+(Tile_W*BlockW)-1, AbsCurY+(Tile_H*BlockH)-1, colWhite);

  Panel(State);

  if (!nomouse)
  {
    m_Update();
    s_Update(0);
  }
  g_HSync();
  g_Update();
}

void Paint(void)
{
  int paint=0, grab=0;

  int OldCurX = CurX;
  int OldCurY = CurY;

  switch (State)
  {
    case ST_NORMAL :
      if (mk_Down(mkEnd))
      {
        if (mk_Pressed(mkUp))    CurY = 0;
        if (mk_Pressed(mkDown))  CurY = RoomCur->RoomH-1;
        if (mk_Pressed(mkLeft))  CurX = 0;
        if (mk_Pressed(mkRight)) CurX = RoomCur->RoomW-1;
      } else
      {
        if (mk_Pressed(mkUp))    CurY--;
        if (mk_Pressed(mkDown))  CurY++;
        if (mk_Pressed(mkLeft))  CurX--;
        if (mk_Pressed(mkRight)) CurX++;
      }

      if (CurX != OldCurX || CurY != OldCurY)
        BlockW = BlockH = 1;
      break;
    case ST_BLOCKING :
      if (mk_Down(mkAlt) || mk_Down(mkEnd) || mk_Down(mkCtrl))
      {
        if (mk_Pressed(mkUp) && CurY>0) CurY--;
        if (mk_Pressed(mkDown) && CurY+BlockH<RoomCur->RoomH) CurY++;
        if (mk_Pressed(mkLeft) && CurX>0) CurX--;
        if (mk_Pressed(mkRight) && CurX+BlockW<RoomCur->RoomW) CurX++;

        if (CurX<MapSX)
        {
          MapSX--;
          MapX--;
        }
        if (CurY<MapSY)
        {
          MapSY--;
          MapY--;
        }
      } else
      {
        if (mk_Pressed(mkUp))    BlockH--;
        if (mk_Pressed(mkDown))  BlockH++;
        if (mk_Pressed(mkLeft))  BlockW--;
        if (mk_Pressed(mkRight)) BlockW++;
      }
      break;
    default :
      if (mk_Down(mkEnd))
      {
        if (mk_Pressed(mkUp))    CurY = 0;
        if (mk_Pressed(mkDown))  CurY = RoomCur->RoomH-1;
        if (mk_Pressed(mkLeft))  CurX = 0;
        if (mk_Pressed(mkRight)) CurX = RoomCur->RoomW-1;
      } else
      {
        if (mk_Down(mkUp))    CurY--;
        if (mk_Down(mkDown))  CurY++;
        if (mk_Down(mkLeft))  CurX--;
        if (mk_Down(mkRight)) CurX++;
      }

      if (CurX != OldCurX || CurY != OldCurY)
        BlockW = BlockH = 1;
      break;
  }

  if (m_Left())
  {
    if (m_X()<240 && m_Y()<192)
    {
      CurX = (m_X() / Tile_W)+MapX;
      CurY = (m_Y() / Tile_H)+MapY;
      if (State==ST_NORMAL) paint = 1;
    }
  }
  if (m_Right() && State==ST_NORMAL)
  {
    if (m_X()<240 && m_Y()<192)
    {
      CurX = (m_X() / Tile_W)+MapX;
      CurY = (m_Y() / Tile_H)+MapY;
    }
    grab = 1;
  }
  if (m_Right() && State==ST_MOUSING)
  {
    if (m_X()==0) CurX--;
    if (m_Y()==0) CurY--;
    if (m_X()==g_MaxX()) CurX++;
    if (m_Y()==g_MaxY()) CurY++;
  }
  if (m_Middle())
  {
    if (m_X()<240 && m_Y()<192)
    {
      int DX = (m_X() / Tile_W)+MapX;
      int DY = (m_Y() / Tile_H)+MapY;
      if (DX>=CurX) BlockW = DX-CurX+1;
      if (DY>=CurY) BlockH = DY-CurY+1;
    }
  }

  if (BlockW<1) BlockW = 1;
  if (BlockH<1) BlockH = 1;
  if (CurX<0) CurX = 0;
  if (CurY<0) CurY = 0;
  if (CurX+BlockW-1>=RoomCur->RoomW)
  {
    if (CurX>=RoomCur->RoomW) CurX = RoomCur->RoomW-1;
    BlockW = RoomCur->RoomW-CurX;
    CurX = RoomCur->RoomW-BlockW;
  }
  if (CurY+BlockH-1>=RoomCur->RoomH)
  {
    if (CurY>=RoomCur->RoomH) CurY = RoomCur->RoomH-1;
    BlockH = RoomCur->RoomH-CurY;
    CurY = RoomCur->RoomH-BlockH;
  }

  if (CurX+BlockW-1<MapSX)
  {
    MapSX--;
    MapX--;
  }
  if (CurY+BlockH-1<MapSY)
  {
    MapSY--;
    MapY--;
  }
  if (CurX+BlockW-1>MapSX+MapW-1)
  {
    MapSX++;
    MapX++;
  }
  if (CurY+BlockH-1>MapSY+MapH-1)
  {
    MapSY++;
    MapY++;
  }

  SetPtr();

  if (mk_Pressed(mkSpace) || PaintMode || paint)
  {
    if (!CheckBlock() && CheckTile() && !PaintMode)
      PlotTile(0); else
      PlotTile(CurTile);
    paint = 0;
  }

  if (mk_Pressed(mkEnter))
  {
    if (mk_Alt() || mk_Shift())
    {
      mk_Clear();
      pObject o = GetObjXY(CurX, CurY, RoomCur);
      if (o) ProcessEditScript(o, 0);
      mk_Clear();
    } CurTile = GrabXY(CurX, CurY);
  }

  if (grab)
  {
    CurTile = GrabXY(CurX, CurY);
    grab = 0;
  }
}

void SelectTile(int x, int y, int w, int h)
{
  int ShowFirst;
  int First, Last, Delta;
  int SelTile, ShowTile;
  int ShowMax=w*h;
  int i, j, ax, ay;

  int AdjTileW = Tile_W/((Tile_W>24)+1);
  int AdjTileH = Tile_H/((Tile_H>24)+1);

  // download current tile from editor (if it points to an object frame,
  // default to 0.)
  SelTile = CurTile;
  if (SelTile>TILE_MAX) SelTile = 0;
  Delta   = ((h-1)/2)*w;

  // handle keys
  if (mk_Pressed(mkUp))   SelTile -= w;
  if (mk_Pressed(mkDown)) SelTile += w;
  if (mk_Pressed(mkHome)) SelTile = 0;
  if (mk_Pressed(mkEnd))  SelTile = TotTiles-1;
  if (mk_Pressed(mkPgUp)) SelTile -= Delta;
  if (mk_Pressed(mkPgDn)) SelTile += Delta;

  while (SelTile<0) SelTile += w;
  while (SelTile>TotTiles-1) SelTile -= w;

  First = SelTile/w;
  First *= w;
  ShowFirst = First-Delta;
  if (ShowFirst<0) ShowFirst = 0;
  
  // determine horizontal min and max
  Last = First+(w-1);
  if (Last>TotTiles-1) Last = TotTiles-1;   

  // handle left/right
  if (mk_Pressed(mkLeft) && SelTile>First) SelTile--;
  if (mk_Pressed(mkRight) && SelTile<Last) SelTile++;

  if (First/w==(ShowFirst/w)+h) ShowFirst += w;
  if (First/w<ShowFirst/w)      ShowFirst -= w;

  // reset ShowTile to display visible tiles in list, and then display
  ShowTile = ShowFirst;

  // upload new current tile
  CurTile = SelTile;

  for (j=0; j<h; j++)
    for (i=0; i<w; i++, ShowTile++)
    {
      ax = x+(i*(AdjTileW+6));
      ay = y+(j*(AdjTileH+6));
      if (ShowTile<TotTiles) DispTile(ax, ay, ShowTile);
      if (ShowTile==CurTile) g_Rect(ax, ay, ax+AdjTileW+3, ay+AdjTileH+3, colWhite);
    }

/* debugging

  g_FontColor(14);
  g_sprintf(3, 180, "ShF:%d SelT:%d F:%d L:%d",
            ShowFirst, SelTile, First, Last);
  g_FontColor(15);
*/
}

void SetupWindow(void)
{
  int AbsMapX, AbsMapY;
  int AbsCurX, AbsCurY;

  g_SetCurPage(SpriPage);
  g_SuperBlack();

  AbsMapX = MapX*Tile_W;
  AbsMapY = MapY*Tile_H;
  AbsCurX = (CurX*Tile_W)-AbsMapX;
  AbsCurY = (CurY*Tile_H)-AbsMapY;

  RoomData.SetPos(AbsMapX, AbsMapY);
  if (DrawFloor) RoomData.DrawBase();
  if (DrawObject)
  {
    RoomData.DrawFringe();
    // draw separate objects
    pObject Cur = RoomCur->ObjHead;
    while (Cur)
    {
      g_ctBitmap(RoomData.Tile2ScrX(Cur->x),
                 RoomData.Tile2ScrY(Cur->y),
                 &Tiles[Cur->frame]);
      Cur = Cur->next;
    }
  }
  if (RoomCur->RoomCeiling && DrawCeiling) RoomData.DrawRoof();

  g_Rect(AbsCurX, AbsCurY, AbsCurX+Tile_W-1, AbsCurY+Tile_H-1, colWhite);

  Panel(State);

  g_FontColor(14);
  g_Font(1);
}

void ProcessMessage(char *str)
{
  Rect r;
  SetupWindow();

  int len = g_strlen(str)+10;
  if (len<70) len = 70;
  r.Set(len, 40);
  Dialog Win(r);
  Win.SetTitle(str);
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+((Win.r.w-60)/2), Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  Win.Add(&Ok);

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagMessage");
      UpdateObj(Z2Help);
    }
  } while (!Ok.WasPressed() || mk_Ctrl(mkEnter));

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

int ProcessYesNo(char *str)
{
  Rect r;
  SetupWindow();

  int len = g_strlen(str)+10;
  if (len<135) len = 135;
  r.Set(len, 40);
  Dialog Win(r);
  Win.SetTitle(str);
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+((Win.r.w-125)/2), Win.r.y2-15, 60, 12);
  Button Yes(r);
  Yes.SetTitle("Yes");

  r.MoveRel(65, 0);
  Button No(r);
  No.SetTitle("No");

  Win.Add(&Yes);
  Win.Add(&No);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagYesNo");
      UpdateObj(Z2Help);
    }

    if (Yes.WasPressed()) Result = 1;
    if (No.WasPressed()) Result = 2;
  } while (!Result);

  Win.Done();
  UpdateEditor();

  return (Result==1);
}

int ProcessOkCan(char *str)
{
  Rect r;
  SetupWindow();

  int len = g_strlen(str)+10;
  if (len<135) len = 135;
  r.Set(len, 40);
  Dialog Win(r);
  Win.SetTitle(str);
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+((Win.r.w-125)/2), Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagOkCan");
      UpdateObj(Z2Help);
    }

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  Win.Done();
  UpdateEditor();
  mk_Clear();

  return (Result==1);
}

void ProcessAddRoom(void)
{
  pRoom rm;

  Rect r;
  SetupWindow();

  r.Set(200, 95);
  Dialog Win(r);
  Win.SetTitle("Add a Room");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y+34);
  Input Room(r);
  Room.SetTitle("Enter room keyname:");
  Room.SetMax(RNAME_MAX);
  Room.AllowSpaces(0);

  r.MoveRel(0, 20);
  Input Desc(r);
  Desc.SetTitle("Enter room description:");
  Room.SetMax(RDESC_MAX);

  r.MoveRel(0, 17);
  Check Jump(r);
  Jump.Add("Edit the new room", 1);
  Jump.SetTitle("");

  Win.Add(&Room);
  Win.Add(&Desc);
  Win.Add(&Jump);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagAddRoom");
      UpdateObj(Z2Help);
    }

    if (Ok.WasPressed() || mk_Ctrl(mkEnter))
    {
      Result = 1;

      // validate
      rm = FindRoom(Room.GetDataStr());
      if (rm)
      {
        ProcessMessage("Duplicate name found, use another name.");
        Ok.ResetPress();
        Result = 0;
      }
    }
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result == 1)
  {
    rm = AddRoom(Room.GetDataStr(), Desc.GetDataStr());
    DefaultRoom(rm);
    if (!Jump.GetState(0)) SwitchToRoom(rm);      // check the check-box
//    ProcessMessage("Room added successfully.");
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

int ProcessCatList(int c)
{
  if (CatList[c][0]=='*' || c>=CatNum) return -1;

  Rect r;
  SetupWindow();

  r.Set(300, 150);
  Dialog Win(r);
  Win.SetTitle(CatList[c]);
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  ListTile Items(r);
  Items.SetTitle("Select an item:");
  Items.SetNoMsg("Error: No items in this category!");

  for (int i=0; i<ItemNum; i++)
    if (ItemList[i].cat == c)
      Items.Add(ItemList[i].desc, ItemList[i].frame);

  Win.Add(&Items);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagCatList");
      UpdateObj(Z2Help);
    }

    if (Ok.WasPressed() || mk_Ctrl(mkEnter) || Items.WasPressed()) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  int retval;
  if (Result == 1)
  {
    pItem it = FindItemDesc(Items.GetName(Items.GetItem()));

    // check for special processing
    switch (it->note)
    {
      case SN_NONE :
        break;
      case SN_ROOM :
        break;
      case SN_DIALOG :
        break;
      case SN_COORD :
        break;
      case SN_SCROLL :  // get specific code, compile, copy
      case SN_OBJECT :
        break;
      case SN_WATER :
        break;
      case SN_FOREST :
        break;
      case SN_FAKE :
        break;
      case SN_PLAYER :
        break;
      case SN_GENERIC :
        break;
    }

    retval = TILE_MAX+it->id;
  } else retval = -1;

  Win.Done();
  UpdateEditor();
  mk_Clear();

  return retval;
}

void ProcessSelectRoom(char *title, char *msg, char *none, char *room)
{
  Rect r;
  SetupWindow();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle(title);
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.MoveRel(65, 0);
  Button GO(r);
  GO.SetTitle("Global");

  r.MoveRel(65, 0);
  Button SO(r);
  SO.SetTitle("System");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Rooms(r);
  Rooms.SetTitle(msg);
  Rooms.SetNoMsg(none);

  pRoom Cur = RoomHead;
  char buffer[256];
  while (Cur)
  {
    sprintf(buffer, "%s (%s)", Cur->name, Cur->Desc);
    Rooms.Add(buffer, Cur->name);
    Cur = Cur->next;
  }

  Win.Add(&Rooms);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&GO);
  Win.Add(&SO);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagSelRoom");
      UpdateObj(Z2Help);
    }

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (SO.WasPressed())
    {
      SO.ResetPress();
      Result = 3;
    }
    if (GO.WasPressed())
    {
      GO.ResetPress();
      Result = 4;
    }
  } while (!Result);

  if (Result == 1)
  {
    Cur = FindRoom(Rooms.GetInfo(Rooms.GetItem()));
    if (Cur)
    {
      Win.Done();
      UpdateEditor();
      strcpy(room, Cur->name);
      return;
    }
  }
  if (Result==3) SwitchToRoom(SystemRoom);
  if (Result==4) SwitchToRoom(GlobalRoom);

  Win.Done();
  UpdateEditor();
  mk_Clear();

  strcpy(room, "none");
  return;
}

int ProcessRoomLink(char *title, char *msg, char *none, char *def)
{
  int way=0;

  Rect r;
  SetupWindow();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle(title);
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.MoveRel(65, 0);
  Button None(r);
  None.SetTitle("None");

  r.MoveRel(65, 0);
  Button Dual(r);
  Dual.SetTitle("One Way");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Rooms(r);
  Rooms.SetTitle(msg);
  Rooms.SetNoMsg(none);

  pRoom Cur = RoomHead;
  char buffer[256];
  while (Cur)
  {
    sprintf(buffer, "%s (%s)", Cur->name, Cur->Desc);
    Rooms.Add(buffer, Cur->name);
    Cur = Cur->next;
  }
  Rooms.SetDefault(def);

  Win.Add(&Rooms);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&None);
  Win.Add(&Dual);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagRoomLink");
      UpdateObj(Z2Help);
    }

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (Dual.WasPressed())
    {
      Dual.ResetPress();
      way = 1-way;
      if (way)
        Dual.SetTitle("Two Way"); else
        Dual.SetTitle("One Way");
    }
    if (None.WasPressed()) Result = 3;
  } while (!Result);

  int ret;

  if (Result == 1)
  {
    Cur = FindRoom(Rooms.GetInfo(Rooms.GetItem()));
    if (Cur)
    {
      strcpy(def, Cur->name);
      ret = way;
    }
  }
  if (Result == 2) ret = -1;
  if (Result == 3) ret = -2;

  Win.Done();
  UpdateEditor();
  mk_Clear();

  return ret;
}

int ProcessModSelect(char *title, char *msg, char *none, char *def)
{
  Rect r;
  SetupWindow();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle(title);
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.MoveRel(65, 0);
  Button None(r);
  None.SetTitle("None");

  r.MoveRel(65, 0);
  Button Silence(r);
  Silence.SetTitle("Silence");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Mods(r);
  Mods.SetTitle(msg);
  Mods.SetNoMsg(none);

  // display wait
  char waitstr[] = "Scanning directory for music...";
  Rect r2;

  r2.Set(g_strlen(waitstr)+10, 20);
  Dialog Wait(r2);
  Wait.SetTitle(waitstr);
  Wait.SetHelp(0);

  g_SetCurPage(1);
  g_CopyPage(SpriPage, 1);
  Wait.Draw();
  g_HSync();
  g_Update();

  // scan files
  struct find_t fileinfo;
  ifstream rm;
  unsigned rc;
  pRoom Room;
  char *str;
  char buffer[256];
  char name[256];
  char desc[256];
  char id[20], ch;

  rc = _dos_findfirst("*.xm", _A_NORMAL, &fileinfo);
  while (rc==0)
  {
    if (fileinfo.size==0)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }
    rm.open(fileinfo.name, ios::in | ios::binary);
    rm.read(id, 16);
    if (strncmp(id, "Extended Module:", 16)==0)
    {
      strcpy(name, fileinfo.name);

      rm.seekg(17, ios::beg);
      str = desc;
      while ((ch = rm.get()) != 26)
      {
        if (ch=='\n' || ch==TABSTOP || ch==COLORCH) continue;
        *str++ = ch;
      }
      *str = 0;

      sprintf(buffer, "%sP%s (%s)", fileinfo.name, name, desc);
      Mods.Add(buffer, fileinfo.name);
    }
    rm.close();
    rc = _dos_findnext(&fileinfo);
  }

  Wait.Done();

  Win.Add(&Mods);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&None);
  Win.Add(&Silence);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("DiagModSelect");
      UpdateObj(Z2Help);
    }

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (None.WasPressed()) Result = 3;
    if (Silence.WasPressed()) Result = 4;
  } while (!Result);

  int ret;

  switch (Result)
  {
    case 1 :
      strcpy(def, Mods.GetInfo(Mods.GetItem()));
      ret = 1;
      break;
    case 2 : ret = -1; break;
    case 3 : ret = -2; break;
    case 4 : ret = -3; break;
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();

  return ret;
}

void ProcessChangeRoom(void)
{
  char *room = new char[RNAME_MAX];
  ProcessSelectRoom("Change to Room", "Select a room to edit", "No rooms to edit!", room);

  if (strcmp(room, "none"))
  {
    pRoom Cur = FindRoom(room);
    SwitchToRoom(Cur);
  }

  delete[] room;
}

void ProcessImportRoom(void)
{
  Rect r;
  SetupWindow();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle("Import Room");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Rooms(r);
  Rooms.SetTitle("Select a room to import:");
  Rooms.SetNoMsg("No rooms available for import.");

  // display wait
  char waitstr[] = "Scanning directory for rooms...";
  Rect r2;

  r2.Set(g_strlen(waitstr)+10, 20);
  Dialog Wait(r2);
  Wait.SetTitle(waitstr);
  Wait.SetHelp(0);

  g_SetCurPage(1);
  g_CopyPage(SpriPage, 1);
  Wait.Draw();
  g_HSync();
  g_Update();

  // scan files
  struct find_t fileinfo;
  ifstream rm;
  unsigned rc;
  pRoom Room;
  char *str;
  char buffer[256];
  char name[256];
  char desc[256];
  char id[4], ch;

  rc = _dos_findfirst("*.zr", _A_NORMAL, &fileinfo);
  while (rc==0)
  {
    if (fileinfo.size==0)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }
    rm.open(fileinfo.name, ios::in | ios::binary);
    rm.read(id, 4);
    if (strncmp(id, RoomVer, 4)==0)
    {
      str = name;
      while (ch = rm.get())
        *str++ = ch;
      *str = 0;

      str = desc;
      while (ch = rm.get())
        *str++ = ch;
      *str = 0;

      sprintf(buffer, "%sP%s (%s)", fileinfo.name, name, desc);
      Rooms.Add(buffer, fileinfo.name);
    }
    rm.close();
    rc = _dos_findnext(&fileinfo);
  }

  Wait.Done();

  Win.Add(&Rooms);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    strcpy(LastImportFile, Rooms.GetInfo(Rooms.GetItem()));
    LoadRoom(LastImportFile, RoomCur);
    SwitchToRoom(RoomCur);
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessExportRoom(void)
{
  Rect r;
  SetupWindow();

  r.Set(200, 60);
  Dialog Win(r);
  Win.SetTitle("Export Room");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y+34);
  Input SaveAs(r);
  SaveAs.SetTitle("Enter room filename:");
  SaveAs.AllowSpaces(0);
  SaveAs.SetData(LastImportFile);

  Win.Add(&SaveAs);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    char *str = SaveAs.GetDataStr();
    AutoAppend(str, ".zr");
    SaveRoom(str, RoomCur);
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessImportTile(BITMAP *t)
{
  Rect r;
  SetupWindow();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle("Import Tiles");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Tiles(r);
  Tiles.SetTitle("Select a tile library to load:");
  Tiles.SetNoMsg("No tile libraries available to load.");

  // display wait
  char waitstr[] = "Scanning directory for tiles...";
  Rect r2;

  r2.Set(g_strlen(waitstr)+10, 20);
  Dialog Wait(r2);
  Wait.SetTitle(waitstr);
  Wait.SetHelp(0);

  g_SetCurPage(1);
  g_CopyPage(SpriPage, 1);
  Wait.Draw();
  g_HSync();
  g_Update();

  // scane files
  struct find_t fileinfo;
  ifstream rm;
  unsigned rc;
  pRoom Room;
  char *str;
  char buffer[256];
  char desc[256];
  char id[4], ch;
  unsigned short w, h;

  rc = _dos_findfirst("*.vsp", _A_NORMAL, &fileinfo);
  while (rc==0)
  {
    if (fileinfo.size==0)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }

    rm.open(fileinfo.name, ios::in | ios::binary);
    rm.read(&w, 2);
    rm.read(&h, 2);
    rm.close();
    if (fileinfo.size != w*h+4)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }

    sprintf(buffer, "%sP%d %dx%d tiles, %d bytes", fileinfo.name, w, h,
      fileinfo.size/(w*h+4), fileinfo.size);
    Tiles.Add(buffer, fileinfo.name);
    rc = _dos_findnext(&fileinfo);
  }

  Wait.Done();

  Win.Add(&Tiles);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    strcpy(LastTileFile, Tiles.GetInfo(Tiles.GetItem()));
    if (LoadTiles(LastTileFile)==-1)
    {
      ProcessMsg("Invalid tileset selected");
      LoadTiles(TileFile);
    }
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessExportTile(BITMAP *t)
{
  Rect r;
  SetupWindow();

  r.Set(200, 60);
  Dialog Win(r);
  Win.SetTitle("Export Tiles");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y+34);
  Input SaveAs(r);
  SaveAs.SetTitle("Enter tile library filename:");
  SaveAs.AllowSpaces(0);
  SaveAs.SetData(TileFile);

  Win.Add(&SaveAs);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    char *str = SaveAs.GetDataStr();
    AutoAppend(str, ".vsp");
    g_SaveBitmap(str, Tiles, TotTiles);
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessImportObj(pObject o)
{
  Rect r;
  SetupWindow();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle("Import Object Code");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Obj(r);
  Obj.SetTitle("Select an object to load:");
  Obj.SetNoMsg("No objects available to load.");

  // display wait
  char waitstr[] = "Scanning directory for objects...";
  Rect r2;

  r2.Set(g_strlen(waitstr)+10, 20);
  Dialog Wait(r2);
  Wait.SetTitle(waitstr);
  Wait.SetHelp(0);

  g_SetCurPage(1);
  g_CopyPage(SpriPage, 1);
  Wait.Draw();
  g_HSync();
  g_Update();

  // scane files
  struct find_t fileinfo;
  unsigned rc;
  pRoom Room;
  char *str;
  char buffer[256];
  char desc[256];
  char id[4], ch;

  rc = _dos_findfirst("*.oz", _A_NORMAL, &fileinfo);
  while (rc==0)
  {
    if (fileinfo.size==0)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }

    sprintf(buffer, "%sP%d bytes", fileinfo.name, fileinfo.size);
    Obj.Add(buffer, fileinfo.name);
    rc = _dos_findnext(&fileinfo);
  }

  Wait.Done();

  Win.Add(&Obj);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    strcpy(LastObjFile, Obj.GetInfo(Obj.GetItem()));
    LoadObj(LastObjFile, o);
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessLoadWorld(void)
{
  Rect r;
  SetupWindow();

  r.Set(300, 100);
  Dialog Win(r);
  Win.SetTitle("Load World");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y2-20);
  List Worlds(r);
  Worlds.SetTitle("Select a world to load:");
  Worlds.SetNoMsg("No worlds available to load.");

  // display wait
  char waitstr[] = "Scanning directory for worlds...";
  Rect r2;

  r2.Set(g_strlen(waitstr)+10, 20);
  Dialog Wait(r2);
  Wait.SetTitle(waitstr);
  Wait.SetHelp(0);

  g_SetCurPage(1);
  g_CopyPage(SpriPage, 1);
  Wait.Draw();
  g_HSync();
  g_Update();

  // scane files
  struct find_t fileinfo;
  ifstream rm;
  unsigned rc;
  pRoom Room;
  char *str;
  char buffer[256];
  char desc[256];
  char id[4], ch;

  rc = _dos_findfirst("*.z2", _A_NORMAL, &fileinfo);
  while (rc==0)
  {
    if (fileinfo.size==0)
    {
      rc = _dos_findnext(&fileinfo);
      continue;
    }
    rm.open(fileinfo.name, ios::in | ios::binary);
    rm.read(id, 4);
    if (strncmp(id, WorldVer, 4)==0)
    {
      str = desc;
      while (ch = rm.get())
        *str++ = ch;
      *str = 0;

      sprintf(buffer, "%sP%s", fileinfo.name, desc);
      Worlds.Add(buffer, fileinfo.name);
    }
    rm.close();
    rc = _dos_findnext(&fileinfo);
  }

  Wait.Done();

  Win.Add(&Worlds);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    strcpy(LastWorldFile, Worlds.GetInfo(Worlds.GetItem()));
    LoadWorld(LastWorldFile);
    SwitchToRoom(RoomHead);
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessSaveWorld(void)
{
  Rect r;
  SetupWindow();

  r.Set(200, 80);
  Dialog Win(r);
  Win.SetTitle("Save World");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y+34);
  Input SaveAs(r);
  SaveAs.SetTitle("Enter world filename:");
  SaveAs.AllowSpaces(0);
  SaveAs.SetData(LastWorldFile);

  r.MoveRel(0, 20);
  Input Desc(r);
  Desc.SetTitle("Enter world description:");
  Desc.SetMax(80);
  Desc.SetData(LastWorldDesc);

  Win.Add(&SaveAs);
  Win.Add(&Desc);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    char *str = SaveAs.GetDataStr();
    AutoAppend(str, ".z2");
    SaveWorld(str, Desc.GetDataStr());
    strcpy(LastWorldFile, str);
    strcpy(LastWorldDesc, Desc.GetDataStr());
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessWorldProp(void)
{
  Rect r;
  SetupWindow();

  r.Set(200, 160);
  Dialog Win(r);
  Win.SetTitle("World Properties");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+24);
  r.SetLR(Win.r.x2-5, Win.r.y+34);
  Input SaveAs(r);
  SaveAs.SetTitle("Enter world filename:");
  SaveAs.AllowSpaces(0);
  SaveAs.SetData(LastWorldFile);

  r.MoveRel(0, 20);
  Input Desc(r);
  Desc.SetTitle("Enter world description:");
  Desc.SetMax(80);
  Desc.SetData(LastWorldDesc);

  Win.Add(&SaveAs);
  Win.Add(&Desc);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    char *str = SaveAs.GetDataStr();
    AutoAppend(str, ".z2");
    strcpy(LastWorldFile, str);
    strcpy(LastWorldDesc, Desc.GetDataStr());
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessObjectProp(pObject o)
{
  Rect r;
  SetupWindow();

  r.Set(200, 140);
  Dialog Win(r);
  Win.SetTitle("Object Properties");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+25);
  r.SetLR(Win.r.x2-5, Win.r.y+35);
  Input ObjName(r);
  ObjName.SetTitle("Change object's name:");
  ObjName.SetData(o->name);
  ObjName.SetMax(OBJNAME_MAX);
  ObjName.AllowSpaces(0);

  r.MoveRel(0, 23);
  Check Prop(r);
  Prop.SetFont(2);
  Prop.SetTitle("Object is:");
  Prop.Add("the player?", o->objPlayer);
  Prop.Add("capable of flight?", o->objFlight);
  Prop.Add("an item?", o->objItem);
  Prop.Add("invisible?", o->objInvisible);
  Prop.Add("pushable by player?", o->objPushP);
  Prop.Add("pushable vertically?", o->objPushNS);
  Prop.Add("pushable horizontally?", o->objPushEW);
  Prop.Add("a swimmer?", o->objPusher);
  Prop.Add("breakable?", o->objBreakable);
  Prop.Add("solid?", o->objSolid);
  Prop.Add("bound?", o->objBound);
  Prop.Add("locked?", o->objLocked);

  r.Move(Win.r.x2-40, Win.r.y+43);
  r.SetLR(Win.r.x2-5, Win.r.y+53);
  Input Cycle(r);
  Cycle.SetTitleLeft(25);
  Cycle.SetTitle("Cycle:");
  Cycle.SetData(o->cycle);

  Win.Add(&ObjName);
  Win.Add(&Cycle);
  Win.Add(&Prop);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    strcpy(o->name, ObjName.GetDataStr());
    o->cycle = Cycle.GetDataInt();
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessEditScript(pObject o, int n)
{
  char *newcode = new char[SCRATCH_SIZE];
  memset(newcode, 0, SCRATCH_SIZE);
  int newcodesize=0;
  char buffer[OBJNAME_MAX];

  Rect r;
  SetupWindow();

  r.Set(252, 180);
  Dialog Win(r);
  Win.SetTitle("Script Editor");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Set(Win.r.x+5, Ok.r.y-15, 60, 12);
  Button Prop(r);
  Prop.SetTitle("Properties");

  r.MoveRel(65, 0);
  Button Copy(r);
  Copy.SetTitle("Copy");

  r.MoveRel(65, 0);
  Button Import(r);
  Import.SetTitle("Import");

  r.MoveRel(0, 15);
  Button Export(r);
  Export.SetTitle("Export");

  r.Move(Win.r.x+5, Win.r.y+25);
  r.Size(242, 120);
  Editor Scr(r);
  Scr.SetTitle("Object Code:");
  char *str = Scr.GetData();
  if (o->code && !n) DeCompile(o, str);

  Win.Add(&Scr);
  Win.Add(&Prop);
  Win.Add(&Copy);
  Win.Add(&Import);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&Export);

  int Result=0;
  int OldTile = CurTile;
  int SetTile = o->frame;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();
    DispTile(Win.r.x2-TileW+8, Win.r.y2-TileH+6, SetTile);

    if (mk_Ctrl())
    {
      Panel(ST_TILESEL);
      SetTile = CurTile;
    }

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (mk_Pressed(mkF1))
    {
      SendHelpMessage("LangHelp");
      UpdateObj(Z2Help);
    }
    if (Ok.WasPressed()) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (Prop.WasPressed())
    {
      ProcessObjectProp(o);
      Prop.ResetPress();
      sprintf(buffer, "%s", o->name);
      Scr.SetTitle(buffer);
    }
    if (Import.WasPressed())
    {
      Import.ResetPress();
      ProcessImportObj(o);
      sprintf(buffer, "%s", o->name);
      Scr.SetTitle(buffer);
      char *str = Scr.GetData();
      if (o->code && !n) DeCompile(o, str);
    }
    if (Export.WasPressed())
    {
      ProcessMessage("Export Object to Disk goes here. :)");
      Export.ResetPress();
    }
    if (Copy.WasPressed())
    {
      ProcessMessage("Copy Another Object goes here. :)");
      Copy.ResetPress();
    }

    if (Result==1)
    {
      int ret = CompileMem(Scr.GetData(), newcode);
      if (!ret)
      {
        ProcessMessage("Error compiling!");
        Result = 0;
      } else
      {
        if (o->code) delete[] o->code;
        o->codesize = ret;
        o->code     = new char[ret];
        o->frame    = SetTile;
        memcpy(o->code, newcode, ret);
        delete[] newcode;
      }
    }
  } while (!Result);

  CurTile = OldTile;

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessLink(void)
{
  pRoom targ;
  Rect r;
  SetupWindow();

  r.Set(260, 115);
  Dialog Win(r);
  Win.SetTitle("Room Links");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Move(Win.r.x+5, Win.r.y+25);
  r.SetLR(Win.r.x2-65, Win.r.y+35);
  Input North(r);
  North.SetTitle("North Exit:");
  if (RoomCur->LinkN->name)
    North.SetData(RoomCur->LinkN->name); else
    North.SetData("None");
  North.SetMax(RNAME_MAX);
  North.AllowSpaces(0);

  r.MoveRel(0, 20);
  Input East(r);
  East.SetTitle("East Exit:");
  if (RoomCur->LinkE->name)
    East.SetData(RoomCur->LinkE->name); else
    East.SetData("None");
  East.SetMax(RNAME_MAX);
  East.AllowSpaces(0);

  r.MoveRel(0, 20);
  Input South(r);
  South.SetTitle("South Exit:");
  if (RoomCur->LinkS->name)
    South.SetData(RoomCur->LinkS->name); else
    South.SetData("None");
  South.SetMax(RNAME_MAX);
  South.AllowSpaces(0);

  r.MoveRel(0, 20);
  Input West(r);
  West.SetTitle("West Exit:");
  if (RoomCur->LinkW->name)
    West.SetData(RoomCur->LinkW->name); else
    West.SetData("None");
  West.SetMax(RNAME_MAX);
  West.AllowSpaces(0);

  r.Move(Win.r.x2-60, Win.r.y+23);
  r.Size(55, 12);
  Button NSel(r);
  NSel.SetTitle("Select");

  r.MoveRel(0, 20);
  r.Size(55, 12);
  Button ESel(r);
  ESel.SetTitle("Select");

  r.MoveRel(0, 20);
  r.Size(55, 12);
  Button SSel(r);
  SSel.SetTitle("Select");

  r.MoveRel(0, 20);
  r.Size(55, 12);
  Button WSel(r);
  WSel.SetTitle("Select");

  Win.Add(&North);
  Win.Add(&NSel);
  Win.Add(&East);
  Win.Add(&ESel);
  Win.Add(&South);
  Win.Add(&SSel);
  Win.Add(&West);
  Win.Add(&WSel);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (NSel.WasPressed())
    {
      char *room = North.GetDataStr();
      int ret = ProcessRoomLink("North Exit", "Select the northern exit", "No rooms to link!", room);
      switch (ret)
      {
        case 0 : // one-way
          North.SetData(room);
          break;
        case 1 : // two-way
          North.SetData(room);
          targ = FindRoom(room);
          targ->LinkS = RoomCur;
          break;
        case -2 : // none
          North.SetData("None");
          break;
      }
      NSel.ResetPress();
    }
    if (ESel.WasPressed())
    {
      char *room = East.GetDataStr();
      int ret = ProcessRoomLink("East Exit", "Select the eastern exit", "No rooms to link!", room);
      switch (ret)
      {
        case 0 : // one-way
          East.SetData(room);
          break;
        case 1 : // two-way
          East.SetData(room);
          targ = FindRoom(room);
          targ->LinkW = RoomCur;
          break;
        case -2 : // none
          East.SetData("None");
          break;
      }
      ESel.ResetPress();
    }
    if (SSel.WasPressed())
    {
      char *room = South.GetDataStr();
      int ret = ProcessRoomLink("South Exit", "Select the southern exit", "No rooms to link!", room);
      switch (ret)
      {
        case 0 : // one-way
          South.SetData(room);
          break;
        case 1 : // two-way
          South.SetData(room);
          targ = FindRoom(room);
          targ->LinkN = RoomCur;
          break;
        case -2 : // none
          South.SetData("None");
          break;
      }
      SSel.ResetPress();
    }
    if (WSel.WasPressed())
    {
      char *room = West.GetDataStr();
      int ret = ProcessRoomLink("West Exit", "Select the western exit", "No rooms to link!", room);
      West.SetData(room);
      switch (ret)
      {
        case 0 : // one-way
          West.SetData(room);
          break;
        case 1 : // two-way
          West.SetData(room);
          targ = FindRoom(room);
          targ->LinkE = RoomCur;
          break;
        case -2 : // none
          West.SetData("None");
          break;
      }
      WSel.ResetPress();
    }
  } while (!Result);

  if (Result==1)
  {
    char *str;
    pRoom Cur;

    // Check North
    str = North.GetDataStr();
    Cur = FindRoom(str);
    if (stricmp(str, "none")==0)
      RoomCur->LinkN = NULL; else
      if (Cur) RoomCur->LinkN = Cur;

    // Check East
    str = East.GetDataStr();
    Cur = FindRoom(str);
    if (stricmp(str, "none")==0)
      RoomCur->LinkE = NULL; else
      if (Cur) RoomCur->LinkE = Cur;

    // Check South
    str = South.GetDataStr();
    Cur = FindRoom(str);
    if (stricmp(str, "none")==0)
      RoomCur->LinkS = NULL; else
      if (Cur) RoomCur->LinkS = Cur;

    // Check West
    str = West.GetDataStr();
    Cur = FindRoom(str);
    if (stricmp(str, "none")==0)
      RoomCur->LinkW = NULL; else
      if (Cur) RoomCur->LinkW = Cur;
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessRoomProp(void)
{
  Rect r;
  SetupWindow();

  r.Set(200, 141);
  Dialog Win(r);
  Win.SetTitle("Room Properties");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.MoveRel(65, 0);
  Button Link(r);
  Link.SetTitle("Links");

  r.Move(Win.r.x+5, Win.r.y+25);
  r.SetLR(Win.r.x2-5, Win.r.y+35);
  Input Room(r);
  Room.SetTitle("Change room keyname:");
  Room.SetData(RoomCur->name);
  Room.SetMax(RNAME_MAX);
  Room.AllowSpaces(0);

  r.MoveRel(0, 20);
  Input Desc(r);
  Desc.SetTitle("Change room description:");
  Desc.SetData(RoomCur->Desc);
  Desc.SetMax(RDESC_MAX);

  r.MoveRel(0, 20);
  r.SetLR(Win.r.x2-65, Win.r.y+75);
  Input Mod(r);
  Mod.SetTitle("Music to play on entry:");
  Mod.SetData(RoomCur->Mod);
  Mod.SetMax(FNAME_MAX);
  Mod.AllowSpaces(0);

  r.MoveRel(Mod.r.w+5, -2);
  r.Size(55, 12);
  Button Select(r);
  Select.SetTitle("Select");

  r.Move(Win.r.x+55, Win.r.y+79);
  r.SetLR(Win.r.x+80, Win.r.y+89);
  Input Width(r);
  Width.SetTitleLeft(50);
  Width.SetTitle("Room Width:");
  Width.SetData(RoomCur->RoomW);

  r.MoveRel(0, 12);
  Input Height(r);
  Height.SetTitleLeft(50);
  Height.SetTitle("Room Height:");
  Height.SetData(RoomCur->RoomH);

  r.MoveRel(0, 12);
  Input Light(r);
  Light.SetTitleLeft(50);
  Light.SetTitle("Light Level:");
  Light.SetData(RoomCur->Light);

  r.Move(Win.r.x+5, Win.r.y2-26);
  Check Ceiling(r);
  Ceiling.Add("Enable ceiling layer", RoomCur->RoomCeiling>NULL);
  Ceiling.SetTitle("");
  Ceiling.SetFont(2);

  r.Move(Win.r.x+170, Win.r.y+79);
  r.SetLR(Win.r.x+195, Win.r.y+89);
  Input Time(r);
  Time.SetTitleLeft(75);
  Time.SetTitle("Time Limit:");
  Time.SetData(RoomCur->Time);

  r.MoveRel(0, 12);
  Input Shots(r);
  Shots.SetTitleLeft(75);
  Shots.SetTitle("Maximum Shots:");
  Shots.SetData(RoomCur->MaxShots);

  r.MoveRel(0, 12);
  YesNo ReEntry(r);
  ReEntry.SetTitleLeft(75);
  ReEntry.SetTitle("Re-Enter if zapped:");
  ReEntry.SetData(RoomCur->ReEntry);

  Win.Add(&Room);
  Win.Add(&Desc);
  Win.Add(&Mod);
  Win.Add(&Select);
  Win.Add(&Width);
  Win.Add(&Height);
  Win.Add(&Light);
  Win.Add(&Time);
  Win.Add(&Shots);
  Win.Add(&ReEntry);
  Win.Add(&Ceiling);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&Link);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (Link.WasPressed())
    {
      ProcessLink();
      Link.ResetPress();
    }
    if (Select.WasPressed())
    {
      char *mod = Mod.GetDataStr();
      int ret = ProcessModSelect("Select Music",
        "Select to music to play on entry", "No music files found!", mod);

      switch (ret)
      {
        case 0 : // picked
          Mod.SetData(mod);
          break;
        case -2 : // none
          Mod.SetData("None");
          break;
        case -3 : // silence
          Mod.SetData("Silence");
          break;
      }

      Select.ResetPress();
    }
  } while (!Result);

  if (Result==1)
  {
    // update information
    strcpy(RoomCur->name, Room.GetDataStr());
    strcpy(RoomCur->Desc, Desc.GetDataStr());
    strcpy(RoomCur->Mod, Mod.GetDataStr());
    RoomCur->Light = Light.GetDataInt();
    RoomCur->Time = Time.GetDataInt();
    RoomCur->MaxShots = Shots.GetDataInt();
    RoomCur->ReEntry = ReEntry.GetDataAns();

    // these require some memory adjustment
    int changesize=0;
    int w = Width.GetDataInt();
    int h = Height.GetDataInt();

    if (w != RoomCur->RoomW || h != RoomCur->RoomH)
      changesize = 1;

    if (w <= 0 || h <= 0)
    {
      ProcessMessage("Invalid room width or height! (No changes made)");
      changesize = 0;
    } else
    {
      if (w < RoomCur->RoomW || h < RoomCur->RoomH)
        changesize = ProcessOkCan("Shrinking the room will clip the west/south edges!");
    }

    if (changesize)
    {
      short *TempFloor;
      short *TempObject;
      short *TempCeiling;
      int cw, ch;

      if (CheckFree(w*h*sizeof(short)*2))
      {
        if (w>RoomCur->RoomW)
          cw = RoomCur->RoomW; else
          cw = w;
        if (h>RoomCur->RoomH)
          ch = RoomCur->RoomH; else
          ch = h;

        // We have to check if any objects are out-of-bounds (if so, kill em)
        pObject Cur = RoomCur->ObjHead;
        while (Cur)
        {
          if (Cur->x>w || Cur->y>h) KillObj(RoomCur, Cur);
          Cur = Cur->next;
        }

        TempFloor = new short[w*h];
        TempObject = new short[w*h];
        memset16(TempFloor, 0, w*h*sizeof(short));
        memset16(TempObject, 0, w*h*sizeof(short));
        if (RoomCur->RoomCeiling)
        {
          TempCeiling = new short[w*h];
          memset16(TempCeiling, 0, w*h*sizeof(short));
        } else TempCeiling = NULL;

        for (int i=0; i<ch; i++)
        {
          memcpy(TempFloor + i*w, RoomCur->RoomFloor + i*RoomCur->RoomW, cw*2);
          memcpy(TempObject + i*w, RoomCur->RoomObject + i*RoomCur->RoomW, cw*2);
        }
        if (RoomCur->RoomCeiling)
        {
          for (int i=0; i<ch; i++)
            memcpy(TempCeiling + i*w, RoomCur->RoomCeiling + i*RoomCur->RoomW, cw*2);
        }

        if (RoomCur->RoomFloor) delete[] RoomCur->RoomFloor;
        if (RoomCur->RoomObject) delete[] RoomCur->RoomObject;
        if (RoomCur->RoomCeiling) delete[] RoomCur->RoomCeiling;
        RoomCur->RoomFloor = (short *) TempFloor;
        RoomCur->RoomObject = (short *) TempObject;
        RoomCur->RoomCeiling = (short *) TempCeiling;
        RoomCur->RoomW = w;
        RoomCur->RoomH = h;

        MapSX = MapSY = 0;             // Reset cursor position
        CurX = CurY = 0;
        MapX = MapY = 0;

        RoomData.InitMap(w, h, RoomCur->RoomFloor, RoomCur->RoomObject, RoomCur->RoomCeiling);
      } else ProcessMessage("Room dimensions too large!");
    }

    if (Ceiling.GetState(0)==1)
    {                                  // we want a ceiling
      if (RoomCur->RoomCeiling) delete[] RoomCur->RoomCeiling;
      RoomCur->RoomCeiling = new short[RoomCur->RoomW*RoomCur->RoomH];
      RoomData.InitRoof(RoomCur->RoomCeiling);
      RoomData.ClearRoof(0);
      MAXLAYER = 3;
      DrawCeiling = 1;
    } else
    {                                  // no ceiling
      if (RoomCur->RoomCeiling) delete[] RoomCur->RoomCeiling;
      RoomData.InitRoof(NULL);
      MAXLAYER = 2;
      DrawCeiling = 0;
    }
  }

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

void ProcessGlobalProp(void)
{
  Rect r;
  SetupWindow();

  r.Set(200, 95);
  Dialog Win(r);
  Win.SetTitle("Global Properties");
  Win.SetShade(Shadow);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Set(Win.r.x+5, Win.r.y+40, 100, 12);
  Button GO(r);
  GO.SetTitle("Edit Global Objects");

  r.MoveRel(0, 15);
  Button SO(r);
  SO.SetTitle("Edit System Objects");

  r.Move(Win.r.x+5, Win.r.y+25);
  r.SetLR(Win.r.x2-65, Win.r.y+35);
  Input Start(r);
  Start.SetTitle("Start room:");
  Start.SetData(RoomCur->name);
  Start.SetMax(RNAME_MAX);
  Start.AllowSpaces(0);

  r.MoveRel(Start.r.w+5, -2);
  r.Size(55, 12);
  Button Select(r);
  Select.SetTitle("Select");

  Win.Add(&Start);
  Win.Add(&Select);
  Win.Add(&GO);
  Win.Add(&SO);
  Win.Add(&Ok);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Ok.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
    if (SO.WasPressed())
    {
      SO.ResetPress();
      Result = 3;
    }
    if (GO.WasPressed())
    {
      GO.ResetPress();
      Result = 4;
    }
  } while (!Result);

  if (Result==1)
  {
    pRoom r = FindRoom(Start.GetDataStr());
    if (r) StartRoom = r;
  }
  if (Result==3) SwitchToRoom(SystemRoom);
  if (Result==4) SwitchToRoom(GlobalRoom);

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

int ProcessIns(void)
{
  Rect r;
  SetupWindow();

  r.Set(200, 40);
  Dialog Win(r);
  Win.SetTitle("Insert where?");
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Here(r);
  Here.SetTitle("Here");

  r.MoveRel(65, 0);
  Button AtEnd(r);
  AtEnd.SetTitle("At End");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  Win.Add(&Here);
  Win.Add(&AtEnd);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Here.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (AtEnd.WasPressed()) Result = 2;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 3;
  } while (!Result);

  Win.Done();
  UpdateEditor();
  mk_Clear();

  return Result;
}

int ProcessDel(void)
{
  Rect r;
  SetupWindow();

  r.Set(200, 40);
  Dialog Win(r);
  Win.SetTitle("Delete how?");
  Win.SetShade(Shadow);
  Win.SetHelp(0);

  r.Set(Win.r.x+5, Win.r.y2-15, 60, 12);
  Button Del(r);
  Del.SetTitle("Delete");

  r.MoveRel(65, 0);
  Button Erase(r);
  Erase.SetTitle("Erase");

  r.MoveRel(65, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  Win.Add(&Del);
  Win.Add(&Erase);
  Win.Add(&Can);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);
    
    Win.Process();

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }
    g_HSync();
    g_Update();

    if (Del.WasPressed() || mk_Ctrl(mkEnter)) Result = 1;
    if (Erase.WasPressed()) Result = 2;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 3;
  } while (!Result);

  Win.Done();
  UpdateEditor();
  mk_Clear();

  return Result;
}

void TileEditor(void)
{
  int OldCurTile = CurTile;
  char *TileClip;

  TileClip = new char[Tile_S];
  memset(TileClip, 0, Tile_S);

  Rect r;
  SetupWindow();

  r.Set(214, 155);
  Dialog Win(r);
  Win.SetTitle("Tile Editor");
  Win.SetShade(Shadow);
  
  r.Set(Win.r.x+5, Win.r.y2-15, 38, 12);
  Button Ok(r);
  Ok.SetTitle("Ok");

  r.MoveRel(41, 0);
  Button Can(r);
  Can.SetTitle("Cancel");

  r.Set(Win.r.x2-54, Win.r.y2-30, 49, 12);
  Button Import(r);
  Import.SetTitle("Import");

  r.MoveRel(0, 15);
  Button Export(r);
  Export.SetTitle("Export");

  r.Set(Win.r.x+5, Win.r.y+25, 96, 96);
  TileEdit Tile(r);
  Tile.SetTile(CurTile);

  r.Set(Win.r.x+110, Win.r.y+25, 99, 25+10);
  Palette Pal(r);
  Pal.SetTitle("Palette");

  r.MoveRel(50, 47);
  r.Size(49, 49);
  TileView Preview(r);
  Preview.SetTitle("Preview");
  Preview.SetTileLib(Tile.GetTileLib());
  Preview.SetTile(CurTile);

  char iconUp[2]     = {24, 0};
  char iconDown[2]   = {25, 0};
  char iconLeft[2]   = {27, 0};
  char iconRight[2]  = {26, 0};
  char iconFlip[2]   = {19, 0};
  char iconMirror[2] = {20, 0};
  char iconRLeft[2]  = {21, 0};
  char iconRRight[2] = {22, 0};

  r.Set(Preview.r.x-48, Preview.r.y-5, 12, 12);
  ToolButton UL(r);
  UL.SetTitle(iconFlip);

  r.MoveRel(14, 0);
  ToolButton U(r);
  U.SetTitle(iconUp);

  r.MoveRel(14, 0);
  ToolButton UR(r);
  UR.SetTitle(iconMirror);

  r.Set(Preview.r.x-48, Preview.r.y+9, 12, 12);
  ToolButton L(r);
  L.SetTitle(iconLeft);

  r.MoveRel(14, 0);
  ToolButton M(r);
  M.SetTitle("5");

  r.MoveRel(14, 0);
  ToolButton R(r);
  R.SetTitle(iconRight);

  r.Set(Preview.r.x-48, Preview.r.y+23, 12, 12);
  ToolButton DL(r);
  DL.SetTitle(iconRLeft);

  r.MoveRel(14, 0);
  ToolButton D(r);
  D.SetTitle(iconDown);

  r.MoveRel(14, 0);
  ToolButton DR(r);
  DR.SetTitle(iconRRight);

  r.Set(Preview.r.x-48, Preview.r.y+37, 40, 12);
  Button Lock(r);
  Lock.SetTitle("Lock");

  r.Set(Win.r.x+5, Win.r.y2-30, 20, 12);
  Button Prev(r);
  Prev.SetTitle("");

  r.MoveRel(25, 0);
  Button Next(r);
  Next.SetTitle("");

  r.MoveRel(25, 0);
  r.Size(30, 12);
  Button Ins(r);
  Ins.SetTitle("Ins");

  r.MoveRel(35, 0);
  Button Del(r);
  Del.SetTitle("Del");

  r.MoveRel(0, 15);
  Button Grid(r);
  Grid.SetTitle("Grid");

  r.MoveRel(35, -15);
  Button Copy(r);
  Copy.SetTitle("Copy");

  r.MoveRel(0, 15);
  Button Paste(r);
  Paste.SetTitle("Paste");

  Win.Add(&Tile);
  Win.Add(&Pal);
  Win.Add(&Preview);
  Win.Add(&UL);
  Win.Add(&U);
  Win.Add(&UR);
  Win.Add(&L);
  Win.Add(&M);
  Win.Add(&R);
  Win.Add(&DL);
  Win.Add(&D);
  Win.Add(&DR);
  Win.Add(&Lock);
  Win.Add(&Prev);
  Win.Add(&Next);
  Win.Add(&Ins);
  Win.Add(&Del);
  Win.Add(&Copy);
  Win.Add(&Import);
  Win.Add(&Ok);
  Win.Add(&Can);
  Win.Add(&Grid);
  Win.Add(&Paste);
  Win.Add(&Export);

  int Result=0;

  do {
    g_SetCurPage(1);
    g_CopyPage(SpriPage, 1);

    if (mk_Shift(mkUp))    Tile.ShiftUp();
    if (mk_Shift(mkDown))  Tile.ShiftDown();
    if (mk_Shift(mkLeft))  Tile.ShiftLeft();
    if (mk_Shift(mkRight)) Tile.ShiftRight();

    if (CurTile != Tile.GetTile())
    {
      CurTile = Tile.GetTile();
      Preview.SetTile(CurTile);
    }
    
    Win.Process();

    if (Pal.Updated()) Tile.SetColor(Pal.GetLeft(), Pal.GetRight());
    if (Pal.GetLeft() != Tile.GetLeftG()) Pal.SetLeft(Tile.GetLeftG());
    if (Pal.GetRight() != Tile.GetRightG()) Pal.SetRight(Tile.GetRightG());

    if (U.WasPressed())
    {
      U.ResetPress();
      Tile.ShiftUp();
    }
    if (D.WasPressed())
    {
      D.ResetPress();
      Tile.ShiftDown();
    }
    if (L.WasPressed())
    {
      L.ResetPress();
      Tile.ShiftLeft();
    }
    if (R.WasPressed())
    {
      R.ResetPress();
      Tile.ShiftRight();
    }

    if (M.WasPressed())
    {
      M.ResetPress();
      ProcessMessage("This button for rent.");
    }
    if (UL.WasPressed())
    {
      UL.ResetPress();
      Tile.Flip();
    }
    if (UR.WasPressed())
    {
      UR.ResetPress();
      Tile.Mirror();
    }
    if (DL.WasPressed())
    {
      DL.ResetPress();
      Tile.RotateR();
    }
    if (DR.WasPressed())
    {
      DR.ResetPress();
      Tile.RotateL();
    }

    if (Next.WasPressed())
    {
      Next.ResetPress();
      if (CurTile<TotTiles)
      {
        Tile.SetTile(++CurTile);
        Preview.SetTile(CurTile);
      }
    }
    if (Prev.WasPressed())
    {
      Prev.ResetPress();
      if (CurTile>0)
      {
        Tile.SetTile(--CurTile);
        Preview.SetTile(CurTile);
      }
    }
    if (Import.WasPressed())
    {
      Import.ResetPress();
      ProcessImportTile(Tile.backup);
    }
    if (Export.WasPressed())
    {
      Export.ResetPress();
      ProcessExportTile(Tile.backup);
    }
    if (Grid.WasPressed())
    {
      Grid.ResetPress();
      if (Tile.GetGrid())
        Tile.SetGrid(0); else
        Tile.SetGrid(1);
    }
    if (Lock.WasPressed())
    {
      Lock.ResetPress();
      Preview.ToggleLock();
      if (Preview.GetLock())
        Lock.SetTitle("Unlock"); else
        Lock.SetTitle("Lock");
    }
    if (Copy.WasPressed())
    {
      Copy.ResetPress();
      Tile.GetTileData(CurTile, TileClip);
    }
    if (Paste.WasPressed())
    {
      Paste.ResetPress();
      Tile.SetTileData(CurTile, TileClip);
    }
    if (Ins.WasPressed())
    {
      Ins.ResetPress();
      switch (ProcessIns())
      {
        case 1 :
          Tile.InsTile(CurTile);
          Tile.SetTile(CurTile);
          break;
        case 2 :
          Tile.InsTile(TotTiles);
          Tile.SetTile(CurTile);
          break;
      }
    }
    if (Del.WasPressed())
    {
      Del.ResetPress();
      switch (ProcessDel())
      {
        case 1 :
          Tile.DelTile(CurTile);
          Tile.SetTile(CurTile);
          break;
        case 2 :
          Tile.EraseTile(CurTile);
          break;
      }
    }

    if (mk_Ctrl())
    {
      Panel(ST_TILESEL);
      Tile.SetTile(CurTile);
      Preview.SetTile(CurTile);
    }

    if (!nomouse)
    {
      m_Update();
      s_Update(0);
    }

    g_HSync();
    g_Update();

    if (Ok.WasPressed()) Result = 1;
    if (Can.WasPressed() || mk_Pressed(mkEsc)) Result = 2;
  } while (!Result);

  if (Result==1)
  {
    Tile.Accept();
    CurTile = 0;
  } else CurTile = OldCurTile;

  delete[] TileClip;

  Win.Done();
  UpdateEditor();
  mk_Clear();
}

short Edit(pRoom r)
{
  if (r) SwitchToRoom(r);
  
  UpdateEditor();
  zzt_Fade(Black, zztPal);

  quiteditor = 0;
  ineditor   = 1;
  do {
    UpdateEditor();

    State = ST_NORMAL;
    if (mk_Ctrl()) State = ST_TILESEL;
    if (mk_Alt()) State = ST_MOUSING;
    if (mk_Shift() || m_Middle()) State = ST_BLOCKING;
    if (mk_Pressed(mkTab)) PaintMode = !PaintMode;

    if (mk_Alt(mkH))                   // help screen
    {
      SendHelpMessage("EditorHelp");
      UpdateObj(Z2Help);
    }

    if (mk_Alt(mkK))                   // hotkey help
    {
      SendHelpMessage("HotkeyHelp");
      UpdateObj(Z2Help);
    }

    if (mk_Alt(mkX))                   // export a room
    {
      mk_Clear();
      ProcessExportRoom();
    }

    if (mk_Alt(mkI))                   // import a room
    {
      mk_Clear();
      ProcessImportRoom();
    }

    if (mk_Alt(mkL))                   // load a world
    {
      mk_Clear();
      ProcessLoadWorld();
    }

    if (mk_Alt(mkS))                   // save a world
    {
      mk_Clear();
      ProcessSaveWorld();
    }

    if (mk_Alt(mkT))                   // test a world
    {
      mk_Clear();

      // get rid of fade temp.
      int oldfade = Fade;
      Fade = 0;

      // save world and play it
      SaveWorld("$Z2$TEMP.$$$", "Temporary Z2 World");
      short RoomHash = Play(RoomCur);

      // load a fresh copy
      LoadWorld("$Z2$TEMP.$$$");
      remove("$Z2$TEMP.$$$");

      // return to editor, restore fade
      SwitchToRoom(FindRoom(RoomHash));
      Fade = oldfade;
    }

    if (mk_Alt(mkG))          // global properties
    {
      mk_Clear();
      ProcessGlobalProp();
    }

    if (mk_Alt(mkN))          // clear room
    {
      mk_Clear();
      if (ProcessYesNo("Clear out the room?")) DefaultRoom(RoomCur);
    }

    if (mk_Pressed(mkF))               // draw floor layer
      DrawFloor = 1-DrawFloor;

    if (mk_Pressed(mkO))               // draw object layer
      DrawObject = 1-DrawObject;

    if (mk_Pressed(mkC))               // draw ceiling layer
      if (RoomCur->RoomCeiling) DrawCeiling = 1-DrawCeiling;

    if (mk_Pressed(mkG))      // tile editor
    {
      mk_Clear();
      TileEditor();
    }

    if (mk_Pressed(mkZ))      // changes room
    {
      mk_Clear();
      ProcessChangeRoom();
    }

    if (mk_Pressed(mkA))      // add a room
    {
      mk_Clear();
      ProcessAddRoom();
    }

    if (mk_Pressed(mkE))      // room links
    {
      mk_Clear();
      ProcessLink();
    }

    if (mk_Pressed(mkR))      // room properties
    {
      mk_Clear();
      ProcessRoomProp();
    }

    if (mk_Alt(mkIns))                           // transparent paste block
    {
      mk_Clear();
      TransPasteClip();
    }
    if (mk_Pressed(mkHome) || mk_Ctrl(mkC))      // copy block
    {
      mk_Clear();
      CopyBlock();
    }
    if (mk_Pressed(mkDel) || mk_Ctrl(mkX))       // cut block
    {
      mk_Clear();
      CutBlock();
    }
    if (mk_Pressed(mkIns) || mk_Ctrl(mkV))       // paste block
    {
      mk_Clear();
      PasteClip();
    }

    if (mk_Pressed(mkX))               // fill
    {
      SetPtr();
      switch (EditLayer)
      {
        case L_FLOOR  :
          SeedVal = RoomCur->RoomFloor[CurY*rw+CurX];
          break;
        case L_OBJECT :
          SeedVal = RoomCur->RoomObject[CurY*rw+CurX];
          break;
        case L_CEILING :
          SeedVal = RoomCur->RoomCeiling[CurY*rw+CurX];
          break;
      }

      if (SeedVal != CurTile) Fill(CurX, CurY, 1, CurX, CurX);

      // Convert objects
      int x, y;
      for (x=0; x<rw; x++)
      {
        for (y=0; y<RoomCur->RoomH; y++)
        {
          int TestTile = RoomCur->RoomObject[y*RoomCur->RoomW+x];
          if (TestTile>TILE_MAX)
          {
            RoomCur->RoomObject[y*RoomCur->RoomW+x] = 0;
            AddObjXY(x, y, ItemList[TestTile-TILE_MAX].name, RoomCur, EditLayer);
          }
        }
      }
    }

    if (mk_Pressed(mkL))               // change layer
    {
      EditLayer++;
      if (EditLayer>=MAXLAYER) EditLayer = 0;
      switch (EditLayer)
      {
        case L_FLOOR  : DrawFloor = 1; break;
        case L_OBJECT : DrawObject = 1; break;
        case L_CEILING : DrawCeiling = 1; break;
      }
    }

    for (int i=0; i<12; i++)
    {
      if (mk_Pressed(mkF1+i))
      {
        mk_Clear();
        int ret = ProcessCatList(i+1);
        if (ret != -1)
        {
          CurTile = ret;
          EditLayer = L_OBJECT;
        }
      }
    }

    if (mk_Pressed(mkEsc))             // quit editor
      quiteditor = 1;
  } while (!quiteditor);
  mk_Clear();

  ineditor = 0;

  zzt_Fade(zztPal, Black);

  return RoomCur->NameHash;
}
