// Font Manager, version 2.0 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version
// Version 2.0 -- Implemented object-oriented design; Separated font-
//                loading and displaying into separate classes; A font
//                registry now keeps all fonts in a central location,
//                which the text display routines tap into
// Version 2.1 -- Added TextBox, for clipped, wrapped text in a
//                rectangle
// Version 2.2 -- Added g_Font(), g_FontColor(), and g_printf() for quick
//                memory-efficient font displaying
// Version 2.3 -- Added special codes: '' for a tab, and '' for a color
//                change.
// Version 2.5 -- Added multi-color font support (good for antialiasing)
//                Two versions: MULTI_COLOR, FULL_COLOR
//                MULTI_COLOR: Allows simple shading/aliasing
//                FULL_COLOR:  Full-color fonts, no color changing

#ifndef __FONT__
#define __FONT__

#include "vgl.h"

#define TABSTOP 16
#define COLORCH 28

#define NO_FONT     0
#define MONO_COLOR  1
#define MULTI_COLOR 2
#define FULL_COLOR  3

class FontManager {
  int TotalFonts;

public:
  BITMAP Fonts[16][256];
  int    FontType[16];

  FontManager();
  ~FontManager();

  void LoadFont(char *f, int id, int type=MONO_COLOR);
  void KillFont(int id);

  int  SetFontSize(int size);
  int  CharWidth(int id, int ch)     { return (ch != TABSTOP && ch != COLORCH) ? Fonts[id][ch].x : 0; }
  int  CharHeight(int id, int ch)    { return Fonts[id][ch].y; }
  int  FontWidth(int id)             { return Fonts[id]['m'].x; }
  int  FontHeight(int id)            { return Fonts[id]['j'].y; }
};

extern FontManager FontMan;

class TextLine {
protected:
  int CharSpacing;
  int FontID;
  char *Text;
  int BackClear;
  int TabStop;

public:
  TextLine(int id=0);
  ~TextLine();

  void SetSpacing(int sp)              { CharSpacing = sp; }
  int  GetSpacing(void)                { return CharSpacing; }
  void SetFontID(int id)               { FontID = id; }
  int  GetFontID(void)                 { return FontID; }
  void SetClear(int c)                 { BackClear = c; }
  int  GetClear(void)                  { return BackClear; }
  void SetTabs(int t)                  { TabStop = t; }
  int  GetTabs(void)                   { return TabStop; }

  int  StrLen(void);
  void SetText(char *fmt, ...);
  void SetPage(VirtPage &out);
  void Print(int x, int y, int c);
  void cenPrint(int y, int c);
  void btwPrint(int x1, int x2, int y, int c);
  void sPrint(int x, int y, int c);
  void scenPrint(int y, int c);
  void sbtwPrint(int x1, int x2, int y, int c);
};

class TextBox {
protected:
  int CharSpacing;
  int FontID;
  int OCX1, OCY1;
  int OCX2, OCY2;
  int CX1, CY1;
  int CX2, CY2;
  int Width, Height;
  int TabStop;

public:
  char *Text;

  TextBox();
  TextBox(int x1, int y1, int x2, int y2, int id=0);
  ~TextBox();

  void SetClip(int x1, int y1, int x2, int y2);
  void SetSpacing(int sp)              { CharSpacing = sp; }
  int  GetSpacing(void)                { return CharSpacing; }
  void SetFontID(int id)               { FontID = id; }
  int  GetFontID(void)                 { return FontID; }
  int  GetHeight(void)                 { return Height; }
  void SetTabs(int t)                  { TabStop = t; }
  int  GetTabs(void)                   { return TabStop; }

  int  LineHeight(char *text);
  void SetText(char *fmt, ...);
  void SetPage(VirtPage &out);
  void Print(int dx, int c);
  void sPrint(int dx, int c);
};

void g_Font(int fontid);
void g_FontColor(int c);
int  g_GetFont(void);
int  g_GetFontColor(void);
int  g_GetFontHeight(void);
void g_gotoxy(int x, int y);
int  g_FontX(void);
int  g_FontY(void);
void g_SetTabs(int t);
int  g_GetTabs(void);

int  g_strlen(char *str);
int  g_mstrlen(char *str);

void g_printf(int x, int y, char *fmt, ...);
void g_sprintf(int x, int y, char *fmt, ...);
void g_printf(char *fmt, ...);
void g_sprintf(char *fmt, ...);
void g_mprintf(int x, int y, char *fmt, ...);
void g_msprintf(int x, int y, char *fmt, ...);
void g_mprintf(char *fmt, ...);
void g_msprintf(char *fmt, ...);

void g_cenprintf(char *fmt, ...);
void g_scenprintf(char *fmt, ...);
void g_cenprintf(int y, char *fmt, ...);
void g_scenprintf(int y, char *fmt, ...);
void g_mcenprintf(char *fmt, ...);
void g_mscenprintf(char *fmt, ...);
void g_mcenprintf(int y, char *fmt, ...);
void g_mscenprintf(int y, char *fmt, ...);

void g_btwprintf(int x1, int x2, int y, char *fmt, ...);
void g_sbtwprintf(int x1, int x2, int y, char *fmt, ...);
void g_mbtwprintf(int x1, int x2, int y, char *fmt, ...);
void g_msbtwprintf(int x1, int x2, int y, char *fmt, ...);

#endif
