#include <stdio.h>
#include <i86.h>

#include "vgl.h"
#include "pal.h"

#include "compiler.h"
#include "objects.h"
#include "editor.h"
#include "global.h"
#include "sound.h"

char ZZTVersion[]  = "ZZT2";
char WorldVer[]    = "Z2W4";
char PWorldVer[]   = "Z2P4";
char SavedVer[]    = "Z2S4";
char RoomNULL[]    = "RN42";
char RoomVer[]     = "ZRC4";
char ZZTModule[]   = "ZMOD";

PALETTE zztPal;
INDEX WinDark, PalPurp;
int colWhite, colLGray, colDGray, colBlack;
int colPurple, colLPurp, colDPurp, colYellow;
int SwapPage, ConsPage, SpriPage, FreeMem;
int Fade=1, OldScroll=0, Shadow=1, consolemode=0;
int totalitems=0, totalsounds=0, ErrorLog=0;
int Tile_W, Tile_H, Tile_S;
char DefaultMOD[80];

int CoreFree(void)
{
  struct tMemInfo {
    unsigned LargestBlockAvail;
    unsigned MaxUnlockedPage;
    unsigned LargestLockablePage;
    unsigned LinAddrSpace;
    unsigned NumFreePagesAvail;
    unsigned NumPhysicalPagesFree;
    unsigned TotalPhysicalPages;
    unsigned FreeLinAddrSpace;
    unsigned SizeOfPageFile;
    unsigned Reserved[3];
  } MemInfo;

  union REGS r;
  struct SREGS s;

  r.x.eax = 0x0500;
  memset(&s, 0, sizeof(s));
  s.es = FP_SEG(&MemInfo);
  r.x.edi = FP_OFF(&MemInfo);
  int386x(0x31, &r, &r, &s);

  return MemInfo.LargestBlockAvail;
}

int CheckFree(int size)
{
  return (size<CoreFree()-(2*1024*1024));
}

void ReadCFG(void)
{
  FILE *cfg;

  if ((cfg = fopen("z2.cfg", "rt"))==NULL)
  {
    g_Done();
    printf("Z2: Error loading configuration file.\n");
    exit(1);
  }

  // first line is def. file
  fscanf(cfg, "%s\n", DefFile);

  // second line is mod file to load
  fscanf(cfg, "%s\n", DefaultMOD);

  // third line is options
  fscanf(cfg, "%d %d %d %d\n", &Fade, &OldScroll, &Shadow, &consolemode);

  // fourth line is sound data
  fscanf(cfg, "%d %d %d %d %d\n", &zm_MixFreq, &zm_Bits, &zm_Mono, &zm_Interp, &zm_Device);

  fclose(cfg);
}

void WriteCFG(void)
{
  FILE *cfg;

  if ((cfg = fopen("z2.cfg", "wt"))==NULL)
  {
    g_Done();
    printf("Z2: Error loading configuration file.\n");
    exit(1);
  }

  // first line is def. file
  fprintf(cfg, DefFile);

  // second line is mod file to load
  fprintf(cfg, DefaultMOD);

  // third line is options
  fprintf(cfg, "%d %d %d %d", Fade, OldScroll, Shadow, consolemode);

  // fourth line is sound data
  fprintf(cfg, "%d %d %d %d %d", zm_MixFreq, zm_Bits, zm_Mono, zm_Interp, zm_Device);

  fclose(cfg);
}

void ReadDEF(char *df)
{
  int totline, itemid;
  int cat=1;
  int totvars;
  short varval;
  char *line, ch;
  char str[80];

  ItemList = NULL;
  itemid = 0;
  totalitems = 0;

  line = new char[1024];
  if (!line)
  {
    g_Done();
    printf("Z2: Error allocating line buffer.\n");
    exit(1);
  }

  ifstream def(df);
  if (!def)
  {
    g_Done();
    printf("Z2: Error loading definition file %s.\n", df);
    exit(1);
  }

  // Clear out old items/tiles/variables
  if (TotTiles) KillTiles();
  if (ItemList) delete[] ItemList;

  // First run looks for defname, deftile, categories, item lists
  for (totline=0; def.getline(line, 1024); totline++)
  {
    // skip comments
    if (line[0]==';') continue;

    if (strnicmp(line, "defname", 7)==0)
    {
      strcpy(DefName, line+8);
    }
    if (strnicmp(line, "deftile", 7)==0)
    {
      strcpy(TileFile, line+8);
    }
    if (strnicmp(line, "sysroom", 7)==0)
    {
      strcpy(SysRoom, line+8);
      LoadRoom(SysRoom, SystemRoom);
    }
    if (stricmp(line, "[soundfx]")==0)
    {
      char name[21];
      char wavf[13];
      do {
        def.getline(line, 1024);
        if (line[0] != '-')
        {
          totalsounds++;
          totline++;
          strncpy(name, strtok(line, ","), 20);
          strncpy(wavf, strtok(NULL, " "), 12);
          zm_wavAdd(wavf, name);
        }
      } while (line[0] != '-');
    }
    if (stricmp(line, "[categories]")==0)
    {
      def.getline(line, 1024);
      totline++;
      ItemNum = strtol(line, NULL, 10);
      def.getline(line, 1024);
      totline++;
      CatNum = strtol(line, NULL, 10);

      ItemList = new tItem[ItemNum];
      if (!ItemList)
      {
        g_Done();
        printf("Z2: Error allocating item list.\n");
        exit(1);
      }

      strcpy(CatList[0], "None");
      for (int i=1; i<=CatNum; i++, totline++)
      {
        def.getline(line, 1024);
        CatList[i] = new char[40];
        strncpy(CatList[i], line, 40);
      }
    }

    if (CatList)
    {
      strcpy(str, "[");
      strcat(str, CatList[cat]);
      strcat(str, "]");
      if (stricmp(line, str)==0)
      {
        do {
          def.getline(line, 1024);
          if (line[0] != '-')
          {
            strncpy(ItemList[itemid].name, strtok(line, ","), ITEMSTR_MAX);
            char *str = strtok(NULL, ",");
            while (*str==32) str++;
            strncpy(ItemList[itemid].desc, str, ITEMSTR_MAX);
            ItemList[itemid].frame = strtol(strtok(NULL, ","), NULL, 10);
            ItemList[itemid].note  = strtol(strtok(NULL, ","), NULL, 10);
            strncpy(ItemList[itemid].flag, strtok(NULL, " "), 16);

            ItemList[itemid].NameHash = Hash(ItemList[itemid].name);
            ItemList[itemid].cat = cat;
            ItemList[itemid].id = itemid;
            
            AddItemObj(ItemList[itemid], SystemRoom->ObjHead);
            totalitems++;
            itemid++;
          }
        } while (line[0] != '-');
        cat++;
      }
    }
    if (stricmp(line, "[endlist]")==0) break;
  }

  def.close();
  delete[] line;
}

void zzt_Fade(PALETTE sp, PALETTE dp, int f, int l)
{
  if (Fade)
  {
    for (int i=0; i<=Speed; i++)
    {
      pal_SetIntensity(sp, dp, i, f, l);
      pal_Set(temp);
    }
  }
}

void zzt_ForceFade(PALETTE sp, PALETTE dp, int f, int l)
{
  for (int i=0; i<=Speed; i++)
  {
    pal_SetIntensity(sp, dp, i, f, l);
    pal_Set(temp);
  }
}

char TextXOR = 0x69;

void xor(char *str)
{
/*
  while (*str)
  {
    *str = *str ^ TextXOR;
    str++;
  }
*/
}

void Z2TileMap::DrawBase(void)
{
  int MapX, MapY;
  int ScrX, ScrY;
  int MapX1, MapY1;
  int MapX2, MapY2;
  int TileDX, TileDY;
  int TScrX, TScrY;
  int TileOffset;
  char *ScrnOffset;
  short TilePlot;
  short *tile;

  MapX1  = PosX / Tile_W;
  MapX2  = MapX1 + MapDX;
  TileDX = PosX % Tile_W;
  MapY1  = PosY / Tile_H;
  MapY2  = MapY1 + MapDY;
  TileDY = PosY % Tile_H;

  for (MapY=MapY1, ScrY = 0; MapY<MapY2; MapY++, ScrY += Tile_H)
  {
    TileOffset = MapOffset[MapY];
    TScrY      = ScrY-TileDY;
    ScrnOffset = Pages[CP].Screen + (TScrY << 6) + (TScrY << 8);
    tile       = Base + TileOffset+MapX1;

    if (MapY >= 0 && MapY < MapH)
      for (MapX=MapX1, ScrX = 0; MapX<MapX2; MapX++, ScrX += Tile_W, tile++)
      {
        if (MapX>=MapW) continue;
        TScrX = ScrX-TileDX;

        if (*tile<TILE_MAX)
          TilePlot = *tile; else
          TilePlot = 5;
        
        if (MapX >= 0 && MapX < MapW && Base[TileOffset+MapX])
        {
          if (TScrX>=0 && TScrX<_BlitMaxX && TScrY>=0 && TScrY<_BlitMaxY)
            g_Tile(ScrnOffset+TScrX, Tiles[TilePlot].Data); else
            g_cBitmap(TScrX, TScrY, &Tiles[TilePlot]);
        }
      }
  }
}

void Z2TileMap::DrawFringe(void)
{
  int MapX, MapY;
  int ScrX, ScrY;
  int MapX1, MapY1;
  int MapX2, MapY2;
  int TileDX, TileDY;
  int TScrX, TScrY;
  int TileOffset;
  char *ScrnOffset;
  short TilePlot;
  short *tile;

  MapX1  = PosX / Tile_W;
  MapX2  = MapX1 + MapDX;
  TileDX = PosX % Tile_W;
  MapY1  = PosY / Tile_H;
  MapY2  = MapY1 + MapDY;
  TileDY = PosY % Tile_H;

  for (MapY=MapY1, ScrY = 0; MapY<MapY2; MapY++, ScrY += Tile_H)
  {
    TileOffset = MapOffset[MapY];
    TScrY      = ScrY-TileDY;
    ScrnOffset = Pages[CP].Screen + (TScrY << 6) + (TScrY << 8);
    tile       = Fringe + TileOffset+MapX1;

    if (MapY >= 0 && MapY < MapH)
      for (MapX=MapX1, ScrX = 0; MapX<MapX2; MapX++, ScrX += Tile_W, tile++)
      {
        if (MapX>=MapW) continue;
        TScrX = ScrX-TileDX;

        if (*tile<TILE_MAX)
          TilePlot = *tile; else
          TilePlot = 5;

        if (MapX >= 0 && MapX < MapW && Fringe[TileOffset+MapX])
        {
          if (TScrX>=0 && TScrX<_BlitMaxX && TScrY>=0 && TScrY<_BlitMaxY)
            g_tTile(ScrnOffset+TScrX, Tiles[TilePlot].Data); else
            g_ctBitmap(TScrX, TScrY, &Tiles[TilePlot]);
        }
      }
  }
}

void Z2TileMap::DrawRoof(void)
{
  int MapX, MapY;
  int ScrX, ScrY;
  int MapX1, MapY1;
  int MapX2, MapY2;
  int TileDX, TileDY;
  int TScrX, TScrY;
  int TileOffset;
  char *ScrnOffset;
  short TilePlot;
  short *tile;

  MapX1  = PosX / Tile_W;
  MapX2  = MapX1 + MapDX;
  TileDX = PosX % Tile_W;
  MapY1  = PosY / Tile_H;
  MapY2  = MapY1 + MapDY;
  TileDY = PosY % Tile_H;

  for (MapY=MapY1, ScrY = 0; MapY<MapY2; MapY++, ScrY += Tile_H)
  {
    TileOffset = MapOffset[MapY];
    TScrY      = ScrY-TileDY;
    ScrnOffset = Pages[CP].Screen + (TScrY << 6) + (TScrY << 8);
    tile       = Roof + TileOffset+MapX1;

    if (MapY >= 0 && MapY < MapH)
      for (MapX=MapX1, ScrX = 0; MapX<MapX2; MapX++, ScrX += Tile_W, tile++)
      {
        if (MapX>=MapW) continue;
        TScrX = ScrX-TileDX;
        
        if (*tile<TILE_MAX)
          TilePlot = *tile; else
          TilePlot = 5;

        if (MapX >= 0 && MapX < MapW && Roof[TileOffset+MapX])
        {
          if (TScrX>=0 && TScrX<_BlitMaxX && TScrY>=0 && TScrY<_BlitMaxY)
            g_tTile(ScrnOffset+TScrX, Tiles[TilePlot].Data); else
            g_ctBitmap(TScrX, TScrY, &Tiles[TilePlot]);
        }
      }
  }
}

/*
BITMAP *DAT2VSP(char *data)
{
  BITMAP *temp = new BITMAP;

  memcpy(&temp->x, data, sizeof(short));
  memcpy(&temp->y, data + 2, sizeof(short));
  temp->Data = data + 4;

  return temp;
}
*/

void AutoAppend(char *filename, char *e)
{
  char name[_MAX_FNAME];
  char ext[_MAX_EXT];

  _splitpath(filename, NULL, NULL, name, ext);
  if (!strlen(ext)) strcpy(ext, ".z2");
  _makepath(filename, NULL, NULL, name, ext);
}

int DetectWindows(void)
{
  union REGS r;

  r.w.ax = 0x1600;
  int386(0x2F, &r, &r);
  return r.h.al;
}
