#ifndef __ZZT_GLOBAL__
#define __ZZT_GLOBAL__

#define Z2VERSION "0.62"

#define EMPTYARRAY
#define NOEMS
//#define NOMIDAS
//#define NOEDIT

#include <fstream.h>
#include "vgl.h"
#include "pal.h"
#include "tile.h"
#include "font.h"

#define RE_EXEC     1
#define SUCCESS     0
#define BAD_OBJ     -1
#define BAD_MSG     -2
#define BAD_ROOM    -3
#define BAD_VAR     -4
#define BAD_SOUND   -5
#define NO_MEM      -6
#define FILE_ERR    -7
#define UNK_ERR     -10

#define LABEL_MAX   16
#define OBJNAME_MAX 32
#define VAR_MAX     16
#define FILE_MAX    8
#define FNAME_MAX   13
#define BARK_MAX    50
#define ITEMSTR_MAX 40
#define RNAME_MAX   32
#define RDESC_MAX   80
#define INTER_MAX   512

#define LOCAL_MAX    5
#define TILE_MAX     16384
#define OBJ_MAX      32768
#define SCRATCH_SIZE 64*1024

#define uint   unsigned int
#define ushort unsigned short
#define uchar  unsigned char
#define ulong  unsigned long

#define u16 unsigned short
#define u8  unsigned char
#define u32 unsigned long

/*
objFlight
objItem
objInvisible
objPushP
objPushNS
objPushEW
objPusher
objSwimmer
objBreakable
objSolid
objPlayer
objBound
objLocked
*/

#define SN_NONE    0
#define SN_ROOM    1
#define SN_DIALOG  2
#define SN_COORD   3
#define SN_SCROLL  4
#define SN_OBJECT  5
#define SN_WATER   6
#define SN_FOREST  7
#define SN_FAKE    8
#define SN_PLAYER  9
#define SN_GENERIC 10

extern char ZZTVersion[];
extern char WorldVer[];
extern char SavedVer[];
extern char RoomNULL[];
extern char RoomVer[];
extern char ZZTModule[];
#define BindCode    ""
#define FirstOp     0xEF
#define ObjNameOp   0xEF
#define MoveOp      0xF0
#define PersgoOp    0xF1
#define CamOp       0xF2
#define CommandOp   0xF3
#define LabelOp     0xF4
#define ZapLabelOp  0xF5
#define CommentOp   0xF6
#define VarOp       '!'
#define InterOp     0xF8
#define BarkOp      0xF9
#define PBarkOp     0xFF
#define MessageOp   0xFA
#define MSMessageOp 0xFB
#define CenTextOp   0xFC
#define TextOp      0xFD
#define BreakOp     0xFE

extern PALETTE zztPal;
extern INDEX WinDark, PalPurp;
extern int colWhite, colLGray, colDGray, colBlack;
extern int colPurple, colLPurp, colDPurp, colYellow;
extern int SwapPage, ConsPage, SpriPage, FreeMem;
extern int Fade, OldScroll, Shadow;
extern int totalitems, totalsounds;
extern int consolemode, ErrorLog;
extern int Tile_W, Tile_H, Tile_S;
extern char DefaultMOD[80];
extern ofstream log;
static unsigned z2sync;

typedef struct tItem {
  char  name[ITEMSTR_MAX];
  char  desc[ITEMSTR_MAX];
  short NameHash;
  short frame, id;
  char  cat, cycle;
  char  flag[16];
  char  note;
  char  str[ITEMSTR_MAX];
} *pItem;

typedef struct tObject {
  char   name[OBJNAME_MAX];  // Name of object (used for referencing)
  short  NameHash;
  char   bark[BARK_MAX];     // Object barking
  short  barktime;           // Object bark time
  ushort x, y, z;            // Object position
  short  objFlight    : 1;
  short  objItem      : 1;
  short  objInvisible : 1;
  short  objPushP     : 1;
  short  objPushNS    : 1;
  short  objPushEW    : 1;
  short  objPusher    : 1;
  short  objSwimmer   : 1;
  short  objBreakable : 1;
  short  objSolid     : 1;
  short  objPlayer    : 1;
  short  objBound     : 1;
  short  objLocked    : 1;
  ushort frame;              // Current frame
  char   local[LOCAL_MAX];   // Local stack
  char   *code;              // Object code
  ushort codesize;           // Code size
  ushort codep;              // Object code pointer
  char   cycle, cyclel;      // Object cycle (instructions per cycle)
  char   light;              // Object "Glow"
  char   queue, qsize;       // movement queue
  struct tObject *next;
} *pObject;

typedef struct tRoom {
  char   name[RNAME_MAX];    // Name of room
  char   Desc[RDESC_MAX];    // Description of room
  short  NameHash;
  char   Mod[FNAME_MAX];     // Module to play
  ushort RoomW, RoomH;       // Room width and height
  short  *RoomFloor;         // Room data -- Floor
  short  *RoomObject;        // Room data -- Objects
  short  *RoomCeiling;       // Room data -- Ceiling
  struct tRoom *LinkN;       // Connecting rooms
  struct tRoom *LinkS;       
  struct tRoom *LinkE;
  struct tRoom *LinkW;
  char   StrLinkN[RNAME_MAX];
  char   StrLinkS[RNAME_MAX];
  char   StrLinkE[RNAME_MAX];
  char   StrLinkW[RNAME_MAX];
  uchar  Light;              // Light level
  char   ReEntry;            // Re-enter when zapped?
  ushort Time;               // Time limit (0 = none)
  ushort MaxShots;           // Maximum shots on-screen (0 = unlimited)
  char   Processed;          // Is room "processed?"
  pObject ObjHead;           // Head of object list
  pObject ObjCur;            // Current object in list
  struct tRoom *next;
} *pRoom;

typedef struct tVar {
  char  name[VAR_MAX];       // Name of variable
  short NameHash;
  short value;               // Value
  struct tVar *next;
} *pVar;

class Z2TileMap : public TileMap {
public:
  void DrawBase(void);
  void DrawFringe(void);
  void DrawRoof(void);
};

class Z2ColorTextBox : public TextBox {};

void xor(char *str);
int  CoreFree(void);
int  CheckFree(int size);
void ReadCFG(void);
void ReadDEF(char *df);
void zzt_Fade(PALETTE sp, PALETTE dp, int f=0, int l=255);
void zzt_ForceFade(PALETTE sp, PALETTE dp, int f=0, int l=255);
//BITMAP *DAT2VSP(char *data);
void AutoAppend(char *filename, char *e);
int  DetectWindows(void);

#endif
