// GUI, version 0.1 -- Copyright 1996 Carlos DaSilva
// Version 0.1 -- First version

#include <iostream.h>
#include <graph.h>
#include <stdlib.h>
#include "mouse.h"
#include "gui.h"
#include "vgl.h"
#include "pal.h"

BITMAP Icons[4];

int LiteColor, NormColor, DarkColor;
int DGrayColor, LGrayColor, ContColor;
int BlackColor, WhiteColor;
int IconSize, IconButtonSize;

// cout inserters for our libraries classes

ostream &operator<<(ostream &stream, Button &b)
{
  stream << "Button [(" << b.r.x << "," << b.r.y
         << ")-(" << b.r.x+b.r.w << "," << b.r.y+b.r.h << ")] [";
  if (b.Clicked())
    stream << "C"; else
    stream << "c";
  if (b.GetPush())
    stream << "R"; else
    stream << "P";
  stream << "]  ";

  return stream;
}

ostream &operator<<(ostream &stream, Window &b)
{
  stream << "Window [(" << b.r.x << "," << b.r.y
         << ")-(" << b.r.x+b.r.w << "," << b.r.y+b.r.h << ")]  ";
  return stream;
}

ostream &operator<<(ostream &stream, Dialog &b)
{
  stream << "Dialog [(" << b.r.x << "," << b.r.y
         << ")-(" << b.r.x+b.r.w << "," << b.r.y+b.r.h << ")]  ";
  return stream;
}

ostream &operator<<(ostream &stream, IconButton &b)
{
  stream << "IconButton [(" << b.r.x << "," << b.r.y
         << ")-(" << b.r.x+b.r.w << "," << b.r.y+b.r.h << ")] [";
  if (b.Clicked())
    stream << "c"; else
    stream << "C";
  if (b.GetPush())
    stream << "R"; else
    stream << "P";
  stream << "]  ";

  return stream;
}

ostream &operator<<(ostream &stream, VScrollBar &b)
{
  stream << "VScrollBar [(" << b.r.x << "," << b.r.y
         << ")-(" << b.r.x+b.r.w << "," << b.r.y+b.r.h << ")] ["
         << b.GetCur() << "/" << b.GetMax() << "]  ";
  return stream;
}

ostream &operator<<(ostream &stream, HScrollBar &b)
{
  stream << "HScrollBar [(" << b.r.x << "," << b.r.y
         << ")-(" << b.r.x+b.r.w << "," << b.r.y+b.r.h << ")] ["
         << b.GetCur() << "/" << b.GetMax() << "]  ";
  return stream;
}

Base::Base()
{
  backup.Data = new char[r.w * r.h];   // Allocate memory for the background
  if (!backup.Data)
  {
    g_Done();
    cout << "GUI: Error allocating backup data buffer.\n";
    exit(1);
  }
  Save();
  ID = state = 0;
  update = 1;
}

Base::Base(Rect a)
{
  r = a;                               // Save rectangle
  backup.Data = new char[r.w * r.h];   // Allocate memory for the background
  if (!backup.Data)
  {
    g_Done();
    cout << "GUI: Error allocating backup data buffer.\n";
    exit(1);
  }
  Save();                              // Save background
  ID = state = 0;                      // Initialize variables
  update = 1;
}

void Base::Save(void)
{
  // Save background

  g_cGetBitmapWH(r.x, r.y, r.w, r.h, &backup);
}

void Base::Restore(void)
{
  // Restore background

  g_cBitmap(r.x, r.y, &backup);
}

Base::~Base(void)
{
  Restore();
  delete[] backup.Data;
}

void Base::SetTitle(char *t)
{
  title.SetText(t);                    // Set object title
  SetTitleFont(0);
}

void Base::SetTitleFont(int id)
{
  title.SetFontID(id);                 // Set title font index
}

void BevelOut(Rect r)
{
  // Draw a bevel-out box (sticking out towards user)

  int x = r.x + r.dx;
  int y = r.y + r.dy;
  int w = r.w-1;
  int h = r.h-1;

  g_cfRect(x+1, y+1, x+w, y+h, NormColor);
  g_cHLine(x, x+w-1, y, LiteColor);
  g_cHLine(x, x+w, y+h, DarkColor);
  g_cVLine(x, y, y+h-1, LiteColor);
  g_cVLine(x+w, y, y+h, DarkColor);
  g_cSetPixel(x, y+h, LGrayColor);
  g_cSetPixel(x+w, y, DGrayColor);
}

void BevelIn(Rect r)
{
  // Draw a bevel-in box (chiseled into screen)

  int x = r.x + r.dx;
  int y = r.y + r.dy;
  int w = r.w-1;
  int h = r.h-1;

  g_cfRect(x+1, y+1, x+w, y+h, NormColor);
  g_cHLine(x, x+w, y, DarkColor);
  g_cHLine(x+1, x+w-1, y+1, DGrayColor);
  g_cHLine(x, x+w, y+h, LGrayColor);
  g_cVLine(x, y, y+h, DarkColor);
  g_cVLine(x+1, y+1, y+h-1, DGrayColor);
  g_cVLine(x+w, y+1, y+h, LGrayColor);
}

Window::Window() : Base()
{
  SetTitle(NULL);

  // fontheight is used to determine height of title bar
  fontheight = FontMan.FontHeight(title.GetFontID())+2;
}

Window::Window(Rect a, char *t) : Base(a)
{
  SetTitle(t);

  // fontheight is used to determine height of title bar
  fontheight = FontMan.FontHeight(title.GetFontID())+2;
}

void Window::Draw(void)
{
  int x = r.x + r.dx;
  int y = r.y + r.dy;
  int w = r.w-1;
  int h = r.h-1;

  // Draw window, title bar, and title
  BevelOut(r);
  g_cfRect(x+1, y+1, x+w-1, y+fontheight, ContColor);
  title.btwPrint(x, x+w, y+1, WhiteColor);
  update = 0;

  _settextposition(24, 1);
  cout << *this;
}

void Window::Process(void)
{
  int x = r.x + r.dx;
  int y = r.y + r.dy;
  int w = r.w-1;
  int h = r.h-1;

  if (m_inbounds(x+1, y+1, x+w-1, x+fontheight) && m_Left())
  {
    int ox = 0;
    int oy = 0;

    do {
      if (x != ox || y != oy)
      {
        x = m_X();
        y = m_Y();
        ox = x;
        oy = y;
      }

      Draw();
    } while (m_Left());
    r.x = x;
    r.y = y;
  }
}

Dialog::Dialog() : Window()
{
  BaseList = new Base*[100];
  if (!BaseList)
  {
    g_Done();
    cout << "GUI: Error allocating dialog memory.\n";
    exit(1);
  }

  max = 100;
  cur = 0;
}

Dialog::Dialog(Rect a, char *t) : Window(a, t)
{
  BaseList = new Base*[100];
  if (!BaseList)
  {
    g_Done();
    cout << "GUI: Error allocating dialog memory.\n";
    exit(1);
  }
  max = 100;
  cur = 0;
}

Dialog::~Dialog()
{
  delete[] BaseList;
}

void Dialog::Add(Base *b)
{
  if (cur==max)
  {
    max += 50;
    BaseList = (Base **) realloc(BaseList, max * sizeof(Base));
    if (!BaseList)
    {
      g_Done();
      cout << "GUI: Insufficent memory for base controls reallocation.\n";
      exit(1);
    }
  }

  int x = r.x + r.dx;
  int y = r.y + r.dy;

  b->r.Delta(x+1, y+fontheight);

  BaseList[cur] = b;
  cur++;
}

void Dialog::Draw(void)
{
  int x = r.x + r.dx;
  int y = r.y + r.dy;
  int w = r.w-1;
  int h = r.h-1;

  BevelOut(r);
  g_cfRect(x+1, y+1, x+w-1, y+fontheight, ContColor);
  title.btwPrint(x, x+w, y+1, WhiteColor);

  for (int i=0; i<cur; i++)
    BaseList[i]->DrawUpdate();
  update = 0;

  _settextposition(24, 1);
  cout << *this;
}

void Dialog::Process(void)
{
  for (int i=0; i<cur; i++)
    BaseList[i]->Process();
}

Button::Button() : Base()
{
  SetTitle(NULL);
  SetState(BUTTON_UP);
  SetPush(BUTTON_PUSH);
  update  = 1;
  clicked = 0;
}

Button::Button(Rect a, char *t) : Base(a)
{
  SetTitle(t);
  SetState(BUTTON_UP);
  SetPush(BUTTON_PUSH);
  update  = 1;
  clicked = 0;
}

void Button::Draw(void)
{
  int x = r.x + r.dx;
  int y = r.y + r.dy;
  int w = r.w-1;
  int h = r.h-1;
  int height = (r.h - FontMan.FontHeight(title.GetFontID()))/2;
    
  switch (GetState())
  {
    case BUTTON_UP :
      BevelOut(r);
      title.btwPrint(x, x+w, y+height, BlackColor);
      break;
    case BUTTON_DOWN :
      BevelIn(r);
      title.btwPrint(x+2, x+w, y+height+1, BlackColor);
      break;
  }
  update = 0;

  _settextposition(24, 1);
  cout << *this;
}

void Button::Process(void)
{
  // If the left mouse button is down and the mouse pointer is inbounds
  // as we begin processing events, go into the routine...

  if (m_Left() && m_inbounds(r.x, r.y, r.x+r.w, r.y+r.h))
  {
    // If button is a "push-button"
    if (GetPush() == BUTTON_PUSH)
    {
      // Continually check, update and redraw mouse button until user
      // lets go of the left mouse button
      do {
        // If the mouse is inbounds, draw the button down graphic, otherwise
        // draw the button up graphic.
        if (m_inbounds(r.x, r.y, r.x+r.w, r.y+r.h))
        {
          update = (GetState() != BUTTON_DOWN);
          SetState(BUTTON_DOWN);
        } else
        {
          update = (GetState() != BUTTON_UP);
          SetState(BUTTON_UP);
        }

        // Redraw the mouse pointer, and button according to update flag.
        DrawUpdate();
      } while (m_Left());

      // If we are still inbounds after letting go of the left mouse
      // button, then the button is officially clicked, and we flag it.
      // It is up to the programmer to check each button's Clicked flag
      // and act appropriately.
      clicked = m_inbounds(r.x, r.y, r.x+r.w, r.y+r.h);
    } else
    {
      // If button is a "reps-button", then draw as down, and set
      // Clicked flag
      clicked = 1;
      if (GetState() != BUTTON_DOWN)
      {
        SetState(BUTTON_DOWN);
        Draw();
      }
    }
  }
  if (!m_Left() || !m_inbounds(r.x, r.y, r.x+r.w, r.y+r.h))
  {
    if (GetState() != BUTTON_UP)
    {
      SetState(BUTTON_UP);
      Draw();
    }
    if (GetPush() == BUTTON_REPS) clicked = 0;
  }
}

IconButton::IconButton() : Button()
{
  Icon = NULL;
}

IconButton::IconButton(Rect a, BITMAP *b) : Button(a, NULL)
{
  Icon = b;
}

void IconButton::SetIcon(BITMAP *b)
{
  Icon = b;
}

void IconButton::Draw(void)
{
  int x = r.x + r.dx;
  int y = r.y + r.dy;
  int w = r.w-1;
  int h = r.h-1;
  int height = (r.h - Icon->y)/2;
  int width  = (r.w - Icon->x)/2;
    
  switch (GetState())
  {
    case BUTTON_UP :
      BevelOut(r);
      g_ctmBitmap(x+width, y+height, BlackColor, Icon);
      break;
    case BUTTON_DOWN :
      BevelIn(r);
      g_ctmBitmap(x+width+1, y+height+1, BlackColor, Icon);
      break;
  }
  update = 0;

  _settextposition(24, 1);
  cout << *this;
}

ScrollBar::ScrollBar() : Base()
{
  cur = 1;
  max = 100;
}

ScrollBar::ScrollBar(Rect a, int c, int m) : Base(a)
{
  cur = c;
  max = m;
}

void ScrollBar::Draw(void)
{
  Rect z;

  z = r;
  z.y += IconButtonSize;
  z.h -= IconButtonSize*2;

  int x = z.x + z.dx;
  int y = z.y + z.dy;
  int w = z.w-1;
  int h = z.h-1;


  g_cfRect(x+1, y+1, x+w, y+h, NormColor);
  g_cHLine(x, x+w-1, y, DarkColor);
  g_cHLine(x, x+w, y+h, LGrayColor);
  g_cVLine(x, y, y+h-1, DarkColor);
  g_cVLine(x+w, y+1, y+h, LGrayColor);

  Inc.DrawUpdate();
  Dec.DrawUpdate();

  update = 0;
}

void ScrollBar::DrawMark(int back)
{
  int xx = a.x + a.dx;
  int yy = a.y + a.dy;
  int ww = a.w-1;
  int hh = a.h-1;

  g_cfRect(xx+1, yy+1, xx+ww, yy+hh, NormColor);
  g_cHLine(xx, xx+ww-1, yy, DarkColor);
  g_cHLine(xx, xx+ww, yy+hh, LGrayColor);
  g_cVLine(xx, yy, yy+hh-1, DarkColor);
  g_cVLine(xx+ww, yy+1, yy+hh, LGrayColor);
  switch (back)
  {
    case 1 : g_cHLine(xx, xx+ww-1, yy-1, NormColor); break;
    case 2 : g_cHLine(xx, xx+ww-1, yy+hh+1, NormColor); break;
  }
}

void ScrollBar::Process(void)
{
  Inc.Process();
  if (Inc.Clicked() && cur<max)
  {
    cur++;
    UpdateMark();
    DrawMark(1);
  }
  Dec.Process();
  if (Dec.Clicked() && cur>0)
  {
    cur--;
    UpdateMark();
    DrawMark(2);
  }
}

VScrollBar::VScrollBar() : ScrollBar()
{
  Dec.SetPush(BUTTON_REPS);
  Dec.SetIcon(&Icons[ICON_UP]);
  Inc.SetPush(BUTTON_REPS);
  Inc.SetIcon(&Icons[ICON_DOWN]);
}

VScrollBar::VScrollBar(Rect a, int c, int m) : ScrollBar(a, c, m)
{
  Rect r;

  r.dx = a.dx;
  r.dy = a.dy;
  r.x = a.x;
  r.y = a.y;
  r.w = r.h = IconButtonSize;
  Dec.SetRect(r);
  Dec.SetPush(BUTTON_REPS);
  Dec.SetIcon(&Icons[ICON_UP]);

  r.y = a.y+a.h-IconButtonSize;
  Inc.SetRect(r);
  Inc.SetPush(BUTTON_REPS);
  Inc.SetIcon(&Icons[ICON_DOWN]);

  size  = a.h-(IconButtonSize*3);
  begin = (a.y+IconButtonSize+1);
  pos   = (cur * size)/max;
  r.x++;
  r.y = begin+pos;
  r.w -= 2;
  r.h -= 2;
}

void VScrollBar::UpdateMark()
{
  pos = (cur * size)/max;
  a = r;
  a.x++;
  a.y = begin+pos;
  a.w = a.h = IconSize;
}

HScrollBar::HScrollBar() : ScrollBar()
{
  Dec.SetPush(BUTTON_REPS);
  Dec.SetIcon(&Icons[ICON_LEFT]);
  Inc.SetPush(BUTTON_REPS);
  Inc.SetIcon(&Icons[ICON_RIGHT]);
}

HScrollBar::HScrollBar(Rect a, int c, int m) : ScrollBar(a, c, m)
{
  Rect r;

  r.dx = a.dx;
  r.dy = a.dy;
  r.x = a.x;
  r.y = a.y;
  r.w = r.h = IconButtonSize;
  Dec.SetRect(r);
  Dec.SetPush(BUTTON_REPS);
  Dec.SetIcon(&Icons[ICON_LEFT]);

  r.x = a.x+a.w-IconButtonSize;
  Inc.SetRect(r);
  Inc.SetPush(BUTTON_REPS);
  Inc.SetIcon(&Icons[ICON_RIGHT]);

  size  = a.w-(IconButtonSize*3);
  begin = (a.x+IconButtonSize+1);
  pos   = (cur * size)/max;
  r.x = begin+pos;
  r.y++;
  r.w -= 2;
  r.h -= 2;
}

void HScrollBar::UpdateMark()
{
  pos = (cur * size)/max;
  a = r;
  a.x = begin+pos;
  a.y++;
  a.w = r.h = IconSize;
}

void guiInit(void)
{
  g_LoadBitmap("ICONS.VSP", Icons);
  IconSize = Icons[0].x+2;
  IconButtonSize = IconSize+2;

  LiteColor  = g_rgb(224, 224, 224);
  NormColor  = g_rgb(160, 160, 160);
  DarkColor  = g_rgb(84, 84, 84);
  DGrayColor = g_rgb(112, 112, 112);
  LGrayColor = g_rgb(176, 176, 176);
  ContColor  = g_rgb(40, 40, 192);
  BlackColor = g_rgb(0, 0, 0);
  WhiteColor = g_rgb(255, 255, 255);
}
