// GUI, version 0.1 -- Copyright 1996 Carlos DaSilva
// Version 0.1 -- First version

#ifndef __GUI__
#define __GUI__

#include <iostream.h>
#include "vgl.h"
#include "font.h"

#define ICON_UP    0
#define ICON_DOWN  1
#define ICON_RIGHT 2
#define ICON_LEFT  3

#define BUTTON_DOWN 0
#define BUTTON_UP   1
#define BUTTON_PUSH 0
#define BUTTON_REPS 1

extern BITMAP Icons[4];

class Rect {
public:
  int dx, dy;
  int x, y;
  int w, h;

  Rect()                               { x = y = w = h = dx = dy = 0; }
  Rect(int xx, int yy, int ww, int hh) { Move(xx, yy); Size(ww, hh); Delta(0, 0); }
  ~Rect()                              {}

  int operator==(Rect t)
  { if (x==t.x && y==t.y && w==t.w && h==t.h) return 1; else return 0; }

  Rect operator=(Rect t)
  { x = t.x; y = t.y; w = t.w; h = t.h; return *this; }

  void Delta(int xx, int yy)           { dx = xx; dy = yy; }
  void Move(int xx, int yy)            { x = xx; y = yy; }
  void MoveRel(int xx, int yy)         { x += xx; y += yy; }
  void Size(int ww, int hh)            { w = ww; h = hh; }
  void Set(int xx, int yy, int ww, int hh)
  { Move(xx, yy); Size(ww, hh); }
};

class Base {
  BITMAP backup;
  int ID, state;

protected:
  TextLine title;
  int update;

public:
  Rect r;

  Base();
  Base(Rect a);
  virtual ~Base(void);

  void SetID(int id)         { ID = id; }
  int  GetID(void)           { return ID; }
  void SetState(int s)       { state = s; }
  int  GetState(void)        { return state; }
  void SetRect(Rect a)       { r = a; }
  void SetTitle(char *t);
  void SetTitleFont(int id);
  void SetUpdate(int u)      { update = u; }
  int  GetUpdate(void)       { return update; }
  void Save(void);
  void Restore(void);
  void DrawUpdate(void)      { if (update) Draw(); }

  virtual void Draw(void) {};
  virtual void Process(void) {};
};

class Window : public Base {
protected:
  int fontheight;

public:
  Window();
  Window(Rect a, char *t);
  friend ostream &operator<<(ostream &stream, Window &b);

  void Draw(void);
  void Process(void);
};

class Dialog : public Window {
  Base **BaseList;
  int max, cur;

public:
  Dialog();
  Dialog(Rect a, char *t);
  ~Dialog();
  friend ostream &operator<<(ostream &stream, Dialog &b);

  void Add(Base *b);
  void Draw(void);
  void Process(void);
};

class Button : public Base {
  int clicked;
  int push;
public:
  Button();
  Button(Rect a, char *t=NULL);
  friend ostream &operator<<(ostream &stream, Button &b);

  void SetPush(int p)        { push = p; }
  int  GetPush(void)         { return push; }
  int  Clicked(void)         { return clicked; }
  void Draw(void);
  void Process(void);
};

class IconButton : public Button {
  BITMAP *Icon;
public:
  IconButton();
  IconButton(Rect a, BITMAP *b);
  friend ostream &operator<<(ostream &stream, IconButton &b);

  void SetIcon(BITMAP *b);
  void Draw(void);
};

class ScrollBar : public Base {
protected:
  IconButton Inc, Dec;
  int cur, max, size, begin, pos;
  Rect a;
public:
  ScrollBar();
  ScrollBar(Rect a, int c, int m);
  
  void SetCur(int c)         { cur = c; }
  int  GetCur(void)          { return cur; }
  void SetMax(int c)         { max = c; }
  int  GetMax(void)          { return max; }
  virtual void UpdateMark()  {}
  void DrawMark(int back);
  void Draw(void);
  void Process(void);
};

class VScrollBar : public ScrollBar {
public:
  VScrollBar();
  VScrollBar(Rect a, int c, int m);
  friend ostream &operator<<(ostream &stream, VScrollBar &b);

  void UpdateMark(void);
};

class HScrollBar : public ScrollBar {
public:
  HScrollBar();
  HScrollBar(Rect a, int c, int m);
  friend ostream &operator<<(ostream &stream, HScrollBar &b);

  void UpdateMark(void);
};

void guiInit(void);

#endif
