// Mouse Handler, version 1.2 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version
// Version 1.1 -- Added code and data locks, essential under VMMs
// Version 1.2 -- Added support for mouse button 3 (middle)

#include <i86.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include "sprite.h"
#include "mem32.h"
#include "mouse.h"
#include "vgl.h"

static MOUSEDATA cbd;
BITMAP pointer;
char MouseError[80];

int MouseHidden=1;
int DivX, DivY;
int HotX, HotY;

extern void CopyDS2ES(void);
#pragma aux CopyDS2ES = \
  "push ds" \
  "pop  es";

#pragma off (check_stack)

void _loadds far ClickHandler(int max, int mbx, int mcx, int mdx, int msi, int mdi)
{
#pragma aux ClickHandler parm [EAX][EBX][ECX][EDX][ESI][EDI]

  // Set ES = DS as BIOS code as well as compiler generated code accesses
  // data via both the ES and DS registers. This should always be first.
  CopyDS2ES();

  cbd.mCode = (unsigned short)max;
  cbd.mBut  = (unsigned short)mbx;
  cbd.mDX   = (short)msi;
  cbd.mDY   = (short)mdi;
  cbd.mX    = (mcx*100)/DivX;
  cbd.mY    = (mdx*100)/DivY;
}

void CBend(void) {}

int LockRegion(void *address, unsigned length)
{
  static union REGS r;
  unsigned linear;

  linear = (unsigned)address;

  r.w.ax = 0x600;
  r.w.bx = (unsigned short)(linear >> 16);
  r.w.cx = (unsigned short)(linear & 0xFFFF);
  r.w.si = (unsigned short)(length >> 16);
  r.w.di = (unsigned short)(length & 0xFFFF);
  int386(0x31, &r, &r);
  return (!r.w.cflag);
}

int m_Init(int w, int h, BITMAP *p)
{
  void (far *funcptr)();
  static union  REGS r;
  static struct SREGS s;
  int    result;

  MouseHidden = 1;

  result = m_HardReset();
  if (!result)
  {
    strcpy(MouseError, "Mouse Handler: Mouse driver not detected.");
    return 0;
  }

  atexit(m_Done);

  // Lock data and code segments that we don't want swapped out in
  // an VMM environment.
  if ((!LockRegion(&cbd, sizeof(cbd))) ||
      (!LockRegion((void near *) ClickHandler,
       (char *) CBend - (char near *) ClickHandler)))
  {
    strcpy(MouseError, "Mouse Handler: Code and data locks failed.");
    return 0;
  }

  r.w.ax = 0x26;
  int386(0x33, &r, &r);
  DivX = (r.w.cx+1)*100 / w;
  DivY = (r.w.dx+1)*100 / h;

  r.w.ax = 0x3;
  int386(0x33, &r, &r);
  cbd.mX = (r.w.cx+1)*100 / DivX;
  cbd.mY = (r.w.dx+1)*100 / DivY;

  if (p) m_SetPointer(p);

  segread(&s);
  r.w.ax  = 0xC;
  r.w.cx  = 0x7F;
  funcptr = (void(*)(void)) ClickHandler;
  r.x.edx = FP_OFF(funcptr);
  s.es    = FP_SEG(funcptr);
  int386x(0x33, &r, &r, &s);

  return 1;
}

void m_Done(void)
{
  m_HardReset();

  delete[] pointer.Data;
}

void m_SetHotspot(int x, int y)
{
  HotX = x;
  HotY = y;
}

int m_inbounds(int x1, int y1, int x2, int y2)
{
  return (cbd.mX>=x1 && cbd.mX<=x2 && cbd.mY>=y1 && cbd.mY<=y2);
}

int m_Moved(void)
{
  return cbd.mCode;
}

int m_Left(void)
{
  return cbd.mBut & 0x1 ? 1 : 0;
}

int m_Right(void)
{
  return cbd.mBut & 0x2 ? 1 : 0;
}

int m_Middle(void)
{
  return cbd.mBut & 0x4 ? 1 : 0;
}

int m_X(void)
{
  return cbd.mX;
}

int m_Y(void)
{
  return cbd.mY;
}

int m_SetPointer(BITMAP *p)
{
  pointer.x = p->x;
  pointer.y = p->y;
  if (pointer.Data) delete[] pointer.Data;
  
  pointer.Data = new char[p->x * p->y];
  if (!pointer.Data)
  {
    strcpy(MouseError, "Mouse Handler: Error allocating mouse pointer buffer.");
    return 0;
  }
  memcpy8(pointer.Data, p->Data, p->x * p->y);

  s_AddMouse(pointer);

  return 1;
}

int m_HardReset(void)
{
  static union REGS r;

  // reset mouse
  r.w.ax = 0x0;
  int386(0x33, &r, &r);

  return r.w.ax;
}

void m_Show(void)
{
  if (!MouseHidden) return;
  MouseHidden = 0;
}

void m_Hide(void)
{
  if (MouseHidden) return;
  MouseHidden = 1;
}

void m_SetPos(int x, int y)
{
  static union REGS r;

  // set mouse position
  r.w.ax = 0x4;
  r.w.cx = (x*DivX)/100;
  r.w.dx = (y*DivY)/100;
  int386(0x33, &r, &r);
}

void m_SetBounds(int x1, int y1, int x2, int y2)
{
  static union REGS r;

  // set horizontal boundaries
  r.w.ax = 0x7;
  r.w.cx = (x1*DivX)/100;
  r.w.dx = (x2*DivX)/100;
  int386(0x33, &r, &r);

  // set vertical boundaries
  r.w.ax = 0x8;
  r.w.cx = (y1*DivY)/100;
  r.w.dx = (y2*DivY)/100;
  int386(0x33, &r, &r);
}

void m_SoftReset(void)
{
  static union REGS r;

  // software reset
  r.w.ax = 0x21;
  int386(0x33, &r, &r);
}

void m_Update(void)
{
  s_SetPos(MousePtr, cbd.mX, cbd.mY);
}

