// MultiKey, version 2.72 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version, 16-bit, in Turbo Pascal
// Version 1.5 -- First version, 16-bit, in Borland C
// Version 2.0 -- First version, 32-bit, in Watcom C
// Version 2.1 -- Added code and data locks, essential under VMMs
// Version 2.5 -- Correctly implemented extended keys, and Win95 keys
// Version 2.6 -- Correctly implemented Pause key, and differentiated
//                pad keys and inverted T keys
// Version 2.7 -- Added a method to obtain ASCII keys as if they were
//                typed with a BIOS handler installed.
// Version 2.71 - Fixed annoying bug (pressing special key caused handler
//                to eat next key)
// Version 2.72 - Added WasDown (check if a key was ever pressed)

// Version 3.0  - Total Rewrite. Implement a 256-char ring buffer to
//                prevent keypress misses; reduce embarrassingly huge
//                interrupt handler with small, fast, efficient code;
//                enhance total keyboard support (including panic button
//                with ctrl-alt-del)

#ifndef __MULTIKEY__
#define __MULTIKEY__

#define mkEsc        1
#define mk1          2
#define mk2          3
#define mk3          4
#define mk4          5
#define mk5          6
#define mk6          7
#define mk7          8
#define mk8          9
#define mk9          10
#define mk0          11
#define mkDash       12
#define mkEqual      13
#define mkBackSpace  14
#define mkTab        15
#define mkQ          16
#define mkW          17
#define mkE          18
#define mkR          19
#define mkT          20
#define mkY          21
#define mkU          22
#define mkI          23
#define mkO          24
#define mkP          25
#define mkLBracket   26
#define mkRBracket   27
#define mkEnter      28
#define mkCtrl       29
#define mkLCtrl      29
#define mkA          30
#define mkS          31
#define mkD          32
#define mkF          33
#define mkG          34
#define mkH          35
#define mkJ          36
#define mkK          37
#define mkL          38
#define mkColon      39
#define mkQuote      40
#define mkTilde      41
#define mkLShift     42
#define mkZ          44
#define mkX          45
#define mkC          46
#define mkV          47
#define mkB          48
#define mkN          49
#define mkM          50
#define mkComma      51
#define mkPeriod     52
#define mkQuestion   53
#define mkRShift     54
#define mkStar       55
#define mkAlt        56
#define mkLAlt       56
#define mkSpace      57
#define mkCapsLock   58
#define mkF1         59
#define mkF2         60
#define mkF3         61
#define mkF4         62
#define mkF5         63
#define mkF6         64
#define mkF7         65
#define mkF8         66
#define mkF9         67
#define mkF10        68
#define mkNumLock    69
#define mkScrollLock 70
#define mkHome       71
#define mkUp         72
#define mkPgUp       73
#define mkMinus      74
#define mkLeft       75
#define mkFive       76
#define mkCenter     76
#define mkRight      77
#define mkPlus       78
#define mkEnd        79
#define mkDown       80
#define mkPgDn       81
#define mkIns        82
#define mkDel        83
#define mkF11        87
#define mkF12        88
#define mkLWin95     91
#define mkRWin95     92
#define mkStart      93
#define mkRCtrl      100
#define mkRAlt       101
#define mkReboot     102
#define mkPrint      103
#define mkPause      104
#define mkKeyPad     105

#define mcNone       0
#define mcBackspace  8
#define mcTab        9
#define mcEnter      13
#define mcEsc        27
#define mcF1         200
#define mcF2         201
#define mcF3         202
#define mcF4         203
#define mcF5         204
#define mcF6         205
#define mcF7         206
#define mcF8         207
#define mcF9         208
#define mcF10        209
#define mcF11        210
#define mcF12        211
#define mcHome       220
#define mcUp         221
#define mcPgUp       222
#define mcLeft       224
#define mcRight      226
#define mcEnd        228
#define mcDown       229
#define mcPgDn       230
#define mcIns        231
#define mcDel        232

void mk_Init(void);
void mk_Done(void);
int  mk_Down(int key);
int  mk_PadDown(int key);
int  mk_Pressed(int key);
int  mk_PadPressed(int key);
int  mk_Shift(void);
int  mk_Ctrl(void);
int  mk_Alt(void);
int  mk_Shift(int key);
int  mk_Ctrl(int key);
int  mk_Alt(int key);
void mk_Clear(void);
void mk_Wait(int key);
int  mk_LastKey(void);
char mk_ReadChar(void);
char mk_PollChar(void);
int  mk_Hit(void);

#endif
