// Watcom includes
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>

// HGL includes
#include "mem32.h"
#include "tile.h"

// ZZT ][ includes
#include "compiler.h"
#include "objects.h"
#include "editor.h"
#include "kernel.h"
#include "global.h"
#include "sound.h"

pRoom   RoomHead, RoomCur;
pVar    VarHead;
pObject SystemHead, GlobalHead, LastObj;
char    DefName[80], DefFile[80], TileFile[80];

void InitEngine(void)
{
  RoomHead = RoomCur = NULL;
  VarHead = NULL;
  SystemHead = NULL;
  SoundHead = NULL;

  t_Init(TILE_MAX);
  TotTiles = t_Load("zzt.vsp");
  RoomData.ReInit();
}

void KillRooms(void)
{
  // Clear room list
  pRoom CurR = RoomHead;
  while (CurR)
  {
    // Clear room object list
    pObject CurRO = CurR->ObjListHead;
    while (CurRO)
    {
      KillObj(CurR, CurRO);
      CurRO = CurRO->next;
    }
    CurR->ObjListHead = NULL;

    delete[] CurR;
    CurR = CurR->next;
  }
  RoomHead = NULL;
}

void DoneEngine(void)
{
  KillRooms();

  // Clear system objects list
  pObject CurO = SystemHead;
  while (CurO)
  {
    delete[] CurO;
    CurO = CurO->next;
  }
  SystemHead = NULL;

  // Clear variables list
  pVar CurV = VarHead;
  while (CurV)
  {
    delete[] CurV;
    CurV = CurV->next;
  }
  VarHead = NULL;

  t_Done();
}

pVar AddVar(char *name, short v)
{
  pVar NewVar = NULL;

  // Find end of linked list
  pVar Cur = VarHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  NewVar = new tVar;
  if (!NewVar) return NULL;

  strncpy(NewVar->name, name, VAR_MAX);
  NewVar->value    = v;
  NewVar->NameHash = Hash(name);
  NewVar->next     = NULL;

  if (Cur)
    Cur->next = NewVar; else
    VarHead = NewVar;

  return NewVar;
}

int masterSaveRoom(ofstream &obj, pRoom r)
{
  int   length, rlep, i, saved;
  short seed;
  short *FloorRLE, *ObjectRLE, *CeilingRLE;

  obj.write(RoomVer, 4);
  obj.write(r->name, strlen(r->name));
  obj.put((char)NULL);
  obj.write(r->Desc, strlen(r->Desc));
  obj.put((char)NULL);
  obj.write(r->Mod, strlen(r->Mod));
  obj.put((char)NULL);
  obj.write((unsigned char *)&r->x, sizeof(r->x));
  obj.write((unsigned char *)&r->y, sizeof(r->y));
  obj.write((unsigned char *)&r->RoomW, sizeof(r->RoomW));
  obj.write((unsigned char *)&r->RoomH, sizeof(r->RoomH));

  FloorRLE = new short[r->RoomW*r->RoomH];
  rlep = 1;
  i = 0;
  while (i<r->RoomW*r->RoomH)
  {
    length = 0;
    seed = r->RoomFloor[i];
    while (r->RoomFloor[i]==seed)
    {
      i++;
      length++;
    }
    if (length>1)
    {
      FloorRLE[rlep++] = 0xFFFF;
      FloorRLE[rlep++] = length;
      FloorRLE[rlep++] = seed;
    } else
    if (length==1)
      FloorRLE[rlep++] = seed;
  }
  FloorRLE[0] = rlep-1;

  obj.write((unsigned char *) FloorRLE, (FloorRLE[0]+1)*2);
  if (FloorRLE) delete[] FloorRLE;

  ObjectRLE = new short[r->RoomW*r->RoomH];
  rlep = 1;
  i = 0;
  while (i<r->RoomW*r->RoomH)
  {
    length = 0;
    seed = r->RoomObject[i];
    while (r->RoomObject[i]==seed && i<r->RoomW*r->RoomH)
    {
      i++;
      length++;
    }
    if (length>1)
    {
      ObjectRLE[rlep++] = 0xFFFF;
      ObjectRLE[rlep++] = length;
      ObjectRLE[rlep++] = seed;
    } else
    if (length==1)
      ObjectRLE[rlep++] = seed;
  }
  ObjectRLE[0] = rlep-1;

  obj.write((unsigned char *) ObjectRLE, (ObjectRLE[0]+1)*2);
  if (ObjectRLE) delete[] ObjectRLE;

  if (r->RoomCeiling)
  {
    CeilingRLE = new short[r->RoomW*r->RoomH];
    rlep = 1;
    i = 0;
    while (i<r->RoomW*r->RoomH)
    {
      length = 0;
      seed = r->RoomCeiling[i];
      while (r->RoomCeiling[i]==seed && i<r->RoomW*r->RoomH)
      {
        i++;
        length++;
      }
      if (length>1)
      {
        CeilingRLE[rlep++] = 0xFFFF;
        CeilingRLE[rlep++] = length;
        CeilingRLE[rlep++] = seed;
      } else
      if (length==1)
        CeilingRLE[rlep++] = seed;
    }
    CeilingRLE[0] = rlep-1;

    obj.write((unsigned char *) CeilingRLE, (CeilingRLE[0]+1)*2);
    if (CeilingRLE) delete[] CeilingRLE;
  } else obj.write("NULL\0", 5);

  if (r->LinkN)
  {
    obj.write(r->LinkN->name, strlen(r->LinkN->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);
  if (r->LinkS)
  {
    obj.write(r->LinkS->name, strlen(r->LinkS->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);
  if (r->LinkE)
  {
    obj.write(r->LinkE->name, strlen(r->LinkE->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);
  if (r->LinkW)
  {
    obj.write(r->LinkW->name, strlen(r->LinkW->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);

  obj.write((unsigned char *) &r->Light, sizeof(r->Light));
  obj.write((unsigned char *) &r->ReEntry, sizeof(r->ReEntry));
  obj.write((unsigned char *) &r->Time, sizeof(r->Time));
  obj.write((unsigned char *) &r->MaxShots, sizeof(r->MaxShots));
  obj.write((unsigned char *) &r->Processed, sizeof(r->Processed));

  pObject Cur = r->ObjListHead;
  while (Cur)
  {
    obj.write(Cur->name, strlen(Cur->name));
    obj.put((char)NULL);
    obj.write(Cur->bark, strlen(Cur->bark));
    obj.put((char)NULL);
    obj.write((unsigned char *) &Cur->barktime, sizeof(Cur->barktime));
    obj.write((unsigned char *) &Cur->x, sizeof(Cur->x));
    obj.write((unsigned char *) &Cur->y, sizeof(Cur->y));
    obj.write((unsigned char *) &Cur->z, sizeof(Cur->z));
    obj.write((unsigned char *) &Cur->property, sizeof(Cur->property));
    obj.write((unsigned char *) &Cur->frame, sizeof(Cur->frame));
    obj.write(Cur->local, LOCAL_MAX);
    obj.write((unsigned char *) &Cur->codesize, sizeof(Cur->codesize));
    obj.write((unsigned char *) &Cur->codep, sizeof(Cur->codep));
    obj.write(Cur->code, Cur->codesize);
    obj.write((unsigned char *) &Cur->cycle, sizeof(Cur->cycle));
    obj.write((unsigned char *) &Cur->cyclel, sizeof(Cur->cyclel));
    obj.write((unsigned char *) &Cur->light, sizeof(Cur->light));
    obj.write((unsigned char *) &Cur->queue, sizeof(Cur->queue));
    obj.write((unsigned char *) &Cur->qsize, sizeof(Cur->qsize));
    Cur = Cur->next;
  }
  obj.write("NULL\0", 5);

  return SUCCESS;
}

int masterLoadRoom(ifstream &obj, pRoom r)
{
  char temp[RNAME_MAX];
  char id[4], ch;
  char *str;
  int  oldw, oldh;

  obj.read(id, 4);
  if (strncmp(id, RoomVer, 4)) return FILE_ERR;
  if (strncmp(id, RoomNULL, 4)==0) return -42;

  str = r->name;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  str = r->Desc;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  str = r->Mod;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  obj.read((unsigned char *) &r->x, sizeof(short));
  obj.read((unsigned char *) &r->y, sizeof(short));
  oldw = r->RoomW;
  oldh = r->RoomH;
  obj.read((unsigned char *) &r->RoomW, sizeof(short));
  obj.read((unsigned char *) &r->RoomH, sizeof(short));

  int i;
  short rle_len, in, out, len, j;
  short seed;

  if (oldw != r->RoomW || oldh != r->RoomH && r->RoomFloor)
  {
    if (r->RoomFloor) delete[] r->RoomFloor;
    r->RoomFloor = new short[r->RoomW*r->RoomH];
  }
  if (!r->RoomFloor)
  {
    g_Done();
    printf("ZZT ][: Error initializing room floor map.\n");
    exit(1);
  }
  obj.read((unsigned char *) &rle_len, sizeof(short));

  out = 0;
  for (i=0; i<rle_len; i++)
  {
    obj.read((unsigned char *) &in, sizeof(short));
    switch (in)
    {
      case 0xFFFF :
        obj.read((unsigned char *)&len, sizeof(short));
        obj.read((unsigned char *)&seed, sizeof(short));
        i += 2;
        for (j=0; j<len; j++)
          r->RoomFloor[out++] = seed;
        break;
      default :
        r->RoomFloor[out++] = in;
        break;
    }
  }

  if (oldw != r->RoomW || oldh != r->RoomH && r->RoomObject)
  {
    if (r->RoomObject) delete[] r->RoomObject;
    r->RoomObject = new short[r->RoomW*r->RoomH];
  }
  if (!r->RoomObject)
  {
    printf("ZZT ][: Error initializing room object map.\n");
    exit(1);
  }
  obj.read((unsigned char *) &rle_len, sizeof(short));

  out = 0;
  for (i=0; i<rle_len; i++)
  {
    obj.read((unsigned char *) &in, sizeof(short));
    switch (in)
    {
      case 0xFFFF :
        obj.read((unsigned char *)&len, sizeof(short));
        obj.read((unsigned char *)&seed, sizeof(short));
        i += 2;
        for (j=0; j<len; j++)
          r->RoomObject[out++] = seed;
        break;
      default :
        r->RoomObject[out++] = in;
        break;
    }
  }

  char check[5];
  obj.read(check, 5);
  if (strcmp(check, "NULL")==0) r->RoomCeiling = NULL; else
  {
    obj.seekg(-5, ios::cur);
    if (oldw != r->RoomW || oldh != r->RoomH && r->RoomCeiling)
    {
      if (r->RoomCeiling) delete[] r->RoomCeiling;
      r->RoomCeiling = new short[r->RoomW*r->RoomH];
    }
    if (!r->RoomCeiling)
    {
      g_Done();
      printf("ZZT ][: Error initializing room ceiling map.\n");
      exit(1);
    }
    obj.read((unsigned char *) &rle_len, sizeof(short));

    out = 0;
    for (i=0; i<rle_len; i++)
    {
      obj.read((unsigned char *) &in, sizeof(short));
      switch (in)
      {
        case 0xFFFF :
          obj.read((unsigned char *)&len, sizeof(short));
          obj.read((unsigned char *)&seed, sizeof(short));
          i += 2;
          for (j=0; j<len; j++)
            r->RoomCeiling[out++] = seed;
          break;
        default :
          r->RoomCeiling[out++] = in;
          break;
      }
    }
  }

  str = r->StrLinkN;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(r->StrLinkN, "NULL")==0)
  {
    r->LinkN = NULL;
    memset(r->StrLinkN, 0, RNAME_MAX);
  } else r->LinkN = FindRoom(temp);

  str = r->StrLinkS;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(r->StrLinkS, "NULL")==0)
  {
    r->LinkS = NULL;
    memset(r->StrLinkS, 0, RNAME_MAX);
  } else r->LinkS = FindRoom(temp);

  str = r->StrLinkE;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(r->StrLinkE, "NULL")==0)
  {
    r->LinkE = NULL;
    memset(r->StrLinkE, 0, RNAME_MAX);
  } else r->LinkE = FindRoom(temp);

  str = r->StrLinkW;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(temp, "NULL")==0)
  {
    r->LinkW = NULL;
    memset(r->StrLinkW, 0, RNAME_MAX);
  } else r->LinkW = FindRoom(temp);

  obj.read((unsigned char *) &r->Light, sizeof(r->Light));
  obj.read((unsigned char *) &r->ReEntry, sizeof(r->ReEntry));
  obj.read((unsigned char *) &r->Time, sizeof(r->Time));
  obj.read((unsigned char *) &r->MaxShots, sizeof(r->MaxShots));
  obj.read((unsigned char *) &r->Processed, sizeof(r->Processed));

  // Clear objects list
  pObject CurO = r->ObjListHead;
  while (CurO)
  {
    delete[] CurO;
    CurO = CurO->next;
  }
  r->ObjListHead = NULL;

  int done = 0;
  while (!done)
  {
    str = temp;
    while (ch = obj.get())
      *str++ = ch;
    *str = 0;

    if (strcmp(temp, "NULL"))
    {
      pObject Cur = AddObj(NULL, r->ObjListHead);

      strcpy(Cur->name, temp);
      Cur->NameHash = Hash(Cur->name);
      str = Cur->bark;
      while (ch = obj.get())
        *str++ = ch;
      *str = 0;
      obj.read((unsigned char *) &Cur->barktime, sizeof(Cur->barktime));
      obj.read((unsigned char *) &Cur->x, sizeof(Cur->x));
      obj.read((unsigned char *) &Cur->y, sizeof(Cur->y));
      obj.read((unsigned char *) &Cur->z, sizeof(Cur->y));
      obj.read((unsigned char *) &Cur->property, sizeof(Cur->property));
      obj.read((unsigned char *) &Cur->frame, sizeof(Cur->frame));
      obj.read(Cur->local, LOCAL_MAX);
      obj.read((unsigned char *) &Cur->codesize, sizeof(Cur->codesize));
      obj.read((unsigned char *) &Cur->codep, sizeof(Cur->codep));
      Cur->code = new char[Cur->codesize];
      obj.read(Cur->code, Cur->codesize);
      obj.read((unsigned char *) &Cur->cycle, sizeof(Cur->cycle));
      obj.read((unsigned char *) &Cur->cyclel, sizeof(Cur->cyclel));
      obj.read((unsigned char *) &Cur->light, sizeof(Cur->light));
      obj.read((unsigned char *) &Cur->queue, sizeof(Cur->queue));
      obj.read((unsigned char *) &Cur->qsize, sizeof(Cur->qsize));
      Cur = Cur->next;
    } else done = 1;
  }

  return SUCCESS;
}

int SaveRoom(char *fname, pRoom r)
{
  if (!r) return BAD_ROOM;

  ofstream obj(fname, ios::out | ios::binary);
  if (!obj) return FILE_ERR;

  int ret = masterSaveRoom(obj, r);

  obj.close();
  return ret;
}

void UpdateLinks(void)
{
  pRoom Cur = RoomHead;

  while (Cur)
  {
    if (strlen(Cur->StrLinkN))
    {
      Cur->LinkN = FindRoom(Cur->StrLinkN);
      memset(Cur->StrLinkN, 0, RNAME_MAX);
    }
    if (strlen(Cur->StrLinkS))
    {
      Cur->LinkS = FindRoom(Cur->StrLinkS);
      memset(Cur->StrLinkS, 0, RNAME_MAX);
    }
    if (strlen(Cur->StrLinkE))
    {
      Cur->LinkE = FindRoom(Cur->StrLinkE);
      memset(Cur->StrLinkE, 0, RNAME_MAX);
    }
    if (strlen(Cur->StrLinkW))
    {
      Cur->LinkW = FindRoom(Cur->StrLinkW);
      memset(Cur->StrLinkW, 0, RNAME_MAX);
    }
    Cur = Cur->next;
  }
}

int SaveWorld(char *fname, char *desc)
{
  ofstream obj(fname, ios::out | ios::binary);
  if (!obj) return FILE_ERR;

  // world header
  obj.write(WorldVer, 4);
  obj.write(desc, strlen(desc));
  obj.put((char)NULL);
  
  // save rooms
  pRoom r = RoomHead;
  while (r)
  {
    masterSaveRoom(obj, r);
    r = r->next;
  }
  obj.write(RoomNULL, 4);

  obj.close();
  return SUCCESS;
}

int LoadWorld(char *fname)
{
  char temp[RNAME_MAX];
  char id[4], ch;
  char *str;
  pRoom r;
  int stop=0;

  ifstream obj(fname, ios::in | ios::binary);
  if (!obj || !fname) return FILE_ERR;

  // check for validity
  obj.read(id, 4);
  if (strncmp(id, WorldVer, 4))
    return FILE_ERR;

  // Read world description
  str = temp;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  // Clear room list
  KillRooms();

  // We should be right at primary chunkage.
  while (!stop)
  {
    r = AddRoom("Error", "Error Loading World");
    if (!r) return NO_MEM;
    stop = masterLoadRoom(obj, r);
    if (stop==SUCCESS) stop = 0;
  }
  KillRoom(r);               // gets rid of extra garbage room
  UpdateLinks();             // updates all links

  obj.close();
  return SUCCESS;
}

pRoom AddRoom(char *fname)
{
  char temp[RNAME_MAX];
  char *str, ch;
  char id[4];

  // Find end of linked list
  pRoom Cur = RoomHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pRoom NewRoom = new tRoom;
  if (!NewRoom) return NULL;

  ifstream obj(fname, ios::in | ios::binary);
  if (!obj || !fname)
  {
    delete[] NewRoom;
    return NULL;
  }

  masterLoadRoom(obj, r);
  obj.close();
  r->next = NULL;

  if (Cur)
    Cur->next = NewRoom; else
    RoomHead = NewRoom;

  return NewRoom;
}

int LoadRoom(char *fname, pRoom r)
{
  char temp[RNAME_MAX];
  char *str, ch;
  char id[4];
  int  oldw, oldh;

  ifstream obj(fname, ios::in | ios::binary);
  if (!obj || !fname) return FILE_ERR;

  int ret = masterLoadRoom(obj, r);

  obj.close();
  return ret;
}

pRoom AddRoom(char *name, char *desc)
{
  // Find end of linked list
  pRoom Cur = RoomHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pRoom NewRoom = new tRoom;
  if (!NewRoom) return NULL;

  if (!strlen(name) || !name)
    strcpy(NewRoom->name, "Untitled"); else
    strncpy(NewRoom->name, name, RNAME_MAX);
  if (!strlen(desc) || !desc)
    strncpy(NewRoom->Desc, name, RDESC_MAX); else
    strncpy(NewRoom->Desc, desc, RDESC_MAX);

  NewRoom->NameHash = Hash(NewRoom->name);
  strcpy(NewRoom->Mod, "None");
  NewRoom->x = NewRoom->y = 0;
  NewRoom->RoomW = NewRoom->RoomH = 100;

  NewRoom->RoomFloor = new short[NewRoom->RoomW*NewRoom->RoomH];
  if (!NewRoom->RoomFloor)
  {
    g_Done();
    printf("ZZT ][: Error initializing room floor map.\n");
    exit(1);
  }
  memset16(NewRoom->RoomFloor, 0, sizeof(short)*NewRoom->RoomW*NewRoom->RoomH);

  NewRoom->RoomObject = new short[NewRoom->RoomW*NewRoom->RoomH];
  if (!NewRoom->RoomObject)
  {
    g_Done();
    printf("ZZT ][: Error initializing room object map.\n");
    exit(1);
  }
  memset16(NewRoom->RoomObject, 0, sizeof(short)*NewRoom->RoomW*NewRoom->RoomH);

  NewRoom->RoomCeiling = NULL;
  NewRoom->LinkN = NULL;
  NewRoom->LinkS = NULL;
  NewRoom->LinkE = NULL;
  NewRoom->LinkW = NULL;
  memset(NewRoom->StrLinkN, 0, RNAME_MAX);
  memset(NewRoom->StrLinkS, 0, RNAME_MAX);
  memset(NewRoom->StrLinkE, 0, RNAME_MAX);
  memset(NewRoom->StrLinkW, 0, RNAME_MAX);
  NewRoom->Light     = 1;
  NewRoom->ReEntry   = 0;
  NewRoom->Time      = 0;
  NewRoom->MaxShots  = 0;
  NewRoom->Processed = 0;
  NewRoom->ObjListHead = NULL;
  NewRoom->ObjListCur  = NULL;
  NewRoom->next        = NULL;

  if (Cur)
  {
    Cur->next = NewRoom;
  } else
  {
    RoomHead = NewRoom;
    RoomCur  = RoomHead;
  }

  return NewRoom;
}

pObject AddGlobal(char *name)
{
  // Find end of linked list
  pObject Cur = GlobalHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  strncpy(NewObj->name, name, OBJNAME_MAX);
  memset(NewObj->bark, 0, BARK_MAX);
  NewObj->NameHash = Hash(NewObj->name);
  NewObj->barktime = 0;
  NewObj->x = NewObj->y = 0;
  NewObj->z = 1;
  NewObj->property = 0;
  NewObj->frame = 0;
  for (int i=0; i<LOCAL_MAX; i++)
    NewObj->local[i] = 0;
  NewObj->code = NULL;
  NewObj->codesize = 0;
  NewObj->codep = 0;
  NewObj->cycle = 1;
  NewObj->cyclel = 0;
  NewObj->light = 0;
  NewObj->queue = 0;
  NewObj->qsize = 0;
  NewObj->next = NULL;

  if (Cur)
    Cur->next = NewObj; else
    GlobalHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

int KillGlobal(pObject o)
{
  pObject Prev = GlobalHead;

  if (Prev == o)
  {
    GlobalHead = Prev->next;

    if (Prev->code) delete[] Prev->code;
    delete[] Prev;
  } else
  {
    while (Prev->next != o)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_OBJ;
    }

    pObject Cur = Prev->next;
    pObject Next = Prev->next->next;

    Prev->next = Next;
    if (Prev->code) delete[] Prev->code;
    delete[] Cur;
  }

  return SUCCESS;
}

pObject AddObj(char *objfile, pObject &head)
{
  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  if (objfile)
  {
    ifstream obj(objfile, ios::in | ios::binary);
    if (!obj || !objfile)
    {
      delete[] NewObj;
      return NULL;
    }

    obj.seekg(0, ios::end);
    int filesize = obj.tellg();
    obj.seekg(0, ios::beg);

    obj.read(id, 4);
    if (strncmp(id, ZZTVersion, 4))
    {
      delete[] NewObj;
      return NULL;
    }

    if (obj.get()==ObjNameOp)
    {
      char *str = NewObj->name;
      while (ch = obj.get())
        *str++ = ch;
      *str = 0;
    } else obj.seekg(-1, ios::cur);

    NewObj->code = new char[filesize];
    if (!NewObj->code)
    {
      delete[] NewObj;
      return NULL;
    }
    memset(NewObj->code, 0, filesize);

    obj.read(NewObj->code, filesize);
    NewObj->codesize = filesize;
    obj.close();
  } else
  {
    NewObj->code = NULL;
    NewObj->codesize = 0;
  }

  NewObj->NameHash = Hash(NewObj->name);
  memset(NewObj->bark, 0, BARK_MAX);
  NewObj->barktime = 0;
  NewObj->x = NewObj->y = 0;
  NewObj->z = 1;
  NewObj->property = 0;
  NewObj->frame = 0;
  for (int i=0; i<LOCAL_MAX; i++)
    NewObj->local[i] = 0;
  NewObj->codep = 0;
  NewObj->cycle = 3;
  NewObj->cyclel = 0;
  NewObj->light = 0;
  NewObj->queue = 0;
  NewObj->qsize = 0;
  NewObj->next = NULL;

  if (Cur)
    Cur->next = NewObj; else
    head = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pItem FindItem(char *name)
{
  short hash = Hash(name);

  for (int i=0; i<ItemNum; i++)
    if (ItemList[i].NameHash==hash)
      return &ItemList[i];
  return NULL;
}

pItem FindItem(short hash)
{
  for (int i=0; i<ItemNum; i++)
    if (ItemList[i].NameHash==hash)
      return &ItemList[i];
  return NULL;
}

pItem FindItemDesc(char *name)
{
  for (int i=0; i<ItemNum; i++)
    if (strcmp(name, ItemList[i].desc)==0)
      return &ItemList[i];
  return NULL;
}

pObject AddItemObj(tItem &item, pObject &head)
{
  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  ifstream obj(item.script, ios::in | ios::binary);
  if (!obj)
  {
    NewObj->code     = NULL;
    NewObj->codesize = 0;
  } else
  {
    obj.seekg(0, ios::end);
    int filesize = obj.tellg();
    obj.seekg(0, ios::beg);

    obj.read(id, 4);
    if (strncmp(id, ZZTVersion, 4))
    {
      delete[] NewObj;
      return NULL;
    }

    char *str = NewObj->name;
    while (ch = obj.get())
      *str++ = ch;
    *str = 0;

    NewObj->code = new char[filesize];
    if (!NewObj->code)
    {
      delete[] NewObj;
      return NULL;
    }
    memset(NewObj->code, 0, filesize);

    obj.read(NewObj->code, filesize);
    NewObj->codesize = filesize;
    obj.close();
  }
  NewObj->codep = 0;

  strncpy(NewObj->name, item.name, OBJNAME_MAX);
  NewObj->NameHash = Hash(NewObj->name);
  memset(NewObj->bark, 0, BARK_MAX);
  NewObj->barktime = 0;
  NewObj->x = NewObj->y = 0;
  NewObj->z = 1;
  NewObj->property = 0;

  if (item.flag[0]=='1') NewObj->property += objPlayer;
  if (item.flag[1]=='1') NewObj->property += objSolid;
  if (item.flag[2]=='1') NewObj->property += objBreakable;
  if (item.flag[3]=='1') NewObj->property += objSwimmer;
  if (item.flag[4]=='1') NewObj->property += objPushEW;
  if (item.flag[5]=='1') NewObj->property += objPushNS;
  if (item.flag[6]=='1') NewObj->property += objPushP;
  if (item.flag[7]=='1') NewObj->property += objInvisible;
  if (item.flag[8]=='1') NewObj->property += objItem;
  if (item.flag[9]=='1') NewObj->property += objFlight;

  NewObj->frame = item.frame;
  for (int i=0; i<LOCAL_MAX; i++)
    NewObj->local[i] = 0;
  NewObj->cycle    = 3;
  NewObj->cyclel   = 0;
  NewObj->light    = 0;
  NewObj->queue    = 0;
  NewObj->qsize    = 0;
  NewObj->next     = NULL;

  if (Cur)
    Cur->next = NewObj; else
    head = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject GetObjXY(int x, int y, pRoom r, int z)
{
  pObject Cur = r->ObjListHead;
  while (Cur)
  {
    if (x==Cur->x && y==Cur->y && z==Cur->z) return Cur;
    Cur = Cur->next;
  }
  return NULL;
}

pObject AddObjXY(int x, int y, char *name, pRoom r, int z)
{
  pItem it = FindItem(name);
  pObject src = FindObj(name, SystemHead);
  pObject success = NULL;

  switch (it->note)
  {
    case 0 : case 1 : case 2 : case 3 :
      success = BindObj(src, r);
      break;
    case 4 : // scroll
      success = BindScroll(src, r);
      break;
    case 5 : // object
      success = BindEmptyObj(src, r);
      break;
    case 6 : // water
      RoomCur->RoomObject[y*RoomCur->RoomW+x] = it->frame;
      break;
    case 7 : // forest
      RoomCur->RoomObject[y*RoomCur->RoomW+x] = it->frame;
      break;
    case 8 : // fake
      RoomCur->RoomFloor[y*RoomCur->RoomW+x] = it->frame;
      break;
    case 10 :
      RoomCur->RoomObject[y*RoomCur->RoomW+x] = it->frame;
      break;
  }

  if (success)
  {
    LastObj->x = x;
    LastObj->y = y;
    LastObj->z = z;
    return LastObj;
  }
  return NULL;
}

int KillObjXY(int x, int y, pRoom r)
{
  pObject o = GetObjXY(x, y, r);
  if (!o) return BAD_OBJ;
  KillObj(r, o);
  return SUCCESS;
}

pObject BindObj(pObject Src, pRoom r)
{
  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = r->ObjListHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash = Src->NameHash;
  memset(NewObj->bark, 0, BARK_MAX);
  NewObj->barktime = 0;
  NewObj->x = NewObj->y = 0;
  NewObj->z = Src->z;
  NewObj->property = Src->property;
  NewObj->frame = Src->frame;
  for (int i=0; i<LOCAL_MAX; i++)
    NewObj->local[i] = 0;
  NewObj->cycle = Src->cycle;
  NewObj->cyclel = Src->cyclel;
  NewObj->light = Src->light;
  NewObj->queue = Src->queue;
  NewObj->qsize = Src->qsize;
  NewObj->next = NULL;

  int csize = 5+strlen(Src->name);
  NewObj->code = new char[csize];
  if (!NewObj->code)
  {
    delete[] NewObj;
    return NULL;
  }
  memset(NewObj->code, 0, csize);

  NewObj->codep = 0;
  NewObj->codesize = csize;
  memcpy(NewObj->code, BindCode, 2);
  memcpy(NewObj->code+2, Src->name, strlen(Src->name));
//  NewObj->code[2+csize] = 0;

  if (Cur)
    Cur->next = NewObj; else
    r->ObjListHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject BindScroll(pObject Src, pRoom r)
{
  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = r->ObjListHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash = NewObj->NameHash;
  memset(NewObj->bark, 0, BARK_MAX);
  NewObj->barktime = 0;
  NewObj->x = NewObj->y = 0;
  NewObj->z = Src->z;
  NewObj->property = Src->property;
  NewObj->property |= objLocked;
  NewObj->frame = Src->frame;
  for (int i=0; i<LOCAL_MAX; i++)
    NewObj->local[i] = 0;
  NewObj->cycle = Src->cycle;
  NewObj->cyclel = Src->cyclel;
  NewObj->light = Src->light;
  NewObj->queue = Src->queue;
  NewObj->qsize = Src->qsize;
  NewObj->next = NULL;
  NewObj->codesize = Src->codesize;
  NewObj->codep = 0;
  NewObj->code = new char[NewObj->codesize];
  if (!NewObj->code)
  {
    delete[] NewObj;
    return NULL;
  }
  memset(NewObj->code, 0, NewObj->codesize);
  memcpy(NewObj->code, Src->code, NewObj->codesize);

  if (Cur)
    Cur->next = NewObj; else
    r->ObjListHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject BindEmptyObj(pObject Src, pRoom r)
{
  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = r->ObjListHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash = Src->NameHash;
  memset(NewObj->bark, 0, BARK_MAX);
  NewObj->barktime = 0;
  NewObj->x = NewObj->y = 0;
  NewObj->z = Src->z;
  NewObj->property = Src->property;
  NewObj->property |= objLocked;
  NewObj->frame = Src->frame;
  for (int i=0; i<LOCAL_MAX; i++)
    NewObj->local[i] = 0;
  NewObj->cycle = Src->cycle;
  NewObj->cyclel = Src->cyclel;
  NewObj->light = Src->light;
  NewObj->queue = Src->queue;
  NewObj->qsize = Src->qsize;
  NewObj->next = NULL;

  NewObj->code = NULL;
  NewObj->codep = 0;
  NewObj->codesize = 0;

  if (Cur)
    Cur->next = NewObj; else
    r->ObjListHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject CopyObj(char *name, pObject &head, pRoom r)
{
  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = r->ObjListHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  pObject Src = FindObj(name, head);
  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash = Src->NameHash;
  memset(NewObj->bark, 0, BARK_MAX);
  NewObj->barktime = 0;
  NewObj->x = NewObj->y = 0;
  NewObj->z = Src->z;
  NewObj->property = Src->property;
  NewObj->frame    = Src->frame;
  for (int i=0; i<LOCAL_MAX; i++)
    NewObj->local[i] = 0;
  NewObj->cycle = Src->cycle;
  NewObj->cyclel = Src->cyclel;
  NewObj->light = Src->light;
  NewObj->queue = Src->queue;
  NewObj->qsize = Src->qsize;
  NewObj->next = NULL;

  NewObj->codesize = Src->codesize;
  NewObj->code = new char[Src->codesize];
  if (!NewObj->code)
  {
    delete[] NewObj;
    return NULL;
  }
  memcpy(NewObj->code, Src->code, Src->codesize);

  NewObj->codep = 0;

  if (Cur)
    Cur->next = NewObj; else
    r->ObjListHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

void LoadObj(char *objfile, pObject o)
{
  char id[4];
  char ch;

  ifstream obj(objfile, ios::in | ios::binary);
  if (!obj || !objfile) return;

  obj.seekg(0, ios::end);
  int filesize = obj.tellg();
  obj.seekg(0, ios::beg);

  obj.read(id, 4);
  if (strncmp(id, ZZTVersion, 4)) return;

  char *str = o->name;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  o->NameHash = Hash(o->name);

  o->code = new char[filesize];
  if (!o->code) return;
  obj.read(o->code, filesize);
  o->codesize = filesize;
  obj.close();

  LastObj = o;
}

pRoom FindRoom(char *name)
{
  pRoom Cur;

  if (strnicmp("--", name, RNAME_MAX)==0) return RoomCur;

  short hash = Hash(name);

  Cur = RoomHead;
  while (Cur)
  {
    if (Cur->NameHash==hash) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

pRoom FindRoom(short hash)
{
  pRoom Cur;

  Cur = RoomHead;
  while (Cur)
  {
    if (Cur->NameHash==hash) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

pVar FindVar(char *name)
{
  short hash = Hash(name);

  pVar Cur = VarHead;
  while (Cur)
  {
    if (Cur->NameHash==hash) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

int sizeofVar(pVar v)
{
  if (!v) return BAD_OBJ;

  return sizeof(tVar);
}

int sizeofVarList(void)
{
  int total=0;

  pVar Cur = VarHead;
  while (Cur)
  {
    total += sizeofVar(Cur);
    Cur = Cur->next;
  }

  return total;
}

pObject FindObj(char *name, pObject head)
{
  pObject Cur;

  short hash = Hash(name);

  // Search for globals
  Cur = GlobalHead;
  while (Cur)
  {
    if (Cur->NameHash==hash)
      return Cur;
    Cur = Cur->next;
  }

  // Search room objects
  Cur = head;
  while (Cur)
  {
    if (Cur->NameHash==hash)
    {
      LastObj = Cur;
      return Cur;
    }
    Cur = Cur->next;
  }

  return NULL;
}

pObject FindObj(short hash, pObject head)
{
  pObject Cur;

  // Search for globals
  Cur = GlobalHead;
  while (Cur)
  {
    if (Cur->NameHash==hash)
      return Cur;
    Cur = Cur->next;
  }

  // Search room objects
  Cur = head;
  while (Cur)
  {
    if (Cur->NameHash==hash)
    {
      LastObj = Cur;
      return Cur;
    }
    Cur = Cur->next;
  }

  return NULL;
}

int sizeofObj(pObject o)
{
  if (!o) return BAD_OBJ;

  return sizeof(tObject)+o->codesize;
}

int sizeofObjList(pObject head)
{
  int total=0;

  pObject Cur = head;
  while (Cur)
  {
    total += sizeofObj(Cur);
    Cur = Cur->next;
  }

  return total;
}

int sizeofRoom(pRoom r)
{
  if (!r) return BAD_ROOM;

  int size = sizeof(tRoom);
  if (r->RoomFloor)   size += r->RoomW * r->RoomH * 2;
  if (r->RoomObject)  size += r->RoomW * r->RoomH * 2;
  if (r->RoomCeiling) size += r->RoomW * r->RoomH * 2;
  size += sizeofObjList(r->ObjListHead);

  return size;
}

int sizeofRoomList(void)
{
  int total = 0;

  pRoom Cur = RoomHead;
  while (Cur)
  {
    total += sizeofRoom(Cur);
    Cur = Cur->next;
  }

  return total;
}

int KillVar(pVar v)
{
  pVar Prev = VarHead;

  if (Prev == v)
  {
    VarHead = Prev->next;
    delete[] Prev;
  } else
  {
    while (Prev->next != v)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_VAR;
    }

    pVar Cur = Prev->next;
    pVar Next = Prev->next->next;

    Prev->next = Next;
    delete[] Cur;
  }

  return SUCCESS;
}

int KillRoom(pRoom r)
{
  pRoom Prev = RoomHead;

  if (Prev == r)
  {
    RoomHead = Prev->next;

    if (Prev->RoomFloor)   delete[] Prev->RoomFloor;
    if (Prev->RoomObject)  delete[] Prev->RoomObject;
    if (Prev->RoomCeiling) delete[] Prev->RoomCeiling;
    pObject CurObj = Prev->ObjListHead;
    while (CurObj)
      KillObj(Prev, CurObj);

    delete[] Prev;
  } else
  {
    while (Prev->next != r)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_ROOM;
    }

    pRoom Cur = Prev->next;
    pRoom Next = Prev->next->next;
    Prev->next = Next;

    if (Cur->RoomFloor)   delete[] Cur->RoomFloor;
    if (Cur->RoomObject)  delete[] Cur->RoomObject;
    if (Cur->RoomCeiling) delete[] Cur->RoomCeiling;
    pObject CurObj = Cur->ObjListHead;
    while (CurObj)
      KillObj(Cur, CurObj);

    delete[] Cur;
  }

  return SUCCESS;
}

int KillObj(pRoom r, pObject o)
{
  pObject Prev = r->ObjListHead;

  if (Prev == o)
  {
    r->ObjListHead = Prev->next;

    if (Prev->property & objBound == objBound)
      if (Prev->code) delete[] Prev->code;
    delete[] Prev;
  } else
  {
    while (Prev->next != o)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_OBJ;
    }

    pObject Cur = Prev->next;
    pObject Next = Prev->next->next;

    Prev->next = Next;
    if (Prev->property & objBound != objBound)
      if (Prev->code) delete[] Prev->code;
    delete[] Cur;
  }

  return SUCCESS;
}

int Spawn(char *name, int x, int y)
{
  int w = RoomCur->RoomW;
//int h = RoomCur->RoomH;

  // if another object is in the way, can't spawn
  short *objl = RoomCur->RoomObject;
  pObject touched = GetObjXY(x, y, RoomCur);
  if (objl[y*w+x] || touched) return 0;

  pObject src = FindObj(name, SystemHead);

  // first, try to bind the object to an internal object
  if (BindObj(src, RoomCur)) return 1;

  // secondly, try to copy the object from another in the same room
  if (CopyObj(name, RoomCur->ObjListHead, RoomCur)) return 1;

  // Object doesn't exist
  return 0;
}

int GetFrame(int t)
{
  if (t<TILE_MAX) return t;

  return ItemList[t-TILE_MAX].frame;
}

char *GetName(int t)
{
  if (t<TILE_MAX) return NULL;

  return ItemList[t-TILE_MAX].name;
}

