// Watcom includes
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>

// HGL includes
#include "mem32.h"
#include "tile.h"

// Z2 includes
#include "compiler.h"
#include "objects.h"
#include "editor.h"
#include "kernel.h"
#include "global.h"
#include "sound.h"

pRoom   SystemRoom, GlobalRoom;
pRoom   RoomHead, RoomCur, StartRoom;
pVar    VarHead;
pObject LastObj, Z2Help;
char    DefName[80], DefFile[80], TileFile[80], SysRoom[80];

void InitEngine(void)
{
  if (ErrorLog) printf("objects.cpp -- InitEngine();\n");

  RoomHead = RoomCur = NULL;
  VarHead = NULL;
  SoundHead = NULL;
  Z2Help = new tObject;

  SystemRoom = InitRoom("System", "System Room");
  GlobalRoom = InitRoom("Global", "Global Room");

  SystemRoom->ObjHead = SystemRoom->ObjHead;
  GlobalRoom->ObjHead = GlobalRoom->ObjHead;
}

void InitTiles(void)
{
  if (ErrorLog) printf("objects.cpp -- InitTiles();\n");

  t_Init(TILE_MAX);

  TotTiles = 0;
  KillTiles();
}

int LoadTiles(char *fn)
{
  if (ErrorLog) printf("objects.cpp -- LoadTiles(%s);\n", fn);

  if (!fn || !strlen(fn)) return -1;

  for (int i=0; i<TotTiles; i++)
  {
    Tiles[i].x = 0;
    Tiles[i].y = 0;
    if (Tiles[i].Data) delete[] Tiles[i].Data;
  }

  TotTiles = t_Load(TileFile);

  if (!TotTiles) return -1;
  
  Tile_W = Tiles[0].x;
  Tile_H = Tiles[0].y;
  Tile_S = Tile_W*Tile_H;
  t_TileSize(Tile_W, Tile_H);
  RoomData.ReInit();

  return 0;
}

void KillTiles(void)
{
  if (ErrorLog) printf("objects.cpp -- KillTiles();\n");

  for (int i=0; i<TotTiles; i++)
  {
    Tiles[i].x = 0;
    Tiles[i].y = 0;
    if (Tiles[i].Data) delete[] Tiles[i].Data;
  }
  if (Tile_S)
  {
    Tiles[0].x = Tile_W;
    Tiles[0].y = Tile_H;
    Tiles[0].Data = new char[Tile_S];
    memset(Tiles[0].Data, 0, Tile_S);

    TotTiles = 1;
    RoomData.ReInit();
  }
}

void KillSystem(void)
{
  if (ErrorLog) printf("objects.cpp -- KillSystem();\n");

  // Clear system rooms
  pRoom CurR = SystemRoom;
  pObject CurRO = SystemRoom->ObjHead;
  pObject next_o;
  while (CurRO)
  {
    next_o = CurRO->next;
    KillObj(CurR, CurRO);
    CurRO = next_o;
  }
  SystemRoom->ObjHead = NULL;
}

void KillRooms(void)
{
  if (ErrorLog) printf("objects.cpp -- KillRooms();\n");

  // Clear global rooms
  pRoom CurR = GlobalRoom;
  pObject CurRO = GlobalRoom->ObjHead;
  pObject next_o;
  while (CurRO)
  {
    next_o = CurRO->next;
    KillObj(CurR, CurRO);
    CurRO = next_o;
  }
  GlobalRoom->ObjHead = NULL;

  // Clear room list
  CurR = RoomHead;
  pRoom next;
  while (CurR)
  {
    // Clear room object list
    CurRO = CurR->ObjHead;
    while (CurRO)
    {
      next_o = CurRO->next;
      KillObj(CurR, CurRO);
      CurRO = next_o;
    }
    CurR->ObjHead = NULL;

    next = CurR->next;
    delete[] CurR;
    CurR = next;
  }
  RoomHead = NULL;
}

void KillVars(void)
{
  if (ErrorLog) printf("objects.cpp -- KillVars();\n");

  // Clear variable list
  pVar CurV = VarHead;
  pVar next;
  while (CurV)
  {
    next = CurV->next;
    delete[] CurV;
    CurV = next;
  }
  VarHead = NULL;
}

void DoneEngine(void)
{
  if (ErrorLog) printf("objects.cpp -- DoneEngine();\n");

  KillRooms();
  KillSystem();
  KillVars();
  KillTiles();
  t_Done();
}

pVar AddVar(char *name, short v, int ov)
{
  if (ErrorLog) printf("objects.cpp -- AddVar(%d, %d, %d);\n", name, v, ov);

  // Check if variables exists
  pVar OldVar = FindVar(name);
  if (OldVar)
  {
    if (ov) OldVar->value = v;
    return OldVar;
  }

  pVar NewVar = NULL;

  // Find end of linked list
  pVar Cur = VarHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  NewVar = new tVar;
  if (!NewVar) return NULL;

  strncpy(NewVar->name, name, VAR_MAX);
  NewVar->value    = v;
  NewVar->NameHash = Hash(name);
  NewVar->next     = NULL;

  if (Cur)
    Cur->next = NewVar; else
    VarHead = NewVar;

  return NewVar;
}

int masterSaveRoom(ofstream &obj, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- masterSaveRoom((ofstream), (pRoom)0x%X);\n", r);

  int   length, rlep, i, saved;
  short seed;
  short *FloorRLE, *ObjectRLE, *CeilingRLE;

  obj.write(RoomVer, 4);
  obj.write(r->name, strlen(r->name));
  obj.put((char)NULL);
  obj.write(r->Desc, strlen(r->Desc));
  obj.put((char)NULL);
  obj.write(r->Mod, strlen(r->Mod));
  obj.put((char)NULL);
  obj.write((unsigned char *)&r->RoomW, sizeof(r->RoomW));
  obj.write((unsigned char *)&r->RoomH, sizeof(r->RoomH));

  FloorRLE = new short[r->RoomW*r->RoomH];
  rlep = 1;
  i = 0;
  while (i<r->RoomW*r->RoomH)
  {
    length = 0;
    seed = r->RoomFloor[i];
    while (r->RoomFloor[i]==seed)
    {
      i++;
      length++;
    }
    if (length>1)
    {
      FloorRLE[rlep++] = 0xFFFF;
      FloorRLE[rlep++] = length;
      FloorRLE[rlep++] = seed;
    } else
    if (length==1)
      FloorRLE[rlep++] = seed;
  }
  FloorRLE[0] = rlep-1;

  obj.write((unsigned char *) FloorRLE, (FloorRLE[0]+1)*2);
  if (FloorRLE) delete[] FloorRLE;

  ObjectRLE = new short[r->RoomW*r->RoomH];
  rlep = 1;
  i = 0;
  while (i<r->RoomW*r->RoomH)
  {
    length = 0;
    seed = r->RoomObject[i];
    while (r->RoomObject[i]==seed && i<r->RoomW*r->RoomH)
    {
      i++;
      length++;
    }
    if (length>1)
    {
      ObjectRLE[rlep++] = 0xFFFF;
      ObjectRLE[rlep++] = length;
      ObjectRLE[rlep++] = seed;
    } else
    if (length==1)
      ObjectRLE[rlep++] = seed;
  }
  ObjectRLE[0] = rlep-1;

  obj.write((unsigned char *) ObjectRLE, (ObjectRLE[0]+1)*2);
  if (ObjectRLE) delete[] ObjectRLE;

  if (r->RoomCeiling)
  {
    CeilingRLE = new short[r->RoomW*r->RoomH];
    rlep = 1;
    i = 0;
    while (i<r->RoomW*r->RoomH)
    {
      length = 0;
      seed = r->RoomCeiling[i];
      while (r->RoomCeiling[i]==seed && i<r->RoomW*r->RoomH)
      {
        i++;
        length++;
      }
      if (length>1)
      {
        CeilingRLE[rlep++] = 0xFFFF;
        CeilingRLE[rlep++] = length;
        CeilingRLE[rlep++] = seed;
      } else
      if (length==1)
        CeilingRLE[rlep++] = seed;
    }
    CeilingRLE[0] = rlep-1;

    obj.write((unsigned char *) CeilingRLE, (CeilingRLE[0]+1)*2);
    if (CeilingRLE) delete[] CeilingRLE;
  } else obj.write("NULL\0", 5);

  if (r->LinkN)
  {
    obj.write(r->LinkN->name, strlen(r->LinkN->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);
  if (r->LinkS)
  {
    obj.write(r->LinkS->name, strlen(r->LinkS->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);
  if (r->LinkE)
  {
    obj.write(r->LinkE->name, strlen(r->LinkE->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);
  if (r->LinkW)
  {
    obj.write(r->LinkW->name, strlen(r->LinkW->name));
    obj.put((char)NULL);
  } else obj.write("NULL\0", 5);

  obj.write((unsigned char *) &r->Light, sizeof(r->Light));
  obj.write((unsigned char *) &r->ReEntry, sizeof(r->ReEntry));
  obj.write((unsigned char *) &r->Time, sizeof(r->Time));
  obj.write((unsigned char *) &r->MaxShots, sizeof(r->MaxShots));
  obj.write((unsigned char *) &r->Processed, sizeof(r->Processed));

  pObject Cur = r->ObjHead;
  while (Cur)
  {
    obj.write(Cur->name, strlen(Cur->name));
    obj.put((char)NULL);
    obj.write(Cur->bark, strlen(Cur->bark));
    obj.put((char)NULL);
    obj.write((unsigned char *) &Cur->barktime, sizeof(Cur->barktime));
    obj.write((unsigned char *) &Cur->x, sizeof(Cur->x));
    obj.write((unsigned char *) &Cur->y, sizeof(Cur->y));
    obj.write((unsigned char *) &Cur->z, sizeof(Cur->z));

    ushort property = 0;
    if (Cur->objPlayer)    property += 0x0001;
    if (Cur->objSolid)     property += 0x0002;
    if (Cur->objBreakable) property += 0x0004;
    if (Cur->objSwimmer)   property += 0x0008;
    if (Cur->objPusher)    property += 0x0010;
    if (Cur->objPushEW)    property += 0x0020;
    if (Cur->objPushNS)    property += 0x0040;
    if (Cur->objPushP)     property += 0x0080;
    if (Cur->objInvisible) property += 0x0100;
    if (Cur->objItem)      property += 0x0200;
    if (Cur->objFlight)    property += 0x0400;
    if (Cur->objBound)     property += 0x0800;
    if (Cur->objLocked)    property += 0x1000;

    obj.write((unsigned char *) &property, sizeof(property));
    obj.write((unsigned char *) &Cur->frame, sizeof(Cur->frame));
    obj.write(Cur->local, LOCAL_MAX);
    obj.write((unsigned char *) &Cur->codesize, sizeof(Cur->codesize));
    obj.write((unsigned char *) &Cur->codep, sizeof(Cur->codep));
    obj.write(Cur->code, Cur->codesize);
    obj.write((unsigned char *) &Cur->cycle, sizeof(Cur->cycle));
    obj.write((unsigned char *) &Cur->cyclel, sizeof(Cur->cyclel));
    obj.write((unsigned char *) &Cur->light, sizeof(Cur->light));
    obj.write((unsigned char *) &Cur->queue, sizeof(Cur->queue));
    obj.write((unsigned char *) &Cur->qsize, sizeof(Cur->qsize));
    Cur = Cur->next;
  }
  obj.write("NULL\0", 5);

  return SUCCESS;
}

int masterLoadRoom(ifstream &obj, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- masterSaveRoom((ifstream), (pRoom)0x%X);\n", r);

  char temp[RNAME_MAX];
  char id[4], ch;
  char *str;
  int  oldw, oldh;

  obj.read(id, 4);
  if (strncmp(id, RoomNULL, 4)==0) return -42;
  if (strncmp(id, RoomVer, 4)) return FILE_ERR;

  str = r->name;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  r->NameHash = Hash(r->name);

  str = r->Desc;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  str = r->Mod;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  oldw = r->RoomW;
  oldh = r->RoomH;
  obj.read((unsigned char *) &r->RoomW, sizeof(short));
  obj.read((unsigned char *) &r->RoomH, sizeof(short));

  int i;
  short rle_len, in, out, len, j;
  short seed;

  if (oldw != r->RoomW || oldh != r->RoomH)
  {
    if (r->RoomFloor) delete[] r->RoomFloor;
    r->RoomFloor = new short[r->RoomW*r->RoomH];
  }
  if (!r->RoomFloor)
  {
    g_Done();
    printf("Z2: Error initializing room floor map.\n");
    exit(1);
  }
  obj.read((unsigned char *) &rle_len, sizeof(short));

  out = 0;
  for (i=0; i<rle_len; i++)
  {
    obj.read((unsigned char *) &in, sizeof(short));
    switch (in)
    {
      case 0xFFFF :
        obj.read((unsigned char *)&len, sizeof(short));
        obj.read((unsigned char *)&seed, sizeof(short));
        i += 2;
        for (j=0; j<len; j++)
          r->RoomFloor[out++] = seed;
        break;
      default :
        r->RoomFloor[out++] = in;
        break;
    }
  }

  if (oldw != r->RoomW || oldh != r->RoomH)
  {
    if (r->RoomObject) delete[] r->RoomObject;
    r->RoomObject = new short[r->RoomW*r->RoomH];
  }
  if (!r->RoomObject)
  {
    printf("Z2: Error initializing room object map.\n");
    exit(1);
  }
  obj.read((unsigned char *) &rle_len, sizeof(short));

  out = 0;
  for (i=0; i<rle_len; i++)
  {
    obj.read((unsigned char *) &in, sizeof(short));
    switch (in)
    {
      case 0xFFFF :
        obj.read((unsigned char *)&len, sizeof(short));
        obj.read((unsigned char *)&seed, sizeof(short));
        i += 2;
        for (j=0; j<len; j++)
          r->RoomObject[out++] = seed;
        break;
      default :
        r->RoomObject[out++] = in;
        break;
    }
  }

  char check[5];
  obj.read(check, 5);
  if (strcmp(check, "NULL")==0) r->RoomCeiling = NULL; else
  {
    obj.seekg(-5, ios::cur);

    if (r->RoomCeiling) delete[] r->RoomCeiling;
    r->RoomCeiling = new short[r->RoomW*r->RoomH];

    if (!r->RoomCeiling)
    {
      g_Done();
      printf("Z2: Error initializing room ceiling map.\n");
      printf("    %s -- line %d", __FILE__, __LINE__);
      exit(1);
    }
    obj.read((unsigned char *) &rle_len, sizeof(short));

    out = 0;
    for (i=0; i<rle_len; i++)
    {
      obj.read((unsigned char *) &in, sizeof(short));
      switch (in)
      {
        case 0xFFFF :
          obj.read((unsigned char *)&len, sizeof(short));
          obj.read((unsigned char *)&seed, sizeof(short));
          i += 2;
          for (j=0; j<len; j++)
            r->RoomCeiling[out++] = seed;
          break;
        default :
          r->RoomCeiling[out++] = in;
          break;
      }
    }
  }

  str = r->StrLinkN;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(r->StrLinkN, "NULL")==0)
  {
    r->LinkN = NULL;
    memset(r->StrLinkN, 0, RNAME_MAX);
  } else r->LinkN = FindRoom(temp);

  str = r->StrLinkS;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(r->StrLinkS, "NULL")==0)
  {
    r->LinkS = NULL;
    memset(r->StrLinkS, 0, RNAME_MAX);
  } else r->LinkS = FindRoom(temp);

  str = r->StrLinkE;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(r->StrLinkE, "NULL")==0)
  {
    r->LinkE = NULL;
    memset(r->StrLinkE, 0, RNAME_MAX);
  } else r->LinkE = FindRoom(temp);

  str = r->StrLinkW;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;
  if (stricmp(temp, "NULL")==0)
  {
    r->LinkW = NULL;
    memset(r->StrLinkW, 0, RNAME_MAX);
  } else r->LinkW = FindRoom(temp);

  obj.read((unsigned char *) &r->Light, sizeof(r->Light));
  obj.read((unsigned char *) &r->ReEntry, sizeof(r->ReEntry));
  obj.read((unsigned char *) &r->Time, sizeof(r->Time));
  obj.read((unsigned char *) &r->MaxShots, sizeof(r->MaxShots));
  obj.read((unsigned char *) &r->Processed, sizeof(r->Processed));

  // Clear objects list
  pObject CurO = r->ObjHead;
  while (CurO)
  {
    delete[] CurO;
    CurO = CurO->next;
  }
  r->ObjHead = NULL;

  int done = 0;
  while (!done)
  {
    str = temp;
    while (ch = obj.get())
      *str++ = ch;
    *str = 0;

    if (strcmp(temp, "NULL"))
    {
      pObject Cur = AddObj(NULL, r->ObjHead);

      strcpy(Cur->name, temp);
      Cur->NameHash = Hash(Cur->name);
      str = Cur->bark;
      while (ch = obj.get())
        *str++ = ch;
      *str = 0;
      obj.read((unsigned char *) &Cur->barktime, sizeof(Cur->barktime));
      obj.read((unsigned char *) &Cur->x, sizeof(Cur->x));
      obj.read((unsigned char *) &Cur->y, sizeof(Cur->y));
      obj.read((unsigned char *) &Cur->z, sizeof(Cur->y));

      ushort property;
      obj.read((unsigned char *) &property, sizeof(property));

      Cur->objPlayer    = (property & 0x0001)==0x0001;
      Cur->objSolid     = (property & 0x0002)==0x0002;
      Cur->objBreakable = (property & 0x0004)==0x0004;
      Cur->objSwimmer   = (property & 0x0008)==0x0008;
      Cur->objPusher    = (property & 0x0010)==0x0010;
      Cur->objPushEW    = (property & 0x0020)==0x0020;
      Cur->objPushNS    = (property & 0x0040)==0x0040;
      Cur->objPushP     = (property & 0x0080)==0x0080;
      Cur->objInvisible = (property & 0x0100)==0x0100;
      Cur->objItem      = (property & 0x0200)==0x0200;
      Cur->objFlight    = (property & 0x0400)==0x0400;
      Cur->objBound     = (property & 0x0800)==0x0800;
      Cur->objLocked    = (property & 0x1000)==0x1000;

      obj.read((unsigned char *) &Cur->frame, sizeof(Cur->frame));
      obj.read(Cur->local, LOCAL_MAX);
      obj.read((unsigned char *) &Cur->codesize, sizeof(Cur->codesize));
      obj.read((unsigned char *) &Cur->codep, sizeof(Cur->codep));
      Cur->code = new char[Cur->codesize];
      obj.read(Cur->code, Cur->codesize);
      obj.read((unsigned char *) &Cur->cycle, sizeof(Cur->cycle));
      obj.read((unsigned char *) &Cur->cyclel, sizeof(Cur->cyclel));
      obj.read((unsigned char *) &Cur->light, sizeof(Cur->light));
      obj.read((unsigned char *) &Cur->queue, sizeof(Cur->queue));
      obj.read((unsigned char *) &Cur->qsize, sizeof(Cur->qsize));
      Cur = Cur->next;
    } else done = 1;
  }

  return SUCCESS;
}

int SaveRoom(char *fname, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- SaveRoom(%s, (pRoom)0x%X);\n", fname, r);

  if (!r) return BAD_ROOM;

  ofstream obj(fname, ios::out | ios::binary);
  if (!obj) return FILE_ERR;

  int ret = masterSaveRoom(obj, r);

  obj.close();
  return ret;
}

void UpdateLinks(void)
{
  if (ErrorLog) printf("objects.cpp -- UpdateLinks();\n");

  pRoom Cur = RoomHead;

  while (Cur)
  {
    if (strlen(Cur->StrLinkN))
    {
      Cur->LinkN = FindRoom(Cur->StrLinkN);
      memset(Cur->StrLinkN, 0, RNAME_MAX);
    }
    if (strlen(Cur->StrLinkS))
    {
      Cur->LinkS = FindRoom(Cur->StrLinkS);
      memset(Cur->StrLinkS, 0, RNAME_MAX);
    }
    if (strlen(Cur->StrLinkE))
    {
      Cur->LinkE = FindRoom(Cur->StrLinkE);
      memset(Cur->StrLinkE, 0, RNAME_MAX);
    }
    if (strlen(Cur->StrLinkW))
    {
      Cur->LinkW = FindRoom(Cur->StrLinkW);
      memset(Cur->StrLinkW, 0, RNAME_MAX);
    }
    Cur = Cur->next;
  }
}

int SaveWorld(char *fname, char *desc)
{
  if (ErrorLog) printf("objects.cpp -- SaveWorld(%s, %s);\n", fname, desc);

  ofstream obj(fname, ios::out | ios::binary);
  if (!obj) return FILE_ERR;

  // world header
  obj.write(WorldVer, 4);
  obj.write(desc, strlen(desc));
  obj.put((char)NULL);
  if (StartRoom) StartRoom = RoomHead;
  obj.write(StartRoom->name, strlen(StartRoom->name));
  obj.put((char)NULL);
  masterSaveRoom(obj, SystemRoom);
  masterSaveRoom(obj, GlobalRoom);
  
  // save rooms
  pRoom r = RoomHead;
  while (r)
  {
    masterSaveRoom(obj, r);
    r = r->next;
  }
  obj.write(RoomNULL, 4);

  obj.close();
  return SUCCESS;
}

int LoadWorld(char *fname)
{
  if (ErrorLog) printf("objects.cpp -- LoadWorld(%s);\n", fname);

  char temp[RNAME_MAX];
  char id[4], ch;
  char *str;
  pRoom r;
  int stop=0;

  ifstream obj(fname, ios::in | ios::binary);
  if (!obj || !fname) return FILE_ERR;

  // check for validity
  obj.read(id, 4);
  if (strncmp(id, WorldVer, 4))
    return FILE_ERR;

  // Read world description
  str = temp;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  // Read start room
  str = temp;
  while (ch = obj.get())
    *str++ = ch;
  *str = 0;

  // Clear room list
  KillRooms();
  KillSystem();

  // Load system/global
  masterLoadRoom(obj, SystemRoom);
  masterLoadRoom(obj, GlobalRoom);

  // We should be right at primary chunkage.
  while (!stop)
  {
    r = AddRoom("Error", "Error Loading World");
    if (!r) return NO_MEM;
    stop = masterLoadRoom(obj, r);
    switch (stop)
    {
      case SUCCESS :
        stop = 0;
        break;
      case FILE_ERR :
        g_Done();
        printf("Z2: Error loading world.\n");
        exit(1);
        break;
      case -42 :
        KillRoom(r);
        stop = 1;
        break;
    }
  }

  // define the starting room
  StartRoom = FindRoom(temp);
  if (StartRoom) StartRoom = RoomHead;

  UpdateLinks();             // updates all links

  obj.close();
  return SUCCESS;
}

pRoom AddRoom(char *fname)
{
  if (ErrorLog) printf("objects.cpp -- AddRoom(%s);\n", fname);

  char temp[RNAME_MAX];
  char *str, ch;
  char id[4];

  // Find end of linked list
  pRoom Cur = RoomHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pRoom NewRoom = new tRoom;
  if (!NewRoom) return NULL;

  ifstream obj(fname, ios::in | ios::binary);
  if (!obj || !fname)
  {
    delete[] NewRoom;
    return NULL;
  }

  masterLoadRoom(obj, r);
  obj.close();
  r->next = NULL;

  if (Cur)
    Cur->next = NewRoom; else
    RoomHead = NewRoom;

  return NewRoom;
}

int LoadRoom(char *fname, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- LoadRoom(%s, (pRoom)0x%X);\n", fname, r);

  char temp[RNAME_MAX];
  char *str, ch;
  char id[4];
  int  oldw, oldh;

  ifstream obj(fname, ios::in | ios::binary);
  if (!obj || !fname) return FILE_ERR;

  int ret = masterLoadRoom(obj, r);

  obj.close();
  return ret;
}

pRoom AddRoom(char *name, char *desc)
{
  if (ErrorLog) printf("objects.cpp -- AddRoom(%s, %s);\n", name, desc);

  pRoom NewRoom = new tRoom;
  if (!NewRoom) return NULL;
  memset(NewRoom, 0, sizeof(tRoom));

  if (!strlen(name) || !name)
    strcpy(NewRoom->name, "Untitled"); else
    strncpy(NewRoom->name, name, RNAME_MAX);
  if (!strlen(desc) || !desc)
    strncpy(NewRoom->Desc, name, RDESC_MAX); else
    strncpy(NewRoom->Desc, desc, RDESC_MAX);

  NewRoom->NameHash = Hash(NewRoom->name);
  strcpy(NewRoom->Mod, "None");
  NewRoom->RoomW = NewRoom->RoomH = 100;

  NewRoom->RoomFloor = new short[NewRoom->RoomW*NewRoom->RoomH];
  if (!NewRoom->RoomFloor)
  {
    g_Done();
    printf("Z2: Error initializing room floor map.\n");
    exit(1);
  }
  memset16(NewRoom->RoomFloor, 0, NewRoom->RoomW*NewRoom->RoomH*sizeof(short));

  NewRoom->RoomObject = new short[NewRoom->RoomW*NewRoom->RoomH];
  if (!NewRoom->RoomObject)
  {
    g_Done();
    printf("Z2: Error initializing room object map.\n");
    exit(1);
  }
  memset16(NewRoom->RoomObject, 0, NewRoom->RoomW*NewRoom->RoomH*sizeof(short));

  NewRoom->LinkN     = NULL;
  NewRoom->LinkS     = NULL;
  NewRoom->LinkE     = NULL;
  NewRoom->LinkW     = NULL;
  NewRoom->Light     = 1;
  NewRoom->ReEntry   = 0;
  NewRoom->Time      = 0;
  NewRoom->MaxShots  = 0;
  NewRoom->Processed = 0;
  NewRoom->ObjHead   = NULL;
  NewRoom->ObjCur    = NULL;
  NewRoom->next      = NULL;

  // Find end of linked list
  pRoom Cur = RoomHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  if (Cur)
  {
    Cur->next = NewRoom;
  } else
  {
    RoomHead = NewRoom;
    RoomCur  = RoomHead;
  }

  return NewRoom;
}

pRoom InitRoom(char *name, char *desc)
{
  if (ErrorLog) printf("objects.cpp -- InitRoom(%s, %s);\n", name, desc);

  pRoom NewRoom = new tRoom;
  if (!NewRoom) return NULL;
  memset(NewRoom, 0, sizeof(tRoom));

  if (!strlen(name) || !name)
    strcpy(NewRoom->name, "Untitled"); else
    strncpy(NewRoom->name, name, RNAME_MAX);
  if (!strlen(desc) || !desc)
    strncpy(NewRoom->Desc, name, RDESC_MAX); else
    strncpy(NewRoom->Desc, desc, RDESC_MAX);

  NewRoom->NameHash = Hash(NewRoom->name);
  strcpy(NewRoom->Mod, "None");
  NewRoom->RoomW = NewRoom->RoomH = 100;

  NewRoom->RoomFloor = new short[NewRoom->RoomW*NewRoom->RoomH];
  if (!NewRoom->RoomFloor)
  {
    g_Done();
    printf("Z2: Error initializing room floor map.\n");
    exit(1);
  }
  memset16(NewRoom->RoomFloor, 0, NewRoom->RoomW*NewRoom->RoomH*sizeof(short));

  NewRoom->RoomObject = new short[NewRoom->RoomW*NewRoom->RoomH];
  if (!NewRoom->RoomObject)
  {
    g_Done();
    printf("Z2: Error initializing room object map.\n");
    exit(1);
  }
  memset16(NewRoom->RoomObject, 0, NewRoom->RoomW*NewRoom->RoomH*sizeof(short));

  NewRoom->LinkN     = NULL;
  NewRoom->LinkS     = NULL;
  NewRoom->LinkE     = NULL;
  NewRoom->LinkW     = NULL;
  NewRoom->Light     = 1;
  NewRoom->ReEntry   = 0;
  NewRoom->Time      = 0;
  NewRoom->MaxShots  = 0;
  NewRoom->Processed = 0;
  NewRoom->ObjHead   = NULL;
  NewRoom->ObjCur    = NULL;
  NewRoom->next      = NULL;

  return NewRoom;
}

pObject AddGlobal(char *name)
{
  if (ErrorLog) printf("objects.cpp -- AddGlobal(%s);\n", name);

  // Find end of linked list
  pObject Cur = GlobalRoom->ObjHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;

  // set non-0 values
  memset(NewObj, 0, sizeof(tObject));
  strncpy(NewObj->name, name, OBJNAME_MAX);
  NewObj->NameHash = Hash(NewObj->name);
  NewObj->z        = 1;
  NewObj->cycle    = 1;

  if (Cur)
    Cur->next = NewObj; else
    GlobalRoom->ObjHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

int KillGlobal(pObject o)
{
  if (ErrorLog) printf("objects.cpp -- KillGlobal((pObject)0x%X);\n", o);

  pObject Prev = GlobalRoom->ObjHead;

  if (Prev == o)
  {
    GlobalRoom->ObjHead = Prev->next;

    if (Prev->code) delete[] Prev->code;
    delete[] Prev;
  } else
  {
    while (Prev->next != o)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_OBJ;
    }

    pObject Cur = Prev->next;
    pObject Next = Prev->next->next;

    Prev->next = Next;
    if (Prev->code) delete[] Prev->code;
    delete[] Cur;
  }

  return SUCCESS;
}

pObject AddObj(char *objfile, pObject &head)
{
  if (ErrorLog) printf("objects.cpp -- AddObj(%s, (pObject)0x%X);\n", objfile, head);

  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  if (objfile)
  {
    ifstream obj(objfile, ios::in | ios::binary);
    if (!obj || !objfile)
    {
      delete[] NewObj;
      return NULL;
    }

    obj.seekg(0, ios::end);
    int filesize = obj.tellg();
    obj.seekg(0, ios::beg);

    obj.read(id, 4);
    if (strncmp(id, ZZTVersion, 4))
    {
      delete[] NewObj;
      return NULL;
    }

    if (obj.get()==ObjNameOp)
    {
      char *str = NewObj->name;
      while (ch = obj.get())
        *str++ = ch;
      *str = 0;
    } else obj.seekg(-1, ios::cur);

    NewObj->code = new char[filesize];
    if (!NewObj->code)
    {
      delete[] NewObj;
      return NULL;
    }
    memset(NewObj->code, 0, filesize);

    obj.read(NewObj->code, filesize);
    NewObj->codesize = filesize;
    obj.close();
  } else
  {
    NewObj->code = NULL;
    NewObj->codesize = 0;
  }

  // set non-0 values
  NewObj->NameHash = Hash(NewObj->name);
  NewObj->z     = 1;
  NewObj->cycle = 3;

  if (Cur)
    Cur->next = NewObj; else
    head = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pItem FindItem(char *name)
{
  if (ErrorLog) printf("objects.cpp -- FindItem(%s);\n", name);

  short hash = Hash(name);

  for (int i=0; i<ItemNum; i++)
    if (ItemList[i].NameHash==hash)
      return &ItemList[i];
  return NULL;
}

pItem FindItem(short hash)
{
  if (ErrorLog) printf("objects.cpp -- FindItem(%d);\n", hash);

  for (int i=0; i<ItemNum; i++)
    if (ItemList[i].NameHash==hash)
      return &ItemList[i];
  return NULL;
}

pItem FindItemDesc(char *name)
{
  if (ErrorLog) printf("objects.cpp -- FindItem(%s);\n", name);

  for (int i=0; i<ItemNum; i++)
    if (strcmp(name, ItemList[i].desc)==0)
      return &ItemList[i];
  return NULL;
}

pObject AddItemObj(tItem &item, pObject &head)
{
  if (ErrorLog) printf("objects.cpp -- AddItemObj((tItem)0x%X, (pObject)0x%X);\n", item, head);

  char id[4];
  char ch;

  pObject Cur = FindObj(item.name, head);
  if (Cur)
  {
    Cur->objFlight    = (item.flag[10]=='1');
    Cur->objItem      = (item.flag[9]=='1');
    Cur->objInvisible = (item.flag[8]=='1');
    Cur->objPushP     = (item.flag[7]=='1');
    Cur->objPushNS    = (item.flag[6]=='1');
    Cur->objPushEW    = (item.flag[5]=='1');
    Cur->objPusher    = (item.flag[4]=='1');
    Cur->objSwimmer   = (item.flag[3]=='1');
    Cur->objBreakable = (item.flag[2]=='1');
    Cur->objSolid     = (item.flag[1]=='1');
    Cur->objPlayer    = (item.flag[0]=='1');
    Cur->objBound     = 0;
    Cur->objLocked    = 0;

    Cur->frame = item.frame;
    return Cur;
  }

  // Find end of linked list
  Cur = head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  // set non-0 values
  strncpy(NewObj->name, item.name, OBJNAME_MAX);
  NewObj->NameHash = Hash(NewObj->name);

  NewObj->x = NewObj->y = 0;
  while (GetObjXY(NewObj->x, NewObj->y, SystemRoom))
  {
    NewObj->x++;
    if (NewObj->x==SystemRoom->RoomW)
    {
      if (NewObj->y==SystemRoom->RoomH) return NULL;
      NewObj->x=0;
      NewObj->y++;
    }
  }
  NewObj->z = 1;

  NewObj->objFlight    = (item.flag[10]=='1');
  NewObj->objItem      = (item.flag[9]=='1');
  NewObj->objInvisible = (item.flag[8]=='1');
  NewObj->objPushP     = (item.flag[7]=='1');
  NewObj->objPushNS    = (item.flag[6]=='1');
  NewObj->objPushEW    = (item.flag[5]=='1');
  NewObj->objPusher    = (item.flag[4]=='1');
  NewObj->objSwimmer   = (item.flag[3]=='1');
  NewObj->objBreakable = (item.flag[2]=='1');
  NewObj->objSolid     = (item.flag[1]=='1');
  NewObj->objPlayer    = (item.flag[0]=='1');
  NewObj->objBound     = 0;
  NewObj->objLocked    = 0;

  NewObj->frame = item.frame;
  NewObj->cycle = 1;

  if (Cur)
    Cur->next = NewObj; else
    head = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject GetObjXY(int x, int y, pRoom r, int z)
{
  if (ErrorLog) printf("objects.cpp -- GetObjXY(%d, %d, (pRoom)0x%X, %d);\n", x, y, r, z);

  pObject Cur = r->ObjHead;

  if (z==-1)
  {
    while (Cur)
    {
      if (x==Cur->x && y==Cur->y) return Cur;
      Cur = Cur->next;
    }
  } else
  {
    while (Cur)
    {
      if (x==Cur->x && y==Cur->y && z==Cur->z) return Cur;
      Cur = Cur->next;
    }
  }
  return NULL;
}

pObject AddObjXY(int x, int y, char *name, pRoom r, int z)
{
  if (ErrorLog) printf("objects.cpp -- GetObjXY(%d, %d, %s, (pRoom)0x%X, %d);\n", x, y, name, r, z);

  pItem it = FindItem(name);
  pObject src = FindObj(name, SystemRoom->ObjHead);
  if (!src) return NULL;
  pObject success = NULL;

  switch (it->note)
  {
    case 0 : case 1 : case 2 : case 3 :
      success = BindObj(src, r);
      break;
    case 4 : // scroll
      success = BindScroll(src, r);
      break;
    case 5 : // object
      success = BindEmptyObj(src, r);
      break;
    case 6 : // water
      RoomCur->RoomObject[y*RoomCur->RoomW+x] = it->frame;
      break;
    case 7 : // forest
      RoomCur->RoomObject[y*RoomCur->RoomW+x] = it->frame;
      break;
    case 8 : // fake
      RoomCur->RoomFloor[y*RoomCur->RoomW+x] = it->frame;
      break;
    case 10 :
      RoomCur->RoomObject[y*RoomCur->RoomW+x] = it->frame;
      break;
  }

  if (success)
  {
    LastObj->x = x;
    LastObj->y = y;
    LastObj->z = z;
    return LastObj;
  }
  return NULL;
}

pObject AddObjXY(int x, int y, pObject src, pRoom r, int z)
{
  if (ErrorLog) printf("objects.cpp -- GetObjXY(%d, %d, (pObject)0x%X, (pRoom)0x%X, %d);\n", x, y, src, r, z);

  pObject success = BindObj(src, r);

  if (success)
  {
    LastObj->x = x;
    LastObj->y = y;
    LastObj->z = z;
    return LastObj;
  }
  return NULL;
}

int KillObjXY(int x, int y, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- KillObjXY(%d, %d, (pRoom)0x%X);\n", x, y, r);

  pObject o = GetObjXY(x, y, r);
  if (!o) return BAD_OBJ;
  KillObj(r, o);
  return SUCCESS;
}

int KillObjXY(int x, int y, pRoom r, int z)
{
  if (ErrorLog) printf("objects.cpp -- KillObjXY(%d, %d, (pRoom)0x%X, %d);\n", x, y, r, z);

  pObject o = GetObjXY(x, y, r, z);
  if (!o) return BAD_OBJ;
  KillObj(r, o);
  return SUCCESS;
}

pObject BindObj(pObject Src, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- BindObj((pObject)0x%X, (pRoom)0x%X);\n", Src, r);

  char id[4];
  char ch;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  if (Src->objBound)
  {
    // NEVER BIND A BOUND OBJECT! :> Bind it to the original.

    // set non-0 values
    strncpy(NewObj->name, Src->name, OBJNAME_MAX);
    NewObj->NameHash = Src->NameHash;
    NewObj->x        = Src->x;
    NewObj->y        = Src->y;
    NewObj->z        = Src->z;

    NewObj->objFlight    = Src->objFlight;
    NewObj->objItem      = Src->objItem;
    NewObj->objInvisible = Src->objInvisible;
    NewObj->objPushP     = Src->objPushP;
    NewObj->objPushNS    = Src->objPushNS;
    NewObj->objPushEW    = Src->objPushEW;
    NewObj->objPusher    = Src->objPusher;
    NewObj->objSwimmer   = Src->objSwimmer;
    NewObj->objBreakable = Src->objBreakable;
    NewObj->objSolid     = Src->objSolid;
    NewObj->objPlayer    = Src->objPlayer;
    NewObj->objLocked    = Src->objLocked;
    NewObj->objBound     = 0;

    NewObj->frame  = Src->frame;
    NewObj->cycle  = Src->cycle;
    NewObj->cyclel = Src->cyclel;
    NewObj->light  = Src->light;
    NewObj->queue  = Src->queue;
    NewObj->qsize  = Src->qsize;
    NewObj->next   = NULL;

    NewObj->codesize = Src->codesize;
    NewObj->code     = Src->code;
    NewObj->codep    = 0;
  } else
  {
    // set non-0 values
    strncpy(NewObj->name, Src->name, OBJNAME_MAX);
    NewObj->NameHash = Src->NameHash;
    NewObj->x        = Src->x;
    NewObj->y        = Src->y;
    NewObj->z        = Src->z;

    NewObj->objFlight    = Src->objFlight;
    NewObj->objItem      = Src->objItem;
    NewObj->objInvisible = Src->objInvisible;
    NewObj->objPushP     = Src->objPushP;
    NewObj->objPushNS    = Src->objPushNS;
    NewObj->objPushEW    = Src->objPushEW;
    NewObj->objPusher    = Src->objPusher;
    NewObj->objSwimmer   = Src->objSwimmer;
    NewObj->objBreakable = Src->objBreakable;
    NewObj->objSolid     = Src->objSolid;
    NewObj->objPlayer    = Src->objPlayer;
    NewObj->objLocked    = Src->objLocked;
    NewObj->objBound     = 0;

    NewObj->frame    = Src->frame;
    NewObj->cycle    = Src->cycle;
    NewObj->cyclel   = Src->cyclel;
    NewObj->light    = Src->light;
    NewObj->queue    = Src->queue;
    NewObj->qsize    = Src->qsize;
    NewObj->next     = NULL;

    int csize = 5+strlen(Src->name);
    NewObj->code = new char[csize];
    if (!NewObj->code)
    {
      delete[] NewObj;
      return NULL;
    }
    memset(NewObj->code, 0, csize);

    NewObj->codep    = 0;
    NewObj->codesize = csize;
    memcpy(NewObj->code, BindCode, 2);
    memcpy(NewObj->code+2, Src->name, strlen(Src->name));
  }

  // Find end of linked list
  pObject Cur = r->ObjHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  if (Cur)
    Cur->next = NewObj; else
    r->ObjHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject BindObj(pObject Src, pObject &head)
{
  if (ErrorLog) printf("objects.cpp -- BindObj((pObject)0x%X, (pObject)0x%X);\n", Src, head);

  char id[4];
  char ch;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  if (Src->objBound)
  {
    // NEVER BIND A BOUND OBJECT! :> Bind it to the original.

    // set non-0 values
    strncpy(NewObj->name, Src->name, OBJNAME_MAX);
    NewObj->NameHash = Src->NameHash;
    NewObj->x        = Src->x;
    NewObj->y        = Src->y;
    NewObj->z        = Src->z;

    NewObj->objFlight    = Src->objFlight;
    NewObj->objItem      = Src->objItem;
    NewObj->objInvisible = Src->objInvisible;
    NewObj->objPushP     = Src->objPushP;
    NewObj->objPushNS    = Src->objPushNS;
    NewObj->objPushEW    = Src->objPushEW;
    NewObj->objPusher    = Src->objPusher;
    NewObj->objSwimmer   = Src->objSwimmer;
    NewObj->objBreakable = Src->objBreakable;
    NewObj->objSolid     = Src->objSolid;
    NewObj->objPlayer    = Src->objPlayer;
    NewObj->objLocked    = Src->objLocked;
    NewObj->objBound     = 0;

    NewObj->frame  = Src->frame;
    NewObj->cycle  = Src->cycle;
    NewObj->cyclel = Src->cyclel;
    NewObj->light  = Src->light;
    NewObj->queue  = Src->queue;
    NewObj->qsize  = Src->qsize;
    NewObj->next   = NULL;

    NewObj->codesize = Src->codesize;
    NewObj->code     = Src->code;
    NewObj->codep    = 0;
  } else
  {
    // set non-0 values
    strncpy(NewObj->name, Src->name, OBJNAME_MAX);
    NewObj->NameHash = Src->NameHash;
    NewObj->x        = Src->x;
    NewObj->y        = Src->y;
    NewObj->z        = Src->z;

    NewObj->objFlight    = Src->objFlight;
    NewObj->objItem      = Src->objItem;
    NewObj->objInvisible = Src->objInvisible;
    NewObj->objPushP     = Src->objPushP;
    NewObj->objPushNS    = Src->objPushNS;
    NewObj->objPushEW    = Src->objPushEW;
    NewObj->objPusher    = Src->objPusher;
    NewObj->objSwimmer   = Src->objSwimmer;
    NewObj->objBreakable = Src->objBreakable;
    NewObj->objSolid     = Src->objSolid;
    NewObj->objPlayer    = Src->objPlayer;
    NewObj->objLocked    = Src->objLocked;
    NewObj->objBound     = 0;

    NewObj->frame    = Src->frame;
    NewObj->cycle    = Src->cycle;
    NewObj->cyclel   = Src->cyclel;
    NewObj->light    = Src->light;
    NewObj->queue    = Src->queue;
    NewObj->qsize    = Src->qsize;
    NewObj->next     = NULL;

    int csize = 5+strlen(Src->name);
    NewObj->code = new char[csize];
    if (!NewObj->code)
    {
      delete[] NewObj;
      return NULL;
    }
    memset(NewObj->code, 0, csize);

    NewObj->codep    = 0;
    NewObj->codesize = csize;
    memcpy(NewObj->code, BindCode, 2);
    memcpy(NewObj->code+2, Src->name, strlen(Src->name));
  }

  // Find end of linked list
  pObject Cur = head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  if (Cur)
    Cur->next = NewObj; else
    head = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject BindScroll(pObject Src, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- BindScroll((pObject)0x%X, (pRoom)0x%X);\n", Src, r);

  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = r->ObjHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  // set non-0 values
  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash  = NewObj->NameHash;
  NewObj->z         = Src->z;

  NewObj->objFlight    = Src->objFlight;
  NewObj->objItem      = Src->objItem;
  NewObj->objInvisible = Src->objInvisible;
  NewObj->objPushP     = Src->objPushP;
  NewObj->objPushNS    = Src->objPushNS;
  NewObj->objPushEW    = Src->objPushEW;
  NewObj->objPusher    = Src->objPusher;
  NewObj->objSwimmer   = Src->objSwimmer;
  NewObj->objBreakable = Src->objBreakable;
  NewObj->objSolid     = Src->objSolid;
  NewObj->objPlayer    = Src->objPlayer;
  NewObj->objBound     = 0;
  NewObj->objLocked    = Src->objLocked;

  NewObj->frame     = Src->frame;
  NewObj->cycle     = Src->cycle;
  NewObj->cyclel    = Src->cyclel;
  NewObj->light     = Src->light;
  NewObj->queue     = Src->queue;
  NewObj->qsize     = Src->qsize;
  NewObj->next      = NULL;
  NewObj->codesize  = Src->codesize;
  NewObj->codep     = 0;

  NewObj->code = new char[NewObj->codesize];
  if (!NewObj->code)
  {
    delete[] NewObj;
    return NULL;
  }
  memset(NewObj->code, 0, NewObj->codesize);
  memcpy(NewObj->code, Src->code, NewObj->codesize);

  if (Cur)
    Cur->next = NewObj; else
    r->ObjHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject BindEmptyObj(pObject Src, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- BindEmptyObj((pObject)0x%X, (pRoom)0x%X);\n", Src, r);

  char id[4];
  char ch;

  // Find end of linked list
  pObject Cur = r->ObjHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash  = Src->NameHash;
  NewObj->z         = Src->z;

  NewObj->objFlight    = Src->objFlight;
  NewObj->objItem      = Src->objItem;
  NewObj->objInvisible = Src->objInvisible;
  NewObj->objPushP     = Src->objPushP;
  NewObj->objPushNS    = Src->objPushNS;
  NewObj->objPushEW    = Src->objPushEW;
  NewObj->objPusher    = Src->objPusher;
  NewObj->objSwimmer   = Src->objSwimmer;
  NewObj->objBreakable = Src->objBreakable;
  NewObj->objSolid     = Src->objSolid;
  NewObj->objPlayer    = Src->objPlayer;
  NewObj->objBound     = Src->objBound;

  NewObj->objLocked = 0;
  NewObj->frame     = Src->frame;
  NewObj->cycle     = Src->cycle;
  NewObj->cyclel    = Src->cyclel;
  NewObj->light     = Src->light;
  NewObj->queue     = Src->queue;
  NewObj->qsize     = Src->qsize;
  NewObj->next      = NULL;

  NewObj->code      = NULL;
  NewObj->codep     = 0;
  NewObj->codesize  = 0;

  if (Cur)
    Cur->next = NewObj; else
    r->ObjHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject CopyObj(char *name, pObject &head, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- CopyObj(%s, (pObject)0x%X, (pRoom)0x%X);\n", name, head, r);

  char id[4];
  char ch;

  // create a new object, and init it to 0
  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  // find source object
  pObject Src = FindObj(name, head);

  // copy source data to new object
  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash  = Src->NameHash;
  NewObj->x         = Src->x;
  NewObj->y         = Src->y;
  NewObj->z         = Src->z;

  NewObj->objFlight    = Src->objFlight;
  NewObj->objItem      = Src->objItem;
  NewObj->objInvisible = Src->objInvisible;
  NewObj->objPushP     = Src->objPushP;
  NewObj->objPushNS    = Src->objPushNS;
  NewObj->objPushEW    = Src->objPushEW;
  NewObj->objPusher    = Src->objPusher;
  NewObj->objSwimmer   = Src->objSwimmer;
  NewObj->objBreakable = Src->objBreakable;
  NewObj->objSolid     = Src->objSolid;
  NewObj->objPlayer    = Src->objPlayer;
  NewObj->objBound     = Src->objBound;
  NewObj->objLocked    = Src->objLocked;

  NewObj->frame     = Src->frame;
  NewObj->cycle     = Src->cycle;
  NewObj->cyclel    = Src->cyclel;
  NewObj->light     = Src->light;
  NewObj->queue     = Src->queue;
  NewObj->qsize     = Src->qsize;

  // create a separate code buffer
  NewObj->codesize  = Src->codesize;
  NewObj->codep     = 0;
  NewObj->code      = new char[Src->codesize];
  if (!NewObj->code)
  {
    delete[] NewObj;
    return NULL;
  }
  memcpy(NewObj->code, Src->code, Src->codesize);

  NewObj->next      = NULL;

  // Find end of linked list, and link new object in
  pObject Cur = r->ObjHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  if (Cur)
    Cur->next = NewObj; else
    r->ObjHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject CopyObj(pObject Src, pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- CopyObj((pObject)0x%X, (pRoom)0x%X);\n", Src, r);

  char id[4];
  char ch;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash = Src->NameHash;
  NewObj->x        = Src->x;
  NewObj->y        = Src->y;
  NewObj->z        = Src->z;

  NewObj->objFlight    = Src->objFlight;
  NewObj->objItem      = Src->objItem;
  NewObj->objInvisible = Src->objInvisible;
  NewObj->objPushP     = Src->objPushP;
  NewObj->objPushNS    = Src->objPushNS;
  NewObj->objPushEW    = Src->objPushEW;
  NewObj->objPusher    = Src->objPusher;
  NewObj->objSwimmer   = Src->objSwimmer;
  NewObj->objBreakable = Src->objBreakable;
  NewObj->objSolid     = Src->objSolid;
  NewObj->objPlayer    = Src->objPlayer;
  NewObj->objBound     = Src->objBound;
  NewObj->objLocked    = Src->objLocked;

  NewObj->frame = Src->frame;
  NewObj->cycle = Src->cycle;
  NewObj->light = Src->light;
  NewObj->queue = Src->queue;
  NewObj->qsize = Src->qsize;
  NewObj->next  = NULL;

  NewObj->codesize = Src->codesize;
  NewObj->codep    = 0;
  NewObj->code     = new char[Src->codesize];
  if (!NewObj->code)
  {
    delete[] NewObj;
    return NULL;
  }
  memcpy(NewObj->code, Src->code, Src->codesize);

  // Find end of linked list
  pObject Cur = r->ObjHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  if (Cur)
    Cur->next = NewObj; else
    r->ObjHead = NewObj;

  LastObj = NewObj;
  return NewObj;
}

pObject CopyObj(pObject Src, pObject &head)
{
  if (ErrorLog) printf("objects.cpp -- CopyObj((pObject)0x%X, (pObject)0x%X);\n", Src, head);

  char id[4];
  char ch;

  pObject NewObj = new tObject;
  if (!NewObj) return NULL;
  memset(NewObj, 0, sizeof(tObject));

  strncpy(NewObj->name, Src->name, OBJNAME_MAX);
  NewObj->NameHash = Src->NameHash;
  NewObj->x        = Src->x;
  NewObj->y        = Src->y;
  NewObj->z        = Src->z;

  NewObj->objFlight    = Src->objFlight;
  NewObj->objItem      = Src->objItem;
  NewObj->objInvisible = Src->objInvisible;
  NewObj->objPushP     = Src->objPushP;
  NewObj->objPushNS    = Src->objPushNS;
  NewObj->objPushEW    = Src->objPushEW;
  NewObj->objPusher    = Src->objPusher;
  NewObj->objSwimmer   = Src->objSwimmer;
  NewObj->objBreakable = Src->objBreakable;
  NewObj->objSolid     = Src->objSolid;
  NewObj->objPlayer    = Src->objPlayer;
  NewObj->objBound     = Src->objBound;
  NewObj->objLocked    = Src->objLocked;

  NewObj->frame = Src->frame;
  NewObj->cycle = Src->cycle;
  NewObj->light = Src->light;
  NewObj->queue = Src->queue;
  NewObj->qsize = Src->qsize;
  NewObj->next  = NULL;

  NewObj->codesize = Src->codesize;
  NewObj->codep    = 0;
  NewObj->code     = new char[Src->codesize];
  if (!NewObj->code)
  {
    delete[] NewObj;
    return NULL;
  }
  memcpy(NewObj->code, Src->code, Src->codesize);

  // Find end of linked list
  pObject Cur = head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  if (Cur)
    Cur->next = NewObj; else
    head = NewObj;

  LastObj = NewObj;
  return NewObj;
}

void LoadObj(char *objfile, pObject o)
{
  if (ErrorLog) printf("objects.cpp -- LoadObj(%d, (pObject)0x%X);\n", objfile, o);

  if (!o) return;
  char id[4];
  char ch;

  ifstream obj(objfile, ios::in | ios::binary);
  if (!obj || !objfile) return;

  obj.seekg(0, ios::end);
  int filesize = obj.tellg();
  obj.seekg(0, ios::beg);

  obj.read(id, 4);
  if (strncmp(id, ZZTVersion, 4)) return;

  if (obj.get()==ObjNameOp)
  {
    char *str = o->name;
    while (ch = obj.get())
      *str++ = ch;
    *str = 0;
  } else strcpy(o->name, "Untitled");

  o->NameHash = Hash(o->name);

  o->code = new char[filesize];
  if (!o->code) return;
  obj.read(o->code, filesize);
  o->codesize = filesize;
  obj.close();

  LastObj = o;
}

pRoom FindRoom(char *name)
{
  if (ErrorLog) printf("objects.cpp -- FindRoom(%s);\n", name);

  pRoom Cur;

  if (strnicmp("--", name, RNAME_MAX)==0) return RoomCur;

  short hash = Hash(name);

  if (hash==GlobalRoom->NameHash) return GlobalRoom;
  if (hash==SystemRoom->NameHash) return SystemRoom;

  Cur = RoomHead;
  while (Cur)
  {
    if (Cur->NameHash==hash) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

pRoom FindRoom(short hash)
{
  if (ErrorLog) printf("objects.cpp -- FindRoom(%d);\n", hash);

  pRoom Cur;

  if (hash==GlobalRoom->NameHash) return GlobalRoom;
  if (hash==SystemRoom->NameHash) return SystemRoom;

  Cur = RoomHead;
  while (Cur)
  {
    if (Cur->NameHash==hash) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

int FindVarVal(char *name)
{
  if (ErrorLog) printf("objects.cpp -- FindVarVal(%s);\n", name);

  short hash = Hash(name);

  pVar Cur = VarHead;
  while (Cur)
  {
    if (Cur->NameHash==hash) return Cur->value;
    Cur = Cur->next;
  }

  return 0;
}

pVar FindVar(char *name)
{
  if (ErrorLog) printf("objects.cpp -- FindVar(%s);\n", name);

  short hash = Hash(name);

  pVar Cur = VarHead;
  while (Cur)
  {
    if (Cur->NameHash==hash) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

int sizeofVar(pVar v)
{
  if (!v) return BAD_OBJ;

  return sizeof(tVar);
}

int sizeofVarList(void)
{
  int total=0;

  pVar Cur = VarHead;
  while (Cur)
  {
    total += sizeofVar(Cur);
    Cur = Cur->next;
  }

  return total;
}

pObject FindObj(char *name, pObject head)
{
  if (ErrorLog) printf("objects.cpp -- FindObj(%s, (pObject)0x%X);\n", name, head);

  pObject Cur;

  short hash = Hash(name);

  // Search for globals
  Cur = GlobalRoom->ObjHead;
  while (Cur)
  {
    if (Cur->NameHash==hash)
      return Cur;
    Cur = Cur->next;
  }

  // Search room objects
  Cur = head;
  while (Cur)
  {
    if (Cur->NameHash==hash)
    {
      LastObj = Cur;
      return Cur;
    }
    Cur = Cur->next;
  }

  return NULL;
}

pObject FindObj(short hash, pObject head)
{
  if (ErrorLog) printf("objects.cpp -- FindObj(%d, (pObject)0x%X);\n", hash, head);

  pObject Cur;

  // Search for globals
  Cur = GlobalRoom->ObjHead;
  while (Cur)
  {
    if (Cur->NameHash==hash)
      return Cur;
    Cur = Cur->next;
  }

  // Search room objects
  Cur = head;
  while (Cur)
  {
    if (Cur->NameHash==hash)
    {
      LastObj = Cur;
      return Cur;
    }
    Cur = Cur->next;
  }

  return NULL;
}

int sizeofObj(pObject o)
{
  if (!o) return BAD_OBJ;

  return sizeof(tObject)+o->codesize;
}

int sizeofObjList(pObject head)
{
  int total=0;

  pObject Cur = head;
  while (Cur)
  {
    total += sizeofObj(Cur);
    Cur = Cur->next;
  }

  return total;
}

int sizeofRoom(pRoom r)
{
  if (!r) return BAD_ROOM;

  int size = sizeof(tRoom);
  if (r->RoomFloor)   size += r->RoomW * r->RoomH * 2;
  if (r->RoomObject)  size += r->RoomW * r->RoomH * 2;
  if (r->RoomCeiling) size += r->RoomW * r->RoomH * 2;
  size += sizeofObjList(r->ObjHead);

  return size;
}

int sizeofRoomList(void)
{
  int total = 0;

  pRoom Cur = RoomHead;
  while (Cur)
  {
    total += sizeofRoom(Cur);
    Cur = Cur->next;
  }

  return total;
}

int KillVar(pVar v)
{
  if (ErrorLog) printf("objects.cpp -- KillVar((pVar)0x%X);\n", v);

  pVar Prev = VarHead;

  if (Prev == v)
  {
    VarHead = Prev->next;
    delete[] Prev;
  } else
  {
    while (Prev->next != v)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_VAR;
    }

    pVar Cur = Prev->next;
    pVar Next = Prev->next->next;

    Prev->next = Next;
    delete[] Cur;
  }

  return SUCCESS;
}

int KillRoom(pRoom r)
{
  if (ErrorLog) printf("objects.cpp -- KillVar((pRoom)0x%X);\n", r);

  pRoom Prev = RoomHead;

  if (Prev == r)
  {
    RoomHead = Prev->next;

    if (Prev->RoomFloor)   delete[] Prev->RoomFloor;
    if (Prev->RoomObject)  delete[] Prev->RoomObject;
    if (Prev->RoomCeiling) delete[] Prev->RoomCeiling;
    pObject CurObj = Prev->ObjHead;
    while (CurObj)
      KillObj(Prev, CurObj);

    delete[] Prev;
  } else
  {
    while (Prev->next != r)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_ROOM;
    }

    pRoom Cur = Prev->next;
    pRoom Next = Prev->next->next;
    Prev->next = Next;

    if (Cur->RoomFloor)   delete[] Cur->RoomFloor;
    if (Cur->RoomObject)  delete[] Cur->RoomObject;
    if (Cur->RoomCeiling) delete[] Cur->RoomCeiling;
    pObject CurObj = Cur->ObjHead;
    while (CurObj)
      KillObj(Cur, CurObj);

    delete[] Cur;
  }

  return SUCCESS;
}

int KillObj(pRoom r, pObject o)
{
  if (ErrorLog) printf("objects.cpp -- KillVar((pRoom)0x%X, (pObject)0x%X);\n", r, o);

  pObject Prev = r->ObjHead;

  if (Prev == o)
  {
    r->ObjHead = Prev->next;

//  if (Prev->code) if (!Prev->objBound) delete[] Prev->code;
    delete[] Prev;
  } else
  {
    while (Prev->next != o)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_OBJ;
    }

    pObject Cur = Prev->next;
    pObject Next = Prev->next->next;

    Prev->next = Next;
//  if (Prev->code) if (!Cur->objBound) delete[] Prev->code;
    delete[] Cur;
  }

  return SUCCESS;
}

int Spawn(char *name, int x, int y)
{
  if (ErrorLog) printf("objects.cpp -- Spawn(%s, %d, %d);\n", name, x, y);

  int w = RoomCur->RoomW;
  int retval=0;

  // if another object is in the way, can't spawn
  short *objl = RoomCur->RoomObject;
  if (objl[y*w+x]) return 0;
  pObject touched = GetObjXY(x, y, RoomCur);
  if (touched) return 0;

  pObject src = FindObj(name, SystemRoom->ObjHead);

  // first, try to bind the object to an internal object
  if (BindObj(src, RoomCur)) retval = 1;

  // secondly, try to copy the object from another in the same room
//  if (CopyObj(name, RoomCur->ObjHead, RoomCur)) retval = 1;

  if (retval)
  {
    LastObj->x = x;
    LastObj->y = y;
  }

  return retval;
}

int GetFrame(int t)
{
  if (t<TILE_MAX) return t;

  return ItemList[t-TILE_MAX].frame;
}

char *GetName(int t)
{
  if (t<TILE_MAX) return NULL;

  return ItemList[t-TILE_MAX].name;
}
