// Watcom includes
#include <stdlib.h>
#include <time.h>

// HGL include
#include "vgl.h"

// ZZT ][ includes
#include "sound.h"
#include "global.h"

// MIDAS includes
#include "midas.h"

pSound SoundHead, SoundCur;
gmpModule *mf;
int handle, voice;
long modsize;

int zm_MixFreq = 22050;
int zm_Bits = 8;
int zm_Mono = 0;
int zm_Interp = 0;
int zm_Device = -1;
int zm_Channels = 32;
int zm_Port = -1;
int zm_IRQ  = -1;
int zm_DMA  = -1;
int zm_Card = -1;

void Error(char *msg)
{
  g_Done();
  printf("Z2 Sound: %s\n", msg);
#ifndef NOMIDAS
  midasClose();           /* IMPORTANT! */
#endif
  exit(EXIT_FAILURE);
}

int zm_modLoaded(void)
{
  return (mf != NULL);
}

void zm_SetDef(void)
{
  // Load these from Z2.CFG.
  (zm_Bits==16)? midasOutputMode = sd16bit : midasOutputMode = sd8bit;
  (zm_Mono==1)?  midasOutputMode |= sdMono : midasOutputMode |= sdStereo;
  midasSDNumber = zm_Device;
  midasSDCard   = -1;
  midasSDPort   = zm_Port;
  midasSDIRQ    = zm_IRQ;
  midasSDDMA    = zm_DMA;
  midasMixRate  = zm_MixFreq;
}

int zm_Init(void)
{
  int retval = 1;
  int config;

  midasDisableEMS = 1;
  zm_Channels     = 32;

#ifndef NOMIDAS
  midasInit();
#endif

  modsize = handle = voice = 0;

  return retval;
}

void zm_modStop(void)
{
#ifndef NOMIDAS
  if (mf) midasStopModule(mf);
#endif
}

void zm_modKill(void)
{
#ifndef NOMIDAS
  if (mf)
  {
    int error;

    midasStopModule(mf);
    if ((error = gmpFreeModule(mf)) != OK)
    {
      g_Done();
      Error("Error deallocating module.");
      exit(1);
    }

    mf = NULL;
  }
#endif
}

void zm_Done(void)
{
#ifndef NOMIDAS
  zm_modKill();
  midasClose();
#endif
}

int zm_modLoad(char *filename)
{
#ifndef NOMIDAS
  int error;

  zm_modKill();

  if ((error = gmpLoadXM(filename, 1, NULL, &mf)) != OK)
    return -1;
#endif
  voice = 0;

  return 0;  // success
}

void zm_modStart(void)
{
#ifndef NOMIDAS
  if (mf)
  {
    midasPlayModule(mf, 0);
    voice = 0;
  }
#endif
}

int zm_modSize(void)
{
  return modsize;
}

void zm_modNextPos(void)
{
//MP_NextPosition();
}

void zm_modPrevPos(void)
{
//MP_PrevPosition();
}

void zm_modSetPos(int p)
{
//MP_SetPosition(p);
}

char *zm_modSongName(void)
{
#ifndef NOMIDAS
  if (mf) return (char *)mf->name;
#endif
  return "No module in memory";
}

char *zm_modType(void)
{
  static char name[20];

  strcpy(name, "Unknown");
#ifndef NOMIDAS
  switch (mf->playMode)
  {
    case gmpPT  : strcpy(name, "ProTracker");
    case gmpST3 : strcpy(name, "ScreamTracker 3.xx");
    case gmpFT2 : strcpy(name, "FastTracker 2");
  }
#endif

  return name;
}

char *zm_modPTLimits(void)
{
#ifndef NOMIDAS
  if (mf->playFlags.ptLimits) return "Yes";
#endif
  return "No";
}

char *zm_modVolSlides(void)
{
#ifndef NOMIDAS
  if (mf->playFlags.fastVolSlides) return "Fast";
#endif
  return "Standard speed";
}

char *zm_modOctaves(void)
{
#ifndef NOMIDAS
  if (mf->playFlags.extOctaves) return "Yes";
#endif
  return "No";
}

char *zm_modFreqTable(void)
{
#ifndef NOMIDAS
  if (mf->playFlags.linFreqTable) return "Linear";
#endif
  return "Logarithmic";
}

int zm_modVolume(void)
{
#ifndef NOMIDAS
  unsigned volume;
  midasSD->GetMasterVolume(&volume);
  return volume;
#else
  return 0;
#endif
}

void zm_modSetVolume(int v)
{
#ifndef NOMIDAS
  midasSD->SetMasterVolume(v);
#endif
}

int zm_modLooping(void)
{
  return 0;
}

void zm_modSetLooping(int l)
{
}

char *zm_cardName(void)
{
#ifndef NOMIDAS
  return midasSD->cardNames[midasSD->cardType-1];
#else
  return "Sound system disabled";
#endif
}

char *zm_cardDriver(void)
{
#ifndef NOMIDAS
  return midasSD->name;
#else
  return "No sound drivers installed";
#endif
}

int zm_cardInterp(void)
{
  return 0;
}

int zm_cardBits(void)
{
#ifndef NOMIDAS
  return (midasOutputMode & sd16bit) ? 16 : 8;
#else
  return 0;
#endif
}

int zm_cardStereo(void)
{
#ifndef NOMIDAS
  return (midasOutputMode & sdStereo);
#else
  return 0;
#endif
}

int zm_cardPort(void)
{
#ifndef NOMIDAS
  return midasSD->port;
#else
  return 0;
#endif
}

int zm_cardIRQ(void)
{
#ifndef NOMIDAS
  return midasSD->IRQ;
#else
  return 0;
#endif
}

int zm_cardDMA(void)
{
#ifndef NOMIDAS
  return midasSD->DMA;
#else
  return 0;
#endif
}

int zm_cardMixFreq(void)
{
  return zm_MixFreq;
}

int zm_cardChannels(void)
{
  return zm_Channels;
}

pSound zm_wavFind(int handle)
{
  pSound Cur = SoundHead;
  while (Cur)
  {
    if (Cur->handle == handle) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

pSound zm_wavAdd(char *wavfile, char *name)
{
  pSound Cur = SoundHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  pSound NewSound = new tSound;
  if (!NewSound) return NULL;

  NewSound->snd = NULL;
  if (!NewSound->snd) return NULL;

  if (name)
    strncpy(NewSound->name, name, 20); else
    strncpy(NewSound->name, wavfile, 20);
  NewSound->handle = handle++;
  NewSound->volume = 64;
  NewSound->panning = 0;
  NewSound->freq = WAV_FREQ;
  NewSound->length = 0;
  NewSound->next = NULL;

  if (Cur)
    Cur->next = NewSound; else
    SoundHead = NewSound;

  return NewSound;
}

pSound zm_wavGet(short handle)
{
  pSound Cur = SoundHead;

  while (Cur)
  {
    if (handle==Cur->handle) return Cur;
    Cur = Cur->next;
  }

  return NULL;
}

void zm_wavLoad(char *wavfile, pSound s)
{
  if (!s) return;

//  if (s->snd) MW_FreeWav(s->snd);
//  s->snd = MW_LoadWavFN(wavfile);
}

int zm_wavKill(pSound s)
{
  pSound Prev = SoundHead;

  if (Prev == s)
  {
    SoundHead = Prev->next;
//    MW_FreeWav(Prev->snd);
    delete[] Prev;
  } else
  {
    while (Prev->next != s)
    {
      Prev = Prev->next;
      if (!Prev) return BAD_SOUND;
    }

    pSound Cur = Prev->next;
    pSound Next = Prev->next->next;

    Prev->next = Next;
//    MW_FreeWav(Cur->snd);
    delete[] Cur;
  }

  return SUCCESS;
}

int zm_wavPlay(int handle)
{
  pSound s = zm_wavFind(handle);
  if (!s) return 0;

  int v = zm_GetFreeVoice();

//  MD_VoiceSetVolume(v, s->volume);
//  MD_VoiceSetPanning(v, s->panning);
//  MD_VoiceSetFrequency(v, s->freq);
//  MD_VoicePlay(v, s->snd->handle, 0, s->snd->length, 0, 0, s->snd->flags);

  return v;
}

int zm_wavPlayFX(int handle, int vol, int pan, int frq)
{
  pSound s = zm_wavFind(handle);
  if (!s) return 0;

  int v = zm_GetFreeVoice();

//  MD_VoiceSetVolume(v, vol);
//  MD_VoiceSetPanning(v, pan);
//  MD_VoiceSetFrequency(v, frq);
//  MD_VoicePlay(v, s->snd->handle, 0, s->snd->length, 0, 0, s->snd->flags);

  return v;
}

int zm_wavFile(char *wavfile)
{
//  if (imm_sample) MW_FreeWav(imm_sample);
//  imm_sample = MW_LoadWavFN(wavfile);

  int v = zm_GetFreeVoice();

//  MD_VoiceSetVolume(v, 64);
//  MD_VoiceSetPanning(v, 0);
//  MD_VoiceSetFrequency(v, WAV_FREQ);

//  imm_endtick = clock() + (imm_sample->length*CLK_TCK / WAV_FREQ);
//  MD_VoicePlay(v, imm_sample->handle, 0, imm_sample->length, 0, 0, imm_sample->flags);

  return v;
}

int zm_wavFileFX(char *wavfile, int vol, int pan, int frq)
{
//  if (imm_sample) MW_FreeWav(imm_sample);
//  imm_sample = MW_LoadWavFN(wavfile);

  int v = zm_GetFreeVoice();

//  MD_VoiceSetVolume(v, vol);
//  MD_VoiceSetPanning(v, pan);
//  MD_VoiceSetFrequency(v, frq);

//  imm_endtick = clock() + (imm_sample->length*CLK_TCK / frq);
//  MD_VoicePlay(v, imm_sample->handle, 0, imm_sample->length, 0, 0, imm_sample->flags);

  return v;
}

int zm_wavPlayCh(int handle, int v)
{
  pSound s = zm_wavFind(handle);
  if (!s) return 0;

//  MD_VoiceSetVolume(v, s->volume);
//  MD_VoiceSetPanning(v, s->panning);
//  MD_VoiceSetFrequency(v, s->freq);
//  MD_VoicePlay(v, s->snd->handle, 0, s->snd->length, 0, 0, s->snd->flags);

  return v;
}

int zm_wavPlayChFX(int handle, int v, int vol, int pan, int frq)
{
  pSound s = zm_wavFind(handle);
  if (!s) return 0;

//  MD_VoiceSetVolume(v, vol);
//  MD_VoiceSetPanning(v, pan);
//  MD_VoiceSetFrequency(v, frq);
//  MD_VoicePlay(v, s->snd->handle, 0, s->snd->length, 0, 0, s->snd->flags);

  return v;
}

int zm_sizeofWav(pSound s)
{
  if (!s) return 0;

  return s->length;
}

int zm_sizeofWavList(void)
{
  int total = 0;

  pSound Cur = SoundHead;
  while (Cur)
  {
    total += zm_sizeofWav(Cur);
    Cur = Cur->next;
  }

  return total;
}

int zm_GetFreeVoice(void)
{
  voice++;
  if (voice>=zm_Channels) voice = 0;
  return voice;
}
