// Sprite Manager, version 1.2 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version
// Version 1.1 -- Added provision for manually updated backgrounds (ie.
//                sprites can be instructed not to update themselves.)
// Version 1.2 -- Added special mouse sprite, which is always updated
//                last.

#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include "sprite.h"
#include "vgl.h"

pSprite SpriteHead, SpriteTail, SpriteList;
pSprite MousePtr;
int SprPage;

int s_InitPage(void)
{
  SprPage = g_InitPage();

  return SprPage;
}

void s_Init(void)
{
  SpriteHead = SpriteTail = MousePtr = NULL;
}

int s_Load(char *f, SPRITE &spr)
{
  int SizeF, SizeD;

  static ifstream fp;
  fp.open(f, ios::in | ios::binary);
  if (!fp) return 1;

  fp.seekg(0, ios::end);
  SizeF = fp.tellg();
  fp.seekg(0, ios::beg);

  fp.read(spr.id, 4);
  if (strncmp(spr.id, "SPR1", 4)) return 1;
  fp.read((unsigned char *)&spr.w, sizeof(short));
  fp.read((unsigned char *)&spr.h, sizeof(short));
  fp.read((unsigned char *)&spr.fr, sizeof(short));
  SizeD = (spr.w*spr.h)*spr.fr;
  spr.Data = new char[SizeD];
  if (!spr.Data) return 1;
  fp.read(spr.Data, SizeD);
  fp.close();

  return 0;
}

pSprite s_Add(SPRITE &spr)
{
  pSprite z = NULL;

  // Find end of linked list
  pSprite Cur = SpriteHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  z = new Sprite;
  if (!z) return NULL;

  int SizeD = spr.w*spr.h;

  z->x  = z->y  = 0;
  z->dx = z->dy = 0;
  z->ox = z->oy = 0;
  z->w = spr.w;
  z->h = spr.h;
  z->totfr  = spr.fr;
  z->curfr  = 0;
  z->hidden = 0;
  z->update = 0;
  z->loop   = 0;

  z->Spr = new BITMAP[spr.fr];
  if (!z->Spr)
  {
    g_Done();
    printf("Sprite Manager: Could not initialize sprite buffer.\n");
    exit(1);
  }
  char *s = (char *) spr.Data;
  for (int i=0; i<spr.fr; i++)
  {
    z->Spr[i].x = spr.w;
    z->Spr[i].y = spr.h;
    z->Spr[i].Data = s;
    s += SizeD;
  }
  z->next = NULL;

  if (Cur)
  {
    Cur->next = z;
    SpriteTail = Cur->next;
  } else SpriteHead = SpriteTail = z;

  return z;
}

pSprite s_Add(BITMAP &b)
{
  pSprite z = NULL;
  
  // Find end of linked list
  pSprite Cur = SpriteHead;
  if (Cur) while (Cur->next) Cur = Cur->next;

  z = new Sprite;
  if (!z) return NULL;

  int SizeD = b.x*b.y;

  z->x  = z->y  = 0;
  z->dx = z->dy = 0;
  z->ox = z->oy = 0;
  z->w = b.x;
  z->h = b.y;
  z->totfr  = 1;
  z->curfr  = 0;
  z->hidden = 0;
  z->update = 0;
  z->loop   = 0;

  z->Spr = new BITMAP;
  if (!z->Spr)
  {
    g_Done();
    printf("Sprite Manager: Could not initialize sprite buffer.\n");
    exit(1);
  }
  z->Spr->x = b.x;
  z->Spr->y = b.y;
  z->Spr->Data = b.Data;
  z->next = NULL;

  if (Cur)
  {
    Cur->next = z;
    SpriteTail = Cur->next;
  } else SpriteHead = SpriteTail = z;

  return z;
}

pSprite s_AddMouse(BITMAP &b)
{
  pSprite z = NULL;
  
  z = new Sprite;
  if (!z) return NULL;

  int SizeD = b.x*b.y;

  z->x  = z->y  = 0;
  z->dx = z->dy = 0;
  z->ox = z->oy = 0;
  z->w = b.x;
  z->h = b.y;
  z->totfr  = 1;
  z->curfr  = 0;
  z->hidden = 0;
  z->update = 0;
  z->loop   = 0;

  z->Spr = new BITMAP;
  if (!z->Spr)
  {
    g_Done();
    printf("Sprite Manager: Could not initialize sprite buffer.\n");
    exit(1);
  }
  z->Spr->x = b.x;
  z->Spr->y = b.y;
  z->Spr->Data = b.Data;
  z->next = NULL;

  MousePtr = z;
  return z;
}

void s_Update(int bg)
{
  pSprite Cur;

  // First update all positions
  Cur = SpriteHead;
  while (Cur)
  {
    Cur->x += Cur->dx;
    Cur->y += Cur->dy;

    Cur = Cur->next;
  }

  // Do likewise for special mouse sprite
  if (MousePtr)
  {
    MousePtr->x += MousePtr->dx;
    MousePtr->y += MousePtr->dy;
  }

  // if bg==0, then program must clear sprites off screen manually.
  // Usually this is the case in a tile-based game (where the screen
  // is constantly updated.) if bg==1, then sprites clear themselves
  // automatically (they replace the flippage with info from spritepage)
  if (bg)
  {
    // Second, remove all sprites from the screen
    Cur = SpriteHead;
    while (Cur)
    {
      g_cCopyDiffPage(Cur->ox, Cur->oy, Cur->ox+Cur->w, Cur->oy+Cur->h,
                      Cur->ox, Cur->oy, SprPage, CP);

      Cur = Cur->next;
    }

    // Do likewise for special mouse sprite
    if (MousePtr)
    {
      g_cCopyDiffPage(MousePtr->ox, MousePtr->oy, MousePtr->ox+MousePtr->w, MousePtr->oy+MousePtr->h,
                      MousePtr->ox, MousePtr->oy, SprPage, CP);
    }
  }

  // Now draw all sprites onto screen (note: a special sort can be done
  // here, for example, to draw sprites in Z order, etc...)
  Cur = SpriteHead;
  while (Cur)
  {
    if (!Cur->hidden) 
    {
      if (Cur->loop)
      {
        g_ctBitmap(Cur->x, Cur->y, &Cur->Spr[Cur->curfr++]);
        if (Cur->curfr>=Cur->totfr) Cur->curfr = 0;
      } else g_ctBitmap(Cur->x, Cur->y, &Cur->Spr[Cur->curfr]);

      Cur->ox = Cur->x;
      Cur->oy = Cur->y;
    }

    Cur = Cur->next;
  }

  // Do likewise for special mouse sprite
  if (MousePtr)
  {
    g_ctBitmap(MousePtr->x, MousePtr->y, &MousePtr->Spr[MousePtr->curfr]);

    MousePtr->ox = MousePtr->x;
    MousePtr->oy = MousePtr->y;
  }
}

void s_SetPos(pSprite spr, int x, int y)
{
  spr->x = x;
  spr->y = y;
}

void s_SetMove(pSprite spr, int dx, int dy)
{
  spr->dx = dx;
  spr->dy = dy;
}

void s_SetFrame(pSprite spr, int f)
{
  spr->curfr = f;
}

void s_SetLoop(pSprite spr, int l)
{
  spr->loop = l;
}

void s_Hide(pSprite spr)
{
  spr->hidden = 1;
  spr->update = 1;
}

void s_Show(pSprite spr)
{
  spr->hidden = 0;
  spr->update = 2;
}

