// Tile, version 1.1 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version
// Version 1.1 -- Implemented two custom 32-bit interior tile blitters,
//                for a 200% speed increase.

#include <stdlib.h>
#include <stdio.h>
#include <graph.h>
#include "mem32.h"
#include "vgl.h"
#include "tile.h"

int MapDX, MapDY;
BITMAP *Tiles=NULL;
int TileW, TileH;

TileMap::TileMap()
{
  Base = Fringe = Roof = NULL;
  MapOffset = NULL;
}

TileMap::~TileMap()
{
  delete[] Base;
  delete[] Fringe;
  delete[] Roof;
  delete[] MapOffset;
}

void TileMap::InitMap(int w, int h)
{
  MapW = w;
  MapH = h;

  if (Base) delete[] Base;
  Base = new short[w*h];
  if (!Base)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile base map.\n");
    exit(1);
  }
  ClearBase(0);

  if (Fringe) delete[] Fringe;
  Fringe = new short[w*h];
  if (!Fringe)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile fringe map.\n");
    exit(1);
  }
  ClearFringe(0);

  if (Roof) delete[] Roof;
  Roof = new short[w*h];
  if (!Roof)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile roof map.\n");
    exit(1);
  }
  ClearRoof(0);

  if (MapOffset) delete[] MapOffset;
  MapOffset = new int[h];
  if (!MapOffset)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile map offsets.\n");
    exit(1);
  }
  for (int i=0; i<h; i++)
    MapOffset[i] = i*w;
  MapSize = w*h;
}

void TileMap::InitMap(int w, int h, short *b, short *f, short *r)
{
  MapW   = w;
  MapH   = h;
  Base   = b;
  Fringe = f;
  Roof   = r;

  if (MapOffset) delete[] MapOffset;
  MapOffset = new int[h];
  if (!MapOffset)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile map offsets.\n");
    exit(1);
  }
  for (int i=0; i<h; i++)
    MapOffset[i] = i*w;
  MapSize = w*h;
}

void TileMap::InitRoof(short *roof)
{
  if (Roof) delete[] Roof;
  Roof = roof;
}

void TileMap::ReInit(int w, int h)
{
  MapW = w;
  MapH = h;

  if (Base) delete[] Base;
  Base = new short[w*h];
  if (!Base)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile base map.\n");
    exit(1);
  }
  ClearBase(0);

  if (Fringe) delete[] Fringe;
  Fringe = new short[w*h];
  if (!Fringe)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile fringe map.\n");
    exit(1);
  }
  ClearFringe(0);

  if (Roof) delete[] Roof;
  Roof = new short[w*h];
  if (!Roof)
  {
    g_Done();
    printf("Tile: Insufficient memory for tile roof map.\n");
    exit(1);
  }
  ClearRoof(0);

  if (MapOffset) delete[] MapOffset;
  MapOffset = new int[h];
  if (!MapOffset)
  {
    g_Done();
    printf("Tile: Error initializing tile map offsets.\n");
    exit(1);
  }
  for (int i=0; i<h; i++)
    MapOffset[i] = i*w;
  MapSize = w*h;
}

void TileMap::ReInit(void)
{
  _BlitMaxX = g_ResX() - TileW;
  _BlitMaxY = g_ResY() - TileH;
  MapDX = (g_ResX()/TileW)+2;
  MapDY = (g_ResY()/TileH)+2;
}

void TileMap::BlackOut(void)
{
  if (Base)   blackout(Base, MapSize/4);
  if (Fringe) blackout(Fringe, MapSize/4);
  if (Roof)   blackout(Roof, MapSize/4);
}

void TileMap::ClearBase(int n)
{
  if (!Base) return;
  for (int i=0; i<MapW*MapH; i++)
    Base[i] = n;
}

void TileMap::ClearFringe(int n)
{
  if (!Fringe) return;
  for (int i=0; i<MapW*MapH; i++)
    Fringe[i] = n;
}

void TileMap::ClearRoof(int n)
{
  if (!Roof) return;
  for (int i=0; i<MapW*MapH; i++)
    Roof[i] = n;
}

void TileMap::Border(void)
{
  ClearBase(0);

  for (int i=0; i<MapW; i++)
  {
    Base[i] = 1;
    Base[MapOffset[MapH-1]+i] = 2;
  }

  for (int j=0; j<MapH; j++)
  {
    Base[MapOffset[j]] = 1;
    Base[MapOffset[j]+MapW-1] = 2;
  }

  for (i=0; i<MapW*MapH; i+=3)
    Fringe[i] = 3;
}

void TileMap::DrawBase(void)
{
  int MapX, MapY;
  int ScrX, ScrY;
  int MapX1, MapY1;
  int MapX2, MapY2;
  int TileDX, TileDY;
  int TScrX, TScrY;
  int TileOffset;
  char *ScrnOffset;

  MapX1  = PosX / TileW;
  MapX2  = MapX1 + MapDX;
  TileDX = PosX % TileW;
  MapY1  = PosY / TileH;
  MapY2  = MapY1 + MapDY;
  TileDY = PosY % TileH;

  for (MapY=MapY1, ScrY = 0; MapY<MapY2; MapY++, ScrY += TileH)
  {
    TileOffset = MapOffset[MapY];
    TScrY      = ScrY-TileDY;
    ScrnOffset = Pages[CP].Screen + (TScrY << 6) + (TScrY << 8);

    if (MapY >= 0 && MapY < MapH)
      for (MapX=MapX1, ScrX = 0; MapX<MapX2; MapX++, ScrX += TileW)
      {
        TScrX = ScrX-TileDX;
        
        if (TScrX>=0 && TScrX<_BlitMaxX && TScrY>=0 && TScrY<_BlitMaxY)
          g_Tile(ScrnOffset+TScrX, Tiles[Base[TileOffset+MapX]].Data); else
          g_cBitmap(TScrX, TScrY, &Tiles[Base[TileOffset+MapX]]);
      }
  }
/*
  _settextposition(24, 1);
  printf("x:%d y:%d MapX1:%d MapY1:%d  ", x, y, MapX1, MapY1);
*/
}

void TileMap::DrawFringe(void)
{
  int MapX, MapY;
  int ScrX, ScrY;
  int MapX1, MapY1;
  int MapX2, MapY2;
  int TileDX, TileDY;
  int TScrX, TScrY;
  int TileOffset;
  char *ScrnOffset;

  MapX1  = PosX / TileW;
  MapX2  = MapX1 + MapDX;
  TileDX = PosX % TileW;
  MapY1  = PosY / TileH;
  MapY2  = MapY1 + MapDY;
  TileDY = PosY % TileH;

  for (MapY=MapY1, ScrY = 0; MapY<MapY2; MapY++, ScrY += TileH)
  {
    TileOffset = MapOffset[MapY];
    TScrY      = ScrY-TileDY;
    ScrnOffset = Pages[CP].Screen + (TScrY << 6) + (TScrY << 8);

    if (MapY >= 0 && MapY < MapH)
      for (MapX=MapX1, ScrX = 0; MapX<MapX2; MapX++, ScrX += TileW)
      {
        TScrX = ScrX-TileDX;
        
        if (MapX >= 0 && MapX < MapW && Fringe[TileOffset+MapX])
        {
          if (TScrX>=0 && TScrX<_BlitMaxX && TScrY>=0 && TScrY<_BlitMaxY)
            g_tTile(ScrnOffset+TScrX, Tiles[Fringe[TileOffset+MapX]].Data); else
            g_ctBitmap(TScrX, TScrY, &Tiles[Fringe[TileOffset+MapX]]);
        }
      }
  }
/*
  _settextposition(24, 1);
  printf("x:%d y:%d MapX1:%d MapY1:%d  ", x, y, MapX1, MapY1);
*/
}

void TileMap::DrawRoof(void)
{
  int MapX, MapY;
  int ScrX, ScrY;
  int MapX1, MapY1;
  int MapX2, MapY2;
  int TileDX, TileDY;
  int TScrX, TScrY;
  int TileOffset;
  char *ScrnOffset;

  MapX1  = PosX / TileW;
  MapX2  = MapX1 + MapDX;
  TileDX = PosX % TileW;
  MapY1  = PosY / TileH;
  MapY2  = MapY1 + MapDY;
  TileDY = PosY % TileH;

  for (MapY=MapY1, ScrY = 0; MapY<MapY2; MapY++, ScrY += TileH)
  {
    TileOffset = MapOffset[MapY];
    TScrY      = ScrY-TileDY;
    ScrnOffset = Pages[CP].Screen + (TScrY << 6) + (TScrY << 8);

    if (MapY >= 0 && MapY < MapH)
      for (MapX=MapX1, ScrX = 0; MapX<MapX2; MapX++, ScrX += TileW)
      {
        TScrX = ScrX-TileDX;
        
        if (MapX >= 0 && MapX < MapW && Roof[TileOffset+MapX])
        {
          if (TScrX>=0 && TScrX<_BlitMaxX && TScrY>=0 && TScrY<_BlitMaxY)
            g_tTile(ScrnOffset+TScrX, Tiles[Roof[TileOffset+MapX]].Data); else
            g_ctBitmap(TScrX, TScrY, &Tiles[Roof[TileOffset+MapX]]);
        }
      }
  }
/*
  _settextposition(24, 1);
  printf("x:%d y:%d MapX1:%d MapY1:%d  ", x, y, MapX1, MapY1);
*/
}

void TileMap::Draw(void)
{
  if (Base)   DrawBase();
  if (Fringe) DrawFringe();
  if (Roof)   DrawRoof();
}

void t_Init(int num)
{
  // Initialize tiles
  if (!Tiles)
  {
    Tiles = new BITMAP[num];
    if (!Tiles)
    {
      g_Done();
      printf("Tile: Error initializing tile library.\n");
      exit(1);
    }
  }

  for (int i=0; i<num; i++)
  {
    Tiles[i].x = 0;
    Tiles[i].y = 0;
    Tiles[i].Data = NULL;
  }
}

void t_TileSize(int x, int y)
{
  TileW = x;
  TileH = y;
}

int t_Load(char *f)
{
  int n = g_LoadBitmap(f, Tiles);
  return n;
}

int t_Load(char *f, int start)
{
  int n = g_LoadBitmap(f, &Tiles[start]);

  // Use first tile as width/height for all
  TileW = Tiles->x;
  TileH = Tiles->y;

  return n;
}

void t_Done(void)
{
  if (Tiles) delete[] Tiles;
}

