// Tile, version 1.1 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version
// Version 1.1 -- Implemented custom 32-bit interior tile blitter, for a
//                200% speed increase.

#ifndef __TILE__
#define __TILE__

#include "vgl.h"

extern BITMAP *Tiles;
extern int TileW, TileH;
extern int MapDX, MapDY;

class TileMap {
protected:
  int MapW, MapH;
  int MapSize;
  int PosX, PosY;
  short *Base;
  short *Fringe;
  short *Roof;
  int   *MapOffset;

public:
  TileMap();
  ~TileMap();

  int  MapWidth(void)                  { return MapW; }
  int  MapHeight(void)                 { return MapH; }
  int  TileWidth(void)                 { return TileW; }
  int  TileHeight(void)                { return TileH; }
  void SetPos(int x, int y)            { PosX = x; PosY = y; }
  int  Tile2ScrX(int x)                { return (x*TileW) - PosX; }
  int  Tile2ScrY(int y)                { return (y*TileH) - PosY; }

  void InitMap(int w, int h);
  void InitMap(int w, int h, short *b, short *f, short *r);
  void InitRoof(short *roof);
  void ReInit(int w, int h);
  void ReInit(void);
  void ClearBase(int n);
  void ClearFringe(int n);
  void ClearRoof(int n);
  void BlackOut(void);
  void Border(void);
  virtual void DrawBase(void);
  virtual void DrawFringe(void);
  virtual void DrawRoof(void);
  void Draw(void);
};

void t_Init(int num);
void t_TileSize(int x, int y);
int  t_Load(char *f);
int  t_Load(char *f, int start);
void t_Done(void);

static int _BlitMaxX, _BlitMaxY;

// Blazingly fast "interior" tile blitter. This blitter is optimized for
// 8-bit linear graphics, although alternate ones could be made up for any
// color depth. It is unrolled 4 times (4 loops).
// Can only blit interior tiles, no clipping, no transparency, etc...
extern void g_Tile(void *d, void *s);
/*
#pragma aux g_Tile = \
  "mov ebx, 4"\
  "mov ecx, _BlitMaxX"\
  "reps:"\
  "movsd"\
  "movsd"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "movsd"\
  "movsd"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "movsd"\
  "movsd"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "movsd"\
  "movsd"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "dec ebx"\
  "jnz reps"\
*/
#pragma aux g_Tile = \
  "mov ebx, 4"\
  "mov ecx, _BlitMaxX"\
  "reps:"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "movsd"\
  "movsd"\
  "add edi, ecx"\
  "dec ebx"\
  "jnz reps"\
parm [edi] [esi]\
modify [ebx ecx edi esi];

// Fast "interior" tile blitter. This blitter is optimized for 8-bit linear
// graphics, although alternate ones could be made up for any color depth.
// It is unrolled 4 times (4 loops). Handles transparency
// Can only blit interior tiles, no clipping, etc...
// ulong  _cols;
// uchar  _bcolor;
extern void g_tTile(void *d, void *s);
#pragma aux g_tTile = \
  "mov  ecx, _BlitMaxX"\
  "mov  edx, 8"\
  "xor  eax, eax"\
  "l1:"\
  "mov  ebx, 4"\
  "l2:"\
  "lodsb"\
  "or   al, al"\
  "jz   s1"\
  "mov  [edi],al"\
  "s1:"\
  "inc  edi"\
  "lodsb"\
  "or   al, al"\
  "jz   s2"\
  "mov  [edi],al"\
  "s2:"\
  "inc  edi"\
  "lodsb"\
  "or   al, al"\
  "jz   s3"\
  "mov  [edi],al"\
  "s3:"\
  "inc  edi"\
  "lodsb"\
  "or   al, al"\
  "jz   s4"\
  "mov  [edi],al"\
  "s4:"\
  "inc  edi"\
  "dec  ebx"\
  "jnz  l2"\
  "add  edi, ecx"\
  "dec  edx"\
  "jnz  l1"\
parm [edi] [esi]\
modify [eax ebx ecx edx];

#endif
