// Watcom includes
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <i86.h>

// HGL includes
#include "vgl.h"
#include "font.h"
#include "tile.h"
#include "mouse.h"
#include "multikey.h"

// Z2 includes
#include "user.h"
#include "global.h"
#include "editor.h"

#define MAXCONTROLS 128
char help[] = "F1: Help";

void WinFrame(int x1, int y1, int x2, int y2)
{
  // Top-left inner border
  g_HLine(x1, x2, y1, colDGray);
  g_VLine(x1, y1, y2, colDGray);

  // Bottom-right inner border
  g_HLine(x1, x2, y2, colWhite);
  g_VLine(x2, y1, y2, colWhite);

  // Border
  g_Rect(x1-1, y1-1, x2+1, y2+1, colLGray);

  // Top-left outer border
  g_HLine(x1-2, x2+2, y1-2, colWhite);
  g_VLine(x1-2, y1-2, y2+2, colWhite);

  // Bottom-right outer border
  g_HLine(x1-2, x2+2, y2+2, colDGray);
  g_VLine(x2+2, y1-2, y2+2, colDGray);
}

void OutFrame(int x1, int y1, int x2, int y2)
{
  // Top-left border
  g_HLine(x1, x2, y1, colWhite);
  g_VLine(x1, y1, y2, colWhite);

  // Bottom-right border
  g_HLine(x1, x2, y2, colDGray);
  g_VLine(x2, y1, y2, colDGray);

  // Top-left point, Bottom-right point
  g_SetPixel(x2, y1, colLGray);
  g_SetPixel(x1, y2, colLGray);
}

void OutFrameP(int x1, int y1, int x2, int y2)
{
  // Top-left border
  g_HLine(x1, x2, y1, colLPurp);
  g_VLine(x1, y1, y2, colLPurp);

  // Bottom-right border
  g_HLine(x1, x2, y2, colDPurp);
  g_VLine(x2, y1, y2, colDPurp);

  // Top-left point, Bottom-right point
  g_SetPixel(x2, y1, colPurple);
  g_SetPixel(x1, y2, colPurple);
}

void InFrame(int x1, int y1, int x2, int y2)
{
  // Top-left border
  g_HLine(x1, x2, y1, colDGray);
  g_VLine(x1, y1, y2, colDGray);

  // Bottom-right border
  g_HLine(x1, x2, y2, colWhite);
  g_VLine(x2, y1, y2, colWhite);

  // Top-right point, Bottom-left point
  g_SetPixel(x2, y1, colLGray);
  g_SetPixel(x1, y2, colLGray);
}

View::View(Rect a)
{
  r = a;

  Back.Data = Update.Data = Title = NULL;

  SetState(0);
  SetFrame(1);
  SetActive(0);
  SetTitleAbove();

  g_GetBitmapXY(r.x-2, r.y-2, r.x2+2, r.y2+2, &Back);
  if (!Back.Data)
  {
    g_Done();
    printf("Z2 GUI: Error saving window background.\n");
    exit(1);
  }
}

View::~View(void)
{
  if (Update.Data) delete[] Update.Data;
  if (Back.Data) delete[] Back.Data;
}

void View::Done(void)
{
  g_Bitmap(r.x-2, r.y-2, &Back);
}

void View::Save(void)
{
  g_GetBitmapXY(r.x-2, r.y-2, r.x2+2, r.y2+2, &Update);
  if (!Update.Data)
  {
    g_Done();
    printf("Z2 GUI: Error saving window update area.\n");
    exit(1);
  }
}

void View::Restore(void)
{
  g_Bitmap(r.x-2, r.y-2, &Update);
}

Window::Window(Rect a) : View(a)
{
  Shade = 1;
}

Window::~Window() {}

void Window::Draw(void)
{
  if (GetFrame()) WinFrame(r.x, r.y, r.x2, r.y2);

  // Shading
  if (Shade)
    g_iRect(r.x+1, r.y+1, r.x2-1, r.y2-1, WinDark); else
    g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, 0);
}

void Window::Process(void)
{
  Draw();
}

Dialog::Dialog(Rect a) : View(a)
{
  ViewList = new View*[MAXCONTROLS];
  if (!ViewList)
  {
    g_Done();
    printf("Z2 GUI: Error allocating view list.\n");
    exit(1);
  }

  activity = -1;
  max = MAXCONTROLS;
  cur = 0;

  Help = 1;
}

Dialog::~Dialog()
{
  delete[] ViewList;
}

void Dialog::Add(View *v)
{
  if (cur==max)
  {
    max += 50;
    ViewList = (View **) realloc(ViewList, max * sizeof(View));
    if (!ViewList)
    {
      g_Done();
      printf("Z2 GUI: Error reallocating controls list.\n");
      exit(1);
    }
  }

  ViewList[cur] = v;
  cur++;
}

void Dialog::Draw(void)
{
  if (GetFrame()) WinFrame(r.x, r.y, r.x2, r.y2);

  // Shading
  if (Shade)
    g_iRect(r.x+1, r.y+1, r.x2-1, r.y2-1, Dark); else
    g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, 0);

  // Title
  g_Font(1);
  g_FontColor(colYellow);
  g_printf(r.x+5, r.y+5, GetTitle());

  // Help
  if (Help)
  {
    g_Font(2);
    g_FontColor(colWhite);
    g_printf(r.x2-g_strlen(help)-5, r.y+8, help);
  }
}

void Dialog::Process(void)
{
  int i;
  Draw();

  // check for inactivity
  if (activity==-1)
  {
    for (i=0; i<cur; i++)
      if (ViewList[i]->GetActive()) activity = i;

    if (activity==-1)
    {
      ViewList[0]->SetActive(1);
      activity = 0;
    }
  }

  ViewList[activity]->SetActive(0);

  for (i=0; i<cur; i++)
    if (ViewList[i]->CheckMouse() && m_Left()) activity = i;

  if (mk_Shift() && mk_Pressed(mkTab)) activity--;
  if (mk_Pressed(mkTab)) activity++;
  if (activity<0) activity = cur-1;
  if (activity>=cur) activity = 0;

  ViewList[activity]->SetActive(1);
  for (i=0; i<cur; i++)
  {
    if (i==activity)
      ViewList[i]->Process(); else
      ViewList[i]->Draw();
  }
}

Button::Button(Rect a) : View(a)
{
  ResetPress();
}

Button::~Button() {}

void Button::DrawUp(void)
{
  if (GetFrame()) OutFrameP(r.x, r.y, r.x2, r.y2);

  // Shading
  g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, colPurple);

  // Title
  g_Font(1);
  g_FontColor(colWhite);
  int offset = ((r.y2-r.y)-FontMan.FontHeight(1))/2;
  g_sbtwprintf(r.x, r.x2, r.y+offset, GetTitle());
}

void Button::DrawDown(void)
{
  if (GetFrame())
  {
    g_HLine(r.x+1, r.x2, r.y+1, colLPurp);
    g_VLine(r.x+1, r.y+1, r.y2, colLPurp);
  }

  // Shading
  g_fRect(r.x+2, r.y+2, r.x2, r.y2, colPurple);

  // Title
  g_Font(1);
  g_FontColor(colWhite);
  int offset = ((r.y2-r.y)-FontMan.FontHeight(1))/2;
  g_sbtwprintf(r.x+2, r.x2, r.y+offset+1, GetTitle());
}

void Button::DrawBase(int up)
{
  if (GetActive()) g_Rect(r.x-1, r.y-1, r.x2+1, r.y2+1, colWhite);
  if (up)
    DrawUp(); else
    DrawDown();
}

void Button::Draw(void)
{
  DrawBase(1);
}

int Button::CheckMouse(void)
{
  return (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<=r.y2);
}

void Button::Process(void)
{
  int down=0;

  // process mouse
  int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<=r.y2);
  if (!inrange) mpressed = 0;

  if (m_Left() && inrange)
    down = mpressed = 1; else
  if (!m_Left() && inrange)
    if (mpressed==1) mpressed = 2;

  if (mpressed) pressed = mpressed;

  // process keyboard
  if (mk_Down(mkEnter) || mk_Down(mkSpace))
    down = kpressed = 1; else
  if (!mk_Down(mkEnter) && !mk_Down(mkSpace))
    if (kpressed==1)
      kpressed = 2; else
      kpressed = 0;

  if (kpressed) pressed = kpressed;

  // draw
  if (!down)
    Draw(); else
    DrawBase(0);
}

ToolButton::ToolButton(Rect a) : Button(a) {}
ToolButton::~ToolButton() {}

void ToolButton::DrawUp(void)
{
  if (GetFrame()) OutFrameP(r.x, r.y, r.x2, r.y2);

  // Shading
  g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, colPurple);

  // Title
  g_Font(1);
  g_FontColor(colWhite);
  int offset = ((r.y2-r.y)-FontMan.FontHeight(1))/2;
  g_sbtwprintf(r.x+1, r.x2+1, r.y+1+offset, GetTitle());
}

void ToolButton::DrawDown(void)
{
  if (GetFrame())
  {
    g_HLine(r.x+1, r.x2, r.y+1, colLPurp);
    g_VLine(r.x+1, r.y+1, r.y2, colLPurp);
  }

  // Shading
  g_fRect(r.x+2, r.y+2, r.x2, r.y2, colPurple);

  // Title
  g_Font(1);
  g_FontColor(colWhite);
  int offset = ((r.y2-r.y)-FontMan.FontHeight(1))/2;
  g_sbtwprintf(r.x+3, r.x2+1, r.y+offset+2, GetTitle());
}

IconButton::IconButton(Rect a) : Button(a) {}
IconButton::~IconButton() {}

void IconButton::DrawBase(int up)
{
  if (up)
    DrawUp(); else
    DrawDown();
}

void IconButton::Draw(void)
{
  DrawBase(1);
}

void IconButton::Process(void)
{
  int down=0;

  // process mouse
  int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<=r.y2);
  if (!inrange) mpressed = 0;

  if (m_Left() && inrange)
    down = mpressed = 1; else
  if (!m_Left() && inrange)
    if (mpressed==1) mpressed = 2;

  if (mpressed) pressed = mpressed;

  // process keyboard
  if (mk_Down(mkEnter) || mk_Down(mkSpace))
    down = kpressed = 1; else
  if (!mk_Down(mkEnter) && !mk_Down(mkSpace))
    if (kpressed==1)
      kpressed = 2; else
      kpressed = 0;

  if (kpressed) pressed = kpressed;

  // draw
  if (!down)
    Draw(); else
    DrawBase(0);
}

List::List(Rect a) : View(a)
{
  Head = NULL;
  font = 2;
  maxlist = r.h/FontMan.FontHeight(font);
  listtop = 0;
  listcur = 0;
  maxel   = 0;
  SetNoMsg("No items");
}

List::~List()
{
  pListEl Cur = Head;
  while (Cur)
  {
    delete[] Cur;
    Cur = Cur->next;
  }
  Head = NULL;
}

void List::SetFont(int f)
{
  font = f;
  maxlist = r.h/FontMan.FontHeight(font);
}

void List::Add(char *str, char *str2, int s)
{
  pListEl NewEl = NULL;

  // Find end of linked list
  pListEl Cur = Head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  NewEl = new ListEl;
  if (!NewEl) return;

  strncpy(NewEl->name, str, 256);
  if (str2)
    strncpy(NewEl->info, str2, 256); else
    memset(NewEl->info, 0, 256);
  NewEl->state = s;
  NewEl->next = NULL;

  if (Cur)
    Cur->next = NewEl; else
    Head = NewEl;

  maxel++;
}

void List::Kill(pListEl l)
{
  pListEl Prev = Head;

  if (Prev == l)
  {
    Head = Prev->next;
    delete[] Prev;
  } else
  {
    while (Prev->next != l)
    {
      Prev = Prev->next;
      if (!Prev) return;
    }

    pListEl Cur = Prev->next;
    pListEl Next = Prev->next->next;

    Prev->next = Next;
    delete[] Cur;
  }
}

char *List::GetName(int i)
{
  pListEl Cur = Head;

  for (int z=0; z<i; z++)
    Cur = Cur->next;

  return Cur->name;
}

char *List::GetInfo(int i)
{
  pListEl Cur = Head;

  for (int z=0; z<i; z++)
    Cur = Cur->next;

  return Cur->info;
}

void List::SetDefault(char *str)
{
  pListEl Cur = Head;
  int i=0;

  while (Cur)
  {
    if (stricmp(Cur->info, str)==0)
    {
      listcur = i;

      // adjust other values
      if (listcur<0) listcur = 0;
      if (listcur>=maxel) listcur = maxel-1;
      if (listcur>=listtop+maxlist) listtop = listcur-maxlist+1;
      if (listcur<=listtop) listtop = listcur;
      return;
    }
    i++;
    Cur = Cur->next;
  }
}

int List::CheckMouse(void)
{
  if (!GetTitlePos())
  {
    int yy = FontMan.FontHeight(2);

    return (m_X()>=r.x && m_X()<=r.x2 &&
            m_Y()>=r.y-yy && m_Y()<=r.y2);
  } else
  {
    int xx = GetTitlePos();

    return (m_X()>=r.x-xx && m_X()<=r.x2 &&
            m_Y()>=r.y && m_Y()<=r.y2);
  }
}

void List::Draw(void)
{
  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(font);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(font)-1, GetTitle());

  // List
  if (Head)
  {
    int fh = FontMan.FontHeight(font);
    int i, y = r.y+1;
    g_FontColor(colWhite);
    pListEl Cur = Head;

    for (i=0; i<listtop; i++)
      Cur = Cur->next;

    for (i=0; i<maxlist && Cur; i++, y += fh, Cur = Cur->next)
    {
      if (i+listtop == listcur)
//      g_fRect(r.x+1, y, r.x2-1, y+fh, colPurple);
        g_iRect(r.x+1, y, r.x2-1, y+fh, PalPurp);
      g_printf(r.x+2, y, Cur->name);
    }
  } else
  {
    g_FontColor(colWhite);
    g_printf(r.x+2, r.y+2, NothingMsg);
  }
}

void List::Process(void)
{
  if (Head)
  {
    if (mk_Pressed(mkDown)) listcur++;
    if (mk_Pressed(mkUp))   listcur--;
    if (mk_Pressed(mkPgDn)) listcur += maxlist;
    if (mk_Pressed(mkPgUp)) listcur -= maxlist;
    if (mk_Pressed(mkHome)) listcur = 0;
    if (mk_Pressed(mkEnd))  listcur = maxel-1;

    int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<r.y2);

    if (m_Left() && inrange)
    {
      int y;
      y = (m_Y() - r.y-2)/FontMan.FontHeight(font);
      y += listtop;
      listcur = y;
    }

    if (listcur<0) listcur = 0;
    if (listcur>=maxel) listcur = maxel-1;
    if (listcur>=listtop+maxlist) listtop = listcur-maxlist+1;
    if (listcur<=listtop) listtop = listcur;
  }
  Draw();
}

Group::Group(Rect a) : View(a)
{
  Head    = NULL;
  listcur = 0;
  maxel   = 0;
  gfont   = 1;
  on      = '!';
  off     = '.';
}

Group::~Group()
{
  KillAll();
}

void Group::Add(char *str, int s)
{
  pGroupEl NewEl = NULL;

  // Find end of linked Group
  pGroupEl Cur = Head;
  if (Cur) while (Cur->next) Cur = Cur->next;

  NewEl = new GroupEl;
  if (!NewEl) return;

  strncpy(NewEl->name, str, 256);
  NewEl->state = s;
  NewEl->next = NULL;

  if (Cur)
    Cur->next = NewEl; else
    Head = NewEl;

  maxel++;
}

void Group::Kill(pGroupEl l)
{
  pGroupEl Prev = Head;

  if (Prev == l)
  {
    Head = Prev->next;
    delete[] Prev;
  } else
  {
    while (Prev->next != l)
    {
      Prev = Prev->next;
      if (!Prev) return;
    }

    pGroupEl Cur = Prev->next;
    pGroupEl Next = Prev->next->next;

    Prev->next = Next;
    delete[] Cur;
  }
}

void Group::KillAll(void)
{
  pGroupEl Cur = Head;
  while (Cur)
  {
    delete[] Cur;
    Cur = Cur->next;
  }
  Head = NULL;
}

char *Group::GetName(int i)
{
  pGroupEl Cur = Head;

  for (int z=0; z<i; z++)
    Cur = Cur->next;

  return Cur->name;
}

int Group::GetState(int i)
{
  pGroupEl Cur = Head;

  for (int z=0; z<i; z++)
    Cur = Cur->next;

  return Cur->state;
}

int Group::CheckMouse(void)
{
  if (!GetTitlePos())
  {
    int yy = FontMan.FontHeight(2);

    return (m_X()>=r.x && m_X()<=r.x2 &&
            m_Y()>=r.y-yy && m_Y()<=r.y2);
  } else
  {
    int xx = GetTitlePos();

    return (m_X()>=r.x-xx && m_X()<=r.x2 &&
            m_Y()>=r.y && m_Y()<=r.y2);
  }
}

void Group::Toggle(int i)
{
  pGroupEl Cur = Head;

  for (int z=0; z<i; z++)
    Cur = Cur->next;

  Cur->state = 1-Cur->state;
}

void Group::Draw(void)
{
//  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  // List
  g_Font(gfont);
  if (Head)
  {
    int fh = FontMan.FontHeight(gfont);
    int i, y = r.y;

    i = 0;
    pGroupEl Cur = Head;
    while (Cur)
    {
      if (i == listcur && GetActive())
      {
        g_FontColor(colWhite);
        g_printf(r.x, y, "%c %s", Cur->state ? on : off, Cur->name);
      } else
      {
        g_FontColor(colYellow);
        g_printf(r.x, y, "%c %s", Cur->state ? on : off, Cur->name);
      }

      i++;
      y += fh;
      Cur = Cur->next;
    }
  } else
  {
    g_FontColor(colWhite);
    g_printf(r.x, r.y, "No group defined.");
  }
}

void Group::Process(void)
{
  if (Head)
  {
    if (mk_Pressed(mkDown)) listcur++;
    if (mk_Pressed(mkUp))   listcur--;
    if (mk_Pressed(mkHome)) listcur = 0;
    if (mk_Pressed(mkEnd))  listcur = maxel-1;
    if (mk_Pressed(mkSpace)) Toggle(listcur);

    int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<r.y2);

    if (m_Left() && inrange)
    {
      listcur = (m_Y() - r.y-2)/FontMan.FontHeight(gfont);
      Toggle(listcur);
      delay(150);
    }

    if (listcur<0) listcur = 0;
    if (listcur>=maxel) listcur = maxel-1;
  }
  Draw();
}

Check::Check(Rect a) : Group(a)
{
  on  = 11;
  off = 12;
}

Radio::Radio(Rect a) : Group(a)
{
  on  = 14;
  off = 15;
}

void Radio::Toggle(int i)
{
  pGroupEl Cur = Head;
  while (Cur)
  {
    Cur->state = 0;
    Cur = Cur->next;
  }

  Cur = Head;
  for (int z=0; z<i; z++)
    Cur = Cur->next;

  Cur->state = 1;
}

void Radio::Draw(void)
{
//  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Check for no buttons pressed (if so, press first button)
  pGroupEl Cur = Head;
  int test=0;
  while (Cur)
  {
    test += Cur->state;
    Cur = Cur->next;
  }
  if (!test) Head->state = 1;

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  // List
  g_Font(1);
  if (Head)
  {
    int fh = FontMan.FontHeight(gfont);
    int i, y = r.y+1;

    i = 0;
    pGroupEl Cur = Head;
    while (Cur)
    {
      if (i == listcur && GetActive())
      {
        g_FontColor(colWhite);
        g_printf(r.x, y, "%c %s", Cur->state ? on : off, Cur->name);
      } else
      {
        g_FontColor(colYellow);
        g_printf(r.x, y, "%c %s", Cur->state ? on : off, Cur->name);
      }

      i++;
      y += fh;
      Cur = Cur->next;
    }
  } else
  {
    g_FontColor(colWhite);
    g_printf(r.x+2, r.y+2, "No group defined.");
  }
}

Editor::Editor(Rect a) : View(a)
{
  Data = new char[SCRATCH_SIZE];
  if (!Data)
  {
    g_Done();
    printf("Z2 GUI: Error allocating editor buffer.\n");
    exit(1);
  }
  memset(Data, 0, SCRATCH_SIZE);

  t.SetClip(r.x+2, r.y+1, r.x2-2, r.y2-1);
  t.SetFontID(1);

  maxlist  = (r.y2-r.y)/(FontMan.FontHeight(1)+1);
  listtop  = 0;
  listcur  = 0;
  maxline  = 0;
  Cursor   = 0;
  InsMode  = 0;
  vcursorx = 0;
  vcursory = 0;
}

Editor::~Editor()
{
  if (Data) delete[] Data;
}

int Editor::FindNextLine(int start)
{
  return FindEndLine(start)+1;
}

int Editor::FindPrevLine(int start)
{
  start = FindBegLine(start)-1;
  return FindBegLine(start);
}

int Editor::FindBegLine(int start)
{
  if (t.Text[start] == '\n' && t.Text[start-1] == '\n')
    return start;

  while (start>0 && t.Text[start-1] != '\n')
    start--;

  return start;
}

int Editor::FindEndLine(int start)
{
  while (start<=strlen(t.Text) && t.Text[start] != '\n')
    start++;

  return start;
}

int Editor::FindColumn(int start, int col)
{
  int c = 0;
  while (start<=strlen(t.Text) && t.Text[start] != '\n' && c<col)
  {
    start++;
    c++;
  }

  return start;
}

void Editor::GetXY(int &x, int &y, int c)
{
  x = y = 0;
  for (int i=0; i<c && t.Text[i]; i++)
  {
    x++;
    if (t.Text[i]=='\n')
    {
      x = 0;
      y++;
    }
  }
}

void Editor::DrawBase()
{
  char *str;

  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  if (!GetTitlePos())
    g_printf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle()); else
    g_printf(r.x-GetTitlePos(), r.y+2, GetTitle());

  t.SetText(Data);
  UpdateXY();

  if (GetActive())
  {
    int offsetx = r.x+2;
    int offsety = r.y+2;

    listcur = 0;
    for (int i=0; i<Cursor; i++)
    {
      if (t.Text[i]=='\n')
      {
        offsetx = r.x+2;
        offsety += FontMan.FontHeight(1);
        listcur++;
      } else offsetx += FontMan.CharWidth(1, t.Text[i])+1;
    }

    int x1 = offsetx;
    int x2 = offsetx+FontMan.CharWidth(1, Data[Cursor]);
    int y1 = offsety-listtop;
    int y2 = offsety+FontMan.FontHeight(1)-1-listtop;

    if (InsMode) y1 += (FontMan.FontHeight(1)+1)/2;

    int b = listtop * (FontMan.FontHeight(1)-1);
    if (Cursor == strlen(t.Text) || t.Text[Cursor]=='\n')
      g_fRect(x1, y1-b, x1+4, y2-b, colPurple); else
      g_fRect(x1, y1-b, x2, y2-b, colPurple);
  }

  // Draw text
  g_FontColor(colWhite);
  t.Print(listtop*FontMan.FontHeight(1), 15);
}

void Editor::Draw(void)
{
  DrawBase();

/*
  // draw debug
  g_FontColor(colWhite);
  g_Font(2);
  g_sprintf(1, 1, "VCurX:%d VCurY:%d Cursor:%d DataLen:%d",
    vcursorx, vcursory, Cursor, strlen(Data));
  g_sprintf(1, 8, "ListTop:%d ListCur:%d MaxLine:%d MaxList:%d",
    listtop, listcur, maxline, maxlist);
*/
}

void Editor::Process(void)
{
  int i;

  char c = mk_PollChar();

  switch (c)
  {
    case mcNone : case mcF1 : case mcF2 : case mcF3 : case mcF4 :
    case mcF5 : case mcF6 : case mcF7 : case mcF8 : case mcF9 :
    case mcF10 : case mcF11 : case mcF12 : case mcTab : break;

    case mcIns :
      InsMode = !InsMode;
      break;
    case mcDel :
      if (Cursor<strlen(Data))
        memmove(Data + Cursor, Data + Cursor + 1, strlen(Data)-Cursor);
      break;
    case mcBackspace :
      if (Cursor>0)
      {
        if (Cursor<strlen(Data))
        {
          Cursor--;
          memmove(Data + Cursor, Data + Cursor + 1, strlen(Data)-Cursor);
        } else Data[--Cursor] = 0;
      }
      break;
    case mcEnter :
      // insert line break if in Insert mode, otherwise go to next line
      if (InsMode)
      {
        memmove(Data + Cursor + 1, Data + Cursor, strlen(Data)-Cursor+1);
        Data[Cursor++] = '\n';
      } else
      if (Cursor == strlen(Data))
      {
        Data[Cursor++] = '\n';
      } else Cursor = FindNextLine(Cursor);
      break;
    case mcEsc : break;
    case mcUp :
      if (vcursory>0)
      {
        Cursor = FindPrevLine(Cursor);
        Cursor = FindColumn(Cursor, vcursorx);
      }
      break;
    case mcDown :
      Cursor = FindNextLine(Cursor);
      Cursor = FindColumn(Cursor, vcursorx);
      break;
    case mcLeft :
      if (Cursor>0) Cursor--;
      break;
    case mcRight :
      if (Cursor<strlen(Data)) Cursor++;
      break;
    case mcHome :
      Cursor = FindBegLine(Cursor);
      break;
    case mcEnd :
      Cursor = FindEndLine(Cursor);
      break;
    case mcPgUp :
      for (i=0; i<maxlist; i++)
        Cursor = FindPrevLine(Cursor);
      break;
    case mcPgDn :
      for (i=0; i<maxlist; i++)
        Cursor = FindNextLine(Cursor);
      break;
    default :
      if (InsMode || Data[Cursor]=='\n')
        memmove(Data + Cursor + 1, Data + Cursor, strlen(Data)-Cursor+1);
      Data[Cursor++] = c;
      break;
  }

  if (Cursor<0) Cursor = 0;
  if (Cursor>strlen(Data)) Cursor = strlen(Data);

  GetXY(vcursorx, vcursory, Cursor);

  while (vcursory>listtop+maxlist) listtop++;
  while (vcursory<listtop) listtop--;

  Draw();
}

int Editor::CheckMouse(void)
{
  if (!GetTitlePos())
  {
    int yy = FontMan.FontHeight(2);

    return (m_X()>=r.x && m_X()<=r.x2 &&
            m_Y()>=r.y-yy && m_Y()<=r.y2);
  } else
  {
    int xx = GetTitlePos();

    return (m_X()>=r.x-xx && m_X()<=r.x2 &&
            m_Y()>=r.y && m_Y()<=r.y2);
  }
}

ListTile::ListTile(Rect a) : List(a)
{
  int AdjTileH = Tile_H/((Tile_H>24)+1);

  theight = AdjTileH+1;
  if (theight<10) theight = 10;
  maxlist = r.h/theight;
  pressed = 0;
}

ListTile::~ListTile()
{
  pListEl Cur = Head;
  while (Cur)
  {
    delete[] Cur;
    Cur = Cur->next;
  }
  Head = NULL;
}

int ListTile::CheckMouse(void)
{
  if (!GetTitlePos())
  {
    int yy = FontMan.FontHeight(2);

    return (m_X()>=r.x && m_X()<=r.x2 &&
            m_Y()>=r.y-yy && m_Y()<=r.y2);
  } else
  {
    int xx = GetTitlePos();

    return (m_X()>=r.x-xx && m_X()<=r.x2 &&
            m_Y()>=r.y && m_Y()<=r.y2);
  }
}

void ListTile::Draw(void)
{
  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  // ListTile
  g_Font(1);

  int AdjTileW = Tile_W/((Tile_W>24)+1);
  int AdjTileH = Tile_H/((Tile_H>24)+1);

  if (Head)
  {
    int fh = FontMan.FontHeight(font);
    int i, y = r.y+1;
    g_FontColor(colWhite);
    pListEl Cur = Head;

    for (i=0; i<listtop; i++)
      Cur = Cur->next;

    for (i=0; i<maxlist && Cur; i++, y += theight, Cur = Cur->next)
    {
      if (i+listtop == listcur)
        g_iRect(r.x+1, y, r.x2-1, y+theight, PalPurp);
      g_printf(r.x+AdjTileW+4, y+((AdjTileH-fh)/2), Cur->name);
      g_ctBitmapS(r.x+2, y+1, AdjTileW, AdjTileH, &Tiles[Cur->state]);
    }
  } else
  {
    g_Font(1);
    g_FontColor(colWhite);
    g_printf(r.x+2, r.y+2, NothingMsg);
  }
}

void ListTile::Process(void)
{
  if (Head)
  {
    if (mk_Pressed(mkDown))  listcur++;
    if (mk_Pressed(mkUp))    listcur--;
    if (mk_Pressed(mkPgDn))  listcur += maxlist;
    if (mk_Pressed(mkPgUp))  listcur -= maxlist;
    if (mk_Pressed(mkHome))  listcur = 0;
    if (mk_Pressed(mkEnd))   listcur = maxel-1;
    if (mk_Pressed(mkSpace) || mk_Pressed(mkEnter))
      pressed = 1;

    int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<r.y2);

    if (m_Left() && inrange)
    {
      int y;
//    y = (m_Y() - r.y-2)/FontMan.FontHeight(1);
      y = (m_Y() - r.y-2)/theight;
      y += listtop;
      listcur = y;
    }

    if (listcur<0) listcur = 0;
    if (listcur>=maxel) listcur = maxel-1;
    if (listcur>=listtop+maxlist) listtop = listcur-maxlist+1;
    if (listcur<=listtop) listtop = listcur;
  }
  Draw();
}

MultiCheck::MultiCheck(Rect a) : List(a) {}

MultiCheck::~MultiCheck() {}

int MultiCheck::IsSet(int i)
{
  pListEl Cur = Head;

  for (int z=0; z<i; z++)
    Cur = Cur->next;

  return (Cur->state==1);
}

int MultiCheck::CheckMouse(void)
{
  if (!GetTitlePos())
  {
    int yy = FontMan.FontHeight(2);

    return (m_X()>=r.x && m_X()<=r.x2 &&
            m_Y()>=r.y-yy && m_Y()<=r.y2);
  } else
  {
    int xx = GetTitlePos();

    return (m_X()>=r.x-xx && m_X()<=r.x2 &&
            m_Y()>=r.y && m_Y()<=r.y2);
  }
}

void MultiCheck::Draw(void)
{
  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  // MultiCheck
  if (Head)
  {
    int fh = FontMan.FontHeight(2);
    int i, y = r.y+1;
    g_FontColor(colWhite);
    pListEl Cur = Head;

    for (i=0; i<listtop; i++)
      Cur = Cur->next;

    for (i=0; i<maxlist && Cur; i++, y += fh, Cur = Cur->next)
    {
      if (i+listtop == listcur)
        g_fRect(r.x+1, y, r.x2-1, y+fh, colPurple);
      g_printf(r.x+2, y, "%s %s", Cur->state ? "" : "", Cur->name);
    }
  } else
  {
    g_FontColor(colWhite);
    g_printf(r.x+2, r.y+2, NothingMsg);
  }
}

void MultiCheck::Process(void)
{
  if (Head)
  {
    if (mk_Pressed(mkDown))  listcur++;
    if (mk_Pressed(mkUp))    listcur--;
    if (mk_Pressed(mkPgDn))  listcur += maxlist;
    if (mk_Pressed(mkPgUp))  listcur -= maxlist;
    if (mk_Pressed(mkHome))  listcur = 0;
    if (mk_Pressed(mkEnd))   listcur = maxel-1;
    if (mk_Pressed(mkSpace))
    {
      pListEl Cur = Head;
      for (int z=0; z<listcur; z++)
        Cur = Cur->next;
      Cur->state = !Cur->state;
    }

    int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<r.y2);

    if (m_Left() && inrange)
    {
      int y;
      y = (m_Y() - r.y-2)/FontMan.FontHeight(2);
      y += listtop;
      listcur = y;
    }

    if (listcur<0) listcur = 0;
    if (listcur>=maxel) listcur = maxel-1;
    if (listcur>=listtop+maxlist) listtop = listcur-maxlist+1;
    if (listcur<=listtop) listtop = listcur;
  }
  Draw();
}

MultiRadio::MultiRadio(Rect a) : List(a) {}

MultiRadio::~MultiRadio() {}

int MultiRadio::IsSet(void)
{
  pListEl Cur = Head;

  int z=0;
  while (z<maxlist && Cur->state==0)
    Cur = Cur->next;

  if (z==maxlist) z=0;
  return z;
}

int MultiRadio::CheckMouse(void)
{
  if (!GetTitlePos())
  {
    int yy = FontMan.FontHeight(2);

    return (m_X()>=r.x && m_X()<=r.x2 &&
            m_Y()>=r.y-yy && m_Y()<=r.y2);
  } else
  {
    int xx = GetTitlePos();

    return (m_X()>=r.x-xx && m_X()<=r.x2 &&
            m_Y()>=r.y && m_Y()<=r.y2);
  }
}

void MultiRadio::Draw(void)
{
  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  // MultiRadio
  if (Head)
  {
    int fh = FontMan.FontHeight(2);
    int i, y = r.y+1;
    g_FontColor(colWhite);
    pListEl Cur = Head;

    int check=0;
    for (i=0; i<maxlist; i++)
    {
      check += Cur->state;
      Cur = Cur->next;
    }
    if (!check) Head->state = 1;

    Cur = Head;
    for (i=0; i<listtop; i++)
      Cur = Cur->next;

    for (i=0; i<maxlist && Cur; i++, y += fh, Cur = Cur->next)
    {
      if (i+listtop == listcur)
        g_fRect(r.x+1, y, r.x2-1, y+fh, colPurple);
      g_printf(r.x+2, y, "%s %s", Cur->state ? "" : "", Cur->name);
    }
  } else
  {
    g_FontColor(colWhite);
    g_printf(r.x+2, r.y+2, NothingMsg);
  }
}

void MultiRadio::Process(void)
{
  if (Head)
  {
    if (mk_Pressed(mkDown)) listcur++;
    if (mk_Pressed(mkUp))   listcur--;
    if (mk_Pressed(mkPgDn)) listcur += maxlist;
    if (mk_Pressed(mkPgUp)) listcur -= maxlist;
    if (mk_Pressed(mkHome)) listcur = 0;
    if (mk_Pressed(mkEnd))  listcur = maxel-1;
    if (mk_Pressed(mkSpace))
    {
      pListEl Cur = Head;
      for (int z=0; z<maxlist; z++)
      {
        Cur->state = (z==listcur);
        Cur = Cur->next;
      }
    }

    int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<r.y2);

    if (m_Left() && inrange)
    {
      int y;
      y = (m_Y() - r.y-2)/FontMan.FontHeight(2);
      y += listtop;
      listcur = y;
    }

    if (listcur<0) listcur = 0;
    if (listcur>=maxel) listcur = maxel-1;
    if (listcur>=listtop+maxlist) listtop = listcur-maxlist+1;
    if (listcur<=listtop) listtop = listcur;
  }
  Draw();
}

Input::Input(Rect a) : View(a)
{
  Data = new char[256];
  memset(Data, 0, 256);
  Cursor = Left = InsMode = 0;
  DrawCursor = AlSp = AlEd = 1;
  Max = 256;
}

Input::~Input()
{
  delete[] Data;
}

int Input::CheckMouse(void)
{
  if (!GetTitlePos())
  {
    int yy = FontMan.FontHeight(2);

    return (m_X()>=r.x && m_X()<=r.x2 &&
            m_Y()>=r.y-yy && m_Y()<=r.y2);
  } else
  {
    int xx = GetTitlePos();

    return (m_X()>=r.x-xx && m_X()<=r.x2 &&
            m_Y()>=r.y && m_Y()<=r.y2);
  }
}

void Input::CursorEnd(void)
{
  Cursor = strlen(Data);

  Left = 0;
  int offset = r.x+2;
  for (int i=0; i<Cursor; i++)
    offset += FontMan.CharWidth(2, Data[i])+1;

  while (offset>r.w-4)
    offset -= Data[Left++];
}

void Input::DrawBase(void)
{
  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  if (!GetTitlePos())
    g_printf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle()); else
    g_printf(r.x-GetTitlePos(), r.y+2, GetTitle());

  if (GetActive() && DrawCursor)
  {
    int offset = r.x+2;
    for (int i=0; i<Cursor; i++)
      offset += FontMan.CharWidth(2, Data[i])+1;

    int x1 = offset;
    int x2 = offset+FontMan.CharWidth(2, Data[Cursor]);
    int y1 = r.y+1;
    int y2 = r.y2-1;

    if (InsMode) y1 += (FontMan.FontHeight(2)+1)/2;

    if (Cursor == strlen(Data))
      g_fRect(x1, y1, x1+3, y2, colPurple); else
      g_fRect(x1, y1, x2, y2, colPurple);
  }

  // Draw text
  g_FontColor(colWhite);
  g_printf(r.x+2, r.y+2, Data);
}

void Input::Draw(void)
{
  DrawBase();
}

void Input::Process(void)
{
  if (!AlEd)
  {
    Draw();
    return;
  }

  char c = mk_PollChar();

  switch (c)
  {
    case mcNone : case mcF1 : case mcF2 : case mcF3 : case mcF4 :
    case mcF5 : case mcF6 : case mcF7 : case mcF8 : case mcF9 :
    case mcF10 : case mcF11 : case mcF12 : case mcTab :
    case mcUp : case mcDown : break;

    case mcIns :
      InsMode = !InsMode;
      break;
    case mcDel :
      if (Cursor<strlen(Data))
        memmove(Data + Cursor, Data + Cursor + 1, strlen(Data)-Cursor);
      break;
    case mcBackspace :
      if (Cursor>0)
      {
        if (Cursor<strlen(Data))
        {
          Cursor--;
          memmove(Data + Cursor, Data + Cursor + 1, strlen(Data)-Cursor);
        } else Data[--Cursor] = 0;
      }
      break;
    case mcEnter : break;
    case mcEsc : break;
    case mcLeft :
      if (Cursor>0) Cursor--;
      break;
    case mcRight :
      if (Cursor<strlen(Data)) Cursor++;
      break;
    case mcHome :
      Cursor = 0;
      break;
    case mcEnd :
      Cursor = strlen(Data);
      break;
    case mcPgUp : break;
    case mcPgDn : break;
    default :
      if (c==32 && !AlSp) break;
      if (strlen(Data)>=Max) break;
      if (InsMode)
        memmove(Data + Cursor + 1, Data + Cursor, strlen(Data)-Cursor+1);
      Data[Cursor++] = c;
      break;
  }

  Draw();
}

YesNo::YesNo(Rect a) : Input(a)
{
  answer = 1;
  UpdateAnswer();
  SetCursor(0);
}

YesNo::~YesNo() {};

void YesNo::UpdateAnswer(void)
{
  switch (answer)
  {
    case 0 :
      strcpy(Data, "No");
      break;
    case 1 :
      strcpy(Data, "Yes");
      break;
    default :
      strcpy(Data, "Huh");
      break;
  }
}

void YesNo::Process(void)
{
  if (mk_Pressed(mkSpace) || mk_Pressed(mkEnter))
  {
    answer = 1-answer;
    UpdateAnswer();
  }

  Draw();
}

TileEdit::TileEdit(Rect a) : View(a)
{
  curtile = -1;
  left  = leftg  = 15;
  right = rightg = 0;
  x = y = grid = 0;
  xw = yw = 11;

  backup = new BITMAP[TILE_MAX];

  int size = Tiles[0].x * Tiles[0].y;
  for (int i=0; i<TotTiles; i++)
  {
    backup[i].x = Tiles[i].x;
    backup[i].y = Tiles[i].y;
    backup[i].Data = new char[size];
    memcpy(backup[i].Data, Tiles[i].Data, size);
  }
}

TileEdit::~TileEdit(void)
{
  for (int i=0; i<TotTiles; i++)
    delete[] backup[i].Data;
  delete[] backup;
}

void TileEdit::GetTileData(int num, char *t)
{
  memcpy(t, backup[curtile].Data, Tile_S);
}

void TileEdit::SetTileData(int num, char *t)
{
  memcpy(backup[curtile].Data, t, Tile_S);
}

void TileEdit::InsTile(int num)
{
  backup[TotTiles].Data = new char[Tile_S];
  memset(backup[TotTiles].Data, 0, Tile_S);
  for (int i=TotTiles; i>=num; i--)
    memcpy(backup[i+1].Data, backup[i].Data, Tile_S);
  TotTiles++;
}

void TileEdit::DelTile(int num)
{
  for (int i=num; i<TotTiles; i++)
    memcpy(backup[i].Data, backup[i+1].Data, Tile_S);
  delete[] backup[TotTiles].Data;
  TotTiles--;
}

void TileEdit::EraseTile(int num)
{
  memset(backup[num].Data, 0, Tile_S);
}

void TileEdit::Process(void)
{
  if (!mk_Ctrl())
  {
    // process mouse
    if (m_X()>=r.x+1 && m_X()<=r.x2-1 && m_Y()>=r.y+1 && m_Y()<=r.y2-1)
    {
      if (m_Left())
      {
        x = (m_X()-r.x) / (xw+1);
        y = (m_Y()-r.y) / (yw+1);
        if (x>TileW-1) x = TileW-1;
        if (y>TileH-1) y = TileH-1;
        if (mk_Shift())
          left = leftg = Grab(); else
          Plot(left);
      }
      if (m_Right())
      {
        x = (m_X()-r.x) / (xw+1);
        y = (m_Y()-r.y) / (yw+1);
        if (x>TileW-1) x = TileW-1;
        if (y>TileH-1) y = TileH-1;
        if (mk_Shift())
          right = rightg = Grab(); else
          Plot(right);
      }
    }

    // process keyboard
    if (mk_Pressed(mkLeft))
    {
      if (!mk_Shift())
        x--; else
        ShiftLeft();
    }
    if (mk_Pressed(mkRight))
    {
      if (!mk_Shift())
        x++; else
        ShiftRight();
    }
    if (mk_Pressed(mkUp))
    {
      if (!mk_Shift())
        y--; else
        ShiftUp();
    }
    if (mk_Pressed(mkDown))
    {
      if (!mk_Shift())
        y++; else
        ShiftDown();
    }

    if (mk_Pressed(mkHome))  SetTile(0);
    if (mk_Pressed(mkEnd))   SetTile(TotTiles);
    if (mk_Pressed(mkPgUp) || mk_Pressed(mkMinus) || mk_Pressed(mkDash))
    {
      if (mk_Shift())
        curtile -= 10; else
        curtile--;
      if (curtile<0) curtile = 0;
      if (curtile>=TotTiles) curtile = TotTiles-1;

      SetTile(curtile);
    }
    if (mk_Pressed(mkPgDn) || mk_Pressed(mkPlus) || mk_Pressed(mkEqual))
    {
      if (mk_Shift())
        curtile += 10; else
        curtile++;
      if (curtile<0) curtile = 0;
      if (curtile>=TotTiles) curtile = TotTiles-1;

      SetTile(curtile);
    }

    // Plot
    if (mk_Down(mkSpace)) Plot(left);
    if (mk_Down(mkRAlt))  Plot(right);

    // Grab
    if (mk_Pressed(mkEnter))
      if (mk_Shift())
        right = rightg = Grab(); else
        left = leftg = Grab();

    if (x>TileW-1) x = 0;
    if (x<0) x = TileW-1;
    if (y>TileH-1) y = 0;
    if (y<0) y = TileH-1;
  }
  Draw();
}

void TileEdit::SetTile(int t)
{
  if (t<0 || t>=TotTiles) return;
  curtile = t;
  sprintf(title, "Tile %d of %d", t, TotTiles-1);
  SetTitle(title);
}

void TileEdit::Accept(void)
{
  int size = Tiles[0].x * Tiles[0].y;
  for (int i=0; i<TotTiles; i++)
    memcpy(Tiles[i].Data, backup[i].Data, size);
}

void TileEdit::Draw(void)
{
  int gridc;

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  int x1, ox1 = r.x+1;
  int y1 = r.y+1;
  int xx, yy;

  for (yy=0; yy<TileH; yy++, y1 += yw+1)
  {
    x1 = ox1;
    for (xx=0; xx<TileW; xx++, x1 += xw+1)
      g_fRect(x1, y1, x1+xw, y1+yw, backup[curtile].Data[yy*TileW+xx]);
  }

  // Grid
  if (grid)
  {
    gridc = g_rgb(32, 32, 32);
    for (yy=1; yy<TileH; yy++)
      g_HLine(r.x+1, r.x2-1, r.y+(yy*(yw+1)), gridc);
    for (xx=1; xx<TileW; xx++)
      g_VLine(r.x+(xx*(xw+1)), r.y+1, r.y2-1, gridc);
  }

  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  if (GetActive())
  {
    // Cursor
    x1 = r.x + (x*(xw+1));
    y1 = r.y + (y*(yw+1));

    g_Rect(x1, y1, x1+xw+1, y1+yw+1, colLPurp);
  }
}

int TileEdit::CheckMouse(void)
{
  int yy = FontMan.FontHeight(2);
  return (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y-yy && m_Y()<=r.y2);
}

Palette::Palette(Rect a) : View(a)
{
  right  = 0;
  left   = 15;
  across = 32;
  down   = 8;
}

Palette::~Palette(void) {}

int Palette::Updated(void)
{
  if (updated)
  {
    updated = 0;
    return 1;
  } else return 0;
}

void Palette::Process(void)
{
  // process mouse
  if (m_X()>=r.x+2 && m_X()<=r.x2-2 && m_Y()>=r.y+2 && m_Y()<=r.y2-10)
  {
    if (m_Left())
    {
      int x = (m_X()-(r.x+2)) / 3;
      int y = (m_Y()-(r.y+2)) / 3;
      if (x>=across) x = across-1;
      if (y>=across) y = across-1;
      left = y*across+x;
      updated = 1;
    }
    if (m_Right())
    {
      int x = (m_X()-(r.x+2)) / 3;
      int y = (m_Y()-(r.y+2)) / 3;
      if (x>=across) x = across-1;
      if (y>=across) y = across-1;
      right = y*across+x;
      updated = 1;
    }
  }

  if (mk_Shift())
  {
    if (mk_Pressed(mkDown))
    {
      right += across;
      updated = 1;
    }
    if (mk_Pressed(mkUp))
    {
      right -= across;
      updated = 1;
    }
    if (mk_Pressed(mkLeft))
    {
      right--;
      updated = 1;
    }
    if (mk_Pressed(mkRight))
    {
      right++;
      updated = 1;
    }

    if (right>255) right -= 256;
    if (right<0)   right += 256;
  } else
  {
    if (mk_Pressed(mkDown))
    {
      left += across;
      updated = 1;
    }
    if (mk_Pressed(mkUp))
    {
      left -= across;
      updated = 1;
    }
    if (mk_Pressed(mkLeft))
    {
      left--;
      updated = 1;
    }
    if (mk_Pressed(mkRight))
    {
      left++;
      updated = 1;
    }

    if (left>255) left -= 256;
    if (left<0)   left += 256;
  }

  Draw();
}

void Palette::Draw(void)
{
  int x, y, c;
  int w = 3;
  int h = 3;

  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  c = 0;
  for (int yy=0; yy<down; yy++)
    for (int xx=0; xx<across; xx++, c++)
    {
      x = r.x+(xx*w)+2;
      y = r.y+(yy*h)+2;

      g_fRect(x, y, x+w-1, y+h-1, c);
    }

  c = 0;
  for (yy=0, c=0; yy<down; yy++)
    for (int xx=0; xx<across; xx++, c++)
    {
      x = r.x+(xx*w)+2;
      y = r.y+(yy*h)+2;

      if (c==left)  g_Rect(x-1, y-1, x+w, y+h, colWhite);
      if (c==right) g_Rect(x-1, y-1, x+w, y+h, colPurple);
    }

  g_FontColor(colWhite);
  g_printf(r.x+2, r.y2-8, "Left:   (%d)  Right:   (%d)", left, right);
  g_fRect(r.x+21, r.y2-8, r.x+26, r.y2-2, left);
  g_fRect(r.x+69, r.y2-8, r.x+74, r.y2-2, right);
}

int Palette::CheckMouse(void)
{
  int yy = FontMan.FontHeight(2);
  return (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y-yy && m_Y()<=r.y2);
}

TileView::TileView(Rect a) : View(a)
{
  lock  = 0;
  first = 0;
  type  = 0;
}

TileView::~TileView(void) {}

void TileView::Process(void)
{
  int inrange = (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y && m_Y()<r.y2);

  if (m_Left() && inrange)
  {
    delay(150);
    type++;
  }
  if (mk_Pressed(mkSpace)) type++;
  if (type>3) type = 0;

  Draw();
}

void TileView::Draw(void)
{
  if (GetFrame()) InFrame(r.x, r.y, r.x2, r.y2);

  // Title
  g_Font(2);
  if (GetActive())
    g_FontColor(colWhite); else
    g_FontColor(colYellow);
  g_sprintf(r.x, r.y-FontMan.FontHeight(2)-1, GetTitle());

  int ix = r.x+((r.y2-r.y)-(TileW*3))/2;
  int iy = r.y+((r.x2-r.x)-(TileH*3))/2;
  int a  = 1;
  int b  = TileW+1;
  int c  = (TileW*2)+1;
  int d  = 1;
  int e  = TileH+1;
  int f  = (TileH*2)+1;

  // Draw tiles
  switch (type)
  {
    case 0 :
      g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, 0);
      g_ctBitmap(ix+b, iy+e, &tiles[first]);
      break;
    case 1 :
      g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, colPurple);
      g_ctBitmap(ix+b, iy+e, &tiles[first]);
      break;
    case 2 :
      g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, 0);
      g_ctBitmap(ix+a, iy+d, &tiles[first]);
      g_ctBitmap(ix+b, iy+d, &tiles[first]);
      g_ctBitmap(ix+c, iy+d, &tiles[first]);
      g_ctBitmap(ix+a, iy+e, &tiles[first]);
      g_ctBitmap(ix+b, iy+e, &tiles[first]);
      g_ctBitmap(ix+c, iy+e, &tiles[first]);
      g_ctBitmap(ix+a, iy+f, &tiles[first]);
      g_ctBitmap(ix+b, iy+f, &tiles[first]);
      g_ctBitmap(ix+c, iy+f, &tiles[first]);
      break;
    case 3 :
      g_fRect(r.x+1, r.y+1, r.x2-1, r.y2-1, 0);
      if (first>=4) g_ctBitmap(ix+a, iy+d, &tiles[first-4]);
      if (first>=3) g_ctBitmap(ix+b, iy+d, &tiles[first-3]);
      if (first>=2) g_ctBitmap(ix+c, iy+d, &tiles[first-2]);
      if (first>=1) g_ctBitmap(ix+a, iy+e, &tiles[first-1]);
      g_ctBitmap(ix+b, iy+e, &tiles[first]);
      if (first<TotTiles-1) g_ctBitmap(ix+c, iy+e, &tiles[first+1]);
      if (first<TotTiles-2) g_ctBitmap(ix+a, iy+f, &tiles[first+2]);
      if (first<TotTiles-3) g_ctBitmap(ix+b, iy+f, &tiles[first+3]);
      if (first<TotTiles-4) g_ctBitmap(ix+c, iy+f, &tiles[first+4]);
      break;
  }
}

int TileView::CheckMouse(void)
{
  int yy = FontMan.FontHeight(2);
  return (m_X()>=r.x && m_X()<=r.x2 && m_Y()>=r.y-yy && m_Y()<=r.y2);
}
