#ifndef __ZZT_USER__
#define __ZZT_USER__

#include "global.h"
#include "kernel.h"

#include "font.h"
#include "vgl.h"

class Rect {
public:
  int dx, dy;
  int x, y;
  int x2, y2;
  int w, h;

  Rect()                          { x = y = x2 = y2 = w = h = dx = dy = 0; }
  Rect(int ww, int hh)            { Move((g_ResX()-ww)/2, (g_ResY()-hh)/2);
                                    Size(ww, hh); Delta(0, 0); }
  Rect(int xx, int yy, int ww, int hh) { Move(xx, yy); Size(ww, hh); Delta(0, 0); }
  ~Rect()                         {}

  int operator==(Rect t)
  { return (x==t.x && y==t.y && w==t.w && h==t.h); }

  Rect operator=(Rect t)
  { x = t.x; y = t.y; dx = t.dx; dy = t.dy;
    w = t.w; h = t.h; x2 = t.x2; y2 = t.y2;
    return *this; }

  void Norm(void)                 { x2 = x+dx+w; y2 = y+dy+h; }
  void Delta(int xx, int yy)      { dx = xx; dy = yy; Norm(); }
  void Move(int xx, int yy)       { x = xx;  y = yy;  Norm(); }
  void MoveRel(int xx, int yy)    { x += xx; y += yy; Norm(); }
  void Size(int ww, int hh)       { w = ww; h = hh;   Norm(); }
  void SizeRel(int ww, int hh)    { w += ww; h += hh; Norm(); }
  void Set(int xx, int yy, int ww, int hh)
  { Move(xx, yy); Size(ww, hh); }
  void Set(int ww, int hh)
  { Move((g_ResX()-ww)/2, (g_ResY()-hh)/2); Size(ww, hh); }
  void SetLR(int xx, int yy)
  { Size(xx-x, yy-y); }
};

class View
{
private:
  BITMAP Back, Update;
  char *Title;
  int TitlePos;
  int State;
  int Active;
  int Frame;

protected:
  int X1, Y1;
  int X2, Y2;

public:
  Rect r;

  View(Rect a);
  virtual ~View();
  void Done(void);

  void SetTitle(char *str)        { Title = str; }
  char *GetTitle(void)            { return Title; }
  void SetTitleAbove(void)        { TitlePos = 0; }
  void SetTitleLeft(int n)        { TitlePos = n; }
  int  GetTitlePos(void)          { return TitlePos; }
  void SetState(int s)            { State = s; }
  int  GetState(void)             { return State; }
  void SetActive(int a)           { Active = a; }
  int  GetActive(void)            { return Active; }
  void SetFrame(int f)            { Frame = f; }
  int  GetFrame(void)             { return Frame; }

  virtual void Draw(void)=0;
  virtual void Process(void)=0;
  virtual int  CheckMouse(void)=0;
  void Save(void);
  void Restore(void);
};

class Window : public View
{
  int Shade;

public:
  Window(Rect a);
  ~Window();

  void SetShade(int s)            { Shade = s; }

  void Draw(void);
  void Process(void);
  int  CheckMouse(void)           { return 0; };
};

class Dialog : public View
{
  View **ViewList;
  int max, cur;
  int activity;
  int Shade;
  int Help;

public:
  Dialog(Rect a);
  ~Dialog();
  
  void SetHelp(int h)             { Help = h; }
  void SetShade(int s)            { Shade = s; }

  void Add(View *v);
  void Draw(void);
  void Process(void);
  int  CheckMouse(void)           { return 0; };
};

class Button : public View
{
protected:
  int mpressed;
  int kpressed;
  int pressed;

public:
  Button(Rect a);
  ~Button();

  int  IsPressed(void)            { return (pressed==1); }
  int  WasPressed(void)           { return (pressed==2); }
  void ResetPress(void)           { mpressed = kpressed = pressed = 0; }

  virtual void DrawDown(void);
  virtual void DrawUp(void);
  void DrawBase(int up);
  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

class ToolButton : public Button
{
public:
  ToolButton(Rect a);
  ~ToolButton();

  void DrawDown(void);
  void DrawUp(void);
};

class IconButton : public Button
{
public:
  IconButton(Rect a);
  ~IconButton();

  void DrawBase(int up);
  void Draw(void);
  void Process(void);
};

typedef struct tListEl {
  char  name[256];
  char  info[256];
  int   state;
  struct tListEl *next;
} ListEl, *pListEl;

class List : public View
{
protected:
  char NothingMsg[80];
  pListEl Head;
  int maxel;
  int maxlist, listtop;
  int listcur;
  int font;

public:
  List(Rect a);
  ~List();

  void SetFont(int f);
  int  GetFont(void)              { return font; }
  void Add(char *str, char *str2=NULL, int s=0);
  void Add(char *str, int s=0)    { Add(str, NULL, s); }
  void Kill(pListEl l);
  int  GetItem(void)              { return listcur; }
  char *GetName(int i);
  char *GetInfo(int i);
  void SetNoMsg(char *str)        { strncpy(NothingMsg, str, 80); }
  void SetDefault(char *str);

  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

typedef struct tGroupEl {
  char  name[256];
  int   state;
  struct tGroupEl *next;
} GroupEl, *pGroupEl;

class Group : public View
{
protected:
  pGroupEl Head;
  int maxel;
  int listcur;
  int gfont;
  char on, off;

public:
  Group(Rect a);
  ~Group();

  void Add(char *str, int s=0);
  void Kill(pGroupEl l);
  void KillAll();
  void SetFont(int f)             { gfont = f; }
  int  GetFont(void)              { return gfont; }
  int  GetItem(void)              { return listcur; }
  char *GetName(int i);
  int  GetState(int i);

  void Toggle(int i);
  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

class Check : public Group
{
public:
  Check(Rect a);
};

class Radio : public Group
{
public:
  Radio(Rect a);

  void Toggle(int i);
  void Draw(void);
};

class ListTile : public List
{
protected:
  int pressed;
  int theight;

public:
  ListTile(Rect a);
  ~ListTile();

  int  WasPressed(void)           { return (pressed==1); }

  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

class MultiCheck : public List
{
public:
  MultiCheck(Rect a);
  ~MultiCheck();

  int  IsSet(int i);

  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

class MultiRadio : public List
{
  int set;

public:
  MultiRadio(Rect a);
  ~MultiRadio();

  int  IsSet(void);

  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

class Input : public View
{
  int Cursor, Max, Left;
  int InsMode, AlSp, AlEd;
  int DrawCursor;

protected:
  char *Data;

public:
  Input(Rect a);
  ~Input();

  void SetData(char *data)        { strncpy(Data, data, 256); CursorEnd(); }
  void SetData(int data)          { itoa(data, Data, 10);     CursorEnd(); }
  char *GetDataStr(void)          { return Data; }
  int  GetDataInt(void)           { return atoi(Data); }
  void CursorEnd(void);

  void SetMax(int m)              { Max = m; }
  void AllowSpaces(int a)         { AlSp = a; }
  void AllowEdit(int e)           { AlEd = e; SetCursor(e); }
  void SetCursor(int c)           { DrawCursor = c; }

  void DrawBase();
  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

class Editor : public View
{
  int vcursorx, vcursory;
  int maxline, maxlist, listtop, listcur;
  int Cursor, InsMode;

protected:
  TextBox t;
  char *Data;

public:
  Editor(Rect a);
  ~Editor();

  int FindNextLine(int start);
  int FindPrevLine(int start);
  int FindBegLine(int start);
  int FindEndLine(int start);
  int FindColumn(int start, int col);

  void SetData(char *data)        { strcpy(Data, data); }
  char *GetData(void)             { return Data; }
  void GetXY(int &x, int &y, int c);
  void UpdateXY(void)             { GetXY(vcursorx, vcursory, Cursor); }

  void DrawBase();
  void Draw(void);
  void Process(void);
  int  CheckMouse(void);
};

class YesNo : public Input
{
  int answer;

public:
  YesNo(Rect a);
  ~YesNo();

  void SetData(int a)             { answer = a; UpdateAnswer(); }
  int  GetDataAns(void)           { return answer; }
  void UpdateAnswer(void);
  void CursorEnd(void);

  void Process(void);
};

class TileEdit : public View
{
  char title[80];
  int left, right;
  int leftg, rightg;
  int curtile, grid;
  int x, y, xw, yw;

public:
  BITMAP *backup;

  TileEdit(Rect a);
  ~TileEdit();

  void SetTile(int t);
  int  GetTile(void)              { return curtile; }
  void GetTileData(int num, char *t);
  void SetTileData(int num, char *t);
  void InsTile(int num);
  void DelTile(int num);
  void EraseTile(int num);

  void Accept(void);
  BITMAP *GetTileLib(void)        { return backup; }
  void Plot(int c)                { backup[curtile].Data[y*TileW+x] = c; }
  int  Grab(void)                 { return backup[curtile].Data[y*TileW+x]; }
  int  GetLeftG(void)             { return leftg; }
  int  GetRightG(void)            { return rightg; }
  void SetColor(int l, int r)     { left = leftg = l; right = rightg = r; }
  void SetGrid(int g)             { grid = g; }
  int  GetGrid(void)              { return grid; }

  void ShiftUp(void)              { zShiftUp(backup[curtile]); }
  void ShiftDown(void)            { zShiftDown(backup[curtile]); }
  void ShiftLeft(void)            { zShiftLeft(backup[curtile]); }
  void ShiftRight(void)           { zShiftRight(backup[curtile]); }
  void Flip(void)                 { zFlip(backup[curtile]); }
  void Mirror(void)               { zMirror(backup[curtile]); }
  void RotateL(void)              { zRotateL(backup[curtile]); }
  void RotateR(void)              { zRotateR(backup[curtile]); }

  void Process(void);
  void Draw(void);
  int  CheckMouse(void);
};

class Palette : public View
{
  int updated;
  int right, left;
  int across, down;

public:
  Palette(Rect a);
  ~Palette();

  int  GetLeft(void)         { return left; }
  void SetLeft(int l)        { left = l; }
  int  GetRight(void)        { return right; }
  void SetRight(int r)       { right = r; }
  int  Updated(void);

  void Process(void);
  void Draw(void);
  int  CheckMouse(void);
};

class TileView : public View
{
  BITMAP *tiles;
  int lock;
  int first;
  int type;

public:
  TileView(Rect a);
  ~TileView();

  void SetTileLib(BITMAP *t)   { tiles = t; }
  void SetTile(int t)          { if (!lock) first = t; }
  void ToggleLock(void)        { lock = !lock; }
  int  GetLock(void)           { return lock; }

  void Process(void);
  void Draw(void);
  int  CheckMouse(void);
};

int  FillTile(char *FillPtr, int x, int y, int d, int pxl, int pxr);
void WinFrame(int x1, int y1, int x2, int y2);
void OutFrame(int x1, int y1, int x2, int y2);
void OutFrameP(int x1, int y1, int x2, int y2);
void InFrame(int x1, int y1, int x2, int y2);

#endif
