// Vision Graphics Library, version 2.99 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version, in Turbo Pascal (FX32)
// Version 2.0 -- First 32-bit version, in Watcom C, implemented scores of
//                general optimizations
// Version 2.1 -- Incorporated as core of HGL Game Programming Library
// Version 2.99 - More powerful animation-related tools, including multi-
//                screen buffering, hardware page flipping, and virtual
//                buffers. Vertical sync and horizontal display enable
//                code also added.

#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <i86.h>
#include "debug.h"
//#include "pmode.h"
//#include "pmpro.h"
#include "mouse.h"
#include "mem32.h"
#include "vgl.h"
#include "pal.h"

int OldMode=3;                         // Previous mode
int CurMode;                           // Current mode
int CP=0, AP=0;                        // Current & active page
int NextPage=0;                        // Next available page
int MaxPage=16;                        // Maximum pages allowed
VirtPage Pages[16];                    // Virtual page manager

int t_w, tm_w;
extern void tBlit(void *s, void *d, int w, int h, int BlitMaxX);
#pragma aux tBlit = \
  "xor  eax, eax"\
  "l1:"\
  "mov  ebx, t_w"\
  "l2:"\
  "lodsb"\
  "or   al, al"\
  "jz   s"\
  "mov  [edi], al"\
  "s:"\
  "inc  edi"\
  "dec  ebx"\
  "jnz  l2"\
  "add  edi, ecx"\
  "dec  edx"\
  "jnz  l1"\
parm [edi] [esi] [ebx] [edx] [ecx]\
modify [eax ebx edx];

extern void tmBlit(void *s, void *d, int w, int h, char c, int BlitMaxX);
#pragma aux tmBlit = \
  "l1:"\
  "mov  ebx, tm_w"\
  "l2:"\
  "lodsb"\
  "or   al, al"\
  "jz   s"\
  "mov  [edi], ah"\
  "s:"\
  "inc  edi"\
  "dec  ebx"\
  "jnz  l2"\
  "add  edi, ecx"\
  "dec  edx"\
  "jnz  l1"\
parm [edi] [esi] [ebx] [edx] [ah] [ecx]\
modify [eax ebx edx];

int g_InitPage(char *screen)
{
  Pages[NextPage].ClipX1 = 0;
  Pages[NextPage].ClipY1 = 0;
  Pages[NextPage].ClipX2 = 319;
  Pages[NextPage].ClipY2 = 199;

  if (!screen)
  {
    Pages[NextPage].MemoryBuffer = 1;
    Pages[NextPage].Screen = new char[64000];
    blackout(Pages[NextPage].Screen, 16000);
  } else
  {
    Pages[NextPage].MemoryBuffer = 0;
    Pages[NextPage].Screen = (char *) screen;
  }

  NextPage++;
  return NextPage-1;
}

int g_GetMode(void)
{
  return CurMode;
}

void g_Init(void)
{
  union REGS r;

  r.h.ah = 0xF;
  int386(0x10, &r, &r);
  OldMode = r.h.al;
  CurMode = 0x13;

  // Define Page[0] as direct access video memory
  g_InitPage((char *) 0xA0000);
  blackout(Pages[0].Screen, 16000);

  // Define a virtual page for virtual page flipping (no hardware)
  g_InitPage();
  blackout(Pages[1].Screen, 16000);

  AP = 1;

  // Assign new mode
  r.w.ax = 0x13;
  int386(0x10, &r, &r);
}

void g_Done(void)
{
  union REGS r;
  struct SREGS s;

  segread(&s);
  r.w.ax = OldMode;                  // Reset user's video mode
  int386x(0x10, &r, &r, &s);

  // Free any virtual pages in system RAM
  for (int i=0; i<MaxPage; i++)
    if (Pages[i].MemoryBuffer && Pages[i].Screen) delete[] Pages[i].Screen;
}

void g_cGetBitmapWH(int x, int y, int w, int h, BITMAP *b)
{
  b->x = w;
  b->y = h;
  char *s = (b->Data = new char[b->x*b->y]);

  for (int yy=0; yy<h; yy++)
    for (int xx=0; xx<w; xx++, s++)
      if (g_inbounds(xx+x, yy+y)) *s = g_GetPixel(xx+x, yy+y);
}

void g_cGetBitmapXY(int x1, int y1, int x2, int y2, BITMAP *b)
{
  b->x = x2-x1+1;
  b->y = y2-y1+1;
  char *s = (b->Data = new char[b->x*b->y]);

  for (int yy=0; yy<b->y; yy++)
    for (int xx=0; xx<b->x; xx++, s++)
      if (g_inbounds(xx+x1, yy+y1)) *s = g_GetPixel(xx+x1, yy+y1);
}

void g_cGetBlockWH(int x, int y, int w, int h, void *b)
{
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++)
    for (int xx=0; xx<w; xx++, s++)
      if (g_inbounds(xx+x, yy+y)) *s = g_GetPixel(xx+x, yy+y);
}

void g_cGetBlockXY(int x1, int y1, int x2, int y2, void *b)
{
  int w = x2-x1+1;
  int h = y2-y1+1;
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++)
    for (int xx=0; xx<w; xx++, s++)
      if (g_inbounds(xx+x1, yy+y1)) *s = g_GetPixel(xx+x1, yy+y1);
}

void g_cSetBlockWH(int x, int y, int w, int h, void *b)
{
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++)
    for (int xx=0; xx<w; xx++, s++)
      if (g_inbounds(xx+x, yy+y)) g_SetPixel(xx+x, yy+y, *s);
}

void g_cSetBlockXY(int x1, int y1, int x2, int y2, void *b)
{
  int w = x2-x1+1;
  int h = y2-y1+1;
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++)
    for (int xx=0; xx<w; xx++, s++)
      if (g_inbounds(xx+x1, yy+y1)) g_SetPixel(xx+x1, yy+y1, *s);
}

void g_iRect(int x1, int y1, int x2, int y2, INDEX ip)
{
  BITMAP ir;

  g_GetBitmapXY(x1, y1, x2, y2, &ir);
  int size = ir.x * ir.y;
  char *data = (char *) ir.Data;

  for (int i=0; i<size; i++, data++)
    *data = ip[*data];

  g_Bitmap(x1, y1, &ir);
  delete[] ir.Data;
}

void g_ciRect(int x1, int y1, int x2, int y2, INDEX ip)
{
  BITMAP ir;

  g_cGetBitmapXY(x1, y1, x2, y2, &ir);
  int size = ir.x * ir.y;
  char *data = (char *) ir.Data;

  for (int i=0; i<size; i++, data++)
    *data = ip[*data];

  g_cBitmap(x1, y1, &ir);
  delete[] ir.Data;
}

/*
void g_VSync(void)
{
  do {} while ((inp(0x3DA) & 8) == 0);
  do {} while ((inp(0x3DA) & 8) == 8);
}

void g_HSync(void)
{
  do {} while ((inp(0x3DA) & 1) == 1);
  do {} while ((inp(0x3DA) & 1) == 0);
}
*/
void g_Update(int flippage)
{
  CP = flippage;
  g_CopyPage(CP, 0);
}

void g_CopySamePage(int x1, int y1, int x2, int y2, int dx, int dy)
{
  BITMAP dest;

  g_GetBitmapXY(x1, y1, x2, y2, &dest);
  g_Bitmap(dx, dy, &dest);
}

int g_LoadBitmap(char *f, BITMAP *b)
{
  ifstream fp;
  int i, s, total, cur;

  fp.open(f, ios::in | ios::binary);
  if (!fp) return -1;

  fp.seekg(0, ios::end);
  total = fp.tellg();
  fp.seekg(0, ios::beg);

  for (i=0, cur=0; cur<total; i++)
  {
    fp.read((unsigned char *)&b[i].x, sizeof(short));
    fp.read((unsigned char *)&b[i].y, sizeof(short));
    s = b[i].x * b[i].y;
    b[i].Data = new char[s];
    if (!b[i].Data) return -2;
    fp.read(b[i].Data, s);
    cur += s+4;
  }

  fp.close();
  return i;
}

void g_LoadBitmapNM(char *f, BITMAP *b)
{
  ifstream fp;
  int s, z;

  fp.open(f, ios::in | ios::binary);
  if (!fp)
  {
    g_Done();
    printf("Vision Graphics: Error opening bitmap file %s.\n", f);
    exit(1);
  }

  fp.read((unsigned char *)&b->x, sizeof(short));
  fp.read((unsigned char *)&b->y, sizeof(short));
  s = b->x * b->y;
  fp.read(b->Data, s);
  z = fp.gcount();
  if (z != s)
  {
    g_Done();
    printf("Vision Graphics: Error reading bitmap file %s (read %d instead of %d).\n", f, z, s);
    exit(1);
  }

  fp.close();
}

void g_SaveBitmap(char *f, BITMAP *b, int num)
{
  ofstream fp;
  int i, s, z;

  fp.open(f, ios::out | ios::binary);
  if (!fp)
  {
    g_Done();
    printf("Vision Graphics: Error opening bitmap file %s.\n", f);
    exit(1);
  }

  i = 0;
  for (i=0; i<num; i++)
  {
    fp.write((unsigned char *)&b[i].x, sizeof(short));
    fp.write((unsigned char *)&b[i].y, sizeof(short));
    s = b[i].x * b[i].y;
    fp.write(b[i].Data, s);
  }

  fp.close();
}

void g_AppendBitmap(char *f, BITMAP *b, int num)
{
  ofstream fp;
  int i, s, z;

  fp.open(f, ios::out | ios::binary | ios::app | ios::ate);
  if (!fp)
  {
    g_Done();
    printf("Vision Graphics: Error opening bitmap file %s.\n", f);
    exit(1);
  }

  i = 0;
  for (i=0; i<num; i++)
  {
    fp.write((unsigned char *)&b[i].x, sizeof(short));
    fp.write((unsigned char *)&b[i].y, sizeof(short));
    s = b[i].x * b[i].y;
    fp.write(b[i].Data, s);
  }

  fp.close();
}

void g_InitBitmap(int x, int y, BITMAP *b)
{
  b->x = x;
  b->y = y;
  b->Data = new char[x*y];
}

void g_KillBitmap(BITMAP *b, int num)
{
  int i;

  for (i=0; i<num; i++)
  {
    b[i].x = 0;
    b[i].y = 0;
    delete[] b[i].Data;
  }
}

void g_LoadPCX(char *f, BITMAP *b)
{
  PCXHEADER Header;
  ifstream fp;
  char v;
  int  gofs;
  int  i, j, k, loop;

  fp.open(f, ios::in | ios::binary);
  if (!fp)
  {
    g_Done();
    printf("Vision Graphics: Error opening PCX file %s.\n", f);
    exit(1);
  }

  gofs = 0;
  fp.read((unsigned char *)&Header, sizeof(Header));
  b->x = (Header.xmax-Header.xmin)+1;
  b->y = (Header.ymax-Header.ymin)+1;
  b->Data = new char[b->x * b->y];

  for (i=0; i<b->y; i++)
  {
    j = 0;
    while (j < Header.BytesPerLine)
    {
      fp.read(&v, sizeof(char));
      if ((v & 192) == 192)
      {
        loop = v & 63;
        j += loop;
        fp.read(&v, sizeof(char));
        for (k=0; k<loop; k++)
        {
          b->Data[gofs] = v;
          gofs++;
        }
      } else
      {
        b->Data[gofs] = v;
        gofs++;
        j++;
      }
    }
  }

  fp.close();
}
/*
int g_ResX(void)
{
  return 320;
}

int g_ResY(void)
{
  return 200;
}

int g_MaxX(void)
{
  return 319;
}

int g_MaxY(void)
{
  return 199;
}

void g_SetCurPage(int p)
{
  CP = p;
}

int g_GetCurPage(void)
{
  return CP;
}
*/
void g_SetClip(int x1, int y1, int x2, int y2)
{
  Pages[CP].ClipX1 = x1;
  Pages[CP].ClipY1 = y1;
  Pages[CP].ClipX2 = x2;
  Pages[CP].ClipY2 = y2;
}

void g_GetClip(int &x1, int &y1, int &x2, int &y2)
{
  x1 = Pages[CP].ClipX1;
  y1 = Pages[CP].ClipY1;
  x2 = Pages[CP].ClipX2;
  y2 = Pages[CP].ClipY2;
}

void g_FullClip(void)
{
  Pages[CP].ClipX1 = 0;
  Pages[CP].ClipY1 = 0;
  Pages[CP].ClipX2 = 319;
  Pages[CP].ClipY2 = 199;
}
/*
char g_rgb(uchar r, uchar g, uchar b)
{
  return pal_FindColor(r/4, g/4, b/4);
}

void g_cSetPixel(int x, int y, char c)
{
  if (g_inbounds(x, y)) g_SetPixel(x, y, c);
}

char g_cGetPixel(int x, int y)
{
  return g_inbounds(x, y)? g_GetPixel(x, y) : 0;
}
*/
void g_cHLine(int x1, int x2, int y, char c)
{
  // Make sure Y value is legit
  if (y >= 0 && y <= 199)
  {
    // Make sure x1 < x2
    if (x1 > x2)
    {
      int temp = x1;
      x1 = x2;
      x2 = temp;
    }

    // Make sure at least part of the line is on the screen
    if (!(x1<0 && x2<0) && !(x1>319 && x2>319))
    {
      // Clamp right and left edge values
      if (x1 < 0)   x1 = 0;
      if (x2 > 319) x2 = 319;

      // Render the line
      g_HLine(x1, x2, y, c);
    }
  }
}

void g_cVLine(int x, int y1, int y2, char c)
{
  // Make sure the line is legit
  if (x >= 0 && x < 320)
  {
    // Make sure y1 < y2
    if (y1 > y2)
    {
      int temp = y1;
      y1 = y2;
      y2 = temp;
    }

    // Make sure part of the line is on the screen
    if (!(y1<0 && y2<0) && !(y1>199 && y2>199))
    {
      // Clamp values vertically
      if (y1<0)   y1 = 0;
      if (y2>199) y2 = 199;

      // Render
      g_VLine(x, y1, y2, c);
    }
  }
}

void g_Line(int x1, int y1, int x2, int y2, char c)
{
  int i, dx, dy, num;
  int d, di1, di2;
  int x, xi1, xi2;
  int y, yi1, yi2;

  dx = ABS(x2-x1);
  dy = ABS(y2-y1);

  if (dx >= dy)
  {
    num = dx + 1;
    d   = (dy * 2) - dx;
    di1 = dy << 1;
    di2 = (dy-dx) << 1;
    xi1 = 1;
    xi2 = 1;
    yi1 = 0;
    yi2 = 1;
  } else
  {
    num = dy + 1;
    d   = (dx * 2) - dy;
    di1 = dx << 1;
    di2 = (dx-dy) << 1;
    xi1 = 0;
    xi2 = 1;
    yi1 = 1;
    yi2 = 1;
  }

  if (x1 > x2)
  {
    xi1 = -xi1;
    xi2 = -xi2;
  }
  if (y1 > y2)
  {
    yi1 = -yi1;
    yi2 = -yi2;
  }

  x = x1;
  y = y1;

  for (i=0; i<num; i++)
  {
    g_SetPixel(x, y, c);
    if (d < 0)
    {
      d += di1;
      x += xi1;
      y += yi1;
    } else
    {
      d += di2;
      x += xi2;
      y += yi2;
    }
  }
}

void g_Rect(int x1, int y1, int x2, int y2, char c)
{
  g_HLine(x1, x2, y1, c);
  g_HLine(x1, x2, y2, c);
  g_VLine(x1, y1, y2, c);
  g_VLine(x2, y1, y2, c);
}

void g_cRect(int x1, int y1, int x2, int y2, char c)
{
  g_cHLine(x1, x2, y1, c);
  g_cHLine(x1, x2, y2, c);
  g_cVLine(x1, y1, y2, c);
  g_cVLine(x2, y1, y2, c);
}

void g_fRect(int x1, int y1, int x2, int y2, char c)
{
  for (int y=y1; y<=y2; y++)
    g_HLine(x1, x2, y, c);
}

void g_cfRect(int x1, int y1, int x2, int y2, char c)
{
  for (int y=y1; y<=y2; y++)
    g_cHLine(x1, x2, y, c);
}

void g_Circle(int x, int y, int r, char c)
{
  int xs, ys;
  int da, db, s;

  if (!r) return;
  if (r==1)
  {
    g_SetPixel(x, y, c);
    return;
  }

  xs = 0;
  ys = r;

  do {
    da = ((xs+1)*(xs+1)) + (ys*ys) - (r*r);
    db = ((xs+1)*(xs+1)) + ((ys-1)*(ys-1)) - (r*r);
    s  = da + db;

    xs++;
    if (s>0) ys--;

    g_SetPixel(x+xs-1, y-ys+1, c);
    g_SetPixel(x-xs+1, y-ys+1, c);
    g_SetPixel(x+ys-1, y-xs+1, c);
    g_SetPixel(x-ys+1, y-xs+1, c);
    g_SetPixel(x+xs-1, y+ys-1, c);
    g_SetPixel(x-xs+1, y+ys-1, c);
    g_SetPixel(x+ys-1, y+xs-1, c);
    g_SetPixel(x-ys+1, y+xs-1, c);
  } while (xs < ys);
}

void g_cCircle(int x, int y, int r, char c)
{
  int xs, ys;
  int da, db, s;

  if (!r) return;
  if (r==1)
  {
    g_cSetPixel(x, y, c);
    return;
  }

  xs = 0;
  ys = r;

  do {
    da = ((xs+1)*(xs+1)) + (ys*ys) - (r*r);
    db = ((xs+1)*(xs+1)) + ((ys-1)*(ys-1)) - (r*r);
    s  = da + db;

    xs++;
    if (s>0) ys--;

    g_cSetPixel(x+xs-1, y-ys+1, c);
    g_cSetPixel(x-xs+1, y-ys+1, c);
    g_cSetPixel(x+ys-1, y-xs+1, c);
    g_cSetPixel(x-ys+1, y-xs+1, c);
    g_cSetPixel(x+xs-1, y+ys-1, c);
    g_cSetPixel(x-xs+1, y+ys-1, c);
    g_cSetPixel(x+ys-1, y+xs-1, c);
    g_cSetPixel(x-ys+1, y+xs-1, c);
  } while (xs < ys);
}

void g_ctBitmap(int x, int y, BITMAP *b)
{
  if (!b->Data) return;

  int x1 = x;
  int y1 = y;
  int x2 = x + b->x - 1;
  int y2 = y + b->y - 1;

  // Test whether the sprite is totally invisible
  if (x1>Pages[CP].ClipX2 || y1>Pages[CP].ClipY2 || x1 < (-b->x) || y1 < (-b->y))
    return;

  // If entire bitmap within bounds, blast it
  if (x1>=Pages[CP].ClipX1 && y1>=Pages[CP].ClipY1 &&
      x2<=Pages[CP].ClipX2 && y2<=Pages[CP].ClipY2)
  {
    g_tBitmap(x, y, b);
    return;
  }

  char *s = b->Data;

  for (int yy=0; yy<b->y; yy++)
    for (int xx=0; xx<b->x; xx++, s++)
      if (*s && g_inbounds(xx+x, yy+y)) g_SetPixel(xx+x, yy+y, *s);
}

void g_tmBitmap(int x, int y, char c, BITMAP *b)
{
  if (!b->Data) return;
  tmBlit(Pages[CP].Screen + (y << 6)+(y << 8)+x, b->Data, tm_w=b->x, b->y, c, 320-b->x);
}

void g_ctmBitmap(int x, int y, char c, BITMAP *b)
{
  if (!b->Data) return;

  int x1 = x;
  int y1 = y;
  int x2 = x + b->x - 1;
  int y2 = y + b->y - 1;

  // Test whether the sprite is totally invisible
  if (x1>Pages[CP].ClipX2 || y1>Pages[CP].ClipY2 || x1 < (-b->x) || y1 < (-b->y))
    return;

  // If entire bitmap within bounds, blast it
  if (x1>=Pages[CP].ClipX1 && y1>=Pages[CP].ClipY1 &&
      x2<=Pages[CP].ClipX2 && y2<=Pages[CP].ClipY2)
  {
    g_tmBitmap(x, y, c, b);
    return;
  }

  char *s = b->Data;

  for (int yy=0; yy<b->y; yy++)
    for (int xx=0; xx<b->x; xx++, s++)
      if (*s && g_inbounds(xx+x, yy+y)) g_SetPixel(xx+x, yy+y, c);
}

void g_ctiBitmap(int x, int y, char c, BITMAP *b)
{
  if (!b->Data) return;

  int x1 = x;
  int y1 = y;
  int x2 = x + b->x - 1;
  int y2 = y + b->y - 1;

  // Test whether the sprite is totally invisible
  if (x1>Pages[CP].ClipX2 || y1>Pages[CP].ClipY2 || x1 < (-b->x) || y1 < (-b->y))
    return;

  // If entire bitmap within bounds, blast it
  if (x1>=Pages[CP].ClipX1 && y1>=Pages[CP].ClipY1 &&
      x2<=Pages[CP].ClipX2 && y2<=Pages[CP].ClipY2)
  {
    g_tmBitmap(x, y, c, b);
    return;
  }

  char *s = b->Data;

  for (int yy=0; yy<b->y; yy++)
    for (int xx=0; xx<b->x; xx++, s++)
      if (*s && g_inbounds(xx+x, yy+y)) g_SetPixel(xx+x, yy+y, *s + c);
}

void g_Clear(char c)
{
  clear8(Pages[CP].Screen, c, 16000);
}

void g_SuperBlack(void)
{
  blackout(Pages[CP].Screen, 16000);
}
/*
void g_SetPixel(int x, int y, char c)
{
  Pages[CP].Screen[(y << 6)+(y << 8)+x] = c;
}

int g_GetPixel(int x, int y)
{
  return Pages[CP].Screen[(y << 6)+(y << 8)+x];
}

void g_HLine(int x1, int x2, int y, char c)
{
  memset8(Pages[CP].Screen + ((y << 6)+(y << 8)+x1), c, x2-x1+1);
}
*/
void g_VLine(int x, int y1, int y2, char c)
{
  char *Output = Pages[CP].Screen + (y1 << 6)+(y1 << 8)+x;

  for (int y=y1; y<=y2; y++, Output += 320)
    *Output = c;
}

void g_GetBitmapWH(int x, int y, int w, int h, BITMAP *b)
{
  b->x = w;
  b->y = h;
  char *d = Pages[CP].Screen + (y << 6)+(y << 8)+x;
  char *s = (b->Data = new char[w*h]);

  for (int yy=0; yy<h; yy++, s += w, d += 320)
    memcpy8(s, d, w);
}

void g_GetBitmapXY(int x1, int y1, int x2, int y2, BITMAP *b)
{
  b->x = x2-x1+1;
  b->y = y2-y1+1;
  char *d = Pages[CP].Screen + (y1 << 6)+(y1 << 8)+x1;
  char *s = (b->Data = new char[b->x*b->y]);

  for (int yy=0; yy<b->y; yy++, s += b->x, d += 320)
    memcpy8(s, d, b->x);
}

void g_GetBlockWH(int x, int y, int w, int h, void *b)
{
  char *d = Pages[CP].Screen + (y << 6)+(y << 8)+x;
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++, s += w, d += 320)
    memcpy8(s, d, w);
}

void g_GetBlockXY(int x1, int y1, int x2, int y2, void *b)
{
  int w = x2-x1+1;
  int h = y2-y1+1;
  char *d = Pages[CP].Screen + (y1 << 6)+(y1 << 8)+x1;
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++, s += w, d += 320)
    memcpy8(s, d, w);
}

void g_SetBlockWH(int x, int y, int w, int h, void *b)
{
  char *d = Pages[CP].Screen + (y << 6)+(y << 8)+x;
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++, s += w, d += 320)
    memcpy8(d, s, w);
}

void g_SetBlockXY(int x1, int y1, int x2, int y2, void *b)
{
  int w = x2-x1+1;
  int h = y2-y1+1;
  char *d = Pages[CP].Screen + (y1 << 6)+(y1 << 8)+x1;
  char *s = (char *)b;

  for (int yy=0; yy<h; yy++, s += w, d += 320)
    memcpy8(d, s, w);
}

void g_Bitmap(int x, int y, BITMAP *b)
{
  if (!b->Data) return;

  char *s = b->Data;
  char *d = Pages[CP].Screen + (y << 6)+(y << 8)+x;

  int w = b->x;

  for (int yy=0; yy<b->y; yy++, s += w, d += 320)
    memcpy8(d, s, w);
}

void g_cBitmap(int x, int y, BITMAP *b)
{
  if (!b->Data) return;

  int x1 = x;
  int y1 = y;
  int x2 = x + b->x - 1;
  int y2 = y + b->y - 1;

  // Test whether the sprite is totally invisible
  if (x1>Pages[CP].ClipX2 || y1>Pages[CP].ClipY2 || x1 < (-b->x) || y1 < (-b->y))
    return;

  // If entire bitmap within bounds, blast it
  if (x1>=Pages[CP].ClipX1 && y1>=Pages[CP].ClipY1 &&
      x2<=Pages[CP].ClipX2 && y2<=Pages[CP].ClipY2)
  {
    g_Bitmap(x, y, b);
    return;
  }

  if (x1 < 0) x1 = 0; else
  if (x2 > Pages[CP].ClipX2) x2 = Pages[CP].ClipX2;
  if (y1 < 0) y1 = 0; else
  if (y2 > Pages[CP].ClipY2) y2 = Pages[CP].ClipY2;

  char *s = b->Data;
  char *d = Pages[CP].Screen;

  int cw = x2-x1+1;
  int ch = y2-y1+1;

  s += (y1-y) * b->x;
  d += (y1 << 6)+(y1 << 8)+x1;
  int xoff = (x1-x);
  s += xoff;

  for (int yy=0; yy<ch; yy++)
  {
    memcpy8(d, s, cw);
    s += b->x;
    d += 320;
  }
}

void g_tBitmap(int x, int y, BITMAP *b)
{
  if (!b->Data) return;
  tBlit(Pages[CP].Screen + (y << 6)+(y << 8)+x, b->Data, t_w=b->x, b->y, 320-b->x);
}

void g_BitmapS(int x, int y, int w, int h, BITMAP *b)
{
  if (!b->Data) return;

  // If bitmap is same size, blast it
  if (b->x==w && b->y==h)
  {
    g_Bitmap(x, y, b);
    return;
  }

  int u = 0;
  int v = 0;

  for (int yy=0; yy<h; yy++)
  {
    u = (yy*b->y)/h;
    for (int xx=0; xx<w; xx++)
    {
      v = (xx*b->x)/w;
      g_SetPixel(xx+x, yy+y, b->Data[u*b->x+v]);
    }
  }
}

void g_cBitmapS(int x, int y, int w, int h, BITMAP *b)
{
  if (!b->Data) return;

  int x1 = x;
  int y1 = y;
  int x2 = x + w - 1;
  int y2 = y + h - 1;

  // Test whether the sprite is totally invisible
  if (x1>Pages[CP].ClipX2 || y1>Pages[CP].ClipY2 || x1 < (-b->x) || y1 < (-b->y))
    return;

  // If bitmap is same size, blast it
  if (b->x==w && b->y==h)
  {
    g_cBitmap(x, y, b);
    return;
  }

  int u = 0;
  int v = 0;

  for (int yy=0; yy<h; yy++)
  {
    u = (yy*b->y)/h;
    for (int xx=0; xx<w; xx++)
    {
      v = (xx*b->x)/w;
      if (g_inbounds(xx+x, yy+y)) g_SetPixel(xx+x, yy+y, b->Data[u*b->x+v]);
    }
  }
}

void g_tBitmapS(int x, int y, int w, int h, BITMAP *b)
{
  if (!b->Data) return;

  // If bitmap is same size, blast it
  if (b->x==w && b->y==h)
  {
    g_tBitmap(x, y, b);
    return;
  }

  char *s;
  int u = 0;
  int v = 0;

  for (int yy=0; yy<h; yy++)
  {
    u = (yy*b->y)/h;
    for (int xx=0; xx<w; xx++)
    {
      v = (xx*b->x)/w;
      s = b->Data+(u*b->x+v);
      if (*s) g_SetPixel(xx+x, yy+y, *s);
    }
  }
}

void g_ctBitmapS(int x, int y, int w, int h, BITMAP *b)
{
  if (!b->Data) return;

  int x1 = x;
  int y1 = y;
  int x2 = x + w - 1;
  int y2 = y + h - 1;

  // Test whether the sprite is totally invisible
  if (x1>Pages[CP].ClipX2 || y1>Pages[CP].ClipY2 || x1 < (-b->x) || y1 < (-b->y))
    return;

  // If bitmap is same size, blast it
  if (b->x==w && b->y==h)
  {
    g_cBitmap(x, y, b);
    return;
  }

  char *s;
  int u = 0;
  int v = 0;

  for (int yy=0; yy<h; yy++)
  {
    u = (yy*b->y)/h;
    for (int xx=0; xx<w; xx++)
    {
      v = (xx*b->x)/w;
      s = b->Data+(u*b->x+v);
      if (*s && g_inbounds(xx+x, yy+y)) g_SetPixel(xx+x, yy+y, *s);
    }
  }
}

/*
void g_CopyPage(int sp, int dp)
{
  memcpy32(Pages[dp].Screen, Pages[sp].Screen, 16000);
}
*/
void g_CopyDiffPage(int x1, int y1, int x2, int y2, int dx, int dy, int sp, int dp)
{
  char *s = Pages[sp].Screen + (y1 << 6)+(y1 << 8)+x1;
  char *d = Pages[dp].Screen + (dy << 6)+(dy << 8)+dx;
  int w = x2-x1;
  int h = y2-y1;

  for (int y=0; y<h; y++, s += 320, d += 320)
    memcpy8(d, s, w);
}

void g_cCopyDiffPage(int x1, int y1, int x2, int y2, int dx, int dy, int sp, int dp)
{
  int w = x2-x1+1;
  int h = y2-y1+1;

  if (x1>Pages[sp].ClipX2 || y1>Pages[sp].ClipY2 || x1<-w || y1<-h) return;

  if (x1>=Pages[sp].ClipX1 && y1>=Pages[sp].ClipY1 &&
      x2<=Pages[sp].ClipX2 && y2<=Pages[sp].ClipY2)
  {
    g_CopyDiffPage(x1, y1, x2, y2, dx, dy, sp, dp);
    return;
  }

  if (x1 < 0) x1 = 0; else
  if (x2 > Pages[sp].ClipX2) x2 = Pages[sp].ClipX2;
  if (y1 < 0) y1 = 0; else
  if (y2 > Pages[sp].ClipY2) y2 = Pages[sp].ClipY2;

  char *s = Pages[sp].Screen;
  char *d = Pages[dp].Screen;

  int cw = x2-x1+1;
  int ch = y2-y1+1;

  s += (y1 << 6)+(y1 << 8)+x1;
  d += (dy << 6)+(dy << 8)+dx;

  for (int yy=0; yy<ch; yy++, s += 320, d += 320)
    memcpy8(d, s, cw);
}

