// Vision Graphics Library, version 3.2 -- Copyright 1996 Carlos DaSilva
// Version 1.0 -- First version, in Turbo Pascal (FX32)
// Version 2.0 -- First 32-bit version, in Watcom C, implemented scores of
//                general optimizations
// Version 2.1 -- Incorporated as core of HGL Game Programming Library
// Version 3.0 -- Supports multiple video modes, including SVGA, and VBE2
// Version 3.1 -- More powerful animation-related tools, including multi-
//                screen buffering, hardware page flipping, and virtual
//                buffers. Vertical sync and horizontal display enable
//                code also added.
// Version 3.2 -- Added support for 16-bit color modes.

#ifndef __VGA__
#define __VGA__

#include <string.h>
#include "debug.h"
#include "mem32.h"
#include "pal.h"

extern int AP, CP;

typedef struct
{
  ushort x, y;
  char   *Data;
} BITMAP;

typedef struct
{
  char Manufacturer;
  char Version;
  char Encoding;
  char BitsPerPixel;
  unsigned short xmin, ymin;
  unsigned short xmax, ymax;
  unsigned short hres, vres;
  char Palette[48];
  char res;
  char ColorPlanes;
  unsigned short BytesPerLine;
  unsigned short PaletteType;
  char Filler[58];
} PCXHEADER;

typedef struct
{
  int ClipX1, ClipY1;
  int ClipX2, ClipY2;
  char *Screen;
  int MemoryBuffer;
} VirtPage;

extern VirtPage Pages[16];

#define g_inbounds(x, y) ((x)>=Pages[CP].ClipX1 && (x)<=Pages[CP].ClipX2 && \
                          (y)>=Pages[CP].ClipY1 && (y)<=Pages[CP].ClipY2)

void g_Init(void);
void g_Done(void);
int  g_GetMode(void);

void g_Clear(char c);
void g_SuperBlack(void);
inline void g_SetPixel(int x, int y, char c)
{
  Pages[CP].Screen[(y << 6)+(y << 8)+x] = c;
}
inline int  g_GetPixel(int x, int y)
{
  return Pages[CP].Screen[(y << 6)+(y << 8)+x];
}
inline void g_HLine(int x1, int x2, int y, char c)
{
  memset8(Pages[CP].Screen + ((y << 6)+(y << 8)+x1), c, x2-x1+1);
}
void g_VLine(int x, int y1, int y2, char c);
void g_GetBitmapWH(int x, int y, int w, int h, BITMAP *b);
void g_GetBitmapXY(int x1, int y1, int x2, int y2, BITMAP *b);
void g_GetBlockWH(int x, int y, int w, int h, void *b);
void g_GetBlockXY(int x1, int y1, int x2, int y2, void *b);
void g_SetBlockWH(int x, int y, int w, int h, void *b);
void g_SetBlockXY(int x1, int y1, int x2, int y2, void *b);
void g_Bitmap(int x, int y, BITMAP *b);
void g_cBitmap(int x, int y, BITMAP *b);
void g_tBitmap(int x, int y, BITMAP *b);
void g_BitmapS(int x, int y, int w, int h, BITMAP *b);
void g_cBitmapS(int x, int y, int w, int h, BITMAP *b);
void g_tBitmapS(int x, int y, int w, int h, BITMAP *b);
void g_CopyDiffPage(int x1, int y1, int x2, int y2, int dx, int dy, int sp, int dp);
void g_cCopyDiffPage(int x1, int y1, int x2, int y2, int dx, int dy, int sp, int dp);
inline void g_CopyPage(int sp, int dp)
{
  memcpy32(Pages[dp].Screen, Pages[sp].Screen, 16000);
}

void g_cGetBitmapWH(int x, int y, int w, int h, BITMAP *b);
void g_cGetBitmapXY(int x1, int y1, int x2, int y2, BITMAP *b);
void g_cGetBlockWH(int x, int y, int w, int h, void *b);
void g_cGetBlockXY(int x1, int y1, int x2, int y2, void *b);
void g_cSetBlockWH(int x, int y, int w, int h, void *b);
void g_cSetBlockXY(int x1, int y1, int x2, int y2, void *b);
void g_iRect(int x1, int y1, int x2, int y2, INDEX ip);
void g_ciRect(int x1, int y1, int x2, int y2, INDEX ip);

int  g_InitPage(char *screen=NULL);
inline int  g_ResX(void)     { return 320; }
inline int  g_ResY(void)     { return 200; }
inline int  g_MaxX(void)     { return 319; }
inline int  g_MaxY(void)     { return 199; }

void g_SetClip(int x1, int y1, int x2, int y2);
void g_GetClip(int &x1, int &y1, int &x2, int &y2);
void g_FullClip(void);
inline char g_rgb(uchar r, uchar g, uchar b)
{
  return pal_FindColor(r/4, g/4, b/4);
}
inline void g_SetCurPage(int p)        { CP = p; }
inline int  g_GetCurPage(void)         { return CP; }
void g_Update(int flippage=1);

inline void g_cSetPixel(int x, int y, char c)
{
  if (g_inbounds(x, y)) g_SetPixel(x, y, c);
}
inline char g_cGetPixel(int x, int y)
{
  return g_inbounds(x, y)? g_GetPixel(x, y) : 0;
}
void g_cHLine(int x1, int x2, int y, char c);
void g_cVLine(int x, int y1, int y2, char c);
void g_Line(int x1, int y1, int x2, int y2, char c);
void g_Rect(int x1, int y1, int x2, int y2, char c);
void g_cRect(int x1, int y1, int x2, int y2, char c);
void g_fRect(int x1, int y1, int x2, int y2, char c);
void g_cfRect(int x1, int y1, int x2, int y2, char c);
void g_Circle(int x, int y, int r, char c);
void g_cCircle(int x, int y, int r, char c);
void g_CopySamePage(int x1, int y1, int x2, int y2, int dx, int dy);

void g_Sync(void);
inline void g_VSync(void)
{
  do {} while ((inp(0x3DA) & 8) == 0);
  do {} while ((inp(0x3DA) & 8) == 8);
}
inline void g_HSync(void)
{
  do {} while ((inp(0x3DA) & 1) == 1);
  do {} while ((inp(0x3DA) & 1) == 0);
}
void g_Start(int x, int y);

int  g_LoadBitmap(char *f, BITMAP *b);
void g_LoadBitmapNM(char *f, BITMAP *b);
void g_SaveBitmap(char *f, BITMAP *b, int num);
void g_AppendBitmap(char *f, BITMAP *b, int num);
void g_InitBitmap(int x, int y, BITMAP *b);
void g_KillBitmap(BITMAP *b, int num=1);

void g_ctBitmap(int x, int y, BITMAP *b);
void g_tmBitmap(int x, int y, char c, BITMAP *b);
void g_ctmBitmap(int x, int y, char c, BITMAP *b);
void g_ctiBitmap(int x, int y, char c, BITMAP *b);
void g_ctBitmapS(int x, int y, int w, int h, BITMAP *b);
//void g_tmBitmapS(int x, int y, int w, int h, char c, BITMAP *b);
//void g_ctmBitmapS(int x, int y, int w, int h, char c, BITMAP *b);
//void g_ctiBitmapS(int x, int y, int w, int h, char c, BITMAP *b);

void g_LoadPCX(char *f, BITMAP *b);

inline BITMAP *DAT2VSP(char *data)
{
  BITMAP *temp = new BITMAP;

  memcpy(&temp->x, data, sizeof(short));
  memcpy(&temp->y, data + 2, sizeof(short));
  temp->Data = data + 4;

  return temp;
}

#endif
