// To keep things clear:
//  Page 0 - Visible page
//  Page 1 - Animation page (tiles,obj,etc...)
//  Page SwapPage - Menus, etc...
//  Page SpriPage - Sprites (Mouse pointer), GUI Swap Page
//  Page ConsPage - Console page

// Watcom includes
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// HGL includes
#include "multikey.h"
#include "sprite.h"
#include "mem32.h"
#include "mouse.h"
#include "font.h"
#include "vgl.h"
#include "pal.h"

// ZZT ][ includes
#include "resource.h"
#include "objects.h"
#include "console.h"
#include "kernel.h"
#include "global.h"
#include "editor.h"
#include "sound.h"

#define NOMIDAS

// MIDAS includes
#ifndef NOMIDAS
#include "midas.h"
#include "mconfig.h"
#include "vga.h"
#endif

char title[] = "Z2 v"Z2VERSION" - Copyright 1996 Carlos DaSilva";

int InitMouse=1, InitZZT2=0, ManualSound=0, noise=1, error;

void Init(void)
{
  InitZZT2 = 1;
  FreeMem = CoreFree();

  if (noise) printf(" Reading configuration file...\n");
  ReadCFG();

#ifndef NOMIDAS
  if (ManualSound)
  {
    midasDetectSD();
    midasConfig();
  }
#endif

  pal_Init();
  pal_SetSpeed(25);
  zzt_Fade(Cur, Black);
  
  g_Init();
  pal_SetCur(Black);
  memcpy(zztPal, Z2Pal, 768);

  ConsPage = g_InitPage();
  blackout(Pages[ConsPage].Screen, 16000);

  // Fonts must be loaded before console initialization
  FontMan.LoadFont("08.fnt", 1);
  if (ErrorLog) printf("Loaded file 08.fnt\n");
  FontMan.LoadFont("05.fnt", 2);
  if (ErrorLog) printf("Loaded file 05.fnt\n");

  InitConsole();
  ConsoleBG("internal");

  AddConsLine(title);
  AddConsLine();
  AddConsLine(" Parsing command line...");
  AddConsLine(" Reading configuration file...");
  if (noise) UpdateConsole();

  zzt_Fade(Black, zztPal);
  pal_SetCur(zztPal);
  pal_GetIndex(40, 40, 40, WinDark);
  pal_GetIndex(120, 90, 120, PalPurp);

  AddConsLine(" Initializing Graphics Engine...");
  if (noise) UpdateConsole();

  SwapPage = g_InitPage();
  blackout(Pages[SwapPage].Screen, 16000);

  colWhite  = g_rgb(255, 255, 255);
  colLGray  = g_rgb(180, 180, 180);
  colDGray  = g_rgb(128, 128, 128);
  colBlack  = g_rgb(0, 0, 0);
  colPurple = g_rgb(128, 0, 128);
  colLPurp  = g_rgb(242, 0, 242);
  colDPurp  = g_rgb(64, 0, 64);
  colYellow = g_rgb(224, 224, 0);

  AddConsLine(" Initializing Keyboard driver...");
  if (noise) UpdateConsole();
  mk_Init();

//  AddConsLine(" Initializing Timer driver...");
//  if (noise) UpdateConsole();

  AddConsLine(" Initializing Sprite Engine...");
  if (noise) UpdateConsole();
  s_Init();
  SpriPage = s_InitPage();
  
  AddConsLine(" Initializing Mouse driver...");
  if (noise) UpdateConsole();

  switch (InitMouse)
  {
    case 1 :
      if (!m_Init(320, 200, DAT2VSP(Z2Mouse_VSP)))
      {
        ReAddConsLine(" Initializing Mouse driver... failed");
        AddConsLine("   Mouse not connected, or mouse driver not installed.");
        AddConsLine("   Check: Mouse connection and mouse driver.");
        nomouse = 1;
      } else nomouse = 0;
      break;
    case 0 :
      ReAddConsLine(" Initializing Mouse driver... bypassed");
      nomouse = 1;
      break;
  }

  AddConsLine(" Reading Definition file...");
  if (noise) UpdateConsole();

  InitEngine();
  ReadDEF(DefFile);
  InitTiles();
  LoadTiles(TileFile);

  ReAddConsLine(" Reading Definition file... (%d items, %d sounds, %d tiles)",
    totalitems, totalsounds, TotTiles);
  if (noise) UpdateConsole();

  AddConsLine(" Initializing Z2 globals...");
  if (noise) UpdateConsole();
  
  InitGame();
  LoadObj("z2help.oz", Z2Help);
  AddVar("rand", 42);
  AddVar("randp", 23);
  InitEditor();

  AddConsLine(" Initializing Sound System...");
  if (noise) UpdateConsole();

#ifndef NOMIDAS
  if (!ManualSound)
  {
    midasSetDefaults();
    zm_SetDef();
  }
#endif

  if (!zm_Init())
  {
    ReAddConsLine(" Initializing Sound System... failed");
    AddConsLine("   Invalid sound card, or bad sound card settings.");
    AddConsLine("   Check: BLASTER settings, and \"DIAGNOSE /S\" in AUTOEXEC.");
  } else
  {
    zm_modSetVolume(50);
    switch (zm_modLoad(DefaultMOD))
    {
      case 0 :
        zm_modStart();
        break;
      case -1 :
        AddConsLine("   Warning: Default background MOD not found.");
        break;
      case -2 :
        AddConsLine("   Warning: Default background MOD is an invalid file.");
        AddConsLine("   Check: Use ZMOD to convert modules to Z2 format.");
        break;
    }
  }

  AddConsLine();
  AddConsLine("Type in 'help' for help on using the console.");
  AddConsLine("Type in 'comref' for a command reference.");
  AddConsLine("Press F1 in Console, or Alt-H in Editor for Z2 Manual.");
  AddConsLine();
  AddConsLine("Type in 'loadplay z2demo.z2' to get started.");
  if (noise) UpdateConsole();
}

void Done(void)
{
  zm_Done();
  DoneEngine();
  AddConsLine("Deinitializing...");
  if (ErrorLog) printf("\n\nDeinitializing...");
  UpdateConsole();
  DoneConsole();
  zzt_Fade(zztPal, Black);

  g_Done();
  printf("%s\n", title);
  if (ErrorLog) printf("Successful termination");
}

void main(int argc, char *argv[])
{
  printf("%s\n\n", title);
  printf(" Parsing command line...\n");

  int i;

  for (i=0; i<argc; i++)
  {
    if (strcmpi(argv[i] + 1, "?")==0 || strcmpi(argv[i] + 1, "help")==0)
    {
      printf("\nUsage: ZZT [switches]\n\n");
      printf("Switches (case-insensitive):\n");
      printf("  /E:<world>       Go directly to editor, optionally load <world>.\n");
      printf("  /P:<world>       Go directly to kernel, load <world>, exit to console.\n");
      printf("  /X:<world>       Go directly to kernel, load <world>, exit Z2.\n");
      printf("  /HELP, /?        Display this help screen.\n");
      printf("  /NOMOUSE         Do not initialize mouse driver.\n");
      printf("  /SOUND           Manually configure the sound card.\n");
      printf("  /LOG             Write an error log.\n");
      exit(0);
    }
  }

  noise = 0;

  InitMouse = 1;
  for (i=0; i<argc; i++)
    if (strcmpi(argv[i] + 1, "nomouse")==0)
      InitMouse = 0;

  ManualSound = 0;
  for (i=0; i<argc; i++)
    if (strcmpi(argv[i] + 1, "sound")==0)
      ManualSound = 1;

  ErrorLog = 0;
  for (i=0; i<argc; i++)
    if (strcmpi(argv[i] + 1, "log")==0)
      ErrorLog = 1;

   // jump directly to editor, esc = console
#ifndef NOEDIT
  for (i=0; i<argc; i++)
  {
    if (argv[i][1]=='e' || argv[i][1]=='E')
    {
      Init();
      if (argv[i][2] == ':') LoadWorld(argv[i] + 3);
      if (Fade) zzt_Fade(zztPal, Black);
      Edit();
      UpdateConsole();
      if (Fade) zzt_Fade(Black, zztPal);
    }
  }
#endif

  // jump directly to kernel, esc = console
  for (i=0; i<argc; i++)
  {
    if (argv[i][1]=='p' || argv[i][1]=='P')
    {
      Init();
      if (argv[i][2] == ':')
      {
        if (LoadWorld(argv[i] + 3) == SUCCESS)
        {
          LoadWorld(argv[i] + 3);
          if (Fade) zzt_Fade(zztPal, Black);
          Play();
          UpdateConsole();
          if (Fade) zzt_Fade(Black, zztPal);
        } else
        {
          AddConsLine("World file not found.");
          UpdateConsole();
        }
      }
    }
  }

  // jump directly to kernel, esc = quit
  for (i=0; i<argc; i++)
  {
    if (argv[i][1]=='x' || argv[i][1]=='X')
    {
      Init();
      if (argv[i][2] == ':')
      {
        if (LoadWorld(argv[i] + 3) == SUCCESS)
        {
          if (Fade) zzt_Fade(zztPal, Black);
          Play();
          Done();
        } else
        {
          Done();
          printf("World file not found.");
        }
        exit(0);
      }
    }
  }

  if (!InitZZT2)
  {
    noise = 1;
    Init();
  }
  Console();
  Done();
}
